/*
 *  AActivateApplication.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Sun Dec 16 2001.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "AActivateApplication.h"
#import "XRunningApplicationsList.h"
#import "MMacroDomain.h"
#import "XException.h"
#import "XRunningApplicationsList.h"
#import "KMIconRegistry.h"
#import "XErrorHandler.h"
#import "ProcessSerialNumberUtils.h"

#if KMF_EDITOR_WINDOW
#import "AActivateApplicationEditor.h"
#import "TemplateEngine.h"
#endif

#if KMF_EDITOR_WINDOW
void
AActivateApplication::UpdateToDefaultValues()
{
	if ( !mApplication.HasValue() ) {
		mApplication = XApplicationIdentifier::FinderDefaultApplication();
	}
}
#endif

CFStringRef
AActivateApplication::CopyActionDescriptionString() const
{
	NSString* theNameStr = mApplication.GetName();
	CFContainer<CFStringRef> theFormatStr = CFSTR("kmActivateApplicationTitleFormat")->CopyLocalized();
	return CFString::CreateWithFormat( theFormatStr, theNameStr );
}

XIcon
AActivateApplication::GetActionDescriptionXIcon() const
{
	return GetApplication().GetXIcon();
}

#if KMF_IS_ENGINE

enum AppState {
	kASNotRunning,
	kASBackground,
	kASForeground,
	kASRunning
};

static AppState AppStateOfApplication( NSString* path, ProcessSerialNumber& psn )
{
	bool background;
	bool found = XRunningApplicationsList::FindApplicationWithPath( kFindApplicationDoValidate, path, psn, background );
	if ( !found ) return kASNotRunning;
	if ( background ) return kASBackground;
	ProcessSerialNumber frontPSN;
	verify_noerr( GetFrontProcess( &frontPSN ) );
	if ( PSN::SameProcess( frontPSN, psn ) ) return kASForeground;
	return kASRunning;
}

enum {
	kActionStateFinished,
	kActionStateWaitFront,
	kActionStateWaitNotFront
};

float
AActivateApplication::DoAction()
{
	float time = 0;
	OSStatus err;

	XErrorHandlerMaybeLogMessage( "CPU", CFSTR("AActivateApplication::DoAction") );
	mActionState = kActionStateFinished;
	NSString* theTargetApp = GetApplication().GetPath();
	if ( theTargetApp ) {
		ProcessSerialNumber psn;
		AppState as = AppStateOfApplication( theTargetApp, psn );
		if ( as == kASForeground && mAlreadyActivatedActionType != AlreadyActivatedActionType::kNormal ) {
			switch ( mAlreadyActivatedActionType ) {
				case AlreadyActivatedActionType::kNormal:
					check( false );
					break;
				case AlreadyActivatedActionType::kSwitchToLast: {
					ProcessSerialNumber	lastPSN = XRunningApplicationsList::GetLastApplication();
					verify_noerr( SetFrontProcessWithOptions( &lastPSN, mAllWindows ? 0 : kSetFrontProcessFrontWindowOnly ) );
					mActionState = kActionStateWaitNotFront;
					break;
				}
				case AlreadyActivatedActionType::kHide:
					ShowHideProcess( &psn, false );
					mActionState = kActionStateWaitNotFront;
					break;
				case AlreadyActivatedActionType::kQuit:
					PSN::SendProcessQuitAE( psn );
					mActionState = kActionStateWaitNotFront;
					break;
			}
		} else if ( as != kASNotRunning ) {
			if ( as != kASBackground ) {
				verify_noerr( err = SetFrontProcessWithOptions( &psn, mAllWindows ? 0 : kSetFrontProcessFrontWindowOnly ) );
				if ( err == noErr ) {
					mActionState = kActionStateWaitFront;
				}
				if ( mReopenWindows ) {
					PSN::SendProcessReopenAE( psn );
				}
			}
			time = 2;
		} else {
			XApplicationIdentifier( theTargetApp ).Switch( mAllWindows );
			mActionState = kActionStateWaitFront;
			time = 10;
		}
	}
	return time;
}

Boolean AActivateApplication::IsActionCompleted()
{
	XErrorHandlerMaybeLogMessage( "CPU", CFSTR("AActivateApplication::IsActionCompleted") );
	if ( mActionState != kActionStateFinished ) {
		NSString* theTargetApp = GetApplication().GetPath();
		if ( theTargetApp ) {
			ProcessSerialNumber psn;
			AppState as = AppStateOfApplication( theTargetApp, psn );
			if ( as == kASBackground ) {
				// give up now
				mActionState = kActionStateFinished;
			} else if ( mActionState == kActionStateWaitFront && as == kASForeground ) {
				mActionState = kActionStateFinished;
			} else if ( mActionState == kActionStateWaitNotFront && as != kASForeground ) {
				mActionState = kActionStateFinished;
			} else {
				// Continue waiting
			}
		} else {
			// Lost application
			mActionState = kActionStateFinished;
		}
	}
	return mActionState == kActionStateFinished;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AActivateApplication::ActionController() const
{
	return [AActivateApplicationEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {% if reopen %}Reopen initial windows.<br />{% /if %}
 {% if mAllWindows %}Bring all windows to the front.<br />{% /if %}
 {% if !mAllWindows %}Bring just one window to the front.<br />{% /if %}
 {% if display %}{{ display | htmlencode }}<br />{% /if %}
)RAW";

NSString*
AActivateApplication::GetActionExtraHTML() const
{
	NSString* display;
	switch( mAlreadyActivatedActionType ) {
		case AlreadyActivatedActionType::kNormal:
			display = @"";
			break;
		case AlreadyActivatedActionType::kSwitchToLast:
			display = @"If already at the front, activate last application.";
			break;
		case AlreadyActivatedActionType::kHide:
			display = @"If already at the front, hide the application.";
			break;
		case AlreadyActivatedActionType::kQuit:
			display = @"If already at the front, quit the application.";
			break;
	}

	NSDictionary *variables = @{
									@"display" : display,
									@"reopen" : @(mReopenWindows),
									@"allwindows" : @(mAllWindows),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kAlreadyActivatedActionTypeNames[] = kAlreadyActivatedActionTypeNamesArray;

void
AActivateApplication::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	SafeCFPropertyListRef app = mApplication.CopyToPropertyList();
	dictionary->Set( CFSTR("Application"), app );
	CFRelease( app );
	dictionary->SetBoolean( CFSTR("ReopenWindows"), mReopenWindows );
	dictionary->SetBoolean( CFSTR("AllWindows"), mAllWindows );
	dictionary->Set( CFSTR("AlreadyActivatedActionType"), kAlreadyActivatedActionTypeNames[int(mAlreadyActivatedActionType)] );
}

void
AActivateApplication::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	SafeCFPropertyListRef app = (SafeCFPropertyListRef)dictionary->Get( CFSTR("Application") );
	XAssert( app );
	mApplication.LoadFromPropertyList( app );
	dictionary->GetBoolean( CFSTR("ReopenWindows"), mReopenWindows, false );
	dictionary->GetBoolean( CFSTR("AllWindows"), mAllWindows, true );
	mAlreadyActivatedActionType = dictionary->GetEnumClass( CFSTR("AlreadyActivatedActionType"), kAlreadyActivatedActionTypeNames, AlreadyActivatedActionType::kDEFAULT );
}

