//
//  AApplyStyle.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 5 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AApplyStyle.h"
#import "FileReference.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#import "CSClipboardDomain.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "XErrorHandler.h"
#import "NSImageExtensions.h"
#import "NSStringRegularExpressions.h"
#import "NSAttributedStringExtensions.h"
#import "NSColorExtensions.h"
#import "MMath.h"
#import "Calculate.h"

#if KMF_IS_ENGINE
#import "KMVariables.h"
#import "KMNotifications.h"
#import "UnitTestFrame.h"
#endif

#if KMF_EDITOR_WINDOW
#import "AApplyStyleEditor.h"
#import "TemplateEngine.h"
#import "NSColorExtensions.h"
#endif

static const CFStringRef kStringRangeTypeNames[] = kStringRangeTypeNamesArray;

CFStringRef
AApplyStyle::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> format = CFSTR("kmApplyStyleHTMLDescriptionTitleFormat")->CopyLocalized();
	return CFString::CreateWithFormat( format, mTarget.GetDisplayName() );
}


float
AApplyStyle::Apply( NSMutableAttributedString* theString, bool applyToAll ) const
{
	float time = 10;

	NSRange match = NSMakeRange( 0, theString.length );
	if ( applyToAll || StringRange::FindMatch( match, mStringRange, theString.string, mFirst, mSecond ) ) {
		NSMutableDictionary* attributes = [NSMutableDictionary dictionaryWithDictionary:mAttributes];

		NSString* fontname = attributes[kApplyStyleFontNameKey];
		NSString* fontsizeexpression = attributes[kApplyStyleFontSizeKey];
		BOOL fontbigger = [(NSNumber*)attributes[kApplyStyleFontBiggerKey] boolValue];
		BOOL fontsmaller = [(NSNumber*)attributes[kApplyStyleFontSmallerKey] boolValue];
		if ( fontbigger && fontsmaller ) {
			fontbigger = fontsmaller = NO;
		}
		[attributes removeObjectForKey:kApplyStyleFontNameKey];
		[attributes removeObjectForKey:kApplyStyleFontSizeKey];
		[attributes removeObjectForKey:kApplyStyleFontBiggerKey];
		[attributes removeObjectForKey:kApplyStyleFontSmallerKey];
		if ( fontname || fontsizeexpression || fontbigger || fontsmaller ) {
			NSMutableAttributedString* newString = [theString mutableCopy];
			[theString enumerateAttributesInRange: match
										  options: 0
									   usingBlock: ^(NSDictionary *attributesDictionary,
													 NSRange range,
													 BOOL *stop)
			 {
				 NSFont* font = [attributesDictionary objectForKey:NSFontAttributeName];
				 //						 [textStorage removeAttribute:NSFontAttributeName range:range];
				 if ( !font ) {
					 static NSFont* kDefaultFont = [NSFont fontWithName:@"Helvetica" size:12];
					 font = kDefaultFont;
				 }
				 CGFloat orgsize = [font pointSize];

				 CGFloat fontsize = 0;
				 if ( fontsizeexpression ) {
					 double v;
					 if ( Calculate::CalculateValue( fontsizeexpression, v, @{@"LENGTH": @(theString.length), @"FONTSIZE": @(orgsize)} ) ) {
						 fontsize = v;
					 } else {
						 XErrorHandlerMaybeLogFormat1( "display", CFSTR("Failed to get font size for expressions %@"), fontsizeexpression );
					 }
				 }
				 if ( fontsize <= 0 ) {
					 fontsize = 0;
				 }
				 CGFloat newsize = orgsize;
				 if ( fontsize > 0 ) {
					 newsize = fontsize;
				 }
				 if ( fontbigger ) {
					 newsize += 1;
				 }
				 if ( fontsmaller ) {
					 newsize = MMath::max( 1.0, newsize - 1 );
				 }
				 if ( fontname ) {
					 font = [NSFont fontWithName:fontname size:newsize];
				 } else {
					 font = [[NSFontManager sharedFontManager] convertFont:font toSize:newsize];
				 }
				 if ( font ) {
					 [newString addAttribute:NSFontAttributeName value:font range:range];
				 }
			 }];
			[theString setAttributedString:newString];
		}

		for ( NSString* key in [attributes allKeys] ) {
			NSString* c = attributes[key];
			if ( [c isKindOfClass:[NSString class]] && [NSColor isSavableString:c] ) {
				attributes[key] = [NSColor colorFromSavableString:c];
			}
		}

		if ( attributes.count > 0 ) {
			[theString addAttributes:attributes range:match];
		}
	} else {
		XErrorHandlerMaybeLogFormat2( "display", CFSTR("Failed to get range for expressions %@ and %@"), mFirst, mSecond );
	}
	return time;
}

#if KMF_IS_ENGINE

float
AApplyStyle::DoAction()
{
	float time = 0;

	const XClipboardDataSet* clip = mTarget.GetClipboardDataSet();
	if ( clip ) {
		NSAttributedString* inClipboard = clip->AttributedString( true );
		NSMutableAttributedString* clipboard = [inClipboard mutableCopy];

		time = Apply( clipboard, false );

		mTarget.WriteClipboardDataSet( XClipboardDataSet( clipboard ) );
	}
	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AApplyStyle::ActionController() const
{
	return [AApplyStyleEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {% if font %}Set font to {{ font | htmlencode }}.<br />{% /if %}
 {% if size %}
	Set size to {{ size }}{% if smaller %}, then make smaller{% /if %}{% if bigger %}, then make bigger{% /if %}.<br />
 {% else %}
	 {% if smaller %}Make smaller.<br />{% /if %}
	 {% if bigger %}Make bigger.<br />{% /if %}
 {% /if %}
 {% if foreground %}
	<span class="IconTitle">Set foreground to</span> <img src={{ foreground | inlineimage: 16 }} alt="Swatch" width="16" height="16" />{% if background %}<span class="IconTitle"> and background to</span> <img src={{ background | inlineimage: 16 }} alt="Swatch" width="16" height="16" />{% /if %}<br />
 {% else %}
		{% if background %}<span class="IconTitle">Set background to</span> <img src={{ background | inlineimage: 16 }} alt="Swatch" width="16" height="16" /><br />{% /if %}
 {% /if %}
{% if strikethrough %}
 {% if strikethroughnone %}
	Turn off strikethrough.<br />
 {% else %}
	<span class="IconTitle">Set strikethrough to {{ strikethroughpattern }} {{ strikethrough }}{% if not strikethroughcolor %}.{% /if %}</span>{% if strikethroughcolor %} <img src={{ strikethroughcolor | inlineimage: 16 }} alt="Swatch" width="16" height="16" />{% /if %}<br />
 {% /if %}
{% /if %}
{% if underline %}
 {% if underlinenone %}
	Turn off underline.<br />
 {% else %}
	<span class="IconTitle">Set underline to {{ underlinepattern }} {{ underline }}{% if not underlinecolor %}.{% /if %}</span>{% if underlinecolor %} <img src={{ underlinecolor | inlineimage: 16 }} alt="Swatch" width="16" height="16" />{% /if %}<br />
 {% /if %}
{% /if %}
{% if hassuperscript %}
 {% if superscript == 0 %}
	Turn off sub/superscript.<br />
 {% /if %}
 {% if superscript > 0 %}
	Turn on superscript.<br />
 {% /if %}
 {% if superscript < 0 %}
	Turn on subscript.<br />
 {% /if %}
{% /if %}
 {{ sample }}
 {% if apply %}{{ apply }}.<br />{% /if %}
)RAW";

static void AddVariablesFor( NSMutableDictionary* variables, NSString* which, NSNumber* stylen, NSString* cs )
{
	if ( stylen ) {
		int mask = [stylen integerValue];
		int style = mask & 0xFF;
		int pattern = mask & 0x0FF00;
		if ( style == NSUnderlineStyleNone ) {
			variables[[NSString stringWithFormat:@"%@none",which]] = @YES;
		} else {
			if ( style == NSUnderlineStyleSingle ) {
				variables[which] = @"single";
			} else if ( style == NSUnderlineStyleThick ) {
				variables[which] = @"thick";
			} else if ( style == NSUnderlineStyleDouble ) {
				variables[which] = @"double";
			} else {
				check( false );
			}
			if ( pattern == NSUnderlinePatternSolid ) {
				variables[[NSString stringWithFormat:@"%@pattern",which]] = @"solid";
			} else if ( pattern == NSUnderlinePatternDot ) {
				variables[[NSString stringWithFormat:@"%@pattern",which]] = @"dot";
			} else if ( pattern == NSUnderlinePatternDash ) {
				variables[[NSString stringWithFormat:@"%@pattern",which]] = @"dash";
			} else if ( pattern == NSUnderlinePatternDashDot ) {
				variables[[NSString stringWithFormat:@"%@pattern",which]] = @"dash dot";
			} else if ( pattern == NSUnderlinePatternDashDotDot ) {
				variables[[NSString stringWithFormat:@"%@pattern",which]] = @"dash dot dot";
			} else {
				check( false );
			}
			if ( cs ) {
				NSColor* c = [NSColor colorFromSavableString:cs];
				if ( cs ) {
					NSImage* i = [NSImage imageWithSwatch:c];
					if ( i ) {
						variables[[NSString stringWithFormat:@"%@color",which]] = i;
					}
				}
			}
		}
	}
}

NSString*
AApplyStyle::GetActionExtraHTML() const
{
	NSMutableDictionary *variables = [NSMutableDictionary new];
	if ( mAttributes[kApplyStyleFontNameKey] ) {
		NSFont* f = [NSFont fontWithName:mAttributes[kApplyStyleFontNameKey] size:0];
		if ( f ) {
			variables[@"font"] = f.displayName;
		}
	}
	if ( mAttributes[kApplyStyleFontSizeKey] ) {
		variables[@"size"] = mAttributes[kApplyStyleFontSizeKey];
	}
	if ( [mAttributes[kApplyStyleFontSmallerKey] boolValue] ) {
		variables[@"smaller"] = @YES;
	}
	if ( [mAttributes[kApplyStyleFontBiggerKey] boolValue] ) {
		variables[@"bigger"] = @YES;
	}
	if ( mAttributes[NSSuperscriptAttributeName] ) {
		variables[@"hassuperscript"] = @YES;
		variables[@"superscript"] = mAttributes[NSSuperscriptAttributeName];
	}
	if ( mAttributes[NSForegroundColorAttributeName] ) {
		NSColor* c = [NSColor colorFromSavableString:mAttributes[NSForegroundColorAttributeName]];
		if ( c ) {
			NSImage* i = [NSImage imageWithSwatch:c];
			if ( i ) {
				variables[@"foreground"] = i;
			}
		}
	}
	if ( mAttributes[NSBackgroundColorAttributeName] ) {
		NSColor* c = [NSColor colorFromSavableString:mAttributes[NSBackgroundColorAttributeName]];
		if ( c ) {
			NSImage* i = [NSImage imageWithSwatch:c];
			if ( i ) {
				variables[@"background"] = i;
			}
		}
	}

	AddVariablesFor( variables, @"strikethrough", mAttributes[NSStrikethroughStyleAttributeName], mAttributes[NSStrikethroughColorAttributeName] );
	AddVariablesFor( variables, @"underline", mAttributes[NSUnderlineStyleAttributeName], mAttributes[NSUnderlineColorAttributeName] );

	NSMutableAttributedString* s = [[NSMutableAttributedString alloc] initWithString:@"Sample Text"];
	(void) Apply( s, true );
	variables[@"sample"] = [s HTMLString];

	NSString* apply = [[NSString stringWithFormat:@"kApplyStyleTo%@", kStringRangeTypeNames[int(mStringRange)]] localized];
	if ( NeedSecond() ) {
		apply = [NSString stringWithFormat:apply, mFirst, mSecond];
	} else if ( NeedFirst() ) {
		apply = [NSString stringWithFormat:apply, mFirst];
	}
	variables[@"apply"] = apply;

	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

//static NSDictionary* ToSavableDictionary( NSDictionary* attributes )
//{
//	NSMutableDictionary* savable = [NSMutableDictionary dictionaryWithDictionary:attributes];
//	for ( NSString* key in [attributes allKeys] ) {
//		NSColor* c = attributes[key];
//		if ( [c isKindOfClass:[NSColor class]] ) {
//			savable[key] = [c savableString];
//		}
//	}
//	return savable;
//}

static NSDictionary* FromSavableDictionary( NSDictionary* savable )
{
	NSMutableDictionary* attributes = [NSMutableDictionary dictionaryWithDictionary:savable];
	for ( NSString* key in [savable allKeys] ) {
		NSString* c = savable[key];
		if ( [c isKindOfClass:[NSString class]] && [NSColor isSavableString:c] ) {
			attributes[key] = [NSColor colorFromSavableString:c];
		}
	}
	return attributes;
}

void
AApplyStyle::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("StringRangeType"), kStringRangeTypeNames[int(mStringRange)] );
	mTarget.SaveToDictionary( dictionary, CFSTR("Target") );
	dictionary->Set( CFSTR("First"), CFBridgingToCFString(mFirst) );
	dictionary->Set( CFSTR("Second"), CFBridgingToCFString(mSecond) );
	dictionary->Set( CFSTR("Attributes"), CFBridgingToCF( mAttributes ) );
}

void
AApplyStyle::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mStringRange = dictionary->GetEnumClass( CFSTR("StringRangeType"), kStringRangeTypeNames, StringRangeType::kAll );
	mTarget.InitializeFromDictionary( dictionary, CFSTR("Target") );

	CFStringRef value;
	dictionary->Get( CFSTR("First"), value, CFSTR("2") );
	check( value );
	mFirst = value->GetNS();

	dictionary->Get( CFSTR("Second"), value, CFSTR("4") );
	check( value );
	mSecond = value->GetNS();

	CFDictionaryRef attributes;
	XAssert( dictionary->Get( CFSTR("Attributes"), attributes, CFDictionaryRef(NULL) ) );
	mAttributes = CFBridgingToNS(attributes);
}

