/*
 *  AClipboardSwitcherMacroAction.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Sun Apr 27 2003.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "KMDefines.h"
#import "KMNotifications.h"
#import "AClipboardSwitcherMacroAction.h"
#import "CSClipboardDomain.h"
#import "UCGRemoteControl.h"
#import "SystemDefinedKeys.h"
#import "XClipboardHistory.h"
#import "KMPreferences.h"
#import "NSStringExtensions.h"
#if KMF_EDITOR_WINDOW
#import "AClipboardSwitcherMacroEditor.h"
#endif

AClipboardSwitcherMacroAction::AClipboardSwitcherMacroAction( CSMacroActionType inMacroAction )
	: mAction( inMacroAction )
{

}

CFStringRef
AClipboardSwitcherMacroAction::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> titleFormat;

	switch ( GetAction() ) {
		case CSMacroActionType::kCut:
			titleFormat = CFSTR("aCSMacroAction_CutFormat")->CopyLocalized();
			break;
		case CSMacroActionType::kCopy:
			titleFormat = CFSTR("aCSMacroAction_CopyFormat")->CopyLocalized();
			break;
		case CSMacroActionType::kPaste:
			titleFormat = CFSTR("aCSMacroAction_PasteFormat")->CopyLocalized();
			break;
//		default:
	}
	
	return CFString::CreateWithFormat( titleFormat, mClipboardReference.GetClipboardName() );
}

NSString*
AClipboardSwitcherMacroAction::GetNewActionTitle() const
{
	NSString* titleKey;

	switch ( GetAction() ) {
		case CSMacroActionType::kCut:
			titleKey = @"aCSMacroAction_CutTitle";
			break;
		case CSMacroActionType::kCopy:
			titleKey = @"aCSMacroAction_CopyTitle";
			break;
		case CSMacroActionType::kPaste:
			titleKey = @"aCSMacroAction_PasteTitle";
			break;
	}

	return [titleKey localized];
}

CSMacroActionType
AClipboardSwitcherMacroAction::GetAction() const
{
	return mAction;
}

void
AClipboardSwitcherMacroAction::SetAction(const CSMacroActionType& inAction )
{
	mAction = inAction;
}

#if KMF_IS_ENGINE

enum {
	kStateNone,
	kStateWaitingOnPaste,
	kStateWaitingOnCutCopy,
	kStateWaitingOnReplacedClipboard
};

static EventTime gClipboardDelay;

float
AClipboardSwitcherMacroAction::DoAction()
{
	float time = 0;
	mCopyPasteState = kStateNone;
	gClipboardDelay = XClipboardDataSet::GetClipboardDelay();
	
	mClipboardReference.UpdateClipboardReference();
	CSClipboardDataSet*	pClip = mClipboardReference.GetClipboardDataSet();
	
	if ( pClip ) {
		if ( GetAction() == CSMacroActionType::kPaste ) {
			XClipboardHistory::PlaceInClipboard( *pClip );
			UCGRemoteControl::SimulateKeystroke( cmdKey, SystemDefinedKeys::GetKeyCodeFromDisplayUniChar( kPasteChar ), false );
			mCopyPasteState = kStateWaitingOnPaste;
			time = 20;
		} else {
			char theCommandChar;
			if ( GetAction() == CSMacroActionType::kCopy ) {
				theCommandChar = kCopyChar;
			} else {
				theCommandChar = kCutChar;
			}
			UCGRemoteControl::SimulateKeystroke( cmdKey, SystemDefinedKeys::GetKeyCodeFromDisplayUniChar( theCommandChar ), false );		
			mCopyPasteState = kStateWaitingOnCutCopy;
		}
	}
	mActionTime = GetCurrentEventTime();
	return time;
}

Boolean
AClipboardSwitcherMacroAction::IsActionCompleted()
{	
	switch ( mCopyPasteState ) {
		case kStateWaitingOnPaste: {
			if ( GetCurrentEventTime() > mActionTime + gClipboardDelay ) {
				mCopyPasteState = kStateWaitingOnReplacedClipboard;
				mActionTime = GetCurrentEventTime();
			}
			break;
		}
		case kStateWaitingOnCutCopy: {
			// We can wait twice as long if we are not replacing the clipboard
			EventTime wait = 2*gClipboardDelay;
			if ( GetCurrentEventTime() > mActionTime + wait ) {
				CSClipboardDataSet*	pClip = mClipboardReference.GetClipboardDataSet();
				if ( pClip ) {
					(*pClip) = (*XClipboardHistory::GetCurrent());
					CSClipboardDomain::GetClipboardDomain()->SaveClipData( KMNotifications::kLoadClipDataFromEngineAction );
				}
				mCopyPasteState = kStateNone;
			}
			break;
		}
		case kStateWaitingOnReplacedClipboard: {
			if ( GetCurrentEventTime() > mActionTime + gClipboardDelay ) {
				mCopyPasteState = kStateNone;
			}
			break;
		}
		case kStateNone: {
			break;
		}
	}
	return mCopyPasteState == kStateNone;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AClipboardSwitcherMacroAction::ActionController() const
{
	return [AClipboardSwitcherMacroEditor class];
}

#endif

#pragma mark streaming

static const CFStringRef kCSMacroActionTypeNames[] = kCSMacroActionTypeNamesArray;

void
AClipboardSwitcherMacroAction::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR( "Action" ), kCSMacroActionTypeNames[int(mAction)] );
	mClipboardReference.SaveToDictionary( dictionary );
}

void
AClipboardSwitcherMacroAction::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mAction = dictionary->GetEnumClass( CFSTR("Action"), kCSMacroActionTypeNames, CSMacroActionType::kDEFAULT );
	mClipboardReference.InitializeFromDictionary( dictionary );
}
