//
//  ACompositeImage.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 10 May 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "ACompositeImage.h"
#import "XException.h"
#import "KMDefines.h"
#import "XErrorHandler.h"
#import "NSStringRegularExpressions.h"
#import "Calculate.h"
#import "TokenText.h"
#import "CSClipboardDomain.h"
#import "NSStringExtensions.h"
#import "NSImageExtensions.h"

#if KMF_EDITOR_WINDOW
#import "ACompositeImageEditor.h"
#import "TemplateEngine.h"
#endif
#if KMF_IS_ENGINE
#import "KMNotifications.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#endif

static CFStringRef kDefaultLeft = CFSTR("100");
static CFStringRef kDefaultTop = CFSTR("100");
static CFStringRef kDefaultWidth = CFSTR("0");
static CFStringRef kDefaultHeight = CFSTR("0");

ACompositeImage::ACompositeImage()
	:	HBaseAction(),
		mLeft( kDefaultLeft->Copy() ),
		mTop( kDefaultTop->Copy() ),
		mWidth( kDefaultWidth->Copy() ),
		mHeight( kDefaultHeight->Copy() )
{
}

CFStringRef
ACompositeImage::CopyActionDescriptionString() const
{
	CFStringRef key;
	if ( mUseText ) {
		key = CFSTR("kmCompositeImageText");
	} else {
		key = CFSTR("kmCompositeImageClipboard");
	}
	return key->CopyLocalized();
}

#if KMF_IS_ENGINE

float
ACompositeImage::DoAction()
{
	const XClipboardDataSet* targetClip = mTarget.GetClipboardDataSet();
	if ( targetClip ) {
		NSImage* image = targetClip->Image();
		if ( image ) {

			NSImage* sourceImage = nil;
			NSAttributedString* sourceText = nil;
			if ( mUseText ) {
				sourceText = mStyledText.GetAttributedString();
			} else {
				const XClipboardDataSet* sourceClip = mSource.GetClipboardDataSet();
				if ( sourceClip ) {
					sourceImage = sourceClip->Image();
					if ( !sourceImage ) {
						sourceText = sourceClip->AttributedString();
					}
				}
			}
			if ( sourceImage || sourceText ) {
				NSRect imageRect;
				imageRect.origin = NSZeroPoint;
				imageRect.size = image.size;
				NSDictionary* calculationContext;
				if ( sourceImage ) {
					NSRect sourceImageRect;
					sourceImageRect.origin = NSZeroPoint;
					sourceImageRect.size = sourceImage.size;
					calculationContext = @{ @"IMAGE" : NSStringFromRect(imageRect), @"SOURCEIMAGE" : NSStringFromRect(sourceImageRect) };
				} else {
					calculationContext = @{ @"IMAGE" : NSStringFromRect(imageRect) };
				}

				double left, top, width, height;
				if ( !Calculate::CalculateValue( mLeft->GetNS(), left, calculationContext ) ) {
					XErrorHandlerMaybeLogFormat1("always", CFSTR("Resize Image %@ failed"), mLeft->GetNS() );
					return 0;
				} else if ( !Calculate::CalculateValue( mTop->GetNS(), top, calculationContext ) ) {
					XErrorHandlerMaybeLogFormat1("always", CFSTR("Resize Image %@ failed"), mTop->GetNS() );
					return 0;
				}
				if ( !Calculate::CalculateValue( mWidth->GetNS(), width, calculationContext ) ) {
					width = 0;
				}
				if ( !Calculate::CalculateValue( mHeight->GetNS(), height, calculationContext ) ) {
					height = 0;
				}
				left = roundf( left );
				top = roundf( top );
				width = roundf( width );
				height = roundf( height );

				NSImage* result = [image copy];
				[result lockFocus];
				if ( sourceImage ) {
					if ( width < 1 ) {
						width = sourceImage.size.width;
					}
					if ( height < 1 ) {
						height = sourceImage.size.height;
					}
					[sourceImage drawInRect:NSMakeRect(left, image.size.height-top-height, width, height) fromRect:NSZeroRect operation:NSCompositeSourceOver fraction: 1.0];
				} else {
					if ( width < 1 ) {
						width = 9999;
					}
					if ( height < 1 ) {
						height = 9999;
					}
					[sourceText drawInRect:NSMakeRect(left, image.size.height-top-height, width, height)];
				}
				[result unlockFocus];
				if ( result ) {
					mTarget.WriteClipboardDataSet( XClipboardDataSet( result ) );
				}
			} else {
				XErrorHandlerMaybeLogMessage("always", CFSTR("Composite Image cannot get source data") );
			}
		} else {
			XErrorHandlerMaybeLogMessage("always", CFSTR("Composite Image cannot get target image") );
		}
	}
	
	return 10;
}


#endif

#if KMF_EDITOR_WINDOW

Class
ACompositeImage::ActionController() const
{
	return [ACompositeImageEditor class];
}

NSString*
ACompositeImage::GetActionHTMLDescriptionString() const
{
	if ( mUseText ) {
		return [NSString stringWithFormat:[@"kmCompositeImageHTMLText" localized], mTarget.GetDisplayName()];
	} else {
		return [NSString stringWithFormat:[@"kmCompositeImageHTMLClipboard" localized], mSource.GetDisplayName(), mTarget.GetDisplayName()];
	}
}

static NSString* kActionDisplayHTML = @R"RAW(
 {{ text }}
 {{ where }}.<br />
)RAW";

NSString*
ACompositeImage::GetActionExtraHTML() const
{
	NSString* where;
	if ( mWidth->IsEqual(CFSTR("0")) && mHeight->IsEqual(CFSTR("0")) ) {
		where = [NSString stringWithFormat:@"At (%@,%@)", mLeft->GetNS(), mTop->GetNS()];
	} else {
		where = [NSString stringWithFormat:@"In (%@,%@,%@,%@)", mLeft->GetNS(), mTop->GetNS(), mWidth->GetNS(), mHeight->GetNS()];
	}
	NSDictionary *variables = @{
							@"text" : mUseText ? mStyledText.GetDisplayHTMLContent() : @"",
							@"where" : where,
						};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

void
ACompositeImage::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	mTarget.SaveToDictionary( dictionary, CFSTR("Target") );
	dictionary->Set( CFSTR("Left"), CFStringRef(mLeft) );
	dictionary->Set( CFSTR("Top"), CFStringRef(mTop) );
	dictionary->Set( CFSTR("Width"), CFStringRef(mWidth) );
	dictionary->Set( CFSTR("Height"), CFStringRef(mHeight) );
	mSource.SaveToDictionary( dictionary, CFSTR("Source") );
	mStyledText.SaveToDictionary( dictionary );
	dictionary->SetBoolean( CFSTR("UseText"), mUseText );
}

void
ACompositeImage::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mTarget.InitializeFromDictionary( dictionary, CFSTR("Target") );
	CFStringRef value;
	dictionary->Get( CFSTR("Left"), value, kDefaultLeft );
	check( value );
	mLeft = value->Copy();
	
	dictionary->Get( CFSTR("Top"), value, kDefaultTop );
	check( value );
	mTop = value->Copy();
	
	dictionary->Get( CFSTR("Width"), value, kDefaultWidth );
	check( value );
	mWidth = value->Copy();
	
	dictionary->Get( CFSTR("Height"), value, kDefaultHeight );
	check( value );
	mHeight = value->Copy();
	mSource.InitializeFromDictionary( dictionary, CFSTR("Source") );
	mStyledText.InitializeFromDictionary( dictionary );
	dictionary->GetBoolean( CFSTR("UseText"), mUseText, false );
}
