//
//  ADrawImage.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 10 May 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "ADrawImage.h"
#import "XException.h"
#import "KMDefines.h"
#import "XErrorHandler.h"
#import "NSStringRegularExpressions.h"
#import "Calculate.h"
#import "TokenText.h"
#import "CSClipboardDomain.h"
#import "NSStringExtensions.h"
#import "NSImageExtensions.h"
#import "NSColorExtensions.h"

#if KMF_EDITOR_WINDOW
#import "ADrawImageEditor.h"
#import "TemplateEngine.h"
#endif
#if KMF_IS_ENGINE
#import "KMNotifications.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#endif

static CFStringRef kDefaultLeft = CFSTR("100");
static CFStringRef kDefaultTop = CFSTR("100");
static CFStringRef kDefaultWidth = CFSTR("200");
static CFStringRef kDefaultHeight = CFSTR("200");
static const int kDefaultFillAlpha = 0;
static const RGBColor kDefaultFillColor = { 0, 0, 0 };
static const int kDefaultStrokeAlpha = 100;
static const RGBColor kDefaultStrokeColor = { 255, 0, 0 };
const CFStringRef kDefaultThickness = CFSTR("5");
const CFStringRef kDefaultCorner = CFSTR("0");

ADrawImage::ADrawImage()
	:	HBaseAction(),
		mDrawAction(kDrawImageActionDEFAULT),
		mLeft( kDefaultLeft->Copy() ),
		mTop( kDefaultTop->Copy() ),
		mWidth( kDefaultWidth->Copy() ),
		mHeight( kDefaultHeight->Copy() ),
		mThickness( kDefaultThickness->Copy() ),
		mCorner( kDefaultCorner->Copy() )
{
	mFillColor = [NSColor colorWithCalibratedRed:kDefaultFillColor.red/255.0 green:kDefaultFillColor.green/255.0 blue:kDefaultFillColor.blue/255.0 alpha:kDefaultFillAlpha/100.0];
	mStrokeColor = [NSColor colorWithCalibratedRed:kDefaultStrokeColor.red/255.0 green:kDefaultStrokeColor.green/255.0 blue:kDefaultStrokeColor.blue/255.0 alpha:kDefaultStrokeAlpha/100.0];
}

void
ADrawImage::SetFillColor( NSColor* inFillColor )
{
	mFillColor = [[inFillColor colorUsingColorSpaceName:NSCalibratedRGBColorSpace] copy];
}

void
ADrawImage::SetStrokeColor( NSColor* inStrokeColor )
{
	mStrokeColor = [[inStrokeColor colorUsingColorSpaceName:NSCalibratedRGBColorSpace] copy];
}

CFStringRef
ADrawImage::CopyActionDescriptionShortString() const
{
	CFStringRef description;
	switch ( mDrawAction ) {
		case kDrawImageActionLine:
			description = CFSTR("kmDrawImageActionLine")->CopyLocalized();
			break;
		case kDrawImageActionRectangle:
			description = CFSTR("kmDrawImageActionRectangle")->CopyLocalized();
			break;
		case kDrawImageActionOval:
			description = CFSTR("kmDrawImageActionOval")->CopyLocalized();
			break;
	}

	return description;
}

CFStringRef
ADrawImage::CopyActionDescriptionString() const
{
	return CopyActionDescriptionShortString();
}

const ArrowString::Style*
ADrawImage::GetArrowStyles() const
{
	switch ( mDrawAction ) {
		case kDrawImageActionLine:
			return ArrowString::kRightDownRightDown;
		case kDrawImageActionRectangle:
			return ArrowString::kRightDownWidthHeight;
		case kDrawImageActionOval:
			return ArrowString::kRightDownToCenterWidthHeight;
	}
}

#if KMF_IS_ENGINE

float
ADrawImage::DoAction()
{	
	const XClipboardDataSet* targetClip = mTarget.GetClipboardDataSet();
	if ( targetClip ) {
		NSImage* image = targetClip->Image();
		NSImage* result = nil;
		if ( image ) {
			NSRect imageRect;
			imageRect.origin = NSZeroPoint;
			imageRect.size = image.size;
			NSDictionary* calculationContext = @{ @"IMAGE" : NSStringFromRect(imageRect) };

			double left, top, width, height, thickness, corner=0;
			if ( !Calculate::CalculateValue( mLeft->GetNS(), left, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Draw Image %@ failed"), mLeft->GetNS() );
				return 0;
			} else if ( !Calculate::CalculateValue( mTop->GetNS(), top, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Draw Image %@ failed"), mTop->GetNS() );
				return 0;
			} else if ( !Calculate::CalculateValue( mWidth->GetNS(), width, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Draw Image %@ failed"), mWidth->GetNS() );
				return 0;
			} else if ( !Calculate::CalculateValue( mHeight->GetNS(), height, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Draw Image %@ failed"), mHeight->GetNS() );
				return 0;
			} else if ( !Calculate::CalculateValue( mThickness->GetNS(), thickness, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Draw Image %@ failed"), mThickness->GetNS() );
				return 0;
			} else if ( mDrawAction == kDrawImageActionRectangle && !Calculate::CalculateValue( mCorner->GetNS(), corner, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Draw Image %@ failed"), mCorner->GetNS() );
				return 0;
			}
			if ( mDrawAction == kDrawImageActionOval ) {
				left = left - width / 2;
				top = top - height / 2;
			}
			left = roundf( left );
			top = roundf( top );
			width = roundf( width );
			height = roundf( height );
			
			result = [image copy];
			if ( result ) {
				[result lockFocus];
				NSBezierPath* path = nil;
				switch ( mDrawAction ) {
					case kDrawImageActionLine: {
						path = [NSBezierPath bezierPath];
						[path moveToPoint:NSMakePoint( left, image.size.height-top )];
						[path lineToPoint:NSMakePoint( width, image.size.height-height )];
						break;
					}
					case kDrawImageActionRectangle: {
						if ( corner > 0.01 ) {
							path = [NSBezierPath bezierPathWithRoundedRect:NSMakeRect( left, image.size.height-top-height, width, height ) xRadius:corner yRadius:corner];
						} else {
							path = [NSBezierPath bezierPathWithRect:NSMakeRect( left, image.size.height-top-height, width, height )];
						}
						break;
					}
					case kDrawImageActionOval: {
						path = [NSBezierPath bezierPathWithOvalInRect:NSMakeRect( left, image.size.height-top-height, width, height )];
						break;
					}
				}
				[mFillColor set];
				[path fill];
				if ( thickness > 0 ) {
					[path setLineWidth:thickness];
					[path setLineCapStyle:NSRoundLineCapStyle];
					[mStrokeColor set];
					[path stroke];
				}

				[result unlockFocus];
				mTarget.WriteClipboardDataSet( XClipboardDataSet( result ) );
			}
		} else {
			XErrorHandlerMaybeLogMessage("always", CFSTR("Draw Image cannot get source image") );
		}
	}
	
	return 10;
}


#endif

#if KMF_EDITOR_WINDOW

Class
ADrawImage::ActionController() const
{
	return [ADrawImageEditor class];
}

NSString*
ADrawImage::GetActionHTMLDescriptionString() const
{
	return [NSString stringWithFormat:[@"kmDrawImageActionHTML" localized], mTarget.GetDisplayName()];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {{ what }}<br />
 {%if fill %}<span class="IconLabel">Fill</span> <img src={{ fill | inlineimage: 16 }} alt="Swatch" width="16" height="16" /> {% /if %}<span class="IconLabel">Stroke</span> <img src={{ stroke | inlineimage: 16 }} alt="Swatch" width="16" height="16" />
)RAW";

NSString*
ADrawImage::GetActionExtraHTML() const
{
	NSString* what;
	switch ( mDrawAction ) {
		case kDrawImageActionLine:
			what = [NSString stringWithFormat:[@"kmDrawImageLineHTML" localized], mLeft.Get(), mTop.Get(), mWidth.Get(), mHeight.Get(), mThickness.Get()];
			break;
		case kDrawImageActionRectangle:
			what = [NSString stringWithFormat:[@"kmDrawImageRectangleHTML" localized], mLeft.Get(), mTop.Get(), mWidth.Get(), mHeight.Get(), mThickness.Get(), mCorner.Get()];
			break;
		case kDrawImageActionOval:
			what = [NSString stringWithFormat:[@"kmDrawImageOvalHTML" localized], mLeft.Get(), mTop.Get(), mWidth.Get(), mHeight.Get(), mThickness.Get()];
			break;
	}
	id fill = (mDrawAction == kDrawImageActionLine) ? [NSNull null] : [NSImage imageWithSwatch:mFillColor];
	NSImage* stroke = [NSImage imageWithSwatch:mStrokeColor];
	NSDictionary *variables = @{
							@"what" : what,
							@"fill" : fill,
							@"stroke" : stroke,
						};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kDrawImageActionTypeNames[] = kDrawImageActionTypeNamesArray;

void
ADrawImage::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("DrawAction"), kDrawImageActionTypeNames[mDrawAction] );
	mTarget.SaveToDictionary( dictionary, CFSTR("Target") );
	dictionary->Set( CFSTR("Left"), CFStringRef(mLeft) );
	dictionary->Set( CFSTR("Top"), CFStringRef(mTop) );
	dictionary->Set( CFSTR("Width"), CFStringRef(mWidth) );
	dictionary->Set( CFSTR("Height"), CFStringRef(mHeight) );
	[mFillColor saveWithKey:@"Fill" withAlphaToDictionary:dictionary];
	[mStrokeColor saveWithKey:@"Stroke" withAlphaToDictionary:dictionary];
	dictionary->Set( CFSTR("Thickness"), CFStringRef(mThickness) );
	dictionary->Set( CFSTR("Corner"), CFStringRef(mCorner) );
}

void
ADrawImage::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	SInt32 v = dictionary->GetEnumerated( CFSTR("DrawAction"), kDrawImageActionTypeNames );
	XAssert( v >= 0 );
	mDrawAction = (DrawImageActionType) v;

	mTarget.InitializeFromDictionary( dictionary, CFSTR("Target") );
	CFStringRef value;
	dictionary->Get( CFSTR("Left"), value, kDefaultLeft );
	check( value );
	mLeft = value->Copy();
	
	dictionary->Get( CFSTR("Top"), value, kDefaultTop );
	check( value );
	mTop = value->Copy();
	
	dictionary->Get( CFSTR("Width"), value, kDefaultWidth );
	check( value );
	mWidth = value->Copy();
	
	dictionary->Get( CFSTR("Height"), value, kDefaultHeight );
	check( value );
	mHeight = value->Copy();
	
	mFillColor = [[NSColor colorWithKey:@"Fill" fromDictionary:dictionary defaultColor:kDefaultFillColor defaultAlpha:kDefaultFillAlpha] copy];

	mStrokeColor = [[NSColor colorWithKey:@"Stroke" fromDictionary:dictionary defaultColor:kDefaultStrokeColor defaultAlpha:kDefaultStrokeAlpha] copy];

	dictionary->Get( CFSTR("Thickness"), value, kDefaultThickness );
	check( value );
	mThickness = value->Copy();
	dictionary->Get( CFSTR("Corner"), value, kDefaultCorner );
	check( value );
	mCorner = value->Copy();
}
