//
//  AExecuteBaseScript.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 30 Apr 2010.
//  Copyright (c) 2010 Stairways Software. All rights reserved.
//

#import "AExecuteBaseScript.h"
#import "FileReference.h"
#import "XErrorHandler.h"
#import "NSStringExtensions.h"
#import "KMNotifications.h"
#import "NSStringExtensions.h"

#if KMF_IS_ENGINE
#import "TempScriptHolder.h"
#import "InsertText.h"
#endif
#if KMF_EDITOR_WINDOW
#import "TemplateEngine.h"
#endif

static const CFStringRef MTextResultsWindowKindNames[] = MTextResultsWindowKindNamesArray;

static const MTextResultsWindowKind kDefaultDisplayKind = MTextResultsWindowKind::kWindow;

CFStringRef
AExecuteBaseScript::CopyActionDescriptionString() const
{
	CFStringRef result;
	if ( mUseScriptText ) {
		CFContainer<CFStringRef> theFormatStr = CFSTR("kmExecuteBaseScriptTitle")->CopyLocalized();
		result = CFString::CreateWithFormat( theFormatStr, CFBridgingToCFString(GetScriptTypeName()) );
	} else {
		CFContainer<CFStringRef> theNameStr = (CFStringRef)CFBridgingRetain(FileReference::PathToName( mScriptPath ));
		CFContainer<CFStringRef> theFormatStr = CFSTR("kmExecuteBaseScriptTitleFormat")->CopyLocalized();
		result = CFString::CreateWithFormat( theFormatStr, CFStringRef(theNameStr), CFBridgingToCFString(GetScriptTypeName()) );
	}
	return result;
}

XIcon
AExecuteBaseScript::GetActionDescriptionXIcon() const
{
	XIcon result;

	NSString* path = [FileReference::PathToActualPath( mScriptPath ) existsPath];
	if ( path ) {
		result = XIcon( path, KMIconRegistry::kGenericDocumentIcon );
	} else {
		result = GetActionXIcon();
	}
	check( result.Defined() );
	return result;
}

#if KMF_IS_ENGINE

NSString*
AExecuteBaseScript::GetLongScriptDisplayName() const
{
	if ( mUseScriptText ) {
		return [@"kmExecuteBaseScriptTextLongName" localized];
	} else {
		return [NSString stringWithFormat:[@"kmExecuteBaseScriptFileFormatLongName" localized], GetScriptDisplayName()];
	}
}

NSString*
AExecuteBaseScript::GetShellCommand()
{
	check( !mTempScriptHolder );
	if ( mUseScriptText ) {
		mTempScriptHolder = CreateTempScriptHolderForText( mScriptText );
	} else {
		NSString* scriptPath = [FileReference::PathToActualPath( mScriptPath ) existsPath];
		XAssert( scriptPath );
		mTempScriptHolder = CreateTempScriptHolderForPath( scriptPath );
	}
	XAssert( mTempScriptHolder.Defined() );

	return GetShellCommand( mTempScriptHolder->GetPath() );
}

#endif

#if KMF_EDITOR_WINDOW

NSString*
AExecuteBaseScript::GetActionHTMLDescriptionString() const
{
	return [NSString stringWithFormat:[@"kmExecuteBaseScriptTitle" localized], GetScriptTypeName()];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {% if path %}Path: {{ path | htmlencode }}<br />{% /if %}
 {% if text %}<div class="UserText">{{ text }}</div>{% /if %}
 {{ results | htmlencode }}
)RAW";

NSString*
AExecuteBaseScript::GetActionExtraResults() const
{
	NSString* results = [[NSString stringWithFormat:@"ExecuteBaseScriptResult%@", MTextResultsWindowKindNames[int(mDisplayKind)]] localized];
	if ( mDisplayKind == MTextResultsWindowKind::kSaveToVariable ) {
		results = [NSString stringWithFormat:results, (mTrimResults ? @" trimmed " : @""), mVariable.GetNS()];
	} else if ( mDisplayKind == MTextResultsWindowKind::kSaveToClipboard ) {
		results = [NSString stringWithFormat:results, (mTrimResults ? @" trimmed " : @""), mClipboardTarget.GetDisplayName()];
	}
	return results;
}

NSString*
AExecuteBaseScript::GetActionExtraHTML() const
{
	id path = [NSNull null];
	id text = [NSNull null];
	if ( mUseScriptText ) {
		text = [mScriptText HTMLString];
	} else {
		path = mScriptPath;
	}
	NSDictionary *variables = @{
							@"path" : path,
							@"text" : text,
							@"results" : GetActionExtraResults(),
						};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

void
AExecuteBaseScript::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	FileReference::SavePathToDictionary( dictionary, mScriptPath, CFSTR("Path") );
	dictionary->Set( CFSTR("Text"), CFBridgingToCFString(mScriptText) );
	dictionary->SetBoolean( CFSTR("UseText"), mUseScriptText );
	dictionary->Set( CFSTR("DisplayKind"), MTextResultsWindowKindNames[int(mDisplayKind)] );
	dictionary->SetBoolean( CFSTR("TrimResults"), mTrimResults ); // depricated
	dictionary->SetBoolean( CFSTR("TrimResultsNew"), mTrimResults );
	if ( mDisplayKind == MTextResultsWindowKind::kSaveToVariable ) {
		dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	}
	if ( mDisplayKind == MTextResultsWindowKind::kSaveToClipboard ) {
		mClipboardTarget.SaveToDictionary( dictionary, CFSTR("ClipboardTarget") );
	}
}

void
AExecuteBaseScript::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mScriptPath = FileReference::LoadPathFromDictionary( dictionary, CFSTR("NewFile"), CFSTR("Path") );
	CFStringRef text;
	dictionary->Get( CFSTR("Text"), text, CFSTR("") );
	mScriptText = [CFBridgingToNS(text) copy];
	dictionary->GetBoolean( CFSTR("UseText"), mUseScriptText, false );
	mDisplayKind = dictionary->GetEnumClass( CFSTR("DisplayKind"), MTextResultsWindowKindNames, mDefaultDisplayKind );
	CFStringRef variable;
	dictionary->Get( CFSTR("Variable"), variable, CFSTR("Variable") );
	check( variable );
	mVariable = variable->Copy();
	if ( !dictionary->GetBoolean( CFSTR("TrimResultsNew"), mTrimResults ) ) {
		if ( mDisplayKind == MTextResultsWindowKind::kSaveToVariable ) {
			dictionary->GetBoolean( CFSTR("TrimResults"), mTrimResults, true );
		} else {
			mTrimResults = false;
		}
	}
	mClipboardTarget.InitializeFromDictionary( dictionary, CFSTR("ClipboardTarget") );
}
