//
//  AExecuteRootScript.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 30 Apr 2010.
//  Copyright (c) 2010 Stairways Software. All rights reserved.
//

#import "AExecuteRootScript.h"
#import "FileReference.h"
#import "XErrorHandler.h"
#import "NSStringExtensions.h"
#import "KMNotifications.h"
#import "NSStringExtensions.h"

#if KMF_IS_ENGINE
#import "TempScriptHolder.h"
#import "InsertText.h"
#endif
#if KMF_EDITOR_WINDOW
#import "TemplateEngine.h"
#endif

AExecuteRootScript::AExecuteRootScript( const AExecuteRootScript& inAction )
	:	HBaseAction( inAction )
#if KMF_IS_ENGINE
		, mTask(nil),
		mInsertText( NULL )
#endif
{
}

AExecuteRootScript::~AExecuteRootScript()
{
#if KMF_IS_ENGINE
	if ( mTask ) {
		mTask.receiver = NULL;
		[mTask terminate];
	}
	delete mInsertText;
#endif
}

#if KMF_IS_ENGINE

float
AExecuteRootScript::DoAction()
{
	float time = 0;
	
	try {
		NSString* theCommandStr = [NSString stringWithFormat:@"%@ 2>&1", GetShellCommand()];
		XAssert( theCommandStr );

		mTask = [[AsyncNSTask alloc] initWithReceiver:this];
		mTask.mergeStdErr = YES;
		mTask.executionParameters = GetExecutionParameters();
		XErrorHandlerMaybeLogFormat1( "XTask", CFSTR( "Execute Shell Command “%@”" ), theCommandStr );

		[mTask launchWithCommandAndArguments:@"/bin/sh", @"-c", theCommandStr, nil];
		time = 10;
	} catch( ... ) {
		CFContainer<CFStringRef> exceptionStr = CFString::CreateWithFormat( CFSTR("Execute %@ Exception"), CFBridgingToCFString(GetScriptTypeName()) );
		mTheResultData = CFData::Create( exceptionStr );
		XErrorBeep( exceptionStr );
		if ( mTask ) {
			mTask.receiver = NULL;
			[mTask terminate];
			mTask = nil;
		}
	}
	
	return time;
}

Boolean
AExecuteRootScript::IsActionCompleted()
{
	if ( mTheResultData.Defined() ) {
		ProcessResults();
		mTheResultData = NULL;
	}
	if ( mInsertText ) {
		if ( mInsertText->IsActionCompleted() ) {
			delete mInsertText;
			mInsertText = NULL;
		}
	}	
	return !mInsertText && !mTask;
}

void
AExecuteRootScript::ActionAborted()
{
	mTask.receiver = NULL;
	[mTask terminate];
	mTask = nil;

	delete mInsertText;
	mInsertText = NULL;
}

void
AExecuteRootScript::FinishedWithData( AsyncNSTask* task, NSData* data, NSData* error )
{
	mTheResultData = ((CFDataRef)CFBridgingToCF(data))->Copy();
	mTask.receiver = NULL;
	mTask = nil;
}

void
AExecuteRootScript::DisplayResults( NSString* results )
{
	if ( !IsEmpty( results ) ) {
		CFContainer<CFStringRef> description;
		CFContainer<CFStringRef> format = CFSTR("kmExecuteRootScriptResultsDescription")->CopyLocalized();
		description = CFString::CreateWithFormat( format, GetLongScriptDisplayName() );

		switch ( DisplayKind() ) {
			case MTextResultsWindowKind::kInsertByTyping:
			case MTextResultsWindowKind::kInsertByPasting: {
				mInsertText = new InsertText( DisplayKind() == MTextResultsWindowKind::kInsertByPasting, results );
				mInsertText->DoAction();
				break;
			}
			case MTextResultsWindowKind::kSaveToVariable: {
				KMVariables::Set( SaveToVariable(), results, KMNotifications::kLoadVariableFromEngineAction );
				break;
			}
			case MTextResultsWindowKind::kSaveToClipboard: {
				WriteClipboardDataSet( XClipboardDataSet( results ) );
				break;
			}
			case MTextResultsWindowKind::kNone:
				break;
			case MTextResultsWindowKind::kWindow:
			case MTextResultsWindowKind::kBriefly: {
				NSString* title = [NSString stringWithFormat:[@"kmExecuteBaseScriptResultsTitle" localized], GetScriptTypeName()];
				MTextResultsWindow::Display( DisplayKind(), title, description->GetNS(), results );
				break;
			}
		}

	} else if ( DisplayKind() == MTextResultsWindowKind::kSaveToVariable ) {
		KMVariables::Set( SaveToVariable(), @"", KMNotifications::kLoadVariableFromEngineAction );
	} else if ( DisplayKind() == MTextResultsWindowKind::kSaveToClipboard ) {
		WriteClipboardDataSet( XClipboardDataSet( @"" ) );
	}
}

void
AExecuteRootScript::ProcessResults()
{
	if ( StripTrailingLineFeed() && mTheResultData.Defined() && !mTheResultData->IsEmpty() ) {
		NSInteger len = mTheResultData->GetLength();
		check( len > 0 );
		UInt8 eol;
		mTheResultData->GetBytes( len-1, 1, &eol );
		if ( eol == kLineFeedCharCode ) {
			mTheResultData = CFData::Create( mTheResultData->GetBytePtr(), len-1 );
		}
	}

	CFContainer<CFStringRef> results = CFString::Create( mTheResultData, kCFStringEncodingUTF8 );
	if ( !results.Defined() ) {
		results = CFString::Create( mTheResultData, kCFStringEncodingMacRoman );
	}
	check( results.Defined() );
	if ( results.Defined() && TrimResults() ) {
		CFContainer<CFMutableStringRef> temp = CFString::CreateMutable( results );
		temp->TrimWhitespace();
		results = temp->Copy();
	}
	DisplayResults( results.Defined() ? results->GetNS() : nil );
}

#endif
