//
//  AFile.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Tue 3 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AFile.h"
#import "KMIconRegistry.h"
#import "KMPreferences.h"
#import "FileReference.h"
#import "NSObjectExtensions.h"
#import "NSStringExtensions.h"
#import "XErrorHandler.h"
#import "NSStringRegularExpressions.h"
#if KMF_IS_ENGINE
#import "ThreadFinished.h"
#endif
#if KMF_EDITOR_WINDOW
#import "AFileEditor.h"
#import "TemplateEngine.h"
#endif

AFile::AFile()
	:	HBaseAction(),
		mOperation( kFileOperationDEFAULT ),
		mSource( CFSTR("") ),
		mDestination( CFSTR("") )
#if KMF_IS_ENGINE
		, mThreadToken( ThreadFinished::kNull )
#endif
{
}

AFile::AFile( const AFile& inAction )
	:	HBaseAction( inAction ),
		mOperation( inAction.mOperation ),
		mSource( inAction.mSource ),
		mDestination( inAction.mDestination )
#if KMF_IS_ENGINE
		, mThreadToken( ThreadFinished::kNull )
#endif
{
}

AFile::AFile( const FileOperationType& inOperation )
	:	HBaseAction(),
		mOperation( inOperation ),
		mSource( CFSTR("") ),
		mDestination( CFSTR("") )
#if KMF_IS_ENGINE
		, mThreadToken( ThreadFinished::kNull )
#endif
{
}

AFile::~AFile()
{
#if KMF_IS_ENGINE
	ThreadFinished::AbandonToken( mThreadToken );
#endif
}

CFStringRef
AFile::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> operation;
	switch ( mOperation ) {
		case kFileOperationMove:
			operation = CFSTR("kmFileMoveOperation")->CopyLocalized();
			break;
		case kFileOperationCopy:
			operation = CFSTR("kmFileCopyOperation")->CopyLocalized();
			break;
		case kFileOperationDuplicate:
			operation = CFSTR("kmFileDuplicateOperation")->CopyLocalized();
			break;
		case kFileOperationTrash:
			operation = CFSTR("kmFileTrashOperation")->CopyLocalized();
			break;
		case kFileOperationDelete:
			operation = CFSTR("kmFileDeleteOperation")->CopyLocalized();
			break;
		case kFileOperationRecursiveDelete:
			operation = CFSTR("kmFileRecursiveDeleteOperation")->CopyLocalized();
			break;
	}
	NSString* path = mSource->GetNS();
	CFContainer<CFStringRef> theFormatStr;
	if ( IsEmpty( path ) ) {
		theFormatStr = CFSTR("kmFileTitleFormat")->CopyLocalized();
	} else {
		theFormatStr = CFSTR("kmFileTitleLongFormat")->CopyLocalized();
	}
	return CFString::CreateWithFormat( theFormatStr, CFStringRef(operation), path );
}

XIcon
AFile::GetActionDescriptionXIcon() const
{
	XIcon result;
	
	NSString* path = [FileReference::PathToActualPath( mSource->GetNS() ) existsPath];
	if ( path ) {
		result = XIcon( path, KMIconRegistry::kGenericDocumentIcon );
	} else {
		result = GetActionXIcon();
	}
	check( result.Defined() );
	return result;
}

#if KMF_IS_ENGINE

float
AFile::DoAction()
{
	NSFileManager* fileManager = [[NSFileManager alloc] init];
	bool good = true;
	FileOperationType operation = mOperation;
	NSString* source = FileReference::PathToActualPath( mSource->GetNS() );
	NSString* destination = nil;
	if ( UseDestination() ) {
		destination = FileReference::PathToActualPath( mDestination->GetNS(), [source stringByDeletingLastPathComponent] );
		BOOL dir;
		if ( [fileManager fileExistsAtPath:destination isDirectory:&dir] ) {
			if ( dir ) {
				destination = [destination stringByAppendingPathComponent:[source lastPathComponent]];
				good = ![fileManager fileExistsAtPath:destination];
			} else {
				good = false;
			}
		} else {
			good = true;
		}
	}
	destination = [destination stringByStandardizingPath];
	if ( mOperation == kFileOperationDelete ) {
		BOOL dir;
		good = [fileManager fileExistsAtPath:source isDirectory:&dir] && !dir;
	}
	
	good = good && [source isMatchedByRegex:@"^/.+"] && (!UseDestination() || [destination isMatchedByRegex:@"^/.+"]); 

	mThreadToken = ThreadFinished::Get();
	ThreadFinished::Token token = mThreadToken;
	if ( good ) {
		dispatch_async( dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
			switch ( operation ) {
				case kFileOperationMove: {
					NSError* error = nil;
					BOOL result = [fileManager moveItemAtPath:source toPath:destination error:&error];
					if ( !result ) {
						XErrorHandlerMaybeLogFormat2( "display", CFSTR("Move file %@ failed with %@"), source, error );
					}
					ThreadFinished::Finished( token );
					break;
				}
				case kFileOperationCopy:{
					NSError* error = nil;
					BOOL result = [fileManager copyItemAtPath:source toPath:destination error:&error];
					if ( !result ) {
						XErrorHandlerMaybeLogFormat2( "display", CFSTR("Copy file %@ failed with %@"), source, error );
					}
					ThreadFinished::Finished( token );
					break;
				}
				case kFileOperationDuplicate: {
					dispatch_async( dispatch_get_main_queue(), ^{
						[[NSWorkspace sharedWorkspace] duplicateURLs:[NSArray arrayWithObject:[NSURL fileURLWithPath:source]] completionHandler:^(NSDictionary *newURLs, NSError *error) 
						 {
							if ( error != nil ) {
								XErrorHandlerMaybeLogFormat2( "display", CFSTR("Duplicate file %@ failed with %@"), source, error );
							}
							ThreadFinished::Finished( token );
						 }];
					} );
					break;
				}
				case kFileOperationTrash: {
					dispatch_async( dispatch_get_main_queue(), ^{
						[[NSWorkspace sharedWorkspace] recycleURLs:[NSArray arrayWithObject:[NSURL fileURLWithPath:source]] completionHandler:^(NSDictionary *newURLs, NSError *error) 
						 {
							if ( error != nil ) {
								XErrorHandlerMaybeLogFormat2( "display", CFSTR("Duplicate file %@ failed with %@"), source, error );
							}
							ThreadFinished::Finished( token );
						 }];
					} );
					break;
				}
				case kFileOperationDelete:
				case kFileOperationRecursiveDelete: {
					NSError* error = nil;
					BOOL result = [fileManager removeItemAtPath:source error:&error];
					if ( !result ) {
						XErrorHandlerMaybeLogFormat2( "display", CFSTR("Delete file %@ failed with %@"), source, error );
					}
					ThreadFinished::Finished( token );
					break;
				}
			}
		} );
	} else {
		ThreadFinished::Finished( token );
	}
	
	return 10;
}

Boolean
AFile::IsActionCompleted()
{
	return ThreadFinished::IsFinished( mThreadToken );
}

void
AFile::ActionAborted()
{
	ThreadFinished::AbandonToken( mThreadToken );
	mThreadToken = ThreadFinished::kNull;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AFile::ActionController() const
{
	return [AFileEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {% if usedestination %}To: {{ destination | htmlencode }}<br />{% /if %}
 {% if warning %}<strong>Warning: This can be very destructive. Use caution.</strong><br />{% /if %}
)RAW";

NSString*
AFile::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"destination" : mDestination->GetNS(),
									@"usedestination" : UseDestination() ? @YES : @NO,
									@"warning" : (mOperation == kFileOperationRecursiveDelete) ? @YES : @NO,
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kFileOperationTypeNames[] = kFileOperationTypeNamesArray;

void
AFile::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Operation"), kFileOperationTypeNames[mOperation] );
	FileReference::SavePathToDictionary( dictionary, mSource->GetNS(), CFSTR("Source") );
	FileReference::SavePathToDictionary( dictionary, mDestination->GetNS(), CFSTR("Destination") );
}

void
AFile::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	SInt32 v = dictionary->GetEnumerated( CFSTR("Operation"), kFileOperationTypeNames );
	XAssert( v >= 0 );
	mOperation = (FileOperationType) v;

	FileReference::LoadPathFromDictionary( dictionary, mSource, NULL, CFSTR("Source") );
	FileReference::LoadPathFromDictionary( dictionary, mDestination, NULL, CFSTR("Destination") );
}
