//
//  AFindImage.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 3 May 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AFindImage.h"
#import "NSImageExtensions.h"

#if KMF_EDITOR_WINDOW
#import "AFindImageEditor.h"
#import "TemplateEngine.h"
#import "NSStringExtensions.h"
#endif
#if KMF_IS_ENGINE
#import "KMNotifications.h"
#import "FindImageWindowController.h"
#endif


AFindImage::AFindImage( const AFindImage& inAction )
	:	HBaseAction( inAction ),
		mImage( [inAction.mImage copy] ),
		mFuzz( inAction.mFuzz ),
		mUnique( inAction.mUnique ),
		mVariable( inAction.mVariable ),
		mDisplayMatches( inAction.mDisplayMatches )
{
#if KMF_IS_ENGINE
	mThreadToken = ThreadFinished::kNull;
#endif
}
	
AFindImage::~AFindImage()
{
#if KMF_IS_ENGINE
	if ( mThreadToken != ThreadFinished::kNull ) {
		ThreadFinished::AbandonToken( mThreadToken );
	}
#endif
}

void AFindImage::SetImage( NSImage* inImage )
{
	mImage = [inImage copy];
}

#if KMF_IS_ENGINE

void
AFindImage::MatchCompleted( bool match, CGPoint topleft )
{
	NSString* result = @"";
	if ( match ) {
		result = [NSString stringWithFormat:@"%d,%d", (int)roundf(topleft.x), (int)roundf(topleft.y)];
	}
	KMVariables::Set( mVariable->GetNS(), result, KMNotifications::kLoadVariableFromEngineAction );
	ThreadFinished::AbandonToken( mThreadToken );
	mThreadToken = ThreadFinished::kNull;
}

float
AFindImage::DoAction()
{
	if ( mImage ) {
		mThreadToken = ThreadFinished::Get();
		ThreadFinished::Token token = mThreadToken; // private variable, not part of possibly deleted object
		bool displayMatches = mDisplayMatches;
		MatchImage::MatchDisplay( mImage, mFuzz, !mUnique, ^( bool match, bool match_without_duplicates, const MatchImage::Results results ) {
			if ( !ThreadFinished::IsFinished( token ) ) {
				if ( displayMatches ) {
					[FindImageWindowController addMatches:results matchWithoutDuplicates:match_without_duplicates requiresUniqueMatch:mUnique];
				}
				CGPoint topleft = results.size() > 0 ? results[0].frame.origin : CGPointMake(0,0);
				MatchCompleted( match, topleft );
			}
		});
	} else {
		KMVariables::Set( mVariable->GetNS(), @"", KMNotifications::kLoadVariableFromEngineAction );
	}
	return 10;
}

Boolean
AFindImage::IsActionCompleted()
{
	return mThreadToken == ThreadFinished::kNull;
}

void
AFindImage::ActionAborted()
{
	if ( mThreadToken != ThreadFinished::kNull ) {
		ThreadFinished::AbandonToken( mThreadToken );
		mThreadToken = ThreadFinished::kNull;
	}
}

#endif

#if KMF_EDITOR_WINDOW

Class
AFindImage::ActionController() const
{
	return [AFindImageEditor class];
}

NSString*
AFindImage::GetActionHTMLDescriptionString() const
{
	return [NSString stringWithFormat:[@"kmFindImageHTML" localized], (mUnique ? @" Unique" : @""), mVariable.GetNS()];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {% if image %}<img src={{ image | inlineimage: -1 }} alt="Search Image" width="{{ image | imagewidth }}" height="{{ image | imageheight }}" /> Fuzz: {{ fuzz }}%<br />{% /if %}
)RAW";

NSString*
AFindImage::GetActionExtraHTML() const
{
	NSImage* image = mImage ? [mImage imageForHTMLDisplay] : (id)[NSNull null];
	NSDictionary *variables = @{
							@"fuzz" : @(mFuzz),
							@"image" : image,
						};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

void
AFindImage::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	[mImage saveToDictionary:dictionary];
	dictionary->SetNumber( CFSTR("Fuzz"), mFuzz );
	dictionary->SetBoolean( CFSTR("Unique"), mUnique );
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	dictionary->SetBoolean( CFSTR("DisplayMatches"), mDisplayMatches );
}

void
AFindImage::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mImage = [[NSImage imageFromDictionary:dictionary] copy];
	if ( !dictionary->Get( CFSTR("Fuzz"), mFuzz ) ) mFuzz = MatchImage::kDefaultFuzz;
	dictionary->GetBoolean( CFSTR("Unique"), mUnique, kFindImageDefaultUnique );
	CFStringRef value;
	dictionary->Get( CFSTR("Variable"), value, CFSTR("Variable") );
	check( value );
	mVariable = value->Copy();
	dictionary->GetBoolean( CFSTR("DisplayMatches"), mDisplayMatches, false );
}
