//
//  AFor.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 18 Mar 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "AFor.h"
#import "MMacro.h"
#import "NSStringExtensions.h"
#if KMF_EDITOR_WINDOW
#import "AForEditor.h"
#import "TemplateEngine.h"
#import "MMath.h"
#else
#import "KMPreferences.h"
#endif
#if KMF_IS_ENGINE
#import "HExecuteActionListManager.h"
#import "KMNotifications.h"
#endif

AFor::AFor()
	:	HBaseAction(),
		mVariable( CFSTR("Variable")->Copy() )
{
}

AFor::AFor( const AFor& inAction )
	:	HBaseAction( inAction ),
		mVariable( inAction.mVariable ),
		mCollections( inAction.mCollections ),
		mActionList( inAction.mActionList )
{
}

AFor::~AFor()
{
#if KMF_IS_ENGINE
	HExecuteActionListManager::CancelExecution( mTrackingID );
#endif
}

CFStringRef
AFor::CopyActionDescriptionString() const
{
	return CFSTR("kmForTitle")->CopyLocalized();
}

#if KMF_IS_ENGINE
float
AFor::DoAction()
{
	mTrackingID = [NSUUID UUID];
	mExecuting = false;
	mCollections.CollectionTestStart();
	return 10;
}

Boolean
AFor::IsActionCompleted()
{
	if ( mExecuting && !HExecuteActionListManager::IsExecuting( mTrackingID ) ) {
		mExecuting = false;
	}
	if ( !mExecuting ) {
		NSString* next = mCollections.CollectionGetNext();
		if ( next ) {
			KMVariables::Set( mVariable->GetNS(), next, KMNotifications::kLoadVariableFromEngineAction );
			HExecuteActionListManager::Execute( ExecuteActionListDisplay::kHide, @"Hide", mActionList, mTrackingID );
			mExecuting = true;
		}
	}
	
	return !mExecuting;
}

void
AFor::ActionAborted()
{
	HExecuteActionListManager::CancelExecution( mTrackingID );
	mCollections.CollectionTestAbort();
}

#endif

#if KMF_EDITOR_WINDOW

Class
AFor::ActionController() const
{
	return [AForEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
<ul class="IconList">
	{% for collection in collections %}
		<li><img src={{ collection.icon | inlineimage: 16 }} alt="collection icon" width="16" height="16" /><span class="IconTitle">{{ collection.description | htmlencode }}</span></li>
	{% /for %}
</ul>
{% if actions.@count == 0 %}
	<h3>Do nothing.</h3>
{% else %}
	<h3>Execute the Following Actions:</h3>
	<ul class="IconList">
		{% for action in actions %}
			<li>{{ action }}</li>
		{% /for %}
	</ul>
{% /if %}
)RAW";

NSString*
AFor::GetActionHTMLDescriptionString() const
{
	int c = MMath::min( 2, mCollections.GetCount() );
	return [[NSString stringWithFormat:@"kmForHTMLTitle%d", c] localized];
}

NSString*
AFor::GetActionExtraHTML() const
{
	if ( mCollections.GetCount() == 0 ) return @"";

	NSDictionary *variables = @{
									@"collections" : mCollections.GetCollectionDisplayContent(),
									@"actions" : mActionList.GetDisplayContent(),
								};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

BOOL
AFor::MatchesTextSearch( NSString* searchString ) const
{
	if ( HBaseAction::MatchesTextSearch( searchString ) ) return YES;
	if ( mCollections.MatchesTextSearch( searchString ) ) return YES;
	return mActionList.MatchesTextSearch( searchString );
}

#pragma mark streaming

void
AFor::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	SafeCFPropertyListRef collections = (SafeCFPropertyListRef)mCollections.CopyToPropertyList();
	dictionary->Set( CFSTR("Collections"), collections );
	collections->Release();
	SafeCFPropertyListRef actions = mActionList.CopyActionsToPropertyList();
	dictionary->Set( CFSTR("Actions"), actions );
	CFRelease( actions );
}

void
AFor::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef value;
	dictionary->Get( CFSTR("Variable"), value, CFSTR("Variable") );
	check( value );
	mVariable = value->Copy();
	SafeCFPropertyListRef collections;
	XAssert( dictionary->Get( CFSTR("Collections"), collections ) );
	mCollections.LoadFromPropertyList( collections );
	SafeCFPropertyListRef actions = (SafeCFPropertyListRef)dictionary->Get( CFSTR("Actions") );
	XAssert( actions );
	mActionList.LoadActionsFromPropertyList( actions );
}

