//
//  AGetFileAttribute.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Wed 11 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AGetFileAttribute.h"
#import "FileReference.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "XErrorHandler.h"
#import "NSStringExtensions.h"
#import "NSStringRegularExpressions.h"

#if KMF_IS_ENGINE
#import "KMVariables.h"
#import "KMNotifications.h"
#endif

#if KMF_EDITOR_WINDOW
#import "AGetFileAttributeEditor.h"
#import "TemplateEngine.h"
#endif

static const CFStringRef kFileAttributeTypeNames[] = kFileAttributeTypeNamesArray;

AGetFileAttribute::AGetFileAttribute()
	:	HBaseAction(),
		mAttribute( FileAttributeType::kDEFAULT ),
		mPath( CFSTR("") ),
		mVariable( CFSTR("Variable")->Copy() )
{
}

AGetFileAttribute::AGetFileAttribute( const AGetFileAttribute& inAction )
	:	HBaseAction( inAction ),
		mAttribute( inAction.mAttribute ),
		mPath( inAction.mPath ),
		mVariable( inAction.mVariable )
{
}

AGetFileAttribute::~AGetFileAttribute()
{
}

CFStringRef
AGetFileAttribute::CopyAttributeName( FileAttributeType inAttribute )
{
	CFContainer<CFStringRef> format = CFSTR("kFileAttribute%@");
	CFContainer<CFStringRef> key = CFString::CreateWithFormat( format, kFileAttributeTypeNames[int(inAttribute)] );
	return key->CopyLocalized();
}

CFStringRef
AGetFileAttribute::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> format = CFSTR("kmGetFileAttributeTitleFormat")->CopyLocalized();
	CFContainer<CFStringRef> attribute = CopyAttributeName( mAttribute );
	return CFString::CreateWithFormat( format, CFStringRef(attribute), CFStringRef(mVariable) );
}

static NSString* kFileAttributeNSAttributeKeyArray[] = kFileAttributeNSAttributeKeyNamesArray;
static NSDictionary* kFileTypeMap = [[NSDictionary alloc] initWithObjectsAndKeys:
									 @"Directory", NSFileTypeDirectory,
									 @"File", NSFileTypeRegular,
									 @"Symlink", NSFileTypeSymbolicLink,
									 @"Socket", NSFileTypeSocket,
									 @"Character Special", NSFileTypeCharacterSpecial,
									 @"Block Special", NSFileTypeBlockSpecial,
									 @"Unknown", NSFileTypeUnknown,
									 nil];

CFStringRef
AGetFileAttribute::CopyTransformationResult( NSUInteger* errorPos, float* outTime ) const
{
	NSString* result = nil;
	NSString* path = FileReference::PathToActualPath( mPath->GetNS() );
	NSError* error;
	NSDictionary* attributes = nil;
	if ( [path isMatchedByRegex:@"/.+"] ) {
		switch ( mAttribute ) {
			case FileAttributeType::kFileType:
			case FileAttributeType::kFileSize:
			case FileAttributeType::kCreationDate:
			case FileAttributeType::kModificationDate:
			case FileAttributeType::kOwnerID:
			case FileAttributeType::kOwnerName:
			case FileAttributeType::kGroupName:
			case FileAttributeType::kGroupID:
			case FileAttributeType::kPosixPermissions:
			case FileAttributeType::kExtensionHidden:
			case FileAttributeType::kHFSCreatorCode:
			case FileAttributeType::kHFSTypeCode:
				attributes = [[NSFileManager defaultManager] attributesOfItemAtPath:path error:&error];
				if ( attributes ) {
					id value = [attributes objectForKey:kFileAttributeNSAttributeKeyArray[int(mAttribute)]];
					if ( value ) {
						switch ( mAttribute ) {
							case FileAttributeType::kFileType: {
								// special
								result = [kFileTypeMap objectForKey:value];
								if ( [result isEqual:NSFileTypeDirectory] && [path fileKind] == PathFileKind::kPackage ) {
									result = @"Package";
								}
								break;
							}
							case FileAttributeType::kCreationDate:
							case FileAttributeType::kModificationDate: {
								// date
								NSNumber* n = [NSNumber numberWithInteger:[value timeIntervalSince1970]];
								result = [n stringValue];
								break;
							}
							case FileAttributeType::kFileSize:
							case FileAttributeType::kOwnerID:
							case FileAttributeType::kGroupID:
							case FileAttributeType::kPosixPermissions:
							case FileAttributeType::kExtensionHidden:
								// number
								result = [value stringValue];
								break;
							case FileAttributeType::kOwnerName:
							case FileAttributeType::kGroupName:
								// string
								result = value;
								break;
							case FileAttributeType::kHFSCreatorCode:
							case FileAttributeType::kHFSTypeCode: {
								// type code
								OSType t = [value unsignedLongValue];
								result = [NSString stringWithOSType:t];
								if ( ![result isMatchedByRegex:@"^[ -~][ -~][ -~][ -~]$"] ) {
									result = [NSString stringWithFormat:@"0x%08lX", (unsigned long)t];
								}
								break;
							}
							case FileAttributeType::kParent:
							case FileAttributeType::kName:
							case FileAttributeType::kBaseName:
							case FileAttributeType::kExtension:
								check ( false );
								break;
						}
					}
				}
				break;
			case FileAttributeType::kParent:
				result = [path stringByDeletingLastPathComponent];
				break;
			case FileAttributeType::kName:
				result = [path lastPathComponent];
				break;
			case FileAttributeType::kBaseName:
				result = [[path lastPathComponent] stringByDeletingPathExtension];
				break;
			case FileAttributeType::kExtension:
				result = [path pathExtension];
				break;
		}
	}
	if ( IsEmpty( result ) ) {
		result = @"";
	}
	if ( errorPos ) *errorPos = 0;
	if ( outTime ) *outTime = 5;
	return (CFStringRef)CFBridgingRetain(result);
}


#if KMF_IS_ENGINE

float
AGetFileAttribute::DoAction()
{
	float time;
	NSString* result = GetTransformationResult( NULL, &time );
	KMVariables::Set( mVariable->GetNS(), result, KMNotifications::kLoadVariableFromEngineAction );
	
	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AGetFileAttribute::ActionController() const
{
	return [AGetFileAttributeEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
From file: {{ file | htmlencode }}
)RAW";

NSString*
AGetFileAttribute::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"file" : mPath->GetNS(),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

void
AGetFileAttribute::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Attribute"), kFileAttributeTypeNames[int(mAttribute)] );
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	dictionary->Set( CFSTR("Path"), CFStringRef(mPath) );
}

void
AGetFileAttribute::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	bool good;
	good = dictionary->CanGetEnumClass( CFSTR("Attribute"), kFileAttributeTypeNames, mAttribute );
	XAssert( good );
	
	CFStringRef value;
	dictionary->Get( CFSTR("Variable"), value, CFSTR("Variable") );
	check( value );
	mVariable = value->Copy();
	FileReference::LoadPathFromDictionary( dictionary, mPath, NULL, CFSTR("Path") );
}
