//
//  AHiliteLocation.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Mon 7 May 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AHiliteLocation.h"
#import "XException.h"
#import "KMDefines.h"
#import "XErrorHandler.h"
#import "NSStringRegularExpressions.h"
#import "Calculate.h"
#import "TokenText.h"

#import "XAXApplicationElement.h"
#import "XAXWindowElement.h"
#import "XAXWindowElementList.h"
#import "ScreenInfo.h"

#if KMF_EDITOR_WINDOW
#import "AHiliteLocationEditor.h"
#import "TemplateEngine.h"
#endif
#if KMF_IS_ENGINE
#import "HiliteController.h"
#endif

static CFStringRef kDefaultLeft = CFSTR("MOUSEX()");
static CFStringRef kDefaultTop = CFSTR("MOUSEY()");
static CFStringRef kDefaultWidth = CFSTR("100");
static CFStringRef kDefaultHeight = CFSTR("100");
const Boolean kDefaultAnimate = true;
static CFStringRef kDefaultDisplayTime = CFSTR("5");

AHiliteLocation::AHiliteLocation()
	:	HBaseAction(),
		mStyle( kHiliteLocationStyleDEFAULT ),
		mLeft( kDefaultLeft->Copy() ),
		mTop( kDefaultTop->Copy() ),
		mWidth( kDefaultWidth->Copy() ),
		mHeight( kDefaultHeight->Copy() ),
		mAnimate( kDefaultAnimate ),
		mDisplayTime( kDefaultDisplayTime->Copy() )
{
#if KMF_IS_ENGINE
	mHiliteController = nil;
#endif
}

AHiliteLocation::AHiliteLocation( const AHiliteLocation& inAction )
	:	HBaseAction( inAction ),
		mStyle( inAction.mStyle ),
		mLeft( inAction.mLeft ),
		mTop( inAction.mTop ),
		mWidth( inAction.mWidth ),
		mHeight( inAction.mHeight ),
		mAnimate( inAction.mAnimate ),
		mDisplayTime( inAction.mDisplayTime )
{
#if KMF_IS_ENGINE
	mHiliteController = nil;
#endif
}
	
AHiliteLocation::~AHiliteLocation()
{
}

CFStringRef
AHiliteLocation::CopyActionDescriptionString() const
{
	// first, build window name string
	
	CFContainer<CFStringRef> format;
	
	switch ( GetStyle() ) {
		case kHiliteLocationStyleRectangle:
			format = CFSTR("kmHiliteLocationTitleRectangle")->CopyLocalized();
			break;
		case kHiliteLocationStyleOval:
			format = CFSTR("kmHiliteLocationTitleOval")->CopyLocalized();
			break;
		case kHiliteLocationStyleTarget:
			format = CFSTR("kmHiliteLocationTitleTarget")->CopyLocalized();
			break;
	}
	
	return CFString::CreateWithFormat( format, CFStringRef(mLeft), CFStringRef(mTop), CFStringRef(mWidth), CFStringRef(mHeight) );
}

const ArrowString::Style*
AHiliteLocation::GetArrowStyles() const
{
	switch ( mStyle ) {
		case kHiliteLocationStyleRectangle:
			return ArrowString::kRightDownWidthHeight;
		case kHiliteLocationStyleOval:
		case kHiliteLocationStyleTarget:
			return ArrowString::kRightDownToCenterWidthHeight;
	}
}

#if KMF_IS_ENGINE

float
AHiliteLocation::DoAction()
{
	double left, top, width, height, displayTime;
	if ( !Calculate::CalculateValue( mLeft->GetNS(), left ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Hilite Location %@ failed"), mLeft->GetNS() );
		return 0;
	} else if ( !Calculate::CalculateValue( mTop->GetNS(), top ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Hilite Location %@ failed"), mTop->GetNS() );
		return 0;
	} else if ( !Calculate::CalculateValue( mWidth->GetNS(), width ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Hilite Location %@ failed"), mWidth->GetNS() );
		return 0;
	} else if ( !Calculate::CalculateValue( mHeight->GetNS(), height ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Hilite Location %@ failed"), mHeight->GetNS() );
		return 0;
	} else if ( !Calculate::CalculateValue( mDisplayTime->GetNS(), displayTime ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Hilite Location %@ failed"), mDisplayTime->GetNS() );
		return 0;
	}
	if ( displayTime < 0 ) displayTime = 0;
	if ( displayTime > 600 ) displayTime = 600;
	if ( width < 1 ) width = 1;
	if ( height < 1 ) height = 1;
	
	if ( mStyle == kHiliteLocationStyleOval || mStyle == kHiliteLocationStyleTarget ) {
		left = left - width / 2;
		top = top - height / 2;
	}
	left = roundf( left );
	top = roundf( top );
	width = roundf( width );
	height = roundf( height );

	mHiliteController = [HiliteController newHiliteController];
	mHiliteController.style = mStyle;
	mHiliteController.frame = NSMakeRect( left, top, width, height );
	mHiliteController.displayTime = displayTime;
	mHiliteController.animate = mAnimate;
	[mHiliteController Start];
	
	return 10;
}

Boolean
AHiliteLocation::IsActionCompleted()
{
	if ( !mHiliteController ) return true;
	if ( mHiliteController.displaying ) return false;
	mHiliteController = nil;
	return true;
}

void
AHiliteLocation::ActionAborted()
{
	[mHiliteController Close];
	mHiliteController = nil;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AHiliteLocation::ActionController() const
{
	return [AHiliteLocationEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
{% if animate %}Animate and display for{% else %}Display for{% /if %} {{ display | htmlencode }} seconds.
)RAW";

NSString*
AHiliteLocation::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"animate" : @(mAnimate),
									@"display" : mDisplayTime->GetNS(),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kHiliteLocationStyleTypeNames[] = kHiliteLocationStyleTypeNamesArray;

void
AHiliteLocation::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Style"), kHiliteLocationStyleTypeNames[mStyle] );
	dictionary->Set( CFSTR("Left"), CFStringRef(mLeft) );
	dictionary->Set( CFSTR("Top"), CFStringRef(mTop) );
	dictionary->Set( CFSTR("Width"), CFStringRef(mWidth) );
	dictionary->Set( CFSTR("Height"), CFStringRef(mHeight) );
	dictionary->SetBoolean( CFSTR( "Animate" ), mAnimate );
	dictionary->Set( CFSTR("DisplayTime"), CFStringRef(mDisplayTime) );
}

void
AHiliteLocation::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef value;
	SInt32 v;
	
	v = dictionary->GetEnumerated( CFSTR("Style"), kHiliteLocationStyleTypeNames, kHiliteLocationStyleDEFAULT );
	mStyle = (HiliteLocationStyleType) v;
	
	dictionary->Get( CFSTR("Left"), value, kDefaultLeft );
	check( value );
	mLeft = value->Copy();
	
	dictionary->Get( CFSTR("Top"), value, kDefaultTop );
	check( value );
	mTop = value->Copy();
	
	dictionary->Get( CFSTR("Width"), value, kDefaultWidth );
	check( value );
	mWidth = value->Copy();
	
	dictionary->Get( CFSTR("Height"), value, kDefaultHeight );
	check( value );
	mHeight = value->Copy();
	
	dictionary->GetBoolean( CFSTR( "Animate" ), mAnimate, kDefaultAnimate );

	dictionary->Get( CFSTR("DisplayTime"), value, kDefaultDisplayTime );
	check( value );
	mDisplayTime = value->Copy();
}
