//
//  AIfThenElse.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Mon 28 Mar 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "AIfThenElse.h"
#import "NSStringExtensions.h"
#if KMF_EDITOR_WINDOW
#import "AIfThenElseEditor.h"
#import "TemplateEngine.h"
#endif
#if KMF_IS_ENGINE
#import "HExecuteActionListManager.h"
#endif

AIfThenElse::AIfThenElse()
	:	HBaseAction()
{
}

AIfThenElse::AIfThenElse( const AIfThenElse& inAction )
	:	HBaseAction( inAction ),
		mConditions( inAction.mConditions ),
		mThenActionList( inAction.mThenActionList ),
		mElseActionList( inAction.mElseActionList )
{
}

AIfThenElse::~AIfThenElse()
{
#if KMF_IS_ENGINE
	HExecuteActionListManager::CancelExecution( mTrackingID );
#endif
}

CFStringRef
AIfThenElse::CopyActionDescriptionString() const
{
	if ( mConditions.GetCount() == 0 && (mThenActionList.GetCount() != 0 || mElseActionList.GetCount() != 0 ) ) {
		return CFSTR("kmIfThenElseNoConditionsTitle")->CopyLocalized();
	} else {
		CFContainer<CFStringRef> key = CFSTR("kmIfThenElseTitle")->CopyLocalized();
		return CFString::CreateWithFormat( key, mConditions.GetConditionDisplayText() );
	}
}

#if KMF_IS_ENGINE
float
AIfThenElse::DoAction()
{
	mTrackingID = [NSUUID UUID];
	mExecuting = false;
	return 10;
}

Boolean
AIfThenElse::IsActionCompleted()
{
	Boolean result = false;
	if ( mConditions.GetCount() == 0 ) {
		result = true;
	} else if ( !mExecuting ) {
		ConditionTestResult cresult = mConditions.ConditionResult();
		if ( cresult == ConditionTestResult::kTrue ) {
			HExecuteActionListManager::Execute( ExecuteActionListDisplay::kHide, @"Hide", mThenActionList, mTrackingID );
			mExecuting = true;
		} else if ( cresult == ConditionTestResult::kFalse ) {
			HExecuteActionListManager::Execute( ExecuteActionListDisplay::kHide, @"Hide", mElseActionList, mTrackingID );
			mExecuting = true;
		}
	} else {
		if ( !HExecuteActionListManager::IsExecuting( mTrackingID ) ) {
			result = true;
		}
	}
	
	return result;
}

void
AIfThenElse::ActionAborted()
{
	HExecuteActionListManager::CancelExecution( mTrackingID );
	mConditions.ConditionTestAbort();
}

#endif

#if KMF_EDITOR_WINDOW

Class
AIfThenElse::ActionController() const
{
	return [AIfThenElseEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
<ul class="IconList">
	{% for condition in conditions %}
		<li><img src={{ condition.icon | inlineimage: 16 }} alt="Condition icon" width="16" height="16" /><span class="IconTitle">{{ condition.description | htmlencode }}</span>{% if condition.extraicon %}<img src={{ condition.extraicon | inlineimage: 16 }} alt="Swatch" width="16" height="16" />{% /if %}</li>
	{% /for %}
</ul>
{% if thenactions.@count == 0 %}
	<h3>Do nothing.</h3>
{% else %}
	<h3>Execute the Following Actions:</h3>
	<ul class="IconList">
		{% for action in thenactions %}
			<li>{{ action }}</li>
		{% /for %}
	</ul>
{% /if %}
{% if elseactions.@count != 0 %}
	<h3>Otherwise, Execute the Following Actions:</h3>
	<ul class="IconList">
		{% for action in elseactions %}
			<li>{{ action }}</li>
		{% /for %}
	</ul>
{% /if %}
)RAW";

NSString*
AIfThenElse::GetActionHTMLDescriptionString() const
{
	if ( mConditions.GetCount() == 0 ) return [@"kmIfThenElseNoConditionsTitle" localized];
	return [NSString stringWithFormat:@"If %@",mConditions.GetConditionDisplayText()];
}

NSString*
AIfThenElse::GetActionExtraHTML() const
{
	if ( mConditions.GetCount() == 0 ) return @"";
	
	NSDictionary *variables = @{
									@"conditions" : mConditions.GetConditionDisplayContent(),
									@"thenactions" : mThenActionList.GetDisplayContent(),
									@"elseactions" :  mElseActionList.GetDisplayContent()
								};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}


#endif

BOOL
AIfThenElse::MatchesTextSearch( NSString* searchString ) const
{
	if ( HBaseAction::MatchesTextSearch( searchString ) ) return YES;
	if ( mConditions.MatchesTextSearch( searchString ) ) return YES;
	return mThenActionList.MatchesTextSearch( searchString ) || mElseActionList.MatchesTextSearch( searchString );
}

#pragma mark streaming

void
AIfThenElse::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	SafeCFPropertyListRef conditions = (SafeCFPropertyListRef)mConditions.CopyToPropertyList();
	dictionary->Set( CFSTR("Conditions"), conditions );
	conditions->Release();
	SafeCFPropertyListRef actions = mThenActionList.CopyActionsToPropertyList();
	dictionary->Set( CFSTR("ThenActions"), actions );
	CFRelease( actions );
	actions = mElseActionList.CopyActionsToPropertyList();
	dictionary->Set( CFSTR("ElseActions"), actions );
	CFRelease( actions );
}

void
AIfThenElse::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	SafeCFPropertyListRef conditions;
	XAssert( dictionary->Get( CFSTR("Conditions"), conditions ) );
	mConditions.LoadFromPropertyList( conditions );
	SafeCFPropertyListRef actions = (SafeCFPropertyListRef)dictionary->Get( CFSTR("ThenActions") );
	XAssert( actions );
	mThenActionList.LoadActionsFromPropertyList( actions );
	actions = (SafeCFPropertyListRef)dictionary->Get( CFSTR("ElseActions") );
	XAssert( actions );
	mElseActionList.LoadActionsFromPropertyList( actions );
}

