/*
 *  AInsertText.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Sun Dec 16 2001.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "AInsertText.h"
#import "KMDefines.h"
#import "XCoreFoundation.h"
#import "ASimulateKeystroke.h"
#import "SystemDefinedKeys.h"
#import "KMIconRegistry.h"

#if KMF_IS_ENGINE
#import "XClipboardDataSet.h"
#import "MTextResultsWindow.h"
#import "UCGRemoteControl.h"
#import "SystemDefinedKeys.h"
#import "XClipboardHistory.h"
#import "TokenText.h"
#import "FilterText.h"
#import "InsertText.h"
#import "TrippedCaseBehaviour.h"
#import "HExecuteActionListManager.h"
#import "XErrorHandler.h"
#import "NSAttributedStringExtensions.h"
#endif

#if KMF_EDITOR_WINDOW
#import "AInsertTextEditor.h"
#import "NSAttributedStringExtensions.h"
#import "NSStringExtensions.h"
#endif

AInsertText::AInsertText( const AInsertText& inAction )
	:	HBaseAction( inAction ),
		mStyledText( inAction.mStyledText ),
		mAction( inAction.mAction )
#if KMF_IS_ENGINE
		, mInsertText( nullptr )
#endif
{
#if KMF_IS_ENGINE
	check( !mInsertText.get() );
#endif
}

AInsertText::AInsertText( const InsertTextActionType& inActionType )
	:	HBaseAction(),
		mAction( inActionType )
{
}

HBaseAction*
AInsertText::Merge( HBaseAction* action )
{
	check( action );
	HBaseAction *newAction = NULL;
	if ( mAction == InsertTextActionType::kByTyping ) {
		ASimulateKeystroke* simulateKeyAction = dynamic_cast<ASimulateKeystroke*>( action );
		if ( simulateKeyAction ) {
			if ( (simulateKeyAction->GetModifiers() & ~UInt32(shiftKey)) == 0 ) {
				UniChar ch2 = SystemDefinedKeys::GetNormalUniCharFromKeyCode( simulateKeyAction->GetKeyCode(), 0, (simulateKeyAction->GetModifiers() & shiftKey) ? ShiftStyle::kShift : ShiftStyle::kNormal );
				if ( ' ' <= ch2 && ch2 <= 126 ) {
					AInsertText* insertTextAction = new AInsertText( InsertTextActionType::kByTyping );
					insertTextAction->SetPlainText( [NSString stringWithFormat:@"%@%C", mStyledText.GetString(), ch2] );
					newAction = insertTextAction;
				}
			}
		}
	}
	return newAction;
}

CFStringRef
AInsertText::CopyActionTitle() const
{
	CFStringRef key;
	switch ( mAction ) {
		case InsertTextActionType::kByTyping:
		case InsertTextActionType::kByPasting:
		case InsertTextActionType::kByPastingStyles:
			key = CFSTR("aInsertText_InsertText");
			break;
		case InsertTextActionType::kDisplayWindow:
		case InsertTextActionType::kDisplayBriefly:
			key = CFSTR("aInsertText_DisplayText");
			break;
	}
	return key->CopyLocalized();
}

CFStringRef
AInsertText::CopySummaryText( CFStringRef text )
{
	CFStringRef result;
	CFContainer<CFCharacterSetRef> set = CFCharacterSetCreateWithCharactersInString( kCFAllocatorDefault, CFSTR( "\n\r" ) );
	CFRange range;
	if ( text->FindCharacterFromSet( set, &range ) ) {
		CFContainer<CFStringRef> string = text->CopySubstring( CFRangeMake( 0, range.location ) );
		CFContainer<CFStringRef> format;
		if ( range.location == text->GetLength()-1 ) {
			format = CFSTR("aSummaryReturnFormat")->CopyLocalized();
		} else {
			format = CFSTR("aSummaryTextFormat")->CopyLocalized();
		}
		result = CFString::CreateWithFormat( format, CFStringRef(string) );
	} else if ( text->GetLength() > 50 ) {
		CFContainer<CFStringRef> string = text->CopySubstring( CFRangeMake( 0, 50 ) );
		CFContainer<CFStringRef> format = CFSTR("aSummaryTextFormat")->CopyLocalized();
		result = CFString::CreateWithFormat( format, CFStringRef(string) );
	} else {
		result = text->Copy();
	}
	return result;
}

CFStringRef
AInsertText::CopySummaryText( NSString* text )
{
	return CopySummaryText( CFBridgingToCFString(text) );
}

NSString*
AInsertText::GetSummaryText( NSString* text )
{
	CFContainer<CFStringRef> s = CopySummaryText( CFBridgingToCFString(text) );
	return s->GetNS();
}


CFStringRef
AInsertText::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> description;

	CFStringRef key;
	switch ( mAction ) {
		case InsertTextActionType::kByTyping:
			key = CFSTR("aInsertText_ByTyping");
			break;
		case InsertTextActionType::kByPasting:
			key = CFSTR("aInsertText_ByPasting");
			break;
		case InsertTextActionType::kByPastingStyles:
			key = CFSTR("aInsertText_ByPastingStyles");
			break;
		case InsertTextActionType::kDisplayWindow:
			key = CFSTR("aInsertText_DisplayWindow");
			break;
		case InsertTextActionType::kDisplayBriefly:
			key = CFSTR("aInsertText_DisplayBriefly");
			break;
	}

	CFContainer<CFStringRef> summary = CopySummaryText( mStyledText.GetString() );
	
	CFContainer<CFStringRef> format = key->CopyLocalized();
	description = CFString::CreateWithFormat( format, CFStringRef(summary) );

	return description->Copy();
}

XIcon
AInsertText::GetActionDescriptionXIcon() const
{
	XIcon result;
	
	if ( mAction == InsertTextActionType::kDisplayWindow || mAction == InsertTextActionType::kDisplayBriefly ) {
		result = XIcon( KMIconRegistry::kWindowIcon );
	} else {
		result = XIcon( KMIconRegistry::kClipboardAction );
	}
	check( result.Defined() );
	return result;
}


#if KMF_IS_ENGINE

float
AInsertText::DoAction()
{
	NSMutableAttributedString* t = mStyledText.CopyMutableAttributedString();
	bool wasPlainText = t.wasPlainText;
	t.selectionPoint = NSNotFound;
	float time = TokenText::Process( t );
	TrippedCaseBehaviour triggerCase = HExecuteActionListManager::GetTrippedCase();
	XErrorHandlerMaybeLogFormat1( "TypedString", CFSTR("triggerCase is %d"), triggerCase );
	switch ( triggerCase ) {
		case TrippedCaseBehaviour::kNone:
			// do nothing;
			break;
		case TrippedCaseBehaviour::kTitleCase:
			time += FilterText::Process( FilterText::kCapitalize, t );
			break;
		case TrippedCaseBehaviour::kUppercase:
			time += FilterText::Process( FilterText::kUppercase, t );
			break;
	}

	switch ( mAction ) {
		case InsertTextActionType::kByTyping:
			mInsertText.reset( new InsertText( false, t.string ) );
			mInsertText->DoAction();
			break;
		case InsertTextActionType::kByPasting:
			mInsertText.reset( new InsertText( true, t.string ) );
			mInsertText->DoAction();
			break;
		case InsertTextActionType::kByPastingStyles:
			if ( wasPlainText ) {
				mInsertText.reset( new InsertText( true, t.string ) );
			} else {
				mInsertText.reset( new InsertText( true, t ) );
			}
			mInsertText->DoAction();
			break;
		case InsertTextActionType::kDisplayWindow:
			if ( wasPlainText ) {
				MTextResultsWindow::Display( MTextResultsWindowKind::kWindow, @"aInsertText_DisplayTitle", @"aInsertText_DisplayDescription", t.string );
			} else {
				MTextResultsWindow::Display( MTextResultsWindowKind::kWindow, @"aInsertText_DisplayTitle", @"aInsertText_DisplayDescription", t );
			}
			break;
		case InsertTextActionType::kDisplayBriefly:
			MTextResultsWindow::Display( MTextResultsWindowKind::kBriefly, @"aInsertText_DisplayTitle", @"aInsertText_DisplayDescription", t.string );
			break;
	}
	mSimulateLeftArrows = 0;
	switch ( mAction ) {
		case InsertTextActionType::kByTyping:
		case InsertTextActionType::kByPasting:
		case InsertTextActionType::kByPastingStyles:
			if ( t.selectionPoint != NSNotFound && t.selectionPoint < t.length ) {
				mSimulateLeftArrows = t.length - t.selectionPoint;
			}
			break;
		case InsertTextActionType::kDisplayWindow:
		case InsertTextActionType::kDisplayBriefly:
			break;
	}

	return time;
}

Boolean
AInsertText::IsActionCompleted()
{
	if ( mInsertText.get() ) {
		if ( !mInsertText->IsActionCompleted() ) {
			return false;
		}
		if ( mSimulateLeftArrows > 0 ) {
			mInsertText.reset( new InsertText( mSimulateLeftArrows ) );
			mSimulateLeftArrows = 0;
			mInsertText->DoAction();
			return false;
		}
		return true;
	} else {
		return true;
	}
}

void
AInsertText::ActionAborted()
{
	mInsertText.release();
}

#endif

#if KMF_EDITOR_WINDOW

Class
AInsertText::ActionController() const
{
	return [AInsertTextEditor class];
}

NSString*
AInsertText::GetActionHTMLDescriptionString() const
{
	NSString* key;
	switch ( mAction ) {
		case InsertTextActionType::kByTyping:
			key = @"aInsertTextHTML_ByTyping";
			break;
		case InsertTextActionType::kByPasting:
			key = @"aInsertTextHTML_ByPasting";
			break;
		case InsertTextActionType::kByPastingStyles:
			key = @"aInsertTextHTML_ByPastingStyles";
			break;
		case InsertTextActionType::kDisplayWindow:
			key = @"aInsertTextHTML_DisplayWindow";
			break;
		case InsertTextActionType::kDisplayBriefly:
			key = @"aInsertTextHTML_DisplayBriefly";
			break;
	}

	return [key localized];
}

NSString*
AInsertText::GetActionExtraHTML() const
{
	return mStyledText.GetDisplayHTMLContent();
}

#endif

#pragma mark streaming

static const CFStringRef kInsertTextActionNames[] = kInsertTextActionNamesArray;

void
AInsertText::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	mStyledText.SaveToDictionary( dictionary );
	dictionary->SetBoolean( CFSTR("Paste"), (mAction == InsertTextActionType::kByPasting) );
	dictionary->Set( CFSTR("Action"), kInsertTextActionNames[int(mAction)] );
}

void
AInsertText::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mStyledText.InitializeFromDictionary( dictionary );
	if ( !dictionary->CanGetEnumClass( CFSTR("Action"), kInsertTextActionNames, mAction ) ) {
		Boolean pasteMode;
		Boolean justDisplay;
		dictionary->GetBoolean( CFSTR("Paste"), pasteMode, true );
		dictionary->GetBoolean( CFSTR("JustDisplay"), justDisplay, false );
		if ( justDisplay ) {
			mAction = InsertTextActionType::kDisplayWindow;
		} else if ( pasteMode ) {
			mAction = InsertTextActionType::kByPasting;
		} else {
			mAction = InsertTextActionType::kByTyping;
		}
	}
}
