/*
 *  AManipulateWindow.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Sat Mar 15 2003.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "AManipulateWindow.h"
#import "XException.h"
#import "KMDefines.h"
#import "XContainer.h"
#import "XErrorHandler.h"
#import "NSStringRegularExpressions.h"
#import "Calculate.h"
#import "TokenText.h"

#import "XAXApplicationElement.h"
#import "XAXWindowElement.h"
#import "XAXWindowElementList.h"
#import "ScreenInfo.h"

#if KMF_EDITOR_WINDOW
#import "AManipulateWindowEditor.h"
#import "TemplateEngine.h"
#endif

static const CFStringRef kManipulateWindowActionTypeNames[] = kManipulateWindowActionTypeNamesArray;

static bool SimpleValue( NSString* s, long int& v )
{
	if ( ![s isMatchedByRegex:@"^-?\\d+$"] ) return false;
	double doublevalue;
	if ( !Calculate::CalculateValue( s, doublevalue ) ) return false;
	v = lround( doublevalue );
	return true;
}

HBaseAction*
AManipulateWindow::Merge( HBaseAction* action )
{
	check( action );
	AManipulateWindow* newAction = NULL;
	AManipulateWindow* actualAction = dynamic_cast<AManipulateWindow*>( action );

	if ( !actualAction ) {
		// newAction = NULL;
	} else if ( mTargeting != ManipulateWindowTargetType::kFrontWindow || actualAction->mTargeting != ManipulateWindowTargetType::kFrontWindow ) {
		// newAction = NULL;
	} else if ( !(mTargetApplication == actualAction->mTargetApplication) ) {
		// newAction = NULL;
	} else if ( ((mAction == ManipulateWindowActionType::kResizeWindowBy) || (mAction == ManipulateWindowActionType::kResizeWindowTo)) && actualAction->mAction == ManipulateWindowActionType::kResizeWindowTo ) {
		newAction = new AManipulateWindow( *this );
		newAction->mAction = actualAction->mAction;
		newAction->mHorizontalNumber = actualAction->mHorizontalNumber;
		newAction->mVerticalNumber = actualAction->mVerticalNumber;
	} else if ( ( ((mAction == ManipulateWindowActionType::kResizeWindowBy) || (mAction == ManipulateWindowActionType::kResizeWindowTo)) && actualAction->mAction == ManipulateWindowActionType::kResizeWindowBy )
			   || ( ((mAction == ManipulateWindowActionType::kMoveWindowBy) || (mAction == ManipulateWindowActionType::kMoveWindowTo)) && actualAction->mAction == ManipulateWindowActionType::kMoveWindowBy ) ) {
		long int h1, h2, v1, v2;
		if ( SimpleValue( mHorizontalNumber, h1 ) && SimpleValue( mVerticalNumber, v1 ) && SimpleValue( actualAction->mHorizontalNumber, h2 ) && SimpleValue( actualAction->mVerticalNumber, v2 ) ) {
			newAction = new AManipulateWindow( *this );
			newAction->mHorizontalNumber = [NSString stringWithLongInt:h1+h2];
			newAction->mVerticalNumber = [NSString stringWithLongInt:v1+v2];
		}
	} else if ( ((mAction == ManipulateWindowActionType::kMoveWindowBy) || (mAction == ManipulateWindowActionType::kMoveWindowTo)) && actualAction->mAction == ManipulateWindowActionType::kMoveWindowTo ) {
		newAction = new AManipulateWindow( *this );
		newAction->mAction = actualAction->mAction;
		newAction->mHorizontalNumber = actualAction->mHorizontalNumber;
		newAction->mVerticalNumber = actualAction->mVerticalNumber;
	} else {
		// newAction = NULL;
	}
	
	return newAction;
}

CFStringRef
AManipulateWindow::CopyActionDescriptionString() const
{
	// first, build window name string
	
	CFContainer<CFStringRef> windowName;
	
	switch ( GetWindowTargeting() ) {
		case ManipulateWindowTargetType::kFrontWindow:
			windowName = CFSTR("aManipulateWindow_FrontWindowName")->CopyLocalized();
			break;
		case ManipulateWindowTargetType::kNamedWindow: {
				CFContainer<CFStringRef> windowNameFormat = CFSTR("aManipulateWindow_NamedWindowFormat")->CopyLocalized();
				windowName = CFString::CreateWithFormat( windowNameFormat, mWindowName );
			}
			break;
		case ManipulateWindowTargetType::kWindowNameContaining: {
				CFContainer<CFStringRef> windowNameFormat = CFSTR("aManipulateWindow_WindowWithNameContainingFormat")->CopyLocalized();
				windowName = CFString::CreateWithFormat( windowNameFormat, mWindowName );
			}
			break;
		case ManipulateWindowTargetType::kWindowIndex: {
				CFContainer<CFStringRef> windowNameFormat = CFSTR("aManipulateWindow_WindowWithIndexFormat")->CopyLocalized();
				windowName = CFString::CreateWithFormat( windowNameFormat, mWindowIndex );
			}
			break;
	}
	
	// now get the description format:
	
	CFContainer<CFStringRef> titleFormatKey = CFString::CreateWithFormat( CFSTR("aManipulateWindow_%@Format"), kManipulateWindowActionTypeNames[int(mAction)] );
	CFContainer<CFStringRef> titleFormat = titleFormatKey->CopyLocalized();
	return CFString::CreateWithFormat( titleFormat, CFStringRef(windowName) );
}

bool
AManipulateWindow::NeedHorizontalVertical() const
{
	switch ( GetWindowAction() ) {
		case ManipulateWindowActionType::kResizeWindowByPercent:
		case ManipulateWindowActionType::kResizeWindowBy:
		case ManipulateWindowActionType::kResizeWindowTo:
		case ManipulateWindowActionType::kMoveWindowBy:
		case ManipulateWindowActionType::kMoveWindowTo:
		case ManipulateWindowActionType::kMoveAndResize:
		case ManipulateWindowActionType::kCenterWindowAt:
			return true;
		case ManipulateWindowActionType::kCloseWindow:
		case ManipulateWindowActionType::kZoomWindow:
		case ManipulateWindowActionType::kCenterWindow:
		case ManipulateWindowActionType::kMinimizeWindow:
		case ManipulateWindowActionType::kUnminimizeWindow:
		case ManipulateWindowActionType::kToggleMinimizeWindow:
		case ManipulateWindowActionType::kSelectWindow:
			return false;
	}
}

bool
AManipulateWindow::NeedWidthHeight() const
{
	switch ( GetWindowAction() ) {
		case ManipulateWindowActionType::kMoveAndResize:
			return true;
		case ManipulateWindowActionType::kResizeWindowByPercent:
		case ManipulateWindowActionType::kResizeWindowBy:
		case ManipulateWindowActionType::kResizeWindowTo:
		case ManipulateWindowActionType::kMoveWindowBy:
		case ManipulateWindowActionType::kMoveWindowTo:
		case ManipulateWindowActionType::kCenterWindowAt:
		case ManipulateWindowActionType::kCloseWindow:
		case ManipulateWindowActionType::kZoomWindow:
		case ManipulateWindowActionType::kCenterWindow:
		case ManipulateWindowActionType::kMinimizeWindow:
		case ManipulateWindowActionType::kUnminimizeWindow:
		case ManipulateWindowActionType::kToggleMinimizeWindow:
		case ManipulateWindowActionType::kSelectWindow:
			return false;
	}
}

#if KMF_IS_ENGINE

static inline float
Center( float left, float totalWidth, float width )
{
	return lround( left + (totalWidth - width) / 2 - 0.1 );
}

static void
FindBestDisplayRect( CGRect windowRect, CGRect& displayRect )
{
	const CGDisplayCount kMaxCount = 10;
	CGDirectDisplayID displays[kMaxCount];
	CGDisplayCount count;
	CGDirectDisplayID best = NULL;
	
	if ( CGGetDisplaysWithRect( windowRect, kMaxCount, displays, &count ) == noErr ) {
		float bestArea = 0;
		for ( CGDisplayCount i = 0; i < count; i++ ) {
			CGRect thisDisplayRect = CGDisplayBounds( displays[i] );
			CGRect intersect = CGRectIntersection( windowRect, thisDisplayRect );
			float thisArea = intersect.size.width * intersect.size.height;
			if ( thisArea < bestArea ) {
				best = displays[i];
				bestArea = thisArea;
			}
		}
	}
	if ( !best ) {
		best = CGMainDisplayID();
	}
	displayRect = CGDisplayBounds( best );
}

static void
CenterWindowAt( const XAXWindowElement& inWindowElement, CGPoint center )
{
	CGSize windowSize;
	if ( inWindowElement.GetSize( windowSize ) ) {
		CGPoint thePos;
		thePos.x = lround( center.x - windowSize.width / 2 - 0.1 ); 
		thePos.y = lround( center.y - windowSize.height / 2 - 0.1 ); 
		inWindowElement.SetPosition( thePos );
	}
}

static void
CenterWindow( const XAXWindowElement& inWindowElement )
{
	CGRect windowRect;
	CGRect displayRect;
	if ( inWindowElement.GetPosition( windowRect.origin ) && inWindowElement.GetSize( windowRect.size ) ) {
		FindBestDisplayRect( windowRect, displayRect );
		CGPoint thePos;
		thePos.x = Center( displayRect.origin.x, displayRect.size.width, windowRect.size.width );
		thePos.y = Center( displayRect.origin.y, displayRect.size.height, windowRect.size.height );
		inWindowElement.SetPosition( thePos );
	}
}

void
AManipulateWindow::ExecuteActionOnWindow( ProcessSerialNumber appPSN, const XAXWindowElement& inWindowElement )
{
	NSDictionary* calculationContext = nil;
	if ( !!inWindowElement ) {
		NSRect r;
		if ( inWindowElement.GetPosition( r.origin ) && inWindowElement.GetSize( r.size ) ) {
			calculationContext = @{ @"WINDOW" : NSStringFromRect(r) };
		}
	}

	double horiz = 0;
	double vert = 0;
	double width = 0;
	double height = 0;

	if ( NeedHorizontalVertical() && !Calculate::CalculateValue( mHorizontalNumber, horiz, calculationContext ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Manipulate Window %@ failed"), mHorizontalNumber );
		return;
	}
	if ( NeedHorizontalVertical() && !Calculate::CalculateValue( mVerticalNumber, vert, calculationContext ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Manipulate Window %@ failed"), mVerticalNumber );
		return;
	}
	if ( NeedWidthHeight() && !Calculate::CalculateValue( mWidthNumber, width, calculationContext ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Manipulate Window %@ failed"), mHorizontalNumber );
		return;
	}
	if ( NeedWidthHeight() && !Calculate::CalculateValue( mHeightNumber, height, calculationContext ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Manipulate Window %@ failed"), mVerticalNumber );
		return;
	}

	try {		
		switch ( mAction ) {
			case ManipulateWindowActionType::kResizeWindowByPercent: {
				CGSize theSize;
				if ( inWindowElement.GetSize( theSize ) ) {
					theSize.width *= horiz/100.0;
					theSize.height *= vert/100.0;
					inWindowElement.SetSize( theSize );
				}
				break;
			}
			case ManipulateWindowActionType::kResizeWindowBy: {
				CGSize theSize;
				if ( inWindowElement.GetSize( theSize ) ) {
					theSize.width += lround(horiz);
					theSize.height += lround(vert);
					inWindowElement.SetSize( theSize );
				}
				break;
			}
			case ManipulateWindowActionType::kResizeWindowTo: {
				CGSize theSize;
				theSize.width = lround(horiz);
				theSize.height = lround(vert);
				inWindowElement.SetSize( theSize );
				break;
			}
			case ManipulateWindowActionType::kMoveWindowBy: {
				CGPoint thePos;
				if ( inWindowElement.GetPosition( thePos ) ) {
					thePos.x += lround(horiz);
					thePos.y += lround(vert);
					inWindowElement.SetPosition( thePos );
				}
				break;
			}
			case ManipulateWindowActionType::kMoveWindowTo: {
				CGPoint thePos;
				thePos.x = lround(horiz);
				thePos.y = lround(vert);
				inWindowElement.SetPosition( thePos );
				break;
			}
			case ManipulateWindowActionType::kCenterWindowAt: {
				CGPoint thePos;
				thePos.x = lround(horiz);
				thePos.y = lround(vert);
				CenterWindowAt( inWindowElement, thePos );
				break;
			}
			case ManipulateWindowActionType::kMoveAndResize: {
				CGRect theFrame;
				theFrame.origin.x = lround(horiz);
				theFrame.origin.y = lround(vert);
				theFrame.size.width = lround(width);
				theFrame.size.height = lround(height);
				inWindowElement.SetFrame( theFrame );
				break;
			}
			case ManipulateWindowActionType::kCloseWindow:
				inWindowElement.CloseWindow();
				break;
			case ManipulateWindowActionType::kZoomWindow:
				inWindowElement.ZoomWindow();
				break;
			case ManipulateWindowActionType::kCenterWindow:
				CenterWindow( inWindowElement );
				break;
			case ManipulateWindowActionType::kMinimizeWindow:
				inWindowElement.SetMinimized( true );
				break;
			case ManipulateWindowActionType::kUnminimizeWindow:
				inWindowElement.SetMinimized( false );
				break;
			case ManipulateWindowActionType::kToggleMinimizeWindow:
				inWindowElement.SetMinimized( !inWindowElement.IsMinimized() );
				break;
			case ManipulateWindowActionType::kSelectWindow:
				inWindowElement.SetMainWindow();
				(void) SetFrontProcessWithOptions( &appPSN, kSetFrontProcessFrontWindowOnly );
				break;
		}
		
	} catch (...) {
	
	}
}

float
AManipulateWindow::DoAction()
{
	float time = 0;
	if (!AXAPIEnabled()) {
		XErrorHandlerMaybeLogMessage( "always", CFSTR("Manipulate Window action required Universal Access which is disabled") );
		return time;
	}

	ProcessSerialNumber psn;
	if ( mTargetApplication.GetCurrentlyTargetedPSN( psn ) ) {
		XAXApplicationElement theApp( psn );
		if ( theApp ) {
			if ( mTargeting == ManipulateWindowTargetType::kFrontWindow ) {
				XAXWindowElement windowElement( theApp.GetFocussedOrMainWindow() );
				
				if ( !!windowElement ) {
					ExecuteActionOnWindow( psn, windowElement );
					time = 5;
				}
			} else {
				XAXWindowElementList* theWindowList = ScreenInfo::NewGoodWindowList( theApp );
				
				if ( theWindowList ) {
				
					if ( mTargeting == ManipulateWindowTargetType::kWindowIndex ) {
						double windex;
						if ( !Calculate::CalculateValue( mWindowIndex, windex ) ) {
							XErrorHandlerMaybeLogFormat1("always", @"Manipulate Window %@ failed", mWindowIndex );
						} else if ( 0 < lround(windex) && lround(windex) <= theWindowList->GetCount() ) {
							ExecuteActionOnWindow( psn, theWindowList->GetValueAtIndex( lround(windex)-1 ) );
							time = 10;
						} else if ( -theWindowList->GetCount() <= lround(windex) && lround(windex) < 0 ) {
							ExecuteActionOnWindow( psn, theWindowList->GetValueAtIndex( theWindowList->GetCount()+lround(windex) ) );
							time = 10;
						} else {
							XErrorHandlerMaybeLogFormat1("always", @"Manipulate Window failed to find window with index %ld", lround(windex) );
						}
					} else {
						NSString* wname = TokenText::GetResult( mWindowName, &time );
						for ( CFIndex i = 0; i < theWindowList->GetCount(); i++ ) {
							Boolean winFound = false;
							NSString* winName = theWindowList->GetValueAtIndex( i ).GetTitle( NULL );
							
							if ( winName != NULL ) {
								if ( mTargeting == ManipulateWindowTargetType::kNamedWindow ) {
									winFound = ( [wname caseInsensitiveCompare:winName] == NSOrderedSame );
								} else if ( mTargeting == ManipulateWindowTargetType::kWindowNameContaining ) {
									winFound = [winName containsCaseInsensitiveString:wname];
								}
							}
							
							if ( winFound ) {
								ExecuteActionOnWindow( psn, theWindowList->GetValueAtIndex( i ) );
								time = 10;
							}
						}
					}
					delete theWindowList;
				}
			}
		}
    }
	
	return time;
}


#endif

#if KMF_EDITOR_WINDOW

Class
AManipulateWindow::ActionController() const
{
	return [AManipulateWindowEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
{% if in %}<span class="IconTitle">In</span> {{ in }}<br />{% /if %}
{% if to %}{{ to }}: {{ totext }}<br />{% /if %}
)RAW";

NSString*
AManipulateWindow::GetActionExtraHTML() const
{
	NSMutableDictionary* variables = [NSMutableDictionary new];

	switch ( GetWindowAction() ) {
		case ManipulateWindowActionType::kMoveAndResize:
			variables[@"to"] = @"To";
			variables[@"totext"] = [NSString stringWithFormat:@"(%@,%@, %@,%@)", mHorizontalNumber, mVerticalNumber, mWidthNumber, mHeightNumber];
			break;
		case ManipulateWindowActionType::kResizeWindowByPercent:
			variables[@"to"] = @"By";
			variables[@"totext"] = [NSString stringWithFormat:@"%@ percent horizontally, %@ percent vertically.", mHorizontalNumber, mVerticalNumber];
			break;
		case ManipulateWindowActionType::kResizeWindowBy:
		case ManipulateWindowActionType::kMoveWindowBy:
			variables[@"to"] = @"By";
			variables[@"totext"] = [NSString stringWithFormat:@"%@ horizontally, %@ vertically.", mHorizontalNumber, mVerticalNumber];
			break;
		case ManipulateWindowActionType::kResizeWindowTo:
		case ManipulateWindowActionType::kMoveWindowTo:
			variables[@"to"] = @"To";
			variables[@"totext"] = [NSString stringWithFormat:@"%@ horizontally, %@ vertically.", mHorizontalNumber, mVerticalNumber];
			break;
		case ManipulateWindowActionType::kCenterWindowAt:
			variables[@"to"] = @"At";
			variables[@"totext"] = [NSString stringWithFormat:@"%@ horizontally, %@ vertically.", mHorizontalNumber, mVerticalNumber];
			break;
		case ManipulateWindowActionType::kCloseWindow:
		case ManipulateWindowActionType::kZoomWindow:
		case ManipulateWindowActionType::kCenterWindow:
		case ManipulateWindowActionType::kMinimizeWindow:
		case ManipulateWindowActionType::kUnminimizeWindow:
		case ManipulateWindowActionType::kToggleMinimizeWindow:
		case ManipulateWindowActionType::kSelectWindow:
			break;
	}

	if ( mTargetApplication.GetApplicationTargetType() == ApplicationTargetApplicationType::kSpecific ) {
		variables[@"in"] = mTargetApplication.GetTargetApplication().GetDisplayHTMLContent();
	}
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kManipulateWindowTargetTypeNames[] = kManipulateWindowTargetTypeNamesArray;

void
AManipulateWindow::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Targeting"), kManipulateWindowTargetTypeNames[int(mTargeting)] );
	dictionary->Set( CFSTR("Action"), kManipulateWindowActionTypeNames[int(mAction)] );
	dictionary->Set( CFSTR("WindowName"), CFBridgingToCFString(mWindowName) );
	dictionary->Set( CFSTR("HorizontalExpression"), CFBridgingToCFString(mHorizontalNumber) );
	dictionary->Set( CFSTR("VerticalExpression"), CFBridgingToCFString(mVerticalNumber) );
	dictionary->Set( CFSTR("WidthExpression"), CFBridgingToCFString(mWidthNumber) );
	dictionary->Set( CFSTR("HeightExpression"), CFBridgingToCFString(mHeightNumber) );
	dictionary->Set( CFSTR("WindowIndexExpression"), CFBridgingToCFString(mWindowIndex) );
	mTargetApplication.SaveToDictionary( dictionary, CFSTR("TargetingType"), CFSTR("TargetApplication") );
}

void
AManipulateWindow::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef value;
	SInt32 v;

	bool good;
	good = dictionary->CanGetEnumClass( CFSTR("Targeting"), kManipulateWindowTargetTypeNames, mTargeting );
	XAssert( good );
	
	good = dictionary->CanGetEnumClass( CFSTR("Action"), kManipulateWindowActionTypeNames, mAction );
	XAssert( good );

	CFStringRef newWindowName;
	dictionary->Get( CFSTR("WindowName"), newWindowName, CFSTR("") );
	check( newWindowName );
	mWindowName = newWindowName->GetNS();
	
	dictionary->Get( CFSTR("HorizontalExpression"), value, CFStringRef(NULL) );
	mHorizontalNumber = value ? value->GetNS() : nil;
	if ( !mHorizontalNumber && dictionary->Get( CFSTR("Horizontal"), v ) ) {
		mHorizontalNumber = [NSString stringWithLongInt:v];
	}
	if ( !mHorizontalNumber ) {
		mHorizontalNumber = kManipulateWindowDefaultNumber;
	}
	
	dictionary->Get( CFSTR("VerticalExpression"), value, CFStringRef(NULL) );
	mVerticalNumber = value ? value->GetNS() : nil;
	if ( !mVerticalNumber && dictionary->Get( CFSTR("Vertical"), v ) ) {
		mVerticalNumber = [NSString stringWithLongInt:v];
	}
	if ( !mVerticalNumber ) {
		mVerticalNumber = kManipulateWindowDefaultNumber;
	}
	
	dictionary->Get( CFSTR("WidthExpression"), value, CFBridgingToCFString(kManipulateWindowDefaultSizeNumber) );
	mWidthNumber = value->GetNS();

	dictionary->Get( CFSTR("HeightExpression"), value, CFBridgingToCFString(kManipulateWindowDefaultSizeNumber) );
	mHeightNumber = value->GetNS();

	dictionary->Get( CFSTR("WindowIndexExpression"), value, CFStringRef(NULL) );
	mWindowIndex = value ? value->GetNS() : nil;
	if ( !mWindowIndex && dictionary->Get( CFSTR("WindowIndex"), v ) ) {
		mWindowIndex = [NSString stringWithLongInt:v];
	}
	if ( !mWindowIndex ) {
		mWindowIndex = kManipulateWindowDefaultWindowIndex;
	}
	
	
	if ( mTargeting == ManipulateWindowTargetType::kWindowIndex && [mWindowIndex isEqualToString:@"1"] ) {
		mTargeting = ManipulateWindowTargetType::kFrontWindow;
	}
	mTargetApplication.InitializeFromDictionary( dictionary, CFSTR("TargetingType"), CFSTR("TargetApplication") );
}
