//
//  APlaySound.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 19 May 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "APlaySound.h"
#import "KMDefines.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "AudioOutputDevices.h"
#if KMF_EDITOR_WINDOW
#import "APlaySoundEditor.h"
#import "TemplateEngine.h"
#endif

const int kDefaultVolume = 100;
CFStringRef kPlaySoundDefaultDeviceUID = CFSTR("DEFAULT");
CFStringRef kPlaySoundSoundEffectsDeviceUID = CFSTR("SOUNDEFFECTS");
CFStringRef kPlaySoundBeepPath = CFSTR("/");

APlaySound::APlaySound()
	:	HBaseAction(),
		mOutputDeviceUID( kPlaySoundSoundEffectsDeviceUID->Copy() ),
		mVolume( kDefaultVolume )
{
	NSString* kDefaultPath = @"/System/Library/Sounds/Hero.aiff";
	if ( [kDefaultPath fileKind] == PathFileKind::kFile ) {
		mPath = (CFStringRef)CFBridgingRetain(kDefaultPath);
	} else {
		mPath = kPlaySoundBeepPath->Copy();
	}
#if KMF_IS_ENGINE
	mSound = nil;
#endif
}

APlaySound::APlaySound( const APlaySound& inAction )
	:	HBaseAction( inAction ),
		mPath( inAction.mPath ),
		mOutputDeviceUID( inAction.mOutputDeviceUID ),
		mVolume( inAction.mVolume )
{
#if KMF_IS_ENGINE
	mSound = nil;
#endif
}

APlaySound::~APlaySound()
{
#if KMF_IS_ENGINE
	[mSound stop];
#endif
}

bool
APlaySound::IsBeep() const
{
	return mPath->IsEqual( kPlaySoundBeepPath );
}

CFStringRef
APlaySound::CopyActionDescriptionString() const
{
	if ( IsBeep() ) {
		return CFSTR("hkPlaySoundSystemBeepTitle")->CopyLocalized();
	} else {
		CFContainer<CFStringRef> theFormatStr = CFSTR("hkPlaySoundTitle")->CopyLocalized();
		return CFString::CreateWithFormat( theFormatStr, [mPath->GetNS() displayName] );
	}
}

#if KMF_IS_ENGINE

float
APlaySound::DoAction()
{
	check( mSound == nil );
	if ( IsBeep() ) {
		NSBeep();
	} else {
		mSound = [[NSSound alloc] initWithContentsOfFile:mPath->GetNS() byReference:YES];
		if ( !mSound ) {
			XErrorBeep( "Play Sound action failed" );
		} else {
			NSString* uid = mOutputDeviceUID->GetNS();
			if ( mOutputDeviceUID->IsEqual( kPlaySoundDefaultDeviceUID ) ) {
				uid = AudioOutputDevices::DefaultDeviceUID();
			} else if ( mOutputDeviceUID->IsEqual( kPlaySoundSoundEffectsDeviceUID ) ) {
				uid = AudioOutputDevices::SoundEffectsDeviceUID();
			}
			if ( !IsEmpty( uid ) ) {
				[mSound setPlaybackDeviceIdentifier:uid];
			}
		}
		
		[mSound setVolume:mVolume/100.0];
		[mSound play];
	}
	
	return 10;
}

Boolean
APlaySound::IsActionCompleted()
{
	return ![mSound isPlaying];
}

void
APlaySound::ActionAborted()
{
	[mSound stop];
	mSound = nil;
}

#endif

#if KMF_EDITOR_WINDOW

Class
APlaySound::ActionController() const
{
	return [APlaySoundEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 Out {{ device }} at {{ volume }}% volume.
)RAW";

NSString*
APlaySound::GetActionExtraHTML() const
{
	NSString* device = @"Unknown Device";
	if ( mOutputDeviceUID->IsEqual( kPlaySoundDefaultDeviceUID ) ) {
		device = @"Default Output Device";
	} else if ( mOutputDeviceUID->IsEqual( kPlaySoundSoundEffectsDeviceUID ) ) {
		device = @"Sound Effects Output Device";
	} else {
		NSArray* devices = AudioOutputDevices::List();
		for ( NSDictionary* entry in devices ) {
			if ( mOutputDeviceUID->IsEqual( CFBridgingToCFString([entry objectForKey:AudioOutputDevices::kUIDKey]) ) ) {
				device = [entry objectForKey:AudioOutputDevices::kNameKey];
			}
		}
	}
	NSDictionary *variables = @{
							@"volume" : @(mVolume),
							@"device" : device,
						};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

void
APlaySound::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Path"), CFStringRef(mPath) );
	dictionary->SetNumber( CFSTR("Volume"), mVolume );
	dictionary->Set( CFSTR("DeviceID"), CFStringRef(mOutputDeviceUID) );
}

void
APlaySound::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef newPath;
	dictionary->Get( CFSTR("Path"), newPath, kPlaySoundBeepPath );
	check( newPath );
	mPath = newPath->Copy();
	
	if ( !dictionary->Get( CFSTR("Volume"), mVolume ) ) mVolume = kDefaultVolume;

	CFStringRef newDeviceID;
	dictionary->Get( CFSTR("DeviceID"), newDeviceID, kPlaySoundSoundEffectsDeviceUID );
	check( newDeviceID );
	mOutputDeviceUID = newDeviceID->Copy();
}
