//
//  APlugIn.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Wed 27 Mar 2013.
//  Copyright (c) 2013 Stairways Software. All rights reserved.
//

#import "APlugIn.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "XErrorHandler.h"
#import "KMPreferences.h"
#import "TokenText.h"
#import "KMVariables.h"
#import "NSStringRegularExpressions.h"
#import "KMFileTypes.h"
#import "CSClipboardDomain.h"
#import "CSClipboardDataSet.h"
#import "KMNotifications.h"
#import "Calculate.h"
#include <mutex>

#if KMF_EDITOR_WINDOW
#import "APlugInEditor.h"
#import "EnumIterator.h"
#import "TemplateEngine.h"
#endif

NSString* kReloadThirdPartyActionsNotification = @"ReloadThirdPartyActionsNotification";

static NSMutableDictionary* gInformations = nil;
static NSMutableDictionary* gIcons = nil;
static NSString* kPlugInFolderName = @"Keyboard Maestro Actions";
static NSString* kInformationPlistName = @"Keyboard Maestro Action.plist";

static NSString* kInformationNameKey = @"Name";
static NSString* kInformationScriptKey = @"Script";
static NSString* kInformationIconKey = @"Icon";
static NSString* kInformationTitleKey = @"Title";
static NSString* kInformationTimeOutKey = @"Time Out";
static NSString* kInformationAuthorKey = @"Author";
static NSString* kInformationURLKey = @"URL";
static NSString* kInformationHelpKey = @"Help";
static NSString* kInformationResultsKey = @"Results";
//static NSString* kInformationTrimResultsKey = @"Trim Results";
//#if KMF_IS_ENGINE
//static NSString* kInformationVariableKey = @"Variable";
//static NSString* kResultsParameterVariableName = @"Output Variable";
//static NSString* kInformationClipboardKey = @"Clipboard";
//static NSString* kResultsParameterClipboardName = @"Output Clipboard";
//#endif
static NSString* kInformationParametersKey = @"Parameters";

static NSString* kInformationParametersLabelKey = @"Label";
static NSString* kInformationParametersTypeKey = @"Type";
static NSString* kInformationParametersDefaultKey = @"Default";
static NSString* kInformationParametersMenuKey = @"Menu";

static const CFStringRef kInformationParameterTypesNames[] = kInformationParameterTypesNamesArray;
static const CFStringRef kMTextResultsWindowKindNames[] = MTextResultsWindowKindNamesArray;

static std::mutex& gInformationsMutext()
{
	static std::mutex* g;
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		g = new std::mutex();
	});
	return *g;
}

static std::mutex& gIconsMutext()
{
	static std::mutex* g;
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		g = new std::mutex();
	});
	return *g;
}

static NSDictionary* InformationForFolderName( NSString* folderName )
{
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		gInformations = [NSMutableDictionary new];
	});
	std::unique_lock<std::mutex> lock(gInformationsMutext());
	NSDictionary* result = gInformations[folderName];
	if ( !result ) {
		NSString* path = KMPreferences::FilePath( kPlugInFolderName, folderName );
		path = [path stringByAppendingPathComponent:kInformationPlistName];
		result = [NSDictionary dictionaryWithContentsOfFile:path];
		if ( result ) {
			check( [result isKindOfClass:[NSDictionary class]] );
			gInformations[folderName] = result;
		} else {
			XErrorHandlerMaybeLogFormat1("always", @"Invalid Plug In Action with Name %@", folderName);
		}
	}
	return result;
}

void
APlugIn::ClearCache()
{
	{
		std::unique_lock<std::mutex> lock(gInformationsMutext());
		[gInformations removeAllObjects];
	}
	{
		std::unique_lock<std::mutex> lock(gIconsMutext());
		[gIcons removeAllObjects];
	}
}

void
APlugIn::UpdateToDefaults()
{
	NSMutableDictionary* parameters = [mParameters mutableCopy];
	NSUInteger count = GetInformationParameterCount();
	for ( NSUInteger index = 0; index < count; index++ ) {
		NSString* def = GetInformationParameterEntryDefault( index );
		if ( def ) {
			NSString* label = GetInformationParameterEntryLabel( index );
			InformationParameterTypes type = GetInformationParameterEntryType( index );
			if ( label && (type == InformationParameterTypes::kHidden || !parameters[label]) ) {
				parameters[label] = def;
			}
		}
	}
	mParameters = parameters;
}

APlugIn::APlugIn( NSString* folderName )
	: mFolderName( folderName )
{
	mDisplayKind = GetInformationDefaultDisplayKind();
	UpdateToDefaults();
}

NSString*
APlugIn::WeakGetInformationKeyedString( NSString* key ) const
{
	NSDictionary* information = InformationForFolderName( mFolderName );
	return information[key];
}

NSString*
APlugIn::GetInformationKeyedString( NSString* key ) const
{
	NSDictionary* information = InformationForFolderName( mFolderName );
	NSString* result = information[key];
	if ( ![result isKindOfClass:[NSString class]] ) {
		result = nil;
	}
	return result;
}

NSNumber*
APlugIn::GetInformationKeyedNumber( NSString* key ) const
{
	NSDictionary* information = InformationForFolderName( mFolderName );
	NSNumber* result = information[key];
	if ( ![result isKindOfClass:[NSNumber class]] ) {
		result = nil;
	}
	return result;
}

NSString*
APlugIn::GetInformationName() const
{
	NSString* result = GetInformationKeyedString( kInformationNameKey );
	if ( !result ) {
		XErrorHandlerMaybeLogFormat2( "always", @"Invalid Name %@ for Plug In Action %@", WeakGetInformationKeyedString( kInformationNameKey ), mFolderName );
		result = mFolderName;
	}
	return result;
}

NSString*
APlugIn::GetInformationKeyedFile( NSString* key ) const
{
	NSString* file = GetInformationKeyedString( key );
	if ( ![file isMatchedByRegex:@"^[A-Za-z0-9_]+(?:[.][A-Za-z0-9_]+)?$"] ) {
		file = nil;
	}
	if ( !file ) {
		XErrorHandlerMaybeLogFormat2( "always", @"Invalid File %@ for Plug In Action %@", WeakGetInformationKeyedString( key ), mFolderName );
	}

	NSString* path = KMPreferences::FilePath( kPlugInFolderName, mFolderName );
	if ( path && file ) {
		return [path stringByAppendingPathComponent:file];
	}
	return nil;
}

NSString*
APlugIn::GetInformationScript() const
{
	return GetInformationKeyedFile( kInformationScriptKey );
}

NSString*
APlugIn::GetInformationIcon() const
{
	return GetInformationKeyedFile( kInformationIconKey );
}

NSString*
APlugIn::GetInformationTitle() const
{
	NSString* result = GetInformationKeyedString( kInformationTitleKey );
	if ( !result ) {
		result = GetInformationName();
	}
	return result;
}

//Boolean
//APlugIn::GetInformationTrimResults() const
//{
//	NSDictionary* information = InformationForFolderName( mFolderName );
//	NSString* result = information[kInformationTrimResultsKey];
//	if ( ![result isKindOfClass:[NSNumber class]] ) {
//		result = nil;
//	}
//	return result ? [result boolValue] : true;
//}

//MTextResultsWindowKind
//APlugIn::GetInformationResults() const
//{
//	NSString* results = GetInformationKeyedString( kInformationResultsKey );
//	if ( !results ) {
//		return MTextResultsWindowKind::kNone;
//	}
//	int result = -1;
//	if ( [results isKindOfClass:[NSString class]] ) {
//		CFIndex i = 0;
//		while ( kMTextResultsWindowKindNames[i] ) {
//			if ( [results caseInsensitiveCompare:CFBridgingToNS(kMTextResultsWindowKindNames[i])] == NSOrderedSame ) {
//				result = i;
//				break;
//			}
//			i++;
//		}
//	}
//	if ( result >= 0 ) {
//		return static_cast<MTextResultsWindowKind>(result);
//	}
//	XErrorHandlerMaybeLogFormat2( "always", @"Invalid Results %@ for Plug In Action %@", WeakGetInformationKeyedString( kInformationResultsKey ), mFolderName );
//	return MTextResultsWindowKind::kNone;
//}

MTextResultsWindowKind
APlugIn::GetInformationDefaultDisplayKind() const
{
	NSString* results = GetInformationKeyedString( kInformationResultsKey );
	NSArray* a = [results componentsSeparatedByString:@"|"];
	if ( a.count < 1 ) {
		XErrorHandlerMaybeLogFormat1("always", @"Invalid Plug In Results %@", mFolderName);
		return MTextResultsWindowKind::kNone;
	}
	return CFDictionary::GetEnumClassValue( CFBridgingToCFString( a[0] ), kMTextResultsWindowKindNames, MTextResultsWindowKind::kNone );
}

#if KMF_EDITOR_WINDOW

Boolean
APlugIn::IsAllowedDisplayKind( MTextResultsWindowKind kind ) const
{
	NSString* results = GetInformationKeyedString( kInformationResultsKey );
	if ( !results ) {
		XErrorHandlerMaybeLogFormat1("always", @"Invalid Plug In Results %@", mFolderName);
		return kind == MTextResultsWindowKind::kNone;
	}
	return [results containsString:CFBridgingToNS( kMTextResultsWindowKindNames[int(kind)] )];
}

int
APlugIn::CountAllowedDisplayKind() const
{
	int resultsKindCount = 0;
	for( MTextResultsWindowKind kind : EnumIterator<MTextResultsWindowKind>() ) {
		if ( IsAllowedDisplayKind( kind ) ) {
			resultsKindCount++;
		}
	}
	return resultsKindCount;
}


Boolean
APlugIn::ShouldDisplayKind() const
{
	int resultsKindCount = CountAllowedDisplayKind();
	return ( resultsKindCount > 1 || (resultsKindCount == 1 && !IsAllowedDisplayKind( MTextResultsWindowKind::kNone ) ) );
}

Boolean
APlugIn::NeedDisplayKind() const
{
	int resultsKindCount = CountAllowedDisplayKind();
	return ( resultsKindCount > 1 || (resultsKindCount == 1 && (IsAllowedDisplayKind( MTextResultsWindowKind::kSaveToVariable ) || IsAllowedDisplayKind( MTextResultsWindowKind::kSaveToClipboard ) ) ) );
}

#endif

NSArray*
APlugIn::GetInformationParameters() const
{
	NSDictionary* information = InformationForFolderName( mFolderName );
	NSArray* result = information[kInformationParametersKey];
	if ( !result ) return nil;
	if ( ![result isKindOfClass:[NSArray class]] ) {
		XErrorHandlerMaybeLogFormat2( "always", @"Invalid Parameters %@ for Plug In Action %@", information[kInformationParametersKey], mFolderName );
		return nil;
	}
	return result;
}

NSUInteger
APlugIn::GetInformationParameterCount() const
{
	return [GetInformationParameters() count];
}

NSDictionary*
APlugIn::GetInformationParameterEntry( NSUInteger index ) const
{
	NSArray* paraminfo = GetInformationParameters();
	check( index < paraminfo.count );
	NSDictionary* result = nil;
	if ( index < paraminfo.count ) {
		result = paraminfo[index];
		if ( ![result isKindOfClass:[NSDictionary class]] ) {
			XErrorHandlerMaybeLogFormat3( "always", @"Invalid Parameter Entry %d of parameters %@ for Plug In Action %@", int(index), paraminfo, mFolderName );
			result = nil;
		}
	}
	return result;
}

NSString*
APlugIn::GetInformationParameterEntryLabel( NSUInteger index ) const
{
	NSDictionary* entry = GetInformationParameterEntry( index );
	NSString* result = entry[kInformationParametersLabelKey];
	if ( ![result isKindOfClass:[NSString class]] ) {
		XErrorHandlerMaybeLogFormat3( "always", @"Invalid Parameter Entry %d Label %@ for Plug In Action %@", int(index), result, mFolderName );
		result = nil;
	} else if ( !KMVariables::ValidVariable( result ) ) {
		XErrorHandlerMaybeLogFormat3( "always", @"Invalid Parameter Entry %d Label Format %@ for Plug In Action %@", int(index), result, mFolderName );
		result = nil;
	}
	return result;
}

InformationParameterTypes
APlugIn::GetInformationParameterEntryType( NSUInteger index ) const
{
	NSDictionary* entry = GetInformationParameterEntry( index );
	NSString* results = entry[kInformationParametersTypeKey];
	InformationParameterTypes result = InformationParameterTypes::kInvalid;
	if ( [results isKindOfClass:[NSString class]] ) {
		result = CFDictionary::GetEnumClassValue( CFBridgingToCFString( results ), kInformationParameterTypesNames, InformationParameterTypes::kInvalid );
	}
	if ( result == InformationParameterTypes::kInvalid ) {
		XErrorHandlerMaybeLogFormat3( "always", @"Invalid Parameter Entry %d Type %@ for Plug In Action %@", int(index), results, mFolderName );
	}
	return result;
}

NSString*
APlugIn::GetInformationParameterEntryDefault( NSUInteger index ) const
{
	NSDictionary* entry = GetInformationParameterEntry( index );
	NSString* result = entry[kInformationParametersDefaultKey];
	if ( !result ) return nil;
	if ( ![result isKindOfClass:[NSString class]] ) {
		XErrorHandlerMaybeLogFormat3( "always", @"Invalid Parameter Entry %d Default %@ for Plug In Action %@", int(index), result, mFolderName );
		result = nil;
	}
	return result;
}

NSString*
APlugIn::GetInformationParameterEntryMenu( NSUInteger index ) const
{
	NSDictionary* entry = GetInformationParameterEntry( index );
	NSString* result = entry[kInformationParametersMenuKey];
	if ( !result ) return nil;
	if ( ![result isKindOfClass:[NSString class]] ) {
		XErrorHandlerMaybeLogFormat3( "always", @"Invalid Parameter Entry %d Default %@ for Plug In Action %@", int(index), result, mFolderName );
		result = nil;
	}
	return result;
}

NSString*
APlugIn::GetNewActionTitle() const
{
	return GetInformationName();
}


CFStringRef
APlugIn::CopyActionDescriptionString() const
{
	NSString* title = GetInformationTitle();
	title = TokenText::GetResult( title, NULL, true, false, @{@"Parameters" : mParameters} );
	return CFBridgingToCFString(title)->Copy();
}

NSString*
APlugIn::GetToolTip() const
{
	NSString* author = GetInformationKeyedString( kInformationAuthorKey );
	NSString* url = GetInformationKeyedString( kInformationURLKey );
	NSString* help = GetInformationKeyedString( kInformationHelpKey );
	if ( !author && !url && !help ) return nil;
	if ( !author && !url ) return help;
	NSMutableString* result = [NSMutableString string];
	if ( author ) {
		[result appendFormat:@"Author: %@\r", author];
	}
	if ( url ) {
		[result appendFormat:@"URL: %@\r", url];
	}
	if ( help ) {
		[result appendFormat:@"Help: %@\r", help];
	}
	[result replaceOccurrencesOfString:@"\r\n" withString:@"\r"];
	[result replaceOccurrencesOfString:@"\n" withString:@"\r"];
	return [result stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
}

Boolean
APlugIn::HasTimeOut() const
{
	return DefaultTimeOutPeriod() > 0.01;
}

EventTime
APlugIn::DefaultTimeOutPeriod() const
{
	NSNumber* n = GetInformationKeyedNumber( kInformationTimeOutKey );
	if ( !n ) return kVeryLongDefaultTimeOutValue;
	return [n doubleValue];
}

NSString*
APlugIn::GetParameterValue( NSString* label ) const
{
	check( KMVariables::ValidVariable( label ) );
	NSString* result = nil;
	if ( KMVariables::ValidVariable( label ) ) {
		result = mParameters[label];
	}
	return result ? result : @"";
}

void
APlugIn::SetParameterValue( NSString* label, NSString* value )
{
	check( KMVariables::ValidVariable( label ) );
	NSMutableDictionary* parameters = [mParameters mutableCopy];
	parameters[label] = value;
	mParameters = parameters;
}

XIcon
APlugIn::GetActionXIcon() const
{
	NSString* path = GetInformationIcon();
	if ( IsEmpty( path ) ) return HBaseAction::GetActionXIcon();
	std::unique_lock<std::mutex> lock(gIconsMutext());
	NSImage* i = gIcons[path];
	if ( (id)i == [NSNull null] ) return HBaseAction::GetActionXIcon();
	if ( !i ) {
		i = [[NSImage alloc] initWithContentsOfFile:path];
		[i setSize:NSMakeSize( 32, 32 )];
		if ( !i ) {
			i = (id)[NSNull null];
		}
		gIcons[path] = i;
	}
	if ( (id)i == [NSNull null] ) return HBaseAction::GetActionXIcon();
	return XIcon( i );
}

#if KMF_IS_ENGINE

NSString*
APlugIn::GetLongScriptDisplayName() const
{
	return [NSString stringWithFormat:[@"kmPlugInLongScriptName" localized], GetInformationName()];
}

NSString*
APlugIn::GetShellCommand()
{
	NSString* script = GetInformationScript();
	NSString* command = nil;
	if ( script ) {
		if ( KMFileTypes::IsAppleScriptFilename( script ) ) {
			NSString* osascript = KMPreferences::Get( KMPreferences::kOSAScriptCommandID, @"/usr/bin/osascript" );
			command = [NSString stringWithFormat:@"%@ %@", osascript, [script shellQuotedPath]];
			mStripTrailingLineFeed = true;
		} else {
			NSInteger mode = [script getFileMode];
			if ( (mode & 0100) == 0 ) {
				mode |= 0500;
				mode &= 0777; // no funny business!
				[script setFileMode:mode];
			}
			command = [script shellQuotedPath];
			mStripTrailingLineFeed = false;
		}
	}
	return command;
}

NSDictionary*
APlugIn::GetExecutionParameters()
{
	NSUInteger count = GetInformationParameterCount();
	if ( IsEmpty( mParameters ) || count == 0 ) return nil;

	NSMutableDictionary* params = [NSMutableDictionary new];
	for ( NSUInteger i = 0; i < count; i++ ) {
		NSString* label = GetInformationParameterEntryLabel( i );
		InformationParameterTypes type = GetInformationParameterEntryType( i );
		if ( label && type != InformationParameterTypes::kInvalid ) {
			NSString* value = mParameters[label];
			value = value ? value : @"";
			switch ( type ) {
				case InformationParameterTypes::kInvalid:
					check( false );
					[[clang::fallthrough]];
				case InformationParameterTypes::kString:
				case InformationParameterTypes::kText:
				case InformationParameterTypes::kCheckbox:
				case InformationParameterTypes::kPopupMenu:
					break;
				case InformationParameterTypes::kTokenString:
				case InformationParameterTypes::kTokenText:
				case InformationParameterTypes::kHidden:
					value = TokenText::GetResult( value );
					break;
				case InformationParameterTypes::kCalculation:
					value = Calculate::Calculate( value );
					break;
			}
			params[label] = value;
		}
	}
	return params;
}

//NSString*
//APlugIn::SaveToVariable()
//{
//	NSString* result = GetInformationKeyedString( kInformationVariableKey );
//	if ( !IsEmpty( result ) ) return result;
//	result = GetParameterValue( kResultsParameterVariableName );
//	if ( !IsEmpty( result ) ) {
//		result = TokenText::GetResult( result, NULL, true, false, @{@"Parameters" : mParameters} );
//	}
//	return result;
//}
//
//void
//APlugIn::WriteClipboardDataSet( const XClipboardDataSet& dataset )
//{
//	NSString* result = GetInformationKeyedString( kInformationClipboardKey );
//	if ( IsEmpty( result ) ) {
//		result = GetParameterValue( kResultsParameterClipboardName );
//		if ( !IsEmpty( result ) ) {
//			result = TokenText::GetResult( result, NULL, true, false, @{@"Parameters" : mParameters} );
//		}
//	}
//	ClipboardTarget target;
//	if ( !IsEmpty( result ) ) {
//		CSClipboardDataSet* pExistingClipData = CSClipboardDomain::GetClipboardDomain()->GetClipDataSetByDisplayName( CFBridgingToCFString( result ) );
//		if ( pExistingClipData ) {
//			target.SetGenericNamedClipboardUID( pExistingClipData->GetUID() );
//		} else {
//			std::unique_ptr<CSClipboardDataSet> pNewClipData( new CSClipboardDataSet( CFBridgingToCFString( result ), kCSClipboardDataSetAddDefaultContent ) );
//			CSClipboardDomain::GetClipboardDomain()->InsertClipDataSet( pNewClipData.get(), KMNotifications::kLoadClipDataFromDefaultCreate );
//			target.SetGenericNamedClipboardUID( pNewClipData->GetUID() );
//		}
//	}
//	target.WriteClipboardDataSet( dataset );
//}

#endif

#if KMF_EDITOR_WINDOW

Class
APlugIn::ActionController() const
{
	bool someUnhiddenParameter = false;
	NSUInteger count = GetInformationParameterCount();
	for ( NSUInteger i = 0; i < count; i++ ) {
		InformationParameterTypes type = GetInformationParameterEntryType( i );
		if ( type != InformationParameterTypes::kHidden ) {
			someUnhiddenParameter = true;
			break;
		}
	}

	if ( someUnhiddenParameter || NeedDisplayKind() ) {
		return [APlugInEditor class];
	}
	return HBaseAction::ActionController();
}

static NSString* kActionDisplayHTML = @R"RAW(
{% if params %}
	<ul class="VariableList">
		{% for param in params %}
<li>{{ param.label | htmlencode }}: {% if param.istext %}{{ param.text }}{% else %}{{ param.value | htmlencode }}{% /if %}</li>
		{% /for %}
	</ul>
{% /if %}
{{ results | htmlencode }}
)RAW";

NSString*
APlugIn::GetActionExtraResults() const
{
	NSString* results = [[NSString stringWithFormat:@"ExecuteBaseScriptResult%@", kMTextResultsWindowKindNames[int(mDisplayKind)]] localized];
	if ( mDisplayKind == MTextResultsWindowKind::kSaveToVariable ) {
		results = [NSString stringWithFormat:results, (mTrimResults ? @" trimmed " : @""), mVariable.GetNS()];
	} else if ( mDisplayKind == MTextResultsWindowKind::kSaveToClipboard ) {
		results = [NSString stringWithFormat:results, (mTrimResults ? @" trimmed " : @""), mClipboardTarget.GetDisplayName()];
	}
	return results;
}

NSString*
APlugIn::GetActionExtraHTML() const
{
	NSUInteger count = GetInformationParameterCount();
	NSMutableArray* params = [NSMutableArray new];
	for ( NSUInteger i = 0; i < count; i++ ) {
		NSString* label = GetInformationParameterEntryLabel( i );
		InformationParameterTypes type = GetInformationParameterEntryType( i );
		if ( label && type != InformationParameterTypes::kInvalid ) {
			NSString* value = mParameters[label];
			value = value ? value : @"";
			bool istext;
			switch ( type ) {
				case InformationParameterTypes::kInvalid:
					check( false );
					[[clang::fallthrough]];
				case InformationParameterTypes::kString:
				case InformationParameterTypes::kCheckbox:
				case InformationParameterTypes::kPopupMenu:
				case InformationParameterTypes::kTokenString:
				case InformationParameterTypes::kHidden:
				case InformationParameterTypes::kCalculation:
					istext = false;
					break;
				case InformationParameterTypes::kText:
				case InformationParameterTypes::kTokenText:
					istext = true;
					break;
			}
			[params addObject:@{@"label" : label, @"value" : value, @"text" : UserTextForHTML( value ), @"istext" : @(istext)}];
		}
	}



	NSDictionary *variables = @{
							@"params" : params,
							@"results" : GetActionExtraResults(),
						};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}


NSArray*
APlugIn::ListPlugIns()
{
	NSString* plugInFolderPath = KMPreferences::FilePath( kPlugInFolderName );

	NSMutableArray* folders = [NSMutableArray new];
	NSFileManager* fileManager = [NSFileManager defaultManager];
	for ( NSString* foldername in [fileManager contentsOfDirectoryAtPath:plugInFolderPath error:NULL] ) {
		NSString* informationPath = [[plugInFolderPath stringByAppendingPathComponent:foldername] stringByAppendingPathComponent:kInformationPlistName];
		if ( [informationPath existsPath] ) {
			if ( InformationForFolderName( foldername ) ) {
				[folders addObject:foldername];
			}
		}
	}
	return folders;
}

Boolean
APlugIn::UnzipLikelyContains( NSString* output )
{
	return [output containsString:kInformationPlistName];
}

Boolean
APlugIn::ImportUnzipped( NSString* importPath )
{
	Boolean good = false;
	NSString* foldername = [importPath lastPathComponent];
	NSString* plugInFolderPath = KMPreferences::FilePath( kPlugInFolderName );
	[plugInFolderPath createFolder];
	NSString* folderPath = [plugInFolderPath stringByAppendingPathComponent:foldername];
	NSString* oldFile = [folderPath renameFile:[folderPath stringByAppendingString:[NSString uuidString]]];
	if ( ![folderPath existsPath] ) {
		NSFileManager* fileManager = [NSFileManager defaultManager];
		if ( [fileManager moveItemAtPath:importPath toPath:folderPath error:NULL] ) {
			if ( InformationForFolderName( foldername ) ) {
				[oldFile trashFile];
				[[NSNotificationCenter defaultCenter] postNotificationName:kReloadThirdPartyActionsNotification object:nil];
				good = true;
			} else {
				[folderPath trashFile];
				XErrorBeep( @"Imported action invalid" );
			}
		} else {
			check( false );
		}
	} else {
		XErrorBeep( @"Imported action already exists" );
	}
	return good;
}

#endif

#pragma mark streaming

void
APlugIn::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("PlugInFolderName"), CFBridgingToCFString(mFolderName) );
	dictionary->Set( CFSTR("PlugInParameters"), (CFDictionaryRef)CFBridgingToCF(mParameters) );
	dictionary->Set( CFSTR("DisplayKind"), kMTextResultsWindowKindNames[int(mDisplayKind)] );
	dictionary->SetBoolean( CFSTR("TrimResultsNew"), mTrimResults );
	if ( mDisplayKind == MTextResultsWindowKind::kSaveToVariable ) {
		dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	}
	if ( mDisplayKind == MTextResultsWindowKind::kSaveToClipboard ) {
		mClipboardTarget.SaveToDictionary( dictionary, CFSTR("ClipboardTarget") );
	}
}

void
APlugIn::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef value;
	dictionary->Get( CFSTR("PlugInFolderName"), value, CFSTR("") );
	XAssert( value );
	mFolderName = value->GetNS();
	if ( !InformationForFolderName( mFolderName ) ) {
		XErrorHandlerMaybeLogFormat1("always", @"Invalid Plug In Action with Name %@", mFolderName);
	}
	CFDictionaryRef params;
	dictionary->Get( CFSTR("PlugInParameters"), params, CFDictionaryRef(NULL) );
	XAssert( params );
	mParameters = CFBridgingToNS(params);

	mDisplayKind = dictionary->GetEnumClass( CFSTR("DisplayKind"), kMTextResultsWindowKindNames, GetInformationDefaultDisplayKind() );
	CFStringRef variable;
	dictionary->Get( CFSTR("Variable"), variable, CFSTR("Variable") );
	check( variable );
	mVariable = variable->Copy();
	dictionary->GetBoolean( CFSTR("TrimResultsNew"), mTrimResults, true );
	mClipboardTarget.InitializeFromDictionary( dictionary, CFSTR("ClipboardTarget") );

	UpdateToDefaults();
}
