/*
 *  APressButton.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Thu Apr 24 2003.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "KMDefines.h"
#import "APressButton.h"
#import "XException.h"
#import "XAXApplicationElement.h"
#import "XAXWindowElement.h"
#import "XCoreFoundation.h"
#import "XContainer.h"
#import "XErrorHandler.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "KMVariables.h"
#if KMF_IS_ENGINE
#import "KMVariables.h"
#import "KMNotifications.h"
#import "Growl/GrowlApplicationBridge.h"
#import "HExecuteActionListManager.h"
#endif
#if KMF_EDITOR_WINDOW
#import "APressButtonEditor.h"
#import "TemplateEngine.h"
#endif

APressButton::APressButton()
	:	HBaseAction(),
		mButtonName( CFSTR("")->Copy() ),
		mStopOnFailure( kDefaultPressButtonStopOnFailure )
{
}

APressButton::APressButton( const APressButton& inAction )
	:	HBaseAction( inAction ),
		mButtonName( inAction.mButtonName ),
		mStopOnFailure( inAction.mStopOnFailure )
{
}
	
APressButton::~APressButton()
{
}

#if KMF_EDITOR_WINDOW
void
APressButton::UpdateToDefaultValues()
{
	if ( IsEmpty( mButtonName->GetNS() ) ) {
		mButtonName = CFSTR("kUntitledButton")->CopyLocalized();
	}
}
#endif

CFStringRef
APressButton::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> titleFormat = CFSTR("aPressButton_DescriptionFormat")->CopyLocalized();
	CFContainer<CFStringRef> name = CopyButtonName();
	return CFString::CreateWithFormat( titleFormat, CFStringRef(name) );
}

#if KMF_IS_ENGINE

static bool
FindAndPressButtonRecursive( NSString* mButtonName, const XAXUIElement& element, bool &foundDisabledButton )
{
	CFContainer<CFArrayRef> theChildrenArray = (CFArrayRef)element.CopyAttributeValue( kAXChildrenAttribute );
	if ( theChildrenArray ) {
		for ( CFIndex i = 0; i < theChildrenArray->GetCount(); i++ ) {
			try {
				XAXUIElement theElement( (AXUIElementRef)theChildrenArray->Get( i ) );
				
				CFContainer<CFStringRef> theRole = (CFStringRef)theElement.CopyAttributeValue( kAXRoleAttribute );
				XAssert( theRole );
//					NSLog( @"Role %@", CFStringRef(theRole) );
//					theElement.DumpAttribueNames();
				if ( theRole->IsEqual( kAXButtonRole ) || theRole->IsEqual( kAXRadioButtonRole ) || theRole->IsEqual( kAXCheckBoxRole ) || theRole->IsEqual( kAXMenuButtonRole ) ) {
//						CFTypeRef result;
//						OSStatus err = AXUIElementCopyAttributeValue( theElement, CFSTR("AXTitle"), &result );
//						fprintf( stderr, "AXUIElementCopyAttributeValue " PRINTSInt32 "\n", err );
//						if ( err == noErr ) {
//							NSLog( @"Attribute value %@", result );
//						}
					CFContainer<CFStringRef> theElementName = theElement.CopyTitle( NULL );
//					NSLog( @"Name %@", CFStringRef(theElementName) );
					if ( theElementName && [theElementName->GetNS() vaguelyEqualToString:mButtonName] ) {
						CFContainer<CFBooleanRef> isEnabled = theElement.CopyAttributeValueAsBoolean( kAXEnabledAttribute );
						if ( isEnabled.Defined() && !isEnabled->Get() ) {
							foundDisabledButton = true;
						} else {
							OSStatus err = theElement.PerformAction( kAXPressAction );
							verify_noerr( err );
							if ( err == noErr ) {
								return true;
							}
						}
					}
				}
				if ( FindAndPressButtonRecursive( mButtonName, theElement, foundDisabledButton ) ) {
					return true;
				}
			} catch ( ... ) {
			}
		}
	}
	return false;
}

float
APressButton::DoAction()
{
	float time = 0.0;
	NSString* cancelMessage = @"PressButtonWindowNotFound";
	if ( AXAPIEnabled() ) {
		XAXApplicationElement theFrontApp;
        if ( theFrontApp ) {
            XAXWindowElement windowElement( theFrontApp.GetFocussedOrMainWindow() );
            if ( !!windowElement ) {
				bool foundDisabledButton = false;
                if ( FindAndPressButtonRecursive( mButtonName->GetNS(), windowElement, foundDisabledButton ) ) {
                    cancelMessage = nil;
                    time = 5.0;
                } else if ( foundDisabledButton ) {
					cancelMessage = @"PressButtonDisabled";
				} else {
					cancelMessage = @"PressButtonButtonNotFound";
				}
            }
        }
	} else {
		cancelMessage = @"PressButtonRequiresAccess";
	}

	if ( cancelMessage ) {
		cancelMessage = [NSString stringWithFormat:[cancelMessage localized], mButtonName->GetNS() ];
		XErrorHandlerMaybeLogFormat1( "always", CFSTR("%@"), cancelMessage );
	}
	if ( !mStopOnFailure ) {
		cancelMessage = nil;
	}
	if ( cancelMessage != nil ) {
		HExecuteActionListManager::CancelThisExecutionGroup();
	}
	KMVariables::Set( kPressButtonResultVariable, cancelMessage ? cancelMessage : @"OK", KMNotifications::kLoadVariableFromEngineAction );
	if ( cancelMessage ) {
		[GrowlApplicationBridge 
			 notifyWithTitle:CFBridgingToNS(kmEngineName)
			 description:[NSString stringWithFormat:[@"MacroCancelledFormat" localized], cancelMessage]
			 notificationName:@"Macro Cancelled"
			 iconData:nil priority:0 isSticky:NO clickContext:nil];		
		XErrorHandlerMaybeLogMessage( "display", ([NSString stringWithFormat:[@"MacroCancelledFormat" localized], cancelMessage]) );
	}
	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
APressButton::ActionController() const
{
	return [APressButtonEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
{% if stop %}
 Stop macro if button cannot be pressed.<br />
{% /if %}
 Store action result in variable ‘{{ result | htmlencode }}’.
)RAW";


NSString*
APressButton::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
							@"stop" : @(mStopOnFailure),
							@"result" : kPressButtonResultVariable,
						};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}


#endif

#pragma mark streaming

void
APressButton::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("ButtonName"), CFStringRef(mButtonName) );
	dictionary->SetBoolean( CFSTR("StopOnFailure"), mStopOnFailure );
}

void
APressButton::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef newButtonName;
	dictionary->Get( CFSTR("ButtonName"), newButtonName, CFSTR("") );
	check( newButtonName );
	mButtonName = newButtonName->Copy();

	dictionary->GetBoolean( CFSTR("StopOnFailure"), mStopOnFailure, kDefaultPressButtonStopOnFailure );
}

