//
//  AProgramSwitcher.cpp
//  Keyboard Maestro
//
//  Created by Michael Kamprath.
//  Copyright (c) 2004 Stairways Software. All rights reserved.
//

#import "AProgramSwitcher.h"
#import "NSColorExtensions.h"

#if KMF_IS_ENGINE
#import "PSHUDController.h"
#import "HExecuteActionListManager.h"
#endif
#if KMF_IS_EDITOR
#import "AProgramSwitcherEditor.h"
#import "TemplateEngine.h"
#import "NSImageExtensions.h"
#endif

enum {
	kDefaultIconSize = 64,
	kDefaultAlpha = 50,
};

static const RGBColor kDefaultColor = {0, 128, 128};

AProgramSwitcher::AProgramSwitcher()
	:	HBaseAction(),
		mColor( [NSColor colorWithCalibratedRed:kDefaultColor.red/255.0 green:kDefaultColor.green/255.0 blue:kDefaultColor.blue/255.0 alpha:kDefaultAlpha/100.0] ),
		mIconSize( kDefaultIconSize )
{
}

void AProgramSwitcher::SetIconSize( SInt16 inSize )
{
	check( ( inSize >= 16 ) && ( inSize <= 128 ) );
	if ( ( inSize >= 16 ) && ( inSize <= 128 ) ) {
		mIconSize = inSize;
	}
}

void AProgramSwitcher::SetSwitcherColor( NSColor* inColor )
{
	mColor = [[inColor colorUsingColorSpaceName:NSCalibratedRGBColorSpace] copy];
}

#if KMF_IS_ENGINE

float AProgramSwitcher::DoAction()
{
	return [PSHUDController TriggerFired:this modifiers:HExecuteActionListManager::GetTrippedModifiers()] ? 5 : 0.5;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AProgramSwitcher::ActionController() const
{
	return [AProgramSwitcherEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {{ display | htmlencode }}<br />
 Icon size: {{ iconsize | htmlencode }}<br />
 <span class="IconTitle">Color: </span><img src={{ color | inlineimage: 16 }} alt="Swatch" width="16" height="16" /><br />
 {{ sort | htmlencode }}<br />
 {% if placeundermouse %}Place launcher under mouse.<br />{% /if %}
 {% if hideothers %}Hide other applications when switching.<br />{% /if %}
 {% if rollover %}Mouse rollover enabled.<br />{% /if %}
 {% if includecount > 0 %}
 Always include these applications:<br />
	{{ include }}
 {% /if %}
)RAW";

NSString*
AProgramSwitcher::GetActionExtraHTML() const
{
	NSString* display;
	switch ( mTheme ) {
		case ProgramSwitcherWindowType::kVerticalList:
			display = @"Display in vertical list.";
			break;
		case ProgramSwitcherWindowType::kHorizontalIcons:
			display = @"Display horizontal icons.";
			break;
		case ProgramSwitcherWindowType::kIconGrid:
			display = @"Display in icon grid.";
			break;
		case ProgramSwitcherWindowType::kCoverFlow:
			display = @"Display in Cover Flow.";
			break;
	}
	NSString* sort;
	switch ( mAppSortOrder ) {
		case kPSApplicationSortOrderByAlphabetical:
			sort = @"Sort alphabetically.";
			break;
		case kPSApplicationSortOrderByDepricatedLaunch:
			check( false );
			[[clang::fallthrough]];
		case kPSApplicationSortOrderByUse:
			sort = @"Sort by use order.";
			break;
	}

	NSDictionary *variables = @{
									@"display" : display,
									@"sort" : sort,
									@"iconsize" : @(mIconSize),
									@"placeundermouse" : @(mSwitcherUnderMouse),
									@"hideothers" : @(mHideOthers),
									@"rollover" : @(mMouseRollover),
									@"color" : [NSImage imageWithSwatch:mColor],
									@"includecount" : @(mIncludedApps.GetCount()),
									@"include" : mIncludedApps.GetDisplayHTMLContent(),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kBasePSApplicationSortOrderNames[] = kBadPSApplicationSortOrderNamesArray;
static const CFStringRef kGoodPSApplicationSortOrderNames[] = kGoodPSApplicationSortOrderNamesArray;
static const CFStringRef kProgramSwitcherWindowTypeNames[] = kProgramSwitcherWindowTypeNamesArray;

void
AProgramSwitcher::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("SortOrder"), kGoodPSApplicationSortOrderNames[mAppSortOrder] );
	dictionary->Set( CFSTR("WindowType"), kProgramSwitcherWindowTypeNames[int(mTheme)] );
	[mColor saveWithAlphaToDictionary:dictionary];
	dictionary->SetNumber( CFSTR("IconSize"), mIconSize );
	dictionary->SetBoolean( CFSTR("SwitcherUnderMouse"), mSwitcherUnderMouse );
	dictionary->SetBoolean( CFSTR("HideOthers"), mHideOthers );
	dictionary->SetBoolean( CFSTR("MouseRollover"), mMouseRollover );

	SafeCFPropertyListRef apps = mIncludedApps.CopyToPropertyList();
	dictionary->Set( CFSTR("IncludedApplications"), apps );
	CFRelease( apps );
}

void
AProgramSwitcher::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	int sortorder = dictionary->GetEnumerated( CFSTR("SortOrder"), kGoodPSApplicationSortOrderNames, -1 );
	if ( sortorder < 0 ) {
		sortorder = dictionary->GetEnumerated( CFSTR("SortOrder"), kBasePSApplicationSortOrderNames, kPSApplicationSortOrderDEFAULT );
	}
	mAppSortOrder = (PSApplicationSortOrderType)sortorder;
	if ( mAppSortOrder == kPSApplicationSortOrderByDepricatedLaunch ) {
		mAppSortOrder = kPSApplicationSortOrderDEFAULT;
	}
	mTheme = dictionary->GetEnumClass( CFSTR("WindowType"), kProgramSwitcherWindowTypeNames, ProgramSwitcherWindowType::kDEFAULT );
	mColor = [[NSColor colorFromDictionary:dictionary defaultColor:kDefaultColor defaultAlpha:kDefaultAlpha] copy];
	SInt16 n;
	if ( !dictionary->Get( CFSTR("IconSize"), n ) ) n = kDefaultIconSize;
	SetIconSize( n );
	dictionary->GetBoolean( CFSTR("SwitcherUnderMouse"), mSwitcherUnderMouse, false );
	dictionary->GetBoolean( CFSTR("HideOthers"), mHideOthers, false );
	dictionary->GetBoolean( CFSTR("MouseRollover"), mMouseRollover, true );
	
	SafeCFPropertyListRef apps = (SafeCFPropertyListRef)dictionary->Get( CFSTR("IncludedApplications") );
	mIncludedApps.LoadFromPropertyList( apps );
}

