//
//  APromptForUserInput.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 6 May 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "KMDefines.h"
#import "APromptForUserInput.h"
#import "KMPromptForUserInput.h"
#import "KMVariables.h"
#if KMF_IS_ENGINE
#import "TokenText.h"
#import "HExecuteActionListManager.h"
#endif
#if KMF_EDITOR_WINDOW
#import "APromptForUserInputEditor.h"
#import "TemplateEngine.h"
#endif

const CFStringRef kDefaultTitle = CFSTR("Untitled");
const CFStringRef kDefaultPrompt = CFSTR("Please enter the details for these variables.");

static NSArray* DefaultButtons()
{
	return @[
		@{ KMPromptForUserInput::kButtonKey : @"OK" },
		@{ KMPromptForUserInput::kButtonKey : @"Cancel", KMPromptForUserInput::kCancelKey : @YES },
	];
}

static void UseVariables( NSArray* variables )
{
	for ( NSDictionary* entry in variables ) {
		KMVariables::Use( [entry objectForKey:KMPromptForUserInput::kVariableKey] );
	}
}

static void UnuseVariables( NSArray* variables )
{
	for ( NSDictionary* entry in variables ) {
		KMVariables::Unuse( [entry objectForKey:KMPromptForUserInput::kVariableKey] );
	}
}

APromptForUserInput::APromptForUserInput()
	:	HBaseAction(),
		mTitle( kDefaultTitle->Copy() ),
		mPrompt( kDefaultPrompt->Copy() ),
		mVariables( [NSArray new] ),
		mResultButtonVariable( kPromptResultButtonVariable ),
		mButtons( [DefaultButtons() copy] )
{
	UseVariables( mVariables );
#if KMF_IS_ENGINE
	mPromptController = nil;
#endif
}

APromptForUserInput::APromptForUserInput( const APromptForUserInput& inAction )
	:	HBaseAction( inAction ),
		mTitle( inAction.mTitle ),
		mPrompt( inAction.mPrompt ),
		mVariables( [inAction.mVariables copy] ),
		mResultButtonVariable( inAction.mResultButtonVariable ),
		mButtons( [inAction.mButtons copy] )
{
	UseVariables( mVariables );
#if KMF_IS_ENGINE
	mPromptController = nil;
#endif
}

APromptForUserInput::~APromptForUserInput()
{
	UnuseVariables( mVariables );
#if KMF_IS_ENGINE
	[mPromptController Kill];
#endif
}

void
APromptForUserInput::SetVariables( NSArray* variables )
{
	UnuseVariables( mVariables );
	mVariables = [variables copy];
	UseVariables( mVariables );
}

CFStringRef APromptForUserInput::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> format = CFSTR("aPromptForUserInput_Format")->CopyLocalized();
	
	return CFString::CreateWithFormat( format, CFStringRef(mTitle) );
}

#if KMF_IS_ENGINE

float
APromptForUserInput::DoAction()
{
	float time = 30;
	float a;
	mPromptController = [KMPromptForUserInputController newPromptForUserInput];
	mPromptController.title = TokenText::GetResult( mTitle->GetNS(), &a ); time += a;
	mPromptController.prompt = TokenText::GetResult( mPrompt->GetNS(), &a ); time += a;
	mPromptController.variables = mVariables;
	mPromptController.resultButtonVariable = mResultButtonVariable->GetNS();
	mPromptController.buttons = mButtons;

	[mPromptController Show];
	return time;
}

Boolean
APromptForUserInput::IsActionCompleted()
{
	if ( [mPromptController IsComplete] ) {
		if ( mPromptController.cancel ) {
			HExecuteActionListManager::CancelThisExecutionGroup();
		}
		[mPromptController Kill];
		mPromptController = nil;
	}
	return mPromptController == nil;
}

void
APromptForUserInput::ActionAborted()
{
	[mPromptController Kill];
	mPromptController = nil;
}

#endif

#if KMF_EDITOR_WINDOW

Class
APromptForUserInput::ActionController() const
{
	return [APromptForUserInputEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {{ prompt }}
{% if variables %}
	Input the following variables:
	<ul class="VariableList">
		{% for variable in variables %}
			<li>{{ variable.VARIABLE | htmlencode }}{% if variable.DEFAULT %} (default to ‘{{ variable.DEFAULT | htmlencode }}’){% /if %}</li>
		{% /for %}
	</ul>
{% /if %}
{% if buttons %}
	Finish with the following buttons:
	<ul class="VariableList">
		{% for button in buttons %}
			<li>{{ button.BUTTON | htmlencode }}{% if button.CANCEL %} (cancel macro){% /if %}</li>
		{% /for %}
	</ul>
{% /if %}
 Store button pressed in variable ‘{{ result | htmlencode }}’.
)RAW";

NSString*
APromptForUserInput::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
							@"prompt" : UserTextForHTML( mPrompt->GetNS() ),
							@"variables" : mVariables,
							@"buttons" : mButtons,
							@"result" : mResultButtonVariable.GetNS(),
						};
	NSString* tmpl = kActionDisplayHTML;
	tmpl = [tmpl stringByReplacingOccurrencesOfString:@"VARIABLE" withString:KMPromptForUserInput::kVariableKey];
	tmpl = [tmpl stringByReplacingOccurrencesOfString:@"DEFAULT" withString:KMPromptForUserInput::kDefaultKey];
	tmpl = [tmpl stringByReplacingOccurrencesOfString:@"BUTTON" withString:KMPromptForUserInput::kButtonKey];
	tmpl = [tmpl stringByReplacingOccurrencesOfString:@"CANCEL" withString:KMPromptForUserInput::kCancelKey];
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:tmpl withVariables:variables];
}

#endif

#pragma mark streaming

void
APromptForUserInput::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Title"), CFStringRef(mTitle) );
	dictionary->Set( CFSTR("Prompt"), CFStringRef(mPrompt) );
	dictionary->Set( CFSTR("Variables"), CFBridgingToCF(mVariables) );
	dictionary->Set( CFSTR("ResultButtonVariable"), CFStringRef(mResultButtonVariable) );
	dictionary->Set( CFSTR("Buttons"), CFBridgingToCF(mButtons) );
}

void
APromptForUserInput::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef newTitle;
	dictionary->Get( CFSTR("Title"), newTitle, kDefaultTitle );
	check( newTitle );
	mTitle = newTitle->Copy();
	
	CFStringRef newPrompt;
	dictionary->Get( CFSTR("Prompt"), newPrompt, kDefaultPrompt );
	check( newPrompt );
	mPrompt = newPrompt->Copy();
	
	CFStringRef newResultButtonVariable;
	dictionary->Get( CFSTR("ResultButtonVariable"), newResultButtonVariable, CFBridgingToCFString(kPromptResultButtonVariable) );
	check( newResultButtonVariable );
	mResultButtonVariable = newResultButtonVariable->Copy();
	
	CFArrayRef newVariables;
	dictionary->Get( CFSTR("Variables"), newVariables, CFArrayRef(NULL) );
	if ( newVariables ) {
		SetVariables( CFBridgingToNS(newVariables) );
	}
	
	CFArrayRef newButtons;
	dictionary->Get( CFSTR("Buttons"), newButtons, (CFArrayRef)CFBridgingToCF(DefaultButtons()) );
	if ( newButtons ) {
		SetButtons( CFBridgingToNS(newButtons) );
	}
}

