/*
 *  AQuickMacro.cpp
 *  Keyboard Maestro
 *
 *  Created by Peter N Lewis on Mon 28 Jan 2008.
 *  Copyright (c) 2008 Stairways Software. All rights reserved.
 */

#import "AQuickMacro.h"
#import "KMDefines.h"
#import "XCoreFoundation.h"
#import "SystemDefinedKeys.h"
#if KMF_IS_ENGINE
#import "QMRecorderDomain.h"
#endif
#if KMF_EDITOR_WINDOW
#import "AQuickMacroEditor.h"
#import "XErrorHandler.h"
#import "MMacroPreferenceFile.h"
#import "TemplateEngine.h"
#endif

const int kDefaultFnKey = 1;
const UInt32 kDefaultModifiers = optionKey;

static NSMutableSet* gQuickMacroUIDs = nil;

void
AQuickMacro::ResetQuickMacroUIDSet()
{
	gQuickMacroUIDs = nil;
}

AQuickMacro::AQuickMacro()
	:	HBaseAction(),
		mQuickMacroUID( CFCS( CFString::CreateUUID() ) ),
		mKeyCode( SystemDefinedKeys::FnKeyCode( kDefaultFnKey ) ),
		mModifiers( kDefaultModifiers ),
		mEnableRepeat( false ),
		mAddToMacroPalette( false ),
		mAddToStatusMenu( false )
{
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		gQuickMacroUIDs = [NSMutableSet new];
	});
	[gQuickMacroUIDs addObject:mQuickMacroUID->GetNS()];
}

AQuickMacro::AQuickMacro( const AQuickMacro& inAction )
	:	HBaseAction( inAction ),
		mQuickMacroUID( inAction.mQuickMacroUID ),
		mKeyCode( inAction.mKeyCode ),
		mModifiers( inAction.mModifiers ),
		mEnableRepeat( inAction.mEnableRepeat ),
		mAddToMacroPalette( inAction.mAddToMacroPalette ),
		mAddToStatusMenu( inAction.mAddToStatusMenu )
{
	[gQuickMacroUIDs addObject:mQuickMacroUID->GetNS()];
}

AQuickMacro::~AQuickMacro()
{
}

#if KMF_EDITOR_WINDOW

void
AQuickMacro::UpdateForPasteOrAdd()
{
	mQuickMacroUID = CFString::CreateUUID();
	[gQuickMacroUIDs addObject:mQuickMacroUID->GetNS()];
}

#endif

void
AQuickMacro::SetKeystroke( const CGKeyCode& inKeyCode, const UInt32& inModifiers, Boolean inRepeat )
{
	mKeyCode = inKeyCode;
	mModifiers = inModifiers;
	mEnableRepeat = inRepeat;
}

CFStringRef
AQuickMacro::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> description;

	CFContainer<CFStringRef> key = SystemDefinedKeys::CopyKeystrokeText( mKeyCode, mModifiers, KeyTextFormat::kLong );
	CFContainer<CFStringRef> format = CFSTR("aQuickMacro_DescriptionFormat")->CopyLocalized();
	description = CFString::CreateWithFormat( format, CFStringRef(key) );

	return description->Copy();
}


#if KMF_IS_ENGINE

float
AQuickMacro::DoAction()
{
	QMRecorderDomain::ToggleRecorder( mQuickMacroUID );
	return 30;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AQuickMacro::ActionController() const
{
	return [AQuickMacroEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {% if repeat %}Enable key repeat.<br />{% /if %}
 {% if palette %}Add to global macro palette.<br />{% /if %}
 {% if menu %}Add to status menu.<br />{% /if %}
)RAW";

NSString*
AQuickMacro::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"repeat" : @(mEnableRepeat),
									@"palette" : @(mAddToMacroPalette),
									@"menu" : @(mAddToStatusMenu),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

void
AQuickMacro::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("QuickMacroUID"), CFStringRef(mQuickMacroUID) );
	dictionary->SetNumber( CFSTR("KeyCode"), mKeyCode );
	dictionary->SetNumber( CFSTR("Modifiers"), mModifiers );
	dictionary->SetBoolean( CFSTR("EnableRepeat"), mEnableRepeat );
	dictionary->SetBoolean( CFSTR("AddToMacroPalette"), mAddToMacroPalette );
	dictionary->SetBoolean( CFSTR("AddToStatusMenu"), mAddToStatusMenu );
}

void
AQuickMacro::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	XAssert( dictionary->Get( CFSTR("KeyCode"), mKeyCode ) );
	XAssert( dictionary->Get( CFSTR("Modifiers"), mModifiers ) );
	dictionary->GetBoolean( CFSTR("EnableRepeat"), mEnableRepeat, true );
	dictionary->GetBoolean( CFSTR("AddToMacroPalette"), mAddToMacroPalette, false );
	dictionary->GetBoolean( CFSTR("AddToStatusMenu"), mAddToStatusMenu, false );

	CFStringRef newUID;
	if ( dictionary->Get( CFSTR("QuickMacroUID"), newUID ) ) {
		mQuickMacroUID = newUID->Copy();
	} else if ( dictionary->Get( CFSTR("UID"), newUID ) ) { // Old style
		mQuickMacroUID = newUID->Copy();
	}
#if KMF_EDITOR_WINDOW
	if ( [gQuickMacroUIDs containsObject:mQuickMacroUID->GetNS()] ) {
		XErrorHandlerMaybeLogMessage( "UID", CFSTR("Updating Quick Macro UID to avoid conflict") );
		mQuickMacroUID = CFString::CreateUUID();
		MMacroPreferenceFile::SetLoadAllMacrosNeedsSave();
	}
#endif
	check( ![gQuickMacroUIDs containsObject:mQuickMacroUID->GetNS()] );
	[gQuickMacroUIDs addObject:mQuickMacroUID->GetNS()];
}
