//
//  AQuickTimePlayerControl.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 18 Sep 2009.
//  Copyright (c) 2009 Stairways Software. All rights reserved.
//

#import "AQuickTimePlayerControl.h"
#import "KMIconRegistry.h"
#import "XException.h"
#import "XCoreFoundation.h"
#import "NSStringExtensions.h"
#if KMF_IS_ENGINE
#import "NSAppleScriptExtensions.h"
#import "Calculate.h"
#import "TokenText.h"
#import "XErrorHandler.h"
#endif
#if KMF_EDITOR_WINDOW
#import "AQuickTimePlayerControlEditor.h"
#import "AQuickTimePlayerVolumeControlEditor.h"
#endif


AQuickTimePlayerControl::AQuickTimePlayerControl()
	:	HBaseAction(),
		mCommand( QuickTimePlayerControlCommand::kDEFAULT ),
		mArgumentString( CFSTR( "" )->Copy() )
{
}

AQuickTimePlayerControl::AQuickTimePlayerControl( const AQuickTimePlayerControl& inAction )
	:	HBaseAction( inAction ),
		mCommand( inAction.mCommand ),
		mArgumentString( inAction.mArgumentString )
{
}

AQuickTimePlayerControl::AQuickTimePlayerControl( const QuickTimePlayerControlCommand& inCommand )
	:	HBaseAction(),
		mCommand( inCommand )
{
	switch ( inCommand ) {
		case QuickTimePlayerControlCommand::kStepForward:
		case QuickTimePlayerControlCommand::kStepBackward:
			mArgumentString = CFSTR("100");
			break;
		case QuickTimePlayerControlCommand::kSetVolume:
			mArgumentString = CFSTR("50");
			break;
		case QuickTimePlayerControlCommand::kPlay:
		case QuickTimePlayerControlCommand::kPlayPause:
		case QuickTimePlayerControlCommand::kPause:
		case QuickTimePlayerControlCommand::kIncreaseVolume:
		case QuickTimePlayerControlCommand::kDecreaseVolume:
			mArgumentString = CFSTR("");
			break;
	}
}

AQuickTimePlayerControl::~AQuickTimePlayerControl()
{
}

static NSString*
ActionDescriptionString( QuickTimePlayerControlCommand command, bool newaction )
{
	NSString* theTitleKey;
	switch ( command ) {
		case QuickTimePlayerControlCommand::kPlay:
			theTitleKey = @"QuickTimePlayerCmd_Play_Title";
			break;
		case QuickTimePlayerControlCommand::kPlayPause:
			theTitleKey = @"QuickTimePlayerCmd_PlayPause_Title";
			break;
		case QuickTimePlayerControlCommand::kPause:
			theTitleKey = @"QuickTimePlayerCmd_Pause_Title";
			break;
		case QuickTimePlayerControlCommand::kStepForward:
			theTitleKey = newaction ? @"QuickTimePlayerCmd_StepForward_ShortTitle" : @"QuickTimePlayerCmd_StepForward_Title";
			break;
		case QuickTimePlayerControlCommand::kStepBackward:
			theTitleKey = newaction ? @"QuickTimePlayerCmd_StepBackward_ShortTitle" : @"QuickTimePlayerCmd_StepBackward_Title";
			break;
		case QuickTimePlayerControlCommand::kIncreaseVolume:
			theTitleKey = newaction ? @"QuickTimePlayerCmd_IncreaseVolume_ShortTitle" : @"QuickTimePlayerCmd_IncreaseVolume_Title";
			break;
		case QuickTimePlayerControlCommand::kDecreaseVolume:
			theTitleKey = newaction ? @"QuickTimePlayerCmd_DecreaseVolume_ShortTitle" : @"QuickTimePlayerCmd_DecreaseVolume_Title";
			break;
		case QuickTimePlayerControlCommand::kSetVolume:
			theTitleKey = newaction ? @"QuickTimePlayerCmd_SetVolume_Title" : @"QuickTimePlayerCmd_SetVolume_TitleFormat";
			break;
	}
	return [theTitleKey localized];
}


CFStringRef AQuickTimePlayerControl::CopyActionDescriptionString() const
{
	NSString* theTitle = ActionDescriptionString( mCommand, false );
	if ( [theTitle containsString:@"%@"] ) {
		theTitle = [NSString stringWithFormat:theTitle, mArgumentString->GetNS()];
	} else if ( [theTitle containsString:@"%d"] ) {
		theTitle = [NSString stringWithFormat:theTitle, mArgumentString->GetIntValue()];
	}
	return (CFStringRef)CFBridgingRetain( theTitle );
}

NSString*
AQuickTimePlayerControl::GetNewActionTitle() const
{
	return ActionDescriptionString( mCommand, true );
}

XIcon
AQuickTimePlayerControl::GetActionXIcon() const
{
	KMIconRegistry iconSignature;
	
	switch ( mCommand ) {
		case QuickTimePlayerControlCommand::kPlay:
			iconSignature = KMIconRegistry::kiTunesPlayIcon;
			break;
		case QuickTimePlayerControlCommand::kPlayPause:
			iconSignature = KMIconRegistry::kiTunesPlayIcon;
			break;
		case QuickTimePlayerControlCommand::kPause:
			iconSignature = KMIconRegistry::kiTunesPauseIcon;
			break;
		case QuickTimePlayerControlCommand::kStepForward:
			iconSignature = KMIconRegistry::kiTunesNextIcon;
			break;
		case QuickTimePlayerControlCommand::kStepBackward:
			iconSignature = KMIconRegistry::kiTunesPreviousIcon;
			break;
		case QuickTimePlayerControlCommand::kIncreaseVolume:
		case QuickTimePlayerControlCommand::kDecreaseVolume:
		case QuickTimePlayerControlCommand::kSetVolume:
			iconSignature = KMIconRegistry::kSoundIcon;
			break;
	}
	return XIcon( iconSignature );
}

void
AQuickTimePlayerControl::SetCommand( QuickTimePlayerControlCommand inCommand )
{
	mCommand = inCommand;
}

QuickTimePlayerControlCommand
AQuickTimePlayerControl::GetCommand() const
{
	return mCommand;
}

void
AQuickTimePlayerControl::SetArgumentString( CFStringRef inString )
{
	check( inString );
	mArgumentString = inString->Copy();
}

void
AQuickTimePlayerControl::SetArgumentInteger( NSInteger inValue )
{
	mArgumentString = CFCS( CFString::CreateWithNumber( inValue ) );
}

CFStringRef
AQuickTimePlayerControl::CopyArgumentString() const
{
	return mArgumentString->Copy();
}

#if KMF_IS_ENGINE

#define kTellQuickTimePlayer "tell application \"QuickTime Player\"\r"
#define kEndTell "\r end tell \r"

float
AQuickTimePlayerControl::DoAction()
{
	float time = 0;
	CFContainer<CFStringRef> scriptBase;
	
	SInt32 argument = mArgumentString->GetIntValue();
	
	CFStringRef direction = CFSTR("backward");
	switch ( mCommand ) {
		case QuickTimePlayerControlCommand::kPlay:
			scriptBase = CFSTR( kTellQuickTimePlayer "play document 1" kEndTell );
			break;
		case QuickTimePlayerControlCommand::kPlayPause:
			scriptBase = CFSTR( kTellQuickTimePlayer "if document 1 is playing then \r stop document 1 \r else \r play document 1 \r end if" kEndTell );
			break;
		case QuickTimePlayerControlCommand::kPause:
			scriptBase = CFSTR( kTellQuickTimePlayer "stop document 1" kEndTell );
			break;
		case QuickTimePlayerControlCommand::kStepForward:
			direction = CFSTR("forward");
			[[clang::fallthrough]];
		case QuickTimePlayerControlCommand::kStepBackward: {
			scriptBase = CFString::CreateWithFormat( CFSTR( kTellQuickTimePlayer "step %@ document 1 by %%d" kEndTell ), direction );
			double doubleargument;
			if ( !Calculate::CalculateValue( mArgumentString->GetNS(), doubleargument ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Quick Time Player Control %@ failed"), mArgumentString->GetNS() );
				doubleargument = 1;
			}
			
			argument = lround(doubleargument);
			break;
		}
		case QuickTimePlayerControlCommand::kIncreaseVolume:
			scriptBase = CFSTR( kTellQuickTimePlayer "set a to audio volume of document 1 \r if a <= 0 then \r set audio volume of document 1 to 0.1 \r else if a >= 0.9 then \r set audio volume of document 1 to 1 \r else \r set audio volume of document 1 to (audio volume of document 1) + 0.1 \r end if " kEndTell );
			break;
		case QuickTimePlayerControlCommand::kDecreaseVolume:
			scriptBase = CFSTR( kTellQuickTimePlayer "set a to audio volume of document 1 \r if a <= 0.1 then \r set audio volume of document 1 to 0 \r else if a >= 1 then \r set audio volume of document 1 to 0.9 \r else \r set audio volume of document 1 to (audio volume of document 1) - 0.1 \r end if " kEndTell );
			break;
		case QuickTimePlayerControlCommand::kSetVolume:
			scriptBase = CFSTR( kTellQuickTimePlayer "set audio volume of document 1 to %d / 100" kEndTell );
			break;
	}
	if ( scriptBase->Contains( CFSTR("%@") ) ) {
		NSString* arg = TokenText::GetResult( mArgumentString->GetNS(), &time );
		(void)[NSAppleScript executeWithSource:scriptBase->GetNS() param:arg];
	} else if ( scriptBase->Contains( CFSTR("%d") ) ) {
		(void)[NSAppleScript executeWithSource:scriptBase->GetNS() numericParameter:argument];
	} else {
		(void)[NSAppleScript executeWithSource:scriptBase->GetNS()];
	}

	return time + 10;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AQuickTimePlayerControl::ActionController() const
{
	switch ( mCommand ) {			
		case QuickTimePlayerControlCommand::kStepForward:
		case QuickTimePlayerControlCommand::kStepBackward:
			return [AQuickTimePlayerControlEditor class];
		case QuickTimePlayerControlCommand::kSetVolume:
			return [AQuickTimePlayerVolumeControlEditor class];
		case QuickTimePlayerControlCommand::kPlay:
		case QuickTimePlayerControlCommand::kPlayPause:
		case QuickTimePlayerControlCommand::kPause:
		case QuickTimePlayerControlCommand::kIncreaseVolume:
		case QuickTimePlayerControlCommand::kDecreaseVolume:
			return HBaseAction::ActionController();
	}
}

#endif

#pragma mark streaming

static const CFStringRef kQuickTimePlayerControlCommandNames[] = kQuickTimePlayerControlCommandNamesArray;

void
AQuickTimePlayerControl::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Command"), kQuickTimePlayerControlCommandNames[int(mCommand)] );
	dictionary->Set( CFSTR("Argument"), CFStringRef(mArgumentString) );
}

void
AQuickTimePlayerControl::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	bool good = dictionary->CanGetEnumClass( CFSTR("Command"), kQuickTimePlayerControlCommandNames, mCommand );
	XAssert( good );

	CFStringRef newArgumentString;
	dictionary->Get( CFSTR("Argument"), newArgumentString, CFSTR("") );
	check( newArgumentString );
	mArgumentString = newArgumentString->Copy();
}
