//
//  AReadFile.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 5 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AReadFile.h"
#import "FileReference.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#import "CSClipboardDomain.h"
#import "NSStringRegularExpressions.h"
#import "XErrorHandler.h"

#if KMF_IS_ENGINE
#import "KMVariables.h"
#import "KMNotifications.h"
#endif

#if KMF_EDITOR_WINDOW
#import "AReadFileEditor.h"
#import "TemplateEngine.h"
#endif

CFStringRef
AReadFile::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> format;
	CFContainer<CFStringRef> result;
	switch ( mDestination ) {
		case ReadFileDestinationType::kClipboard: {
			result = CFSTR("kmReadFileToClipboardTitleFormat")->CopyLocalized();
			break;
		}
		case ReadFileDestinationType::kNamedClipboard: {
			format = CFSTR("kmReadFileToNamedClipboardTitleFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, mClipboardReference.GetClipboardName() );
			break;
		}
		case ReadFileDestinationType::kVariable: {
			format = CFSTR("kmReadFileToVariableTitleFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, CFStringRef(mVariable) );
			break;
		}
	}
	return result->Copy();
}

#if KMF_IS_ENGINE

float
AReadFile::DoAction()
{
	float time = 10;
	NSString* source = FileReference::PathToActualPath( mSource->GetNS() );
	if ( ![source isMatchedByRegex:@"^/.+"] ) {
		XErrorHandlerMaybeLogMessage( "display", CFSTR("No file for Read File action") );
		return 0;
	}

	switch ( mDestination ) {
		case ReadFileDestinationType::kClipboard: {
			XClipboardHistory::PlaceInClipboard( XClipboardDataSet( XClipboardDataSetSourceKind::kFileWithPath, source ) );
			break;
		}
		case ReadFileDestinationType::kNamedClipboard: {
			mClipboardReference.UpdateClipboardReference();
			CSClipboardDataSet*	pClip = mClipboardReference.GetClipboardDataSet();
			if ( pClip ) {
				*pClip = XClipboardDataSet( XClipboardDataSetSourceKind::kFileWithPath, source );
				CSClipboardDomain::GetClipboardDomain()->SaveClipData( KMNotifications::kLoadClipDataFromEngineAction );
			}
			break;
		}
		case ReadFileDestinationType::kVariable: {
			NSStringEncoding encoding;
			NSString* result = [NSString stringWithContentsOfFile:source usedEncoding:&encoding error:NULL];
			if ( !result ) {
				result = @"";
			}
			KMVariables::Set( mVariable->GetNS(), result, KMNotifications::kLoadVariableFromEngineAction );
			break;
		}

	}
	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AReadFile::ActionController() const
{
	return [AReadFileEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
File: {{ file | htmlencode }}<br />
)RAW";

NSString*
AReadFile::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"file" : mSource->GetNS(),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kReadFileDestinationTypeNames[] = kReadFileDestinationTypeNamesArray;

void
AReadFile::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Destination"), kReadFileDestinationTypeNames[int(mDestination)] );
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	mClipboardReference.SaveToDictionary( dictionary );
	FileReference::SavePathToDictionary( dictionary, mSource->GetNS(), CFSTR("Source") );
}

void
AReadFile::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mDestination = dictionary->GetEnumClass( CFSTR("Destination"), kReadFileDestinationTypeNames, ReadFileDestinationType::kDEFAULT );

	CFStringRef value;
	dictionary->Get( CFSTR("Variable"), value, CFSTR("Variable") );
	check( value );
	mVariable = value->Copy();
	mClipboardReference.InitializeFromDictionary( dictionary );
	FileReference::LoadPathFromDictionary( dictionary, mSource, NULL, CFSTR("Source") );
}
