//
//  ARepeat.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Mon 28 Mar 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "ARepeat.h"
#import "Calculate.h"
#if KMF_EDITOR_WINDOW
#import "ARepeatEditor.h"
#import "TemplateEngine.h"
#import "NSStringExtensions.h"
#else
#endif
#if KMF_IS_ENGINE
#import "NSStringExtensions.h"
#import "HExecuteActionListManager.h"
#import "XErrorHandler.h"
#endif

const CFStringRef kDefaultCount = CFSTR("1");

ARepeat::ARepeat()
	:	HBaseAction(),
		mCount( kDefaultCount )
{
}

ARepeat::ARepeat( const ARepeat& inAction )
	:	HBaseAction( inAction ),
		mActionList( inAction.mActionList ),
		mCount( inAction.mCount )
{
}

ARepeat::~ARepeat()
{
#if KMF_IS_ENGINE
	HExecuteActionListManager::CancelExecution( mTrackingID );
#endif
}

CFStringRef
ARepeat::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> titleFormat = CFSTR("kmRepeatTitleFormat")->CopyLocalized();
	return CFString::CreateWithFormat( titleFormat, mCount->GetNS() );
}

#if KMF_IS_ENGINE
float
ARepeat::DoAction()
{
	mTrackingID = [NSUUID UUID];
	double count;
	if ( !Calculate::CalculateValue( mCount->GetNS(), count ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Repeat %@ failed"), mCount->GetNS() );
		count = 0;
	}
	mCountRemaining = lround( count );
	return 10;
}

Boolean
ARepeat::IsActionCompleted()
{
	if ( mActionList.GetCount() == 0 ) return true;
	
	if ( HExecuteActionListManager::IsExecuting( mTrackingID ) ) {
		return false;
	}
	if ( mCountRemaining > 0 ) {
		mCountRemaining--;
		HExecuteActionListManager::Execute( ExecuteActionListDisplay::kHide, @"Hide", mActionList, mTrackingID );
		return false;
	}
	return true;
}

void
ARepeat::ActionAborted()
{
	HExecuteActionListManager::CancelExecution( mTrackingID );
}

#endif

#if KMF_EDITOR_WINDOW

Class
ARepeat::ActionController() const
{
	return [ARepeatEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
<ul class="IconList">
	{% for action in actions %}
		<li>{{ action }}</li>
	{% /for %}
</ul>
)RAW";

NSString*
ARepeat::GetActionHTMLDescriptionString() const
{
	if ( mActionList.GetCount() == 0 ) return @"Repeat With No Actions Will Do Nothing";
	return [NSString stringWithFormat:@"Repeat the Following Actions %@ Times", [mCount->GetNS() encodeHTML]];
}

NSString*
ARepeat::GetActionExtraHTML() const
{
	if ( mActionList.GetCount() == 0 ) return @"";

	NSDictionary *variables = @{
									@"actions" : mActionList.GetDisplayContent(),
								};
	NSString *result = [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
	return result;
}


#endif

BOOL
ARepeat::MatchesTextSearch( NSString* searchString ) const
{
	if ( HBaseAction::MatchesTextSearch( searchString ) ) return YES;
	return mActionList.MatchesTextSearch( searchString );
}

#pragma mark streaming

void
ARepeat::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("CountExpression"), CFStringRef(mCount) );
	SafeCFPropertyListRef actions = mActionList.CopyActionsToPropertyList();
	dictionary->Set( CFSTR("Actions"), actions );
	CFRelease( actions );
}

void
ARepeat::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef value;
	dictionary->Get( CFSTR("CountExpression"), value, CFStringRef(NULL) );
	mCount = value ? value->Copy() : NULL;
	
	SInt32 count;
	if ( !mCount && dictionary->Get( CFSTR("Count"), count ) ) {
		mCount = CFString::Create( CFBridgingToCFString(Calculate::DoubleValue( count )) );
	}
	
	if ( !mCount ) {
		mCount = kDefaultCount;
	}

	SafeCFPropertyListRef actions = (SafeCFPropertyListRef)dictionary->Get( CFSTR("Actions") );
	XAssert( actions );
	mActionList.LoadActionsFromPropertyList( actions );
}

