//
//  AResizeImage.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Wed 9 May 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AResizeImage.h"
#import "XException.h"
#import "KMDefines.h"
#import "XErrorHandler.h"
#import "NSStringRegularExpressions.h"
#import "Calculate.h"
#import "TokenText.h"
#import "CSClipboardDomain.h"
#import "NSStringExtensions.h"

#import "XAXApplicationElement.h"
#import "XAXWindowElement.h"
#import "XAXWindowElementList.h"
#import "ScreenInfo.h"
#import "NSImageExtensions.h"

#if KMF_EDITOR_WINDOW
#import "AResizeImageEditor.h"
#import "TemplateEngine.h"
#endif
#if KMF_IS_ENGINE
#import "XRunningApplicationsList.h"
#import "KMNotifications.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#endif

static CFStringRef kDefaultLeft = CFSTR("100");
static CFStringRef kDefaultTop = CFSTR("100");
static CFStringRef kDefaultWidth = CFSTR("100");
static CFStringRef kDefaultHeight = CFSTR("100");

NSString*
AResizeImage::GetNewActionTitle() const
{
	return CFBridgingSafeRelease(CopyActionDescriptionString());
}

CFStringRef
AResizeImage::CopyActionDescriptionString() const
{
	CFStringRef description;
	switch ( mResizeAction ) {
		case ResizeImageActionType::kScaleTo:
			description = CFSTR("kmResizeImageActionScaleTo")->CopyLocalized();
			break;
		case ResizeImageActionType::kScaleToFit:
			description = CFSTR("kmResizeImageActionScaleToFit")->CopyLocalized();
			break;
		case ResizeImageActionType::kResizeTo:
			description = CFSTR("kmResizeImageActionResizeTo")->CopyLocalized();
			break;
		case ResizeImageActionType::kAddMargin:
			description = CFSTR("kmResizeImageActionAddMargin")->CopyLocalized();
			break;
		case ResizeImageActionType::kCrop:
			description = CFSTR("kmResizeImageActionCrop")->CopyLocalized();
			break;
	}
	
	return description;
}

const ArrowString::Style*
AResizeImage::GetArrowStyles() const
{
	switch ( mResizeAction ) {
		case ResizeImageActionType::kScaleTo:
		case ResizeImageActionType::kScaleToFit:
		case ResizeImageActionType::kResizeTo:
			return ArrowString::kWidthHeight;
		case ResizeImageActionType::kAddMargin:
			return ArrowString::kLeftUpRightDown;
		case ResizeImageActionType::kCrop:
			return ArrowString::kRightDownWidthHeight;
	}
}

#if KMF_IS_ENGINE

float
AResizeImage::DoAction()
{	
	const XClipboardDataSet* clip = mTarget.GetClipboardDataSet();
	if ( clip ) {
		NSImage* image = clip->Image();
		NSImage* result = nil;
		if ( image ) {
			NSRect imageRect;
			imageRect.origin = NSZeroPoint;
			imageRect.size = image.size;
			NSDictionary* calculationContext = @{ @"IMAGE" : NSStringFromRect(imageRect) };

			double left=0, top=0, width=0, height=0;
			BOOL needTopLeft = (mResizeAction == ResizeImageActionType::kAddMargin) || (mResizeAction == ResizeImageActionType::kCrop);
			if ( needTopLeft && !Calculate::CalculateValue( mLeft, left, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Resize Image %@ failed"), mLeft );
				return 0;
			} else if ( needTopLeft && !Calculate::CalculateValue( mTop, top, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Resize Image %@ failed"), mTop );
				return 0;
			} else if ( !Calculate::CalculateValue( mWidth, width, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Resize Image %@ failed"), mWidth );
				return 0;
			} else if ( !Calculate::CalculateValue( mHeight, height, calculationContext ) ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Resize Image %@ failed"), mHeight );
				return 0;
			}
			left = roundf( left );
			top = roundf( top );
			width = roundf( width );
			height = roundf( height );

			switch ( mResizeAction ) {
				case ResizeImageActionType::kScaleTo: {
					NSLog( @"ResizeImageActionType::kScaleTo %f, %f", width, height );
					if ( width >= 1 && height >= 1 ) {
						NSLog( @"size %f, %f", width, height );
						result = [image imageWithSize:NSMakeSize(width, height)];
						NSLog( @"ResizeImageActionType::kScaleTo %f, %f", width, height );
					}
					break;
				}
				case ResizeImageActionType::kScaleToFit: {
					NSLog( @"ResizeImageActionType::kScaleTo %f, %f", width, height );
					double scale = 1.0;
					NSSize orgSize = image.size;
					NSLog( @"ResizeImageActionType::kScaleTo image.size %f, %f", image.size.width, image.size.height );
					if ( width >= 1 && orgSize.width > width ) {
						scale = width / orgSize.width;
					}
					if ( height >= 1 && orgSize.height * scale > height ) {
						scale = height / orgSize.height;
					}
					NSLog( @"ResizeImageActionType::kScaleTo scale %f", scale );
					if ( scale < 1 ) {
						NSLog( @"ResizeImageActionType::kScaleTo scale to %@", NSStringFromSize(NSMakeSize(roundf(width*scale), roundf(height*scale))) );
						result = [image imageWithSize:NSMakeSize(roundf(orgSize.width*scale), roundf(orgSize.height*scale))];
					}
					break;
				}
				case ResizeImageActionType::kResizeTo: {
					if ( width >= 1 && height >= 1 ) {
						NSImage *resizedImage = [[NSImage alloc] initWithSize: NSMakeSize(width, height)];
						[resizedImage lockFocus];
						[[NSColor clearColor] set];
						NSRectFill( NSMakeRect(0, 0, resizedImage.size.width, resizedImage.size.height) );
						[image drawAtPoint:NSMakePoint(roundf((width-image.size.width)/2), roundf((height-image.size.height)/2)) fromRect:NSZeroRect operation:NSCompositeCopy fraction: 1.0];
						[resizedImage unlockFocus];
						result = resizedImage;
					}
					break;
				}
				case ResizeImageActionType::kAddMargin: {
					float imageWidth = roundf(image.size.width);
					float imageHeight = roundf(image.size.height);
					float newWidth = left + imageWidth + width;
					float newHeight = top + imageHeight + height;
					if ( newWidth >= 1 && newHeight >= 1 ) {
						NSImage *resizedImage = [[NSImage alloc] initWithSize: NSMakeSize(newWidth, newHeight)];
						[resizedImage lockFocus];
						[[NSColor clearColor] set];
						NSRectFill( NSMakeRect(0, 0, resizedImage.size.width, resizedImage.size.height) );
						[image drawAtPoint:NSMakePoint(left, height) fromRect:NSZeroRect operation:NSCompositeCopy fraction: 1.0];
						[resizedImage unlockFocus];
						result = resizedImage;
					}
					break;
				}
				case ResizeImageActionType::kCrop: {
					if ( width >= 1 && height >= 1 ) {
						NSImage *resizedImage = [[NSImage alloc] initWithSize: NSMakeSize(width, height)];
						[resizedImage lockFocus];
						[[NSColor clearColor] set];
						NSRectFill( NSMakeRect(0, 0, resizedImage.size.width, resizedImage.size.height) );
						[image drawAtPoint:NSMakePoint(-left, top+height-image.size.height) fromRect:NSZeroRect operation:NSCompositeCopy fraction: 1.0];
						[resizedImage unlockFocus];
						result = resizedImage;
					}
					break;
				}
			}
			if ( result ) {
				mTarget.WriteClipboardDataSet( XClipboardDataSet( result ) );
			}
		} else {
			XErrorHandlerMaybeLogMessage("always", CFSTR("Resize Image cannot get source image") );
		}
	}
	
	return 10;
}


#endif

#if KMF_EDITOR_WINDOW

Class
AResizeImage::ActionController() const
{
	return [AResizeImageEditor class];
}

NSString*
AResizeImage::GetActionHTMLDescriptionString() const
{
	NSString* key;
	switch ( mResizeAction ) {
		case ResizeImageActionType::kScaleTo:
			key = @"kmResizeImageActionHTMLScaleTo";
			break;
		case ResizeImageActionType::kScaleToFit:
			key = @"kmResizeImageActionHTMLScaleToFit";
			break;
		case ResizeImageActionType::kResizeTo:
			key = @"kmResizeImageActionHTMLResizeTo";
			break;
		case ResizeImageActionType::kAddMargin:
			key = @"kmResizeImageActionHTMLAddMargin";
			break;
		case ResizeImageActionType::kCrop:
			key = @"kmResizeImageActionHTMLCrop";
			break;
	}

	return [NSString stringWithFormat:[key localized], mTarget.GetDisplayName()];
}

static NSString* kActionDisplayHTML = @R"RAW(
{% if issize %}To {{ fit }} {{ width | htmlencode }} wide, {{ height | htmlencode }} high.{% /if %}
 {% if ismargins %}Add {{ left | htmlencode }} at left,{{ top | htmlencode }} at top, {{ width | htmlencode }} at right, {{ height | htmlencode }} at bottom.{% /if %}
 {% if iscrop %}From ({{ left | htmlencode }},{{ top | htmlencode }}) {{ width | htmlencode }} wide, {{ height | htmlencode }} high.{% /if %}
)RAW";

NSString*
AResizeImage::GetActionExtraHTML() const
{
	BOOL issize, ismargins, iscrop;
	switch ( mResizeAction ) {
		case ResizeImageActionType::kScaleTo:
		case ResizeImageActionType::kScaleToFit:
		case ResizeImageActionType::kResizeTo:
			issize = YES;
			ismargins = NO;
			iscrop = NO;
			break;
		case ResizeImageActionType::kAddMargin:
			issize = NO;
			ismargins = YES;
			iscrop = NO;
			break;
		case ResizeImageActionType::kCrop:
			issize = NO;
			ismargins = NO;
			iscrop = YES;
			break;
	}

	NSDictionary *variables = @{
									@"issize" : @(issize),
									@"ismargins" : @(ismargins),
									@"iscrop" : @(iscrop),
									@"left" : mLeft,
									@"top" : mTop,
									@"width" : mWidth,
									@"height" : mHeight,
									@"fit" : mResizeAction == ResizeImageActionType::kScaleToFit ? @"fit" : @"",
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kResizeImageActionTypeNames[] = kResizeImageActionTypeNamesArray;

void
AResizeImage::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("ResizeAction"), kResizeImageActionTypeNames[int(mResizeAction)] );
	mTarget.SaveToDictionary( dictionary, CFSTR("Target") );
	dictionary->Set( CFSTR("Left"), CFBridgingToCFString(mLeft) );
	dictionary->Set( CFSTR("Top"), CFBridgingToCFString(mTop) );
	dictionary->Set( CFSTR("Width"), CFBridgingToCFString(mWidth) );
	dictionary->Set( CFSTR("Height"), CFBridgingToCFString(mHeight) );
}

void
AResizeImage::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	SInt32 v = dictionary->GetEnumerated( CFSTR("ResizeAction"), kResizeImageActionTypeNames );
	XAssert( v >= 0 );
	mResizeAction = (ResizeImageActionType) v;
	mTarget.InitializeFromDictionary( dictionary, CFSTR("Target") );
	CFStringRef value;
	dictionary->Get( CFSTR("Left"), value, kDefaultLeft );
	check( value );
	mLeft = value->GetNS();
	
	dictionary->Get( CFSTR("Top"), value, kDefaultTop );
	check( value );
	mTop = value->GetNS();
	
	dictionary->Get( CFSTR("Width"), value, kDefaultWidth );
	check( value );
	mWidth = value->GetNS();
	
	dictionary->Get( CFSTR("Height"), value, kDefaultHeight );
	check( value );
	mHeight = value->GetNS();
}
