//
//  ASafariControl.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 18 Sep 2009.
//  Copyright (c) 2009 Stairways Software. All rights reserved.
//

#import "ASafariControl.h"
#import "KMIconRegistry.h"
#import "XException.h"
#import "XCoreFoundation.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "WebBrowserInterface.h"
#if KMF_IS_ENGINE
#import "Calculate.h"
#import "XErrorHandler.h"
#import "TokenText.h"
#endif
#if KMF_EDITOR_WINDOW
#import "ASafariControlEditor.h"
#import "ASafariControlTextEditor.h"
#endif

ASafariControl::ASafariControl( WebBrowserInterface* interface, const SafariCommand& inCommand )
	:	HBaseAction(),
		mCommand( inCommand ),
		mWebBrowserInterface( interface )
{
	switch ( inCommand ) {
		case SafariCommand::kNewWindow:
		case SafariCommand::kNewTab:
		case SafariCommand::kNextTab:
		case SafariCommand::kPreviousTab:
			mArgumentString = @"";
			break;
		case SafariCommand::kSelectTab:
			mArgumentString = @"1";
			break;
		case SafariCommand::kSetTitle:
			mArgumentString = @"My Title";
			break;
		case SafariCommand::kClickLink:
			mArgumentString = @"About";
			break;
		case SafariCommand::kSubmitForm:
		case SafariCommand::kResetForm:
			mArgumentString = @kWebBrowserInterfaceDefaultForm;
			break;
		case SafariCommand::kFocusField:
		case SafariCommand::kSelectField:
			mArgumentString = @kWebBrowserInterfaceDefaultField;
			break;
		case SafariCommand::kWaitForComplete:
			mArgumentString = @"3";
			break;
		case SafariCommand::kSetDocumentURL:
			mArgumentString = [@"kmDefaultOpenURLValue" localized];
			break;
	}
}

static NSString*
ActionDescriptionString( SafariCommand command, NSString* webBrowser, bool newAction, NSString* argument )
{
	NSString* theTitleKey;
	switch ( command ) {
		case SafariCommand::kNewWindow:
			theTitleKey = (newAction || IsEmpty(argument)) ? @"kSafariCommandNewWindowShort" : @"kSafariCommandNewWindowLong";
			break;
		case SafariCommand::kNewTab:
			theTitleKey = (newAction || IsEmpty(argument)) ? @"kSafariCommandNewTabShort" : @"kSafariCommandNewTabLong";
			break;
		case SafariCommand::kNextTab:
			theTitleKey = @"kSafariCommandNextTab";
			break;
		case SafariCommand::kPreviousTab:
			theTitleKey = @"kSafariCommandPreviousTab";
			break;
		case SafariCommand::kSelectTab:
			theTitleKey = newAction ? @"kSafariCommandSelectTabShort" : @"kSafariCommandSelectTabLong";
			break;
		case SafariCommand::kSetTitle:
			theTitleKey = newAction ? @"kSafariCommandSetTitleShort" : @"kSafariCommandSetTitleLong";
			break;
		case SafariCommand::kClickLink:
			theTitleKey = newAction ? @"kSafariCommandClickLinkShort" : @"kSafariCommandClickLinkLong";
			break;
		case SafariCommand::kSubmitForm:
			theTitleKey = newAction ? @"kSafariCommandSubmitFormShort" : @"kSafariCommandSubmitFormLong";
			break;
		case SafariCommand::kResetForm:
			theTitleKey = newAction ? @"kSafariCommandResetFormShort" : @"kSafariCommandResetFormLong";
			break;
		case SafariCommand::kFocusField:
			theTitleKey = newAction ? @"kSafariCommandFocusFieldShort" : @"kSafariCommandFocusFieldLong";
			break;
		case SafariCommand::kSelectField:
			theTitleKey = newAction ? @"kSafariCommandSelectFieldShort" : @"kSafariCommandSelectFieldLong";
			break;
		case SafariCommand::kWaitForComplete:
			theTitleKey = newAction ? @"kSafariCommandWaitForCompleteShort" : @"kSafariCommandWaitForCompleteLong";
			break;
		case SafariCommand::kSetDocumentURL:
			theTitleKey = newAction ? @"kSafariCommandSetDocumentURLShort" : @"kSafariCommandSetDocumentURLLong";
			break;
	}
	NSString* result = [theTitleKey localized];
	result = [NSString stringWithFormat:result, webBrowser, argument];
	return result;
}


CFStringRef ASafariControl::CopyActionDescriptionString() const
{
	return (CFStringRef)CFBridgingRetain( ActionDescriptionString( mCommand, GetWebBrowserName(), false, mArgumentString ) );
}

NSString*
ASafariControl::GetNewActionTitle() const
{
	return ActionDescriptionString( mCommand, GetWebBrowserName(), true, @"" );
}

#if KMF_IS_ENGINE

float
ASafariControl::DoAction()
{
	float time = 0;
	NSString* script;
	
	switch ( mCommand ) {
		case SafariCommand::kNewWindow: {
			NSString* url = TokenText::GetResult( mArgumentString, &time );
			mWebBrowserInterface->NewWindow( url );
			time += 10;
			break;
		}
		case SafariCommand::kNewTab: {
			NSString* url = TokenText::GetResult( mArgumentString, &time );
			mWebBrowserInterface->NewTab( url );
			time += 10;
			break;
		}
		case SafariCommand::kNextTab:
			mWebBrowserInterface->NextTab();
			time += 2;
			break;
		case SafariCommand::kPreviousTab:
			mWebBrowserInterface->PreviousTab();
			time += 2;
			break;
		case SafariCommand::kSelectTab: {
			double doubleargument;
			int intArgument;
			if ( Calculate::CalculateValue( mArgumentString, doubleargument ) ) {
				intArgument = round(doubleargument);
			} else {
				XErrorHandlerMaybeLogFormat2("always", CFSTR("%@ Control argument %@ failed"), GetWebBrowserName(), mArgumentString );
				intArgument = 1;
			}
			mWebBrowserInterface->SelectTab( intArgument );
			time += 4;
			break;
		}
		case SafariCommand::kSetTitle: {
			NSString* title = TokenText::GetResult( mArgumentString, &time );
			mWebBrowserInterface->SetTitle( title );
			time += 10;
			break;
		}
		case SafariCommand::kClickLink: {
			NSString* title = TokenText::GetResult( mArgumentString, &time );
			bool good = mWebBrowserInterface->ClickLink( title );
			if ( !good ) {
				XErrorHandlerMaybeLogFormat2( "always", @"%@ Click Link failed to find link %@", GetWebBrowserName(), title );
			}
			time += 10;
			break;
		}
		case SafariCommand::kSubmitForm: {
			NSString* title = TokenText::GetResult( mArgumentString, &time );
			bool good = mWebBrowserInterface->SubmitForm( title );
			if ( !good ) {
				XErrorHandlerMaybeLogFormat2( "always", @"%@ Submit Form failed to find form %@", GetWebBrowserName(), title );
			}
			time += 10;
			break;
		}
		case SafariCommand::kResetForm: {
			NSString* title = TokenText::GetResult( mArgumentString, &time );
			bool good = mWebBrowserInterface->ResetForm( title );
			if ( !good ) {
				XErrorHandlerMaybeLogFormat2( "always", @"%@ Submit Form failed to find form %@", GetWebBrowserName(), title );
			}
			time += 10;
			break;
		}
		case SafariCommand::kFocusField: {
			NSString* title = TokenText::GetResult( mArgumentString, &time );
			bool good = mWebBrowserInterface->FocusField( title );
			if ( !good ) {
				XErrorHandlerMaybeLogFormat2( "always", @"%@ Focus Field failed to find field %@", GetWebBrowserName(), title );
			}
			time += 10;
			break;
		}
		case SafariCommand::kSelectField: {
			NSString* title = TokenText::GetResult( mArgumentString, &time );
			bool good = mWebBrowserInterface->SelectField( title );
			if ( !good ) {
				XErrorHandlerMaybeLogFormat2( "always", @"%@ Focus Field failed to find field %@", GetWebBrowserName(), title );
			}
			time += 10;
			break;
		}
		case SafariCommand::kWaitForComplete:
			script = nil;
			if ( !Calculate::CalculateValue( mArgumentString, mActionRequiredTime ) ) {
				XErrorHandlerMaybeLogFormat2("always", CFSTR("%@ Control argument %@ failed"), GetWebBrowserName(), mArgumentString );
				mActionRequiredTime = 1;
			}
			mActionLastIncompleteTime = GetCurrentEventTime();
			mActionLastIncompleteCheckTime = GetCurrentEventTime();
			break;
		case SafariCommand::kSetDocumentURL: {
			NSString* url = TokenText::GetResult( mArgumentString, &time );
			mWebBrowserInterface->SetDocumentURL( url );
			time += 10;
			break;
		}
	}

	return time;
}

Boolean
ASafariControl::IsActionCompleted()
{
	if ( mCommand != SafariCommand::kWaitForComplete ) return true;

	EventTime now = GetCurrentEventTime();
	if ( now - mActionLastIncompleteTime > mActionRequiredTime || now - mActionLastIncompleteCheckTime >  0.2 ) {
		Boolean completed = mWebBrowserInterface->IsComplete();
		mActionLastIncompleteCheckTime = now;
		if ( !completed ) {
			mActionLastIncompleteTime = now;
			return false;
		}
	}
	if ( now - mActionLastIncompleteTime > mActionRequiredTime ) {
		return true;
	}
	return false;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASafariControl::ActionController() const
{
	switch ( mCommand ) {
		case SafariCommand::kNextTab:
		case SafariCommand::kPreviousTab:
			return HBaseAction::ActionController();
		case SafariCommand::kSelectTab:
		case SafariCommand::kWaitForComplete:
			return [ASafariControlEditor class];
		case SafariCommand::kSetTitle:
		case SafariCommand::kClickLink:
		case SafariCommand::kSetDocumentURL:
		case SafariCommand::kNewWindow:
		case SafariCommand::kNewTab:
		case SafariCommand::kSubmitForm:
		case SafariCommand::kResetForm:
		case SafariCommand::kFocusField:
		case SafariCommand::kSelectField:
			return [ASafariControlTextEditor class];
	}
}

#endif

#pragma mark streaming

static const CFStringRef kSafariCommandNames[] = kSafariCommandNamesArray;

void
ASafariControl::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Command"), kSafariCommandNames[int(mCommand)] );
	dictionary->Set( CFSTR("Argument"), CFBridgingToCFString(mArgumentString) );
	mWebBrowserInterface->SaveToDictionary( dictionary );
}

void
ASafariControl::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	bool good = dictionary->CanGetEnumClass( CFSTR("Command"), kSafariCommandNames, mCommand );
	XAssert( good );

	CFStringRef newArgumentString;
	dictionary->Get( CFSTR("Argument"), newArgumentString, CFSTR("") );
	mArgumentString = CFBridgingToNS(newArgumentString);
	mWebBrowserInterface = WebBrowserInterface::InitializeFromDictionary( dictionary );
}
