//
//  AScreenCapture.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 19 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AScreenCapture.h"
#import "XException.h"
#import "KMDefines.h"
#import "XContainer.h"
#import "XErrorHandler.h"
#import "NSStringRegularExpressions.h"
#import "Calculate.h"
#import "TokenText.h"
#import "CSClipboardDomain.h"

#import "XAXApplicationElement.h"
#import "XAXWindowElement.h"
#import "XAXWindowElementList.h"
#import "ScreenInfo.h"
#import "NSImageExtensions.h"
#import "XCoreFoundation.h"

#if KMF_EDITOR_WINDOW
#import "AScreenCaptureEditor.h"
#import "TemplateEngine.h"
#endif
#if KMF_IS_ENGINE
#import "XRunningApplicationsList.h"
#import "KMNotifications.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#endif

const CFStringRef kDefaultIndex = CFSTR("2");
const Boolean kDefaultIncludeShadows = true;

AScreenCapture::AScreenCapture()
	:	HBaseAction(),
		mSource( kScreenCaptureSourceDEFAULT ),
		mIncludeShadows( kDefaultIncludeShadows ),
		mIndex( kDefaultIndex->Copy() ),
		mWindowName( CFSTR("")->Copy() )
{
}

AScreenCapture::AScreenCapture( const AScreenCapture& inAction )
	:	HBaseAction( inAction ),
		mSource( inAction.mSource ),
		mIncludeShadows( inAction.mIncludeShadows ),
		mIndex( inAction.mIndex ),
		mWindowName( inAction.mWindowName ),
		mDestination( inAction.mDestination )
{
}
	
AScreenCapture::~AScreenCapture()
{
}

Boolean
AScreenCapture::SourceIsScreen() const
{
	switch ( mSource ) {
		case kScreenCaptureSourceScreenAll:
		case kScreenCaptureSourceScreenMain:
		case kScreenCaptureSourceScreenIndex:
			return true;
		case kScreenCaptureSourceWindowFront:
		case kScreenCaptureSourceWindowIndex:
		case kScreenCaptureSourceWindowName:
		case kScreenCaptureSourceWindowNameContaining:
		case kScreenCaptureSourceWindowNameMatching:
			return false;
	}
}

CFStringRef
AScreenCapture::CopyActionDescriptionString() const
{
	// first, build window name string
	
	CFContainer<CFStringRef> format;
	CFContainer<CFStringRef> windowName;
	
	switch ( mSource ) {
		case kScreenCaptureSourceScreenAll:
			format = CFSTR("kmScreenCaptureSourceScreenAll")->CopyLocalized();
			break;
		case kScreenCaptureSourceScreenMain:
			format = CFSTR("kmScreenCaptureSourceScreenMain")->CopyLocalized();
			break;
		case kScreenCaptureSourceScreenIndex:
			format = CFSTR("kmScreenCaptureSourceScreenIndex")->CopyLocalized();
			windowName = mIndex;
			break;
		case kScreenCaptureSourceWindowFront:
			format = CFSTR("kmScreenCaptureSourceWindowFront")->CopyLocalized();
			break;
		case kScreenCaptureSourceWindowIndex:
			format = CFSTR("kmScreenCaptureSourceWindowIndex")->CopyLocalized();
			windowName = mIndex;
			break;
		case kScreenCaptureSourceWindowName:
			format = CFSTR("kmScreenCaptureSourceWindowName")->CopyLocalized();
			windowName = mWindowName;
			break;
		case kScreenCaptureSourceWindowNameContaining:
			format = CFSTR("kmScreenCaptureSourceWindowNameContaining")->CopyLocalized();
			windowName = mWindowName;
			break;
		case kScreenCaptureSourceWindowNameMatching:
			format = CFSTR("kmScreenCaptureSourceWindowNameMatching")->CopyLocalized();
			windowName = mWindowName;
			break;
	}
	
	return CFString::CreateWithFormat( format, CFStringRef(windowName) );
}

#if KMF_IS_ENGINE

CGImageRef
AScreenCapture::CaptureWindow( const XAXWindowElement& inWindowElement )
{	
	CGImageRef image = NULL;
	try {
		CGPoint			thePos;
		CGSize			theSize;
		
		XAssert( !!inWindowElement );
		XAssert( inWindowElement.GetPosition( thePos ) );
		XAssert( inWindowElement.GetSize( theSize ) );
		CFContainer<CFStringRef> windowName = inWindowElement.CopyTitle( CFSTR("") );

		CGWindowID window;
		int found = 0;
		
		CFContainer<CFArrayRef> windowIdArray = CGWindowListCreate( kCGWindowListOptionOnScreenOnly | kCGWindowListExcludeDesktopElements, kCGNullWindowID );
		CFContainer<CFArrayRef> windowInfoArray = CGWindowListCreateDescriptionFromArray( windowIdArray );
		
		ProcessSerialNumber psn = XRunningApplicationsList::GetFrontApplication();
		pid_t pid;
		XAssert( GetProcessPID( &psn, &pid ) == noErr );
		
		for ( NSInteger i = 0; i < CFArrayGetCount(windowInfoArray); i++ ) {
			CFDictionaryRef info = (CFDictionaryRef)CFArrayGetValueAtIndex( windowInfoArray, i );
			NSNumber *applicationPid = (NSNumber *) CFDictionaryGetValue( info, kCGWindowOwnerPID );
			if ( pid == static_cast<pid_t>([applicationPid unsignedLongValue]) ) {
				CFDictionaryRef boundsRef = (CFDictionaryRef) CFDictionaryGetValue( info, kCGWindowBounds );
				if ( boundsRef ) {
					CGRect bounds;
					if ( CGRectMakeWithDictionaryRepresentation( boundsRef, &bounds ) ) {
						CFStringRef titleref = (CFStringRef) CFDictionaryGetValue( info, kCGWindowName );
						NSString* title = titleref ? CFBridgingToNS(titleref) : @"";
						if ( [title isEqualToString:windowName->GetNS()] ) {
							XErrorHandlerMaybeLogFormat4( "ScreenCapture", CFSTR("Possible match for %@ with %@, and %@ +%@"), windowName->GetNS(), NSStringFromRect( bounds ), NSStringFromPoint( thePos ), NSStringFromSize( theSize ) );
							if ( lround(bounds.origin.x) == lround(thePos.x) && lround(bounds.origin.y) == lround(thePos.y) && lround(bounds.size.width) == lround(theSize.width) && lround(bounds.size.height) == lround(theSize.height) ) {
								found++;
								window = (CGWindowID) NSUInteger(CFArrayGetValueAtIndex( windowIdArray, i ));
							}
						}
					}
				}
			}
		}
		
		if ( found == 1 ) {
			XErrorHandlerMaybeLogMessage( "ScreenCapture", CFSTR("found") );
			image = CGWindowListCreateImage( CGRectZero, kCGWindowListOptionIncludingWindow, window, mIncludeShadows ? kCGWindowImageDefault : kCGWindowImageBoundsIgnoreFraming );
		}
		
	} catch (...) {
	
	}
	return image;
}

float
AScreenCapture::DoAction()
{
	float time = 0;
	
	CGImageRef image = NULL;

	switch ( mSource ) {
		case kScreenCaptureSourceScreenAll:
			image = CGWindowListCreateImage( CGRectInfinite, kCGWindowListOptionOnScreenOnly, kCGNullWindowID, kCGWindowImageDefault );
			break;
		case kScreenCaptureSourceScreenMain:
			[[clang::fallthrough]];
		case kScreenCaptureSourceScreenIndex: {
			NSUInteger index = 0;
			if ( mSource == kScreenCaptureSourceScreenIndex ) {
				double sindex;
				if ( !Calculate::CalculateValue( mIndex->GetNS(), sindex ) ) {
					XErrorHandlerMaybeLogFormat1("always", CFSTR("Screen Capture evaluate index %@ failed"), mIndex->GetNS() );
					sindex = 0;
				}
				index = NSUInteger(sindex);
			}
			if ( index > ScreenInfo::ScreenCount() ) {
				XErrorHandlerMaybeLogFormat1("always", CFSTR("Screen Capture index %d beyond range"), (int)index );
				index = 0;
			}
			CGRect frame = ScreenInfo::ScreenFrame( index );
			XErrorHandlerMaybeLogFormat2( "ScreenCapture", CFSTR("Capture screen %d frame %@"), (int)index, NSStringFromRect( NSRectFromCGRect( frame ) ) );
			image = CGWindowListCreateImage( frame, kCGWindowListOptionOnScreenOnly, kCGNullWindowID, kCGWindowImageDefault );
			break;
		}
		case kScreenCaptureSourceWindowFront:
		case kScreenCaptureSourceWindowIndex:
		case kScreenCaptureSourceWindowName:
		case kScreenCaptureSourceWindowNameContaining:
		case kScreenCaptureSourceWindowNameMatching: {
			if ( !AXAPIEnabled() ) {
				XErrorHandlerMaybeLogMessage( "always", CFSTR("Manipulate Window action required Universal Access which is disabled") );
				return time;
			}

			XAXApplicationElement theFrontApp;
			
			if ( theFrontApp ) {
				if ( mSource == kScreenCaptureSourceWindowFront ) {
					XAXWindowElement windowElement( theFrontApp.GetFocussedOrMainWindow() );
					if ( !!windowElement ) {
						image = CaptureWindow( windowElement );
					}
					break;
				} else {
					XAXApplicationElement frontApp;
					XAXWindowElementList* theWindowList = ScreenInfo::NewGoodWindowList( frontApp );
					if ( theWindowList ) {
						if ( mSource == kScreenCaptureSourceWindowIndex ) {
							double windex;
							if ( !Calculate::CalculateValue( mIndex->GetNS(), windex ) ) {
								XErrorHandlerMaybeLogFormat1("always", CFSTR("Screen Capture %@ failed"), mIndex->GetNS() );
							} else if ( 0 < lround(windex) && lround(windex) <= theWindowList->GetCount() ) {
								image = CaptureWindow( theWindowList->GetValueAtIndex( lround(windex)-1 ) );
							} else if ( -theWindowList->GetCount() <= lround(windex) && lround(windex) < 0 ) {
								image = CaptureWindow( theWindowList->GetValueAtIndex( theWindowList->GetCount()+lround(windex) ) );
							} else {
								XErrorHandlerMaybeLogFormat1("always", CFSTR("Screen Capture failed to find window with index %ld"), lround(windex) );
							}
						} else {
							CFContainer<CFStringRef> tokenTextValue = TokenText::CopyResult( mWindowName, time, mSource != kScreenCaptureSourceWindowNameMatching );
							check( tokenTextValue.Defined() );
							for ( CFIndex i = 0; i < theWindowList->GetCount(); i++ ) {
								Boolean winFound = false;
								CFContainer<CFStringRef> winName = theWindowList->GetValueAtIndex( i ).CopyTitle( NULL );
								if ( winName != NULL ) {
									if ( mSource == kScreenCaptureSourceWindowName ) {
										winFound = ( tokenTextValue->IsEqual( winName, kCFCompareLocalized ) );
									} else if ( mSource == kScreenCaptureSourceWindowNameContaining ) {
										winFound = CFStringFindWithOptions( winName, tokenTextValue, CFRangeMake( 0, CFStringGetLength( winName ) ), kCFCompareLocalized, NULL );
									} else if ( mSource == kScreenCaptureSourceWindowNameMatching ) {
										winFound = [winName->GetNS() isMatchedByRegex:tokenTextValue->GetNS()];
									}
								}
								if ( winFound ) {
									image = CaptureWindow( theWindowList->GetValueAtIndex( i ) );
									break;
								}
							}
						}
						delete theWindowList;
					}
				}
			}

			break;
		}
	}
	
	if ( image ) {
		XErrorHandlerMaybeLogMessage( "ScreenCapture", CFSTR("write image") );
		mDestination.WriteClipboardDataSet( XClipboardDataSet( [NSImage imageWithCGImage:image] ) );
	}
	CFXRelease( image );
	
	return time + 10;
}


#endif

#if KMF_EDITOR_WINDOW

Class
AScreenCapture::ActionController() const
{
	return [AScreenCaptureEditor class];
}

NSString*
AScreenCapture::GetActionHTMLDescriptionString() const
{
	return [NSString stringWithFormat:@"%@ to %@", GetActionDescriptionString(), mDestination.GetDisplayName()];
}

static NSString* kActionDisplayHTML = @R"RAW(
{% if shadow %}Including shadows.<br />{% /if %}
)RAW";

NSString*
AScreenCapture::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"shadow" : @(MustIncludeShadows() || mIncludeShadows),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kScreenCaptureSourceTypeNames[] = kScreenCaptureSourceTypeNamesArray;

void
AScreenCapture::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Source"), kScreenCaptureSourceTypeNames[mSource] );
	dictionary->SetBoolean( CFSTR("IncludeShadows"), mIncludeShadows );
	dictionary->Set( CFSTR("Index"), CFStringRef(mIndex) );
	dictionary->Set( CFSTR("WindowName"), CFStringRef(mWindowName) );
	mDestination.SaveToDictionary( dictionary, CFSTR("Destination") );
}

void
AScreenCapture::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	CFStringRef value;
	
	dictionary->GetBoolean( CFSTR("IncludeShadows"), mIncludeShadows, kDefaultIncludeShadows );

	SInt32 v = dictionary->GetEnumerated( CFSTR("Source"), kScreenCaptureSourceTypeNames );
	XAssert( v >= 0 );
	mSource = (ScreenCaptureSourceType) v;
	
	dictionary->Get( CFSTR("Index"), value, kDefaultIndex );
	check( value );
	mIndex = value->Copy();
	
	dictionary->Get( CFSTR("WindowName"), value, CFSTR("") );
	check( value );
	mWindowName = value->Copy();
	
	mDestination.InitializeFromDictionary( dictionary, CFSTR("Destination") );
}
