/*
 *  AScrollWheelEvent.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Sat Feb 08 2003.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "AScrollWheelEvent.h"
#import "UCGRemoteControl.h"
#import "XException.h"
#import "KMDefines.h"
#import "Calculate.h"
#import "NSStringRegularExpressions.h"
#if KMF_EDITOR_WINDOW
#import "AScrollWheelEventEditor.h"
#endif
#if KMF_IS_ENGINE
#import "XErrorHandler.h"
#endif

static inline SInt32 sign( SInt32 v )
{
	if ( v < 0 ) return -1;
	if ( v > 0 ) return 1;
	return 0;
}

const CFStringRef kDefaultScrollAmount = CFSTR("60");

AScrollWheelEvent::AScrollWheelEvent()
	:	HBaseAction(),
		mScrollAmount( kDefaultScrollAmount ),
		mScrollDirection( ScrollWheelDirectionType::kDEFAULT )
{

}

AScrollWheelEvent::AScrollWheelEvent( const AScrollWheelEvent& inAction )
	:	HBaseAction( inAction ),
		mScrollAmount( inAction.mScrollAmount ),
		mScrollDirection( inAction.mScrollDirection )
{
}
	
AScrollWheelEvent::~AScrollWheelEvent()
{
}

static bool SimpleValue( NSString* s, SInt32& v )
{
	if ( ![s isMatchedByRegex:@"^-?\\d+$"] ) return false;
	double doublevalue;
	if ( !Calculate::CalculateValue( s, doublevalue ) ) return false;
	v = lround( doublevalue );
	return true;
}

HBaseAction*
AScrollWheelEvent::Merge( HBaseAction* action )
{
	check( action );
	AScrollWheelEvent *newAction;
	AScrollWheelEvent* actualAction = dynamic_cast<AScrollWheelEvent*>( action );

	if ( !actualAction ) {
		newAction = NULL;
	} else {
		SInt32 us, them;
		if ( SimpleValue( mScrollAmount->GetNS(), us ) && SimpleValue( actualAction->mScrollAmount->GetNS(), them ) ) {
			if ( us == 0 ) {
				newAction = new AScrollWheelEvent( *this );
				newAction->mScrollAmount = CFString::CreateWithNumber( them );
				newAction->mScrollDirection = actualAction->mScrollDirection;
			} else if ( them == 0 ) {
				newAction = new AScrollWheelEvent( *this );
			} else if ( mScrollDirection == actualAction->mScrollDirection && sign(us) == sign(them) ) {
				newAction = new AScrollWheelEvent( *this );
				newAction->mScrollAmount = CFString::CreateWithNumber( us + them  );
			} else {
				newAction = NULL;
			}
		} else {
			newAction = NULL;
		}
	}
	
	return newAction;
}


CFStringRef
AScrollWheelEvent::CopyActionDescriptionString() const
{
	CFStringRef theFormatKey;
	switch ( mScrollDirection ) {
		case ScrollWheelDirectionType::kUp:
			theFormatKey = CFSTR("aScrollWheel_ActionDescFormatUp");
			break;
		case ScrollWheelDirectionType::kDown:
			theFormatKey = CFSTR("aScrollWheel_ActionDescFormatDown");
			break;
		case ScrollWheelDirectionType::kLeft:
			theFormatKey = CFSTR("aScrollWheel_ActionDescFormatLeft");
			break;
		case ScrollWheelDirectionType::kRight:
			theFormatKey = CFSTR("aScrollWheel_ActionDescFormatRight");
			break;
		case ScrollWheelDirectionType::kObsoleteHorizontal:
		case ScrollWheelDirectionType::kObsoleteVertical:
			XAssert( false );
	}
	return CFString::CreateWithFormat( CFCS(theFormatKey->CopyLocalized()), mScrollAmount.Get() );
}

#if KMF_IS_ENGINE

float
AScrollWheelEvent::DoAction()
{
	double doubleamount;
	if ( !Calculate::CalculateValue( mScrollAmount->GetNS(), doubleamount ) ) {
		XErrorHandlerMaybeLogFormat1("always", CFSTR("Scroll Wheel Event %@ failed"), mScrollAmount->GetNS() );
		return 0;
	}
	
	SInt32 amount = lround(doubleamount);
	bool vertical;
	switch ( mScrollDirection ) {
		case ScrollWheelDirectionType::kUp:
			vertical = true;
			break;
		case ScrollWheelDirectionType::kDown:
			vertical = true;
			amount = -amount;
			break;
		case ScrollWheelDirectionType::kLeft:
			vertical = false;
			break;
		case ScrollWheelDirectionType::kRight:
			vertical = false;
			amount = -amount;
			break;
		case ScrollWheelDirectionType::kObsoleteHorizontal:
		case ScrollWheelDirectionType::kObsoleteVertical:
			XAssert( false );
	}
	UCGRemoteControl::PostScrollWheel( vertical, amount );
	
	return 1;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AScrollWheelEvent::ActionController() const
{
	return [AScrollWheelEventEditor class];
}

#endif

#pragma mark streaming

static const CFStringRef kScrollWheelDirectionNames[] = kScrollWheelDirectionNamesArray;

void
AScrollWheelEvent::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("ScrollAmountExpression"), CFStringRef(mScrollAmount) );
	dictionary->Set( CFSTR("ScrollDirection"), kScrollWheelDirectionNames[int(mScrollDirection)] );
}

void
AScrollWheelEvent::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mScrollDirection = dictionary->GetEnumClass( CFSTR("ScrollDirection"), kScrollWheelDirectionNames, ScrollWheelDirectionType::kDEFAULT );
	if ( mScrollDirection == ScrollWheelDirectionType::kObsoleteHorizontal || mScrollDirection == ScrollWheelDirectionType::kObsoleteVertical ) { // Old Style!
		SInt32 amount;
		if ( dictionary->Get( CFSTR("ScrollAmount"), amount ) ) {
			mScrollAmount = Calculate::CreateDoubleValue( abs(amount) * 60 );
		} else {
			mScrollAmount = kDefaultScrollAmount;
			amount = [(NSString*)kDefaultScrollAmount->GetNS() integerValue];
		}
		if ( mScrollDirection == ScrollWheelDirectionType::kObsoleteHorizontal ) {
			if ( amount >= 0 ) {
				mScrollDirection = ScrollWheelDirectionType::kLeft;
			} else {
				mScrollDirection = ScrollWheelDirectionType::kRight;
			}
		} else {
			if ( amount >= 0 ) {
				mScrollDirection = ScrollWheelDirectionType::kUp;
			} else {
				mScrollDirection = ScrollWheelDirectionType::kDown;
			}
		}
	} else {
		CFStringRef value;
		dictionary->Get( CFSTR("ScrollAmountExpression"), value, CFStringRef(NULL) );
		mScrollAmount = value ? value->Copy() : kDefaultScrollAmount;
	}
}

