//
//  ASearch.h
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 24 Sep 2009.
//  Copyright (c) 2009 Stairways Software. All rights reserved.
//

#import "ASearch.h"
#import "NSColorExtensions.h"
#import "NSStringExtensions.h"

#if KMF_IS_ENGINE
#import "SearchWindowController.h"
#import "TriggerByNameController.h"
#import "TokenText.h"
#endif
#if KMF_IS_EDITOR
#import "ASearchEditor.h"
#import "TemplateEngine.h"
#import "NSImageExtensions.h"
#endif

static const int kDefaultAlpha = 90;
static const RGBColor kDefaultColor = { 0, 152, 255 };

static NSString* GetDefaultSearchTitle()
{
	static NSString* kDefaultSearchTitle = [@"DefaultSearchTitle" localized];
	return kDefaultSearchTitle;
}

static NSString* GetDefaultSearchURL()
{
	static NSString* kDefaultSearchURL = [@"DefaultSearchURL" localized];
	return kDefaultSearchURL;
}

ASearch::ASearch()
	:	HBaseAction(),
		mTitle( GetDefaultSearchTitle() ),
		mURL( GetDefaultSearchURL() ),
		mDefault( @"" )
{
#if KMF_IS_ENGINE
	mSearchWindowController = nil;
#endif
	mColor = [NSColor colorWithCalibratedRed:kDefaultColor.red/255.0 green:kDefaultColor.green/255.0 blue:kDefaultColor.blue/255.0 alpha:kDefaultAlpha/100.0];
}

ASearch::ASearch( const ASearch& inAction )
	:	HBaseAction( inAction ),
		mColor( [inAction.mColor copy] ),
		mTitle( inAction.mTitle ),
		mURL( inAction.mURL ),
		mDefault( inAction.mDefault )
{
#if KMF_IS_ENGINE
	mSearchWindowController = nil;
#endif
}

ASearch::~ASearch()
{
#if KMF_IS_ENGINE
	[mSearchWindowController Kill];
#endif
}

void
ASearch::SetColor( NSColor* inColor )
{
	mColor = [[inColor colorUsingColorSpaceName:NSCalibratedRGBColorSpace] copy];
}

#if KMF_IS_ENGINE

float
ASearch::DoAction()
{
	float time = 0;
	mSearchWindowController = [SearchWindowController newSearchWindow];
	[mSearchWindowController setColor:mColor];
	[mSearchWindowController setTitle:mTitle];
	[mSearchWindowController setURL:mURL];
	NSString* defaulSearch = TokenText::GetResult( mDefault, &time );
	[mSearchWindowController setDefaultSearch:defaulSearch];
	[mSearchWindowController Show];
	return 20 + time;
}

Boolean
ASearch::IsActionCompleted()
{
	if ( [mSearchWindowController IsComplete] ) {
		mSearchWindowController = nil;
	}
	return mSearchWindowController == nil;
}

void
ASearch::ActionAborted()
{
	[mSearchWindowController Kill];
	mSearchWindowController = nil;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASearch::ActionController() const
{
	return [ASearchEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 <span class="IconTitle">Color: </span><img src={{ color | inlineimage: 16 }} alt="Swatch" width="16" height="16" /><br />
 Title: {{ title | htmlencode }}<br />
 URL: {{ URL | htmlencode }}<br />
 {%if default %}Default: {{ default | htmlencode }}<br />{% /if %}
)RAW";

NSString*
ASearch::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"title" : mTitle,
									@"URL" : mURL,
									@"default" : mDefault,
									@"color" : [NSImage imageWithSwatch:mColor],
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

void
ASearch::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	[mColor saveWithAlphaToDictionary:dictionary];
	dictionary->Set( CFSTR("Title"), CFBridgingToCFString(mTitle) );
	dictionary->Set( CFSTR("URL"), CFBridgingToCFString(mURL) );
	dictionary->Set( CFSTR("Default"), CFBridgingToCFString(mDefault) );
}

void
ASearch::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mColor = [[NSColor colorFromDictionary:dictionary defaultColor:kDefaultColor defaultAlpha:kDefaultAlpha] copy];

	CFStringRef theString;
	
	dictionary->Get( CFSTR("Title"), theString, CFBridgingToCFString(GetDefaultSearchTitle()) );
	check( theString );
	mTitle = theString->GetNS();
	
	dictionary->Get( CFSTR("URL"), theString, CFBridgingToCFString(GetDefaultSearchURL()) );
	check( theString );
	mURL = theString->GetNS();

	dictionary->Get( CFSTR("Default"), theString, CFSTR("") );
	check( theString );
	mDefault = theString->GetNS();
}

