//
//  ASearchClipboard.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Mon 4 Feb 2013.
//  Copyright (c) 2013 Stairways Software. All rights reserved.
//

#import "ASearchClipboard.h"
#import "NSObjectExtensions.h"
#import "TokenText.h"
#import "NSStringRegularExpressions.h"
#import "XClipboardDataSet.h"

#if KMF_IS_ENGINE
#import <string>
#import "NSStringExtensions.h"
#import "KMNotifications.h"
#import "KMVariables.h"
#endif
#if KMF_EDITOR_WINDOW
#import "ASearchClipboardEditor.h"
#import "TemplateEngine.h"
#endif

CFStringRef
ASearchClipboard::CopyActionDescriptionString() const
{
	CFStringRef key;
	switch ( mAction ) {
		case SearchClipboardActionType::kCaseSensitiveRegEx:
			key = CFSTR("SearchClipboardActionType::kCaseSensitiveRegEx");
			break;
		case SearchClipboardActionType::kIgnoreCaseRegEx:
			key = CFSTR("SearchClipboardActionType::kIgnoreCaseRegEx");
			break;
	}

	CFContainer<CFStringRef> format = key->CopyLocalized();
	return CFString::CreateWithFormat( format, mTarget.GetDisplayName() );
}

void
ASearchClipboard::SetRequiredCaptures()
{
	if ( IsEmpty( mCaptures ) ) {
		mCaptures = @[@""];
	}
	NSUInteger req = RequiredCaptureCount();
	if ( req > 0 ) {
		while ( mCaptures.count < req ) {
			mCaptures = [mCaptures arrayByAddingObject:@""];
		}
		if ( req < mCaptures.count ) {
			mCaptures = [mCaptures subarrayWithRange:NSMakeRange( 0, req )];
		}
	}
}

NSUInteger
ASearchClipboard::RequiredCaptureCount() const
{
	NSString* search = TokenText::GetResult( mSearch, NULL, false );
	NSRegularExpression* result = [NSRegularExpression regularExpressionWithPattern:search options:0 error:NULL];
	if ( !result ) return 0;
	return result.numberOfCaptureGroups + 1;
}

NSArray*
ASearchClipboard::PrivateGetCurrentResults( float* time, bool testRegex )
{
	float stime = 0;
	NSString* search = TokenText::GetResult( mSearch, &stime, false );
	NSArray* result = nil;
	BOOL good = true;
	if ( testRegex ) {
		NSRegularExpression* result = [NSRegularExpression regularExpressionWithPattern:search options:0 error:NULL];
		good = (result != NULL);
	}
	if ( good ) {
		std::unique_ptr<XClipboardDataSet> clip( mTarget.CopyClipboardDataSet() );
		CFContainer<CFStringRef> clipstr = clip ? clip->CopyIfText() : NULL;
		NSString* value = clipstr ? clipstr->GetNS() : nil;
		if ( !IsEmpty( search ) ) {
			switch ( mAction ) {
				case SearchClipboardActionType::kCaseSensitiveRegEx:
					result = [value captureComponentsMatchedByRegex:search options:0];
					break;
				case SearchClipboardActionType::kIgnoreCaseRegEx:
					result = [value captureComponentsMatchedByRegex:search options:NSRegularExpressionCaseInsensitive];
					break;
			}
		}
	}
//	NSLog( @"search is %@, value is %@, result is %@", search, value, result );
	if ( time ) {
		*time = stime;
	}
	return result;
}

NSArray*
ASearchClipboard::GetCurrentResults()
{
	return PrivateGetCurrentResults( NULL, true );
}

#if KMF_IS_ENGINE

float
ASearchClipboard::DoAction()
{
	float stime;
	NSArray* result;
	result = PrivateGetCurrentResults( &stime, false );

	BOOL changed = false;
	for ( NSInteger i = 0; i < (NSInteger)result.count && i <= (NSInteger)mCaptures.count; i++ ) {
		NSString* variable = mCaptures[i];
		NSString* value = result[i];
		if ( !IsEmpty( variable ) && KMVariables::ValidVariable( variable ) ) {
			KMVariables::Set( variable, value, nil );
			changed = true;
		}
	}
	if ( changed ) {
		KMVariables::Save( KMNotifications::kLoadVariableFromEngineAction );
		stime += 5;
	}

	return stime;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASearchClipboard::ActionController() const
{
	return [ASearchClipboardEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
Search for “{{ search }}”<br />
{% if captures %}
And capture to:<br />
<ol class="VariableList" start="0">
	{% for variable in captures %}
		<li>{% if variable %}{{ variable | htmlencode }}{% else %}<em>Ignored</em>{% /if %}</li>
	{% /for %}
</ol>
{% /if %}
)RAW";

NSString*
ASearchClipboard::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"search" : mSearch,
									@"captures" : mCaptures,
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kSearchClipboardActionTypeNames[] = kSearchClipboardActionTypeNamesArray;

void
ASearchClipboard::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Action"), kSearchClipboardActionTypeNames[int(mAction)] );
	mTarget.SaveToDictionary( dictionary, CFSTR("Target") );
	dictionary->Set( CFSTR("Search"), CFBridgingToCFString(mSearch) );
	dictionary->Set( CFSTR("Captures"), (CFArrayRef)CFBridgingToCF(mCaptures) );
}

void
ASearchClipboard::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mAction = dictionary->GetEnumClass( CFSTR("Action"), kSearchClipboardActionTypeNames, SearchClipboardActionType::kDEFAULT );

	mTarget.InitializeFromDictionary( dictionary, CFSTR("Target") );

	CFStringRef newSearch;
	dictionary->Get( CFSTR("Search"), newSearch, CFSTR("") );
	check( newSearch );
	mSearch = newSearch->GetNS();

	mCaptures = nil;
	CFArrayRef newCaptures;
	dictionary->Get( CFSTR("Captures"), newCaptures, CFArrayRef(NULL) );
	if ( newCaptures ) {
		mCaptures = newCaptures->GetNS();
	}
	SetRequiredCaptures();
}

