//
//  ASearchReplaceClipboard.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Mon 4 Apr 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "ASearchReplaceClipboard.h"

#if KMF_IS_ENGINE
#import <string>
#import "NSStringRegularExpressions.h"
#import "NSStringExtensions.h"
#import "NSAttributedStringExtensions.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#import "TokenText.h"
#import "NSObjectExtensions.h"
#endif
#if KMF_EDITOR_WINDOW
#import "ASearchReplaceClipboardEditor.h"
#import "TemplateEngine.h"
#endif

CFStringRef
ASearchReplaceClipboard::CopyActionDescriptionString() const
{
	CFStringRef key;
	switch ( mAction ) {
		case SearchReplaceClipboardActionType::kCaseSensitiveString:
			key = CFSTR("SearchReplaceClipboardActionType::kCaseSensitiveString");
			break;
		case SearchReplaceClipboardActionType::kIgnoreCaseString:
			key = CFSTR("SearchReplaceClipboardActionType::kIgnoreCaseString");
			break;
		case SearchReplaceClipboardActionType::kCaseSensitiveRegEx:
			key = CFSTR("SearchReplaceClipboardActionType::kCaseSensitiveRegEx");
			break;
		case SearchReplaceClipboardActionType::kIgnoreCaseRegEx:
			key = CFSTR("SearchReplaceClipboardActionType::kIgnoreCaseRegEx");
			break;
	}

	CFContainer<CFStringRef> format = key->CopyLocalized();
	return CFString::CreateWithFormat( format, mTarget.GetDisplayName() );
}

#if KMF_IS_ENGINE

float
ASearchReplaceClipboard::DoAction()
{
	bool normalSearch = (mAction == SearchReplaceClipboardActionType::kCaseSensitiveString) || (mAction == SearchReplaceClipboardActionType::kIgnoreCaseString);
	float stime = 0;
	float rtime = 0;
	const XClipboardDataSet* clip = mTarget.GetClipboardDataSet();
	if ( clip ) {
		NSAttributedString* inClipboard = clip->AttributedString( true );
		bool wasPlainText = inClipboard.wasPlainText;
		NSString* search = TokenText::GetResult( mSearch, &stime, normalSearch );
		NSString* replace = TokenText::GetResult( mReplace, &rtime );
		NSMutableAttributedString* clipboard = [inClipboard mutableCopy];
		if ( search.length > 0 ) {
			switch ( mAction ) {
				case SearchReplaceClipboardActionType::kCaseSensitiveString:
					[clipboard.mutableString replaceOccurrencesOfString:search withString:replace options:0 range:NSMakeRange(0, [clipboard length])];
					break;
				case SearchReplaceClipboardActionType::kIgnoreCaseString:
					[clipboard.mutableString replaceOccurrencesOfString:search withString:replace options:NSCaseInsensitiveSearch range:NSMakeRange(0, [clipboard length])];
					break;
				case SearchReplaceClipboardActionType::kCaseSensitiveRegEx:
					[clipboard.mutableString replaceOccurrencesOfRegex:search withString:replace];
					break;
				case SearchReplaceClipboardActionType::kIgnoreCaseRegEx:
					[clipboard.mutableString replaceOccurrencesOfRegex:search withString:replace options:NSRegularExpressionCaseInsensitive  ];
					break;
			}
		}

		// Avoid adding default formatting to plain text clipboards
		mTarget.WriteClipboardDataSet( wasPlainText, clipboard );
	}
	return stime + rtime + 10;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASearchReplaceClipboard::ActionController() const
{
	return [ASearchReplaceClipboardEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
Search for “{{ search }}”<br />
Replace with “{{ replace }}”
)RAW";

NSString*
ASearchReplaceClipboard::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"search" : mSearch,
									@"replace" : mReplace,
								};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kSearchReplaceClipboardActionTypeNames[] = kSearchReplaceClipboardActionTypeNamesArray;

void
ASearchReplaceClipboard::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Action"), kSearchReplaceClipboardActionTypeNames[int(mAction)] );
	dictionary->Set( CFSTR("Search"), CFBridgingToCFString(mSearch) );
	dictionary->Set( CFSTR("Replace"), CFBridgingToCFString(mReplace) );
	mTarget.SaveToDictionary( dictionary, CFSTR("Target") );
}

void
ASearchReplaceClipboard::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mAction = dictionary->GetEnumClass( CFSTR("Action"), kSearchReplaceClipboardActionTypeNames, SearchReplaceClipboardActionType::kDEFAULT );

	CFStringRef temp;
	dictionary->Get( CFSTR("Search"), temp, CFSTR("") );
	check( temp );
	mSearch = temp->GetNS();

	dictionary->Get( CFSTR("Replace"), temp, CFSTR("") );
	check( temp );
	mReplace = temp->GetNS();
	mTarget.InitializeFromDictionary( dictionary, CFSTR("Target") );
}
