//
//  ASearchReplaceVariable.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Tue 19 Apr 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "ASearchReplaceVariable.h"
#import "KMVariables.h"

#if KMF_IS_ENGINE
#import <string>
#import "NSStringRegularExpressions.h"
#import "NSStringExtensions.h"
#import "TokenText.h"
#import "KMNotifications.h"
#endif
#if KMF_EDITOR_WINDOW
#import "ASearchReplaceVariableEditor.h"
#import "TemplateEngine.h"
#endif

ASearchReplaceVariable::ASearchReplaceVariable()
	:	HBaseAction(),
		mAction( SearchReplaceVariableActionType::kDEFAULT ),
		mVariable( CFSTR("Variable") ),
		mSearch( CFSTR("") ),
		mReplace( CFSTR("") )
{
}

ASearchReplaceVariable::ASearchReplaceVariable( const ASearchReplaceVariable& inAction )
	:	HBaseAction( inAction ),
		mAction( inAction.mAction ),
		mVariable( inAction.mVariable ),
		mSearch( inAction.mSearch ),
		mReplace( inAction.mReplace )
{
}

ASearchReplaceVariable::~ASearchReplaceVariable()
{
}

CFStringRef
ASearchReplaceVariable::CopyActionDescriptionString() const
{
	CFStringRef key;
	switch ( mAction ) {
		case SearchReplaceVariableActionType::kCaseSensitiveString:
			key = CFSTR("SearchReplaceVariableActionType::kCaseSensitiveString");
			break;
		case SearchReplaceVariableActionType::kIgnoreCaseString:
			key = CFSTR("SearchReplaceVariableActionType::kIgnoreCaseString");
			break;
		case SearchReplaceVariableActionType::kCaseSensitiveRegEx:
			key = CFSTR("SearchReplaceVariableActionType::kCaseSensitiveRegEx");
			break;
		case SearchReplaceVariableActionType::kIgnoreCaseRegEx:
			key = CFSTR("SearchReplaceVariableActionType::kIgnoreCaseRegEx");
			break;
	}

	CFContainer<CFStringRef> format = key->CopyLocalized();
	return CFString::CreateWithFormat( format, CFStringRef(mVariable) );
}

#if KMF_IS_ENGINE

float
ASearchReplaceVariable::DoAction()
{
	bool normalSearch = (mAction == SearchReplaceVariableActionType::kCaseSensitiveString) || (mAction == SearchReplaceVariableActionType::kIgnoreCaseString);
	float stime, rtime;
	CFContainer<CFStringRef> search = TokenText::CopyResult( mSearch, stime, normalSearch );
	CFContainer<CFStringRef> replace = TokenText::CopyResult( mReplace, rtime );
	NSMutableString* value = [NSMutableString stringWithString:KMVariables::GetWithPassword( mVariable->GetNS() )];
//	NSLog( @"ASearchReplaceVariable::DoAction %@ %@ %@ %@", mVariable->GetNS(), value, search->GetNS(), replace->GetNS() );
	if ( search->GetLength() > 0 ) {
		switch ( mAction ) {
			case SearchReplaceVariableActionType::kCaseSensitiveString:
				[value replaceOccurrencesOfString:search->GetNS() withString:replace->GetNS() options:0 range:NSMakeRange(0, [value length])];
				break;
			case SearchReplaceVariableActionType::kIgnoreCaseString:
				[value replaceOccurrencesOfString:search->GetNS() withString:replace->GetNS() options:NSCaseInsensitiveSearch range:NSMakeRange(0, [value length])];
				break;
			case SearchReplaceVariableActionType::kCaseSensitiveRegEx:
				[value replaceOccurrencesOfRegex:search->GetNS() withString:replace->GetNS()];
				break;
			case SearchReplaceVariableActionType::kIgnoreCaseRegEx:
				[value replaceOccurrencesOfRegex:search->GetNS() withString:replace->GetNS() options:NSRegularExpressionCaseInsensitive];
				break;
		}
	}
//	NSLog( @"ASearchReplaceVariable::DoAction %@ %@ %@ %@", mVariable->GetNS(), value, mSearch->GetNS(), mReplace->GetNS() );

	KMVariables::Set( mVariable->GetNS(), value, KMNotifications::kLoadVariableFromEngineAction );
	return stime + rtime + 10;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASearchReplaceVariable::ActionController() const
{
	return [ASearchReplaceVariableEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
<ul class="IconList">
	<li>Search for “{{ search }}”</li>
	<li>Replace with “{{ replace }}”</li>
</ul>
)RAW";

NSString*
ASearchReplaceVariable::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"search" : mSearch->GetNS(),
									@"replace" : mReplace->GetNS(),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kSearchReplaceVariableActionTypeNames[] = kSearchReplaceVariableActionTypeNamesArray;

void
ASearchReplaceVariable::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Action"), kSearchReplaceVariableActionTypeNames[int(mAction)] );
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	dictionary->Set( CFSTR("Search"), CFStringRef(mSearch) );
	dictionary->Set( CFSTR("Replace"), CFStringRef(mReplace) );
}

void
ASearchReplaceVariable::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mAction = dictionary->GetEnumClass( CFSTR("Action"), kSearchReplaceVariableActionTypeNames, SearchReplaceVariableActionType::kDEFAULT );

	CFStringRef newVariable;
	dictionary->Get( CFSTR("Variable"), newVariable, CFSTR("Variable") );
	check( newVariable );
	mVariable = newVariable->Copy();
	
	CFStringRef newSearch;
	dictionary->Get( CFSTR("Search"), newSearch, CFSTR("") );
	check( newSearch );
	mSearch = newSearch->Copy();

	CFStringRef newReplace;
	dictionary->Get( CFSTR("Replace"), newReplace, CFSTR("") );
	check( newReplace );
	mReplace = newReplace->Copy();
}
