//
//  ASearchVariable.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 1 Feb 2013.
//  Copyright (c) 2013 Stairways Software. All rights reserved.
//

#import "ASearchVariable.h"
#import "KMVariables.h"
#import "NSObjectExtensions.h"
#import "TokenText.h"
#import "NSStringRegularExpressions.h"

#if KMF_IS_ENGINE
#import <string>
#import "NSStringExtensions.h"
#import "KMNotifications.h"
#endif
#if KMF_EDITOR_WINDOW
#import "ASearchVariableEditor.h"
#import "TemplateEngine.h"
#endif

ASearchVariable::ASearchVariable()
	:	HBaseAction(),
		mAction( SearchVariableActionType::kDEFAULT ),
		mVariable( CFSTR("Variable") ),
		mSearch( @"" ),
		mCaptures( @[@""] )
{
}

CFStringRef
ASearchVariable::CopyActionDescriptionString() const
{
	CFStringRef key;
	switch ( mAction ) {
		case SearchVariableActionType::kCaseSensitiveRegEx:
			key = CFSTR("SearchVariableActionType::kCaseSensitiveRegEx");
			break;
		case SearchVariableActionType::kIgnoreCaseRegEx:
			key = CFSTR("SearchVariableActionType::kIgnoreCaseRegEx");
			break;
	}

	CFContainer<CFStringRef> format = key->CopyLocalized();
	return CFString::CreateWithFormat( format, CFStringRef(mVariable) );
}

void
ASearchVariable::SetRequiredCaptures()
{
	if ( IsEmpty( mCaptures ) ) {
		mCaptures = @[@""];
	}
	NSUInteger req = RequiredCaptureCount();
	if ( req > 0 ) {
		while ( mCaptures.count < req ) {
			mCaptures = [mCaptures arrayByAddingObject:@""];
		}
		if ( req < mCaptures.count ) {
			mCaptures = [mCaptures subarrayWithRange:NSMakeRange( 0, req )];
		}
	}
}

NSUInteger
ASearchVariable::RequiredCaptureCount() const
{
	NSString* search = TokenText::GetResult( mSearch, NULL, false );
	NSRegularExpression* result = [NSRegularExpression regularExpressionWithPattern:search options:0 error:NULL];
	if ( !result ) return 0;
	return result.numberOfCaptureGroups + 1;
}

NSArray*
ASearchVariable::PrivateGetCurrentResults( float* time, bool testRegex )
{
	float stime = 0;
	NSString* search = TokenText::GetResult( mSearch, &stime, false );
	NSArray* result = nil;
	BOOL good = true;
	if ( testRegex ) {
		NSRegularExpression* result = [NSRegularExpression regularExpressionWithPattern:search options:0 error:NULL];
		good = (result != NULL);
	}
	if ( good ) {
		NSString* value = KMVariables::GetWithPassword( mVariable->GetNS() );
		if ( !IsEmpty( search ) ) {
			switch ( mAction ) {
				case SearchVariableActionType::kCaseSensitiveRegEx:
					result = [value captureComponentsMatchedByRegex:search options:0];
					break;
				case SearchVariableActionType::kIgnoreCaseRegEx:
					result = [value captureComponentsMatchedByRegex:search options:NSRegularExpressionCaseInsensitive];
					break;
			}
		}
	}
//	NSLog( @"search is %@, value is %@, result is %@", search, value, result );
	if ( time ) {
		*time = stime;
	}
	return result;
}

NSArray*
ASearchVariable::GetCurrentResults()
{
	return PrivateGetCurrentResults( NULL, true );
}

#if KMF_IS_ENGINE

float
ASearchVariable::DoAction()
{
	float stime;
	NSArray* result;
	result = PrivateGetCurrentResults( &stime, false );

	BOOL changed = false;
	for ( NSInteger i = 0; i < (NSInteger)result.count && i <= (NSInteger)mCaptures.count; i++ ) {
		NSString* variable = mCaptures[i];
		NSString* value = result[i];
		if ( !IsEmpty( variable ) && KMVariables::ValidVariable( variable ) ) {
			KMVariables::Set( variable, value, nil );
			changed = true;
		}
	}
	if ( changed ) {
		KMVariables::Save( KMNotifications::kLoadVariableFromEngineAction );
		stime += 5;
	}

	return stime;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASearchVariable::ActionController() const
{
	return [ASearchVariableEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
Search for “{{ search }}”<br />
{% if captures %}
And capture to:<br />
<ol class="VariableList" start="0">
	{% for variable in captures %}
		<li>{% if variable %}{{ variable | htmlencode }}{% else %}<em>Ignored</em>{% /if %}</li>
	{% /for %}
</ol>
{% /if %}
)RAW";

NSString*
ASearchVariable::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"search" : mSearch,
									@"captures" : mCaptures,
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kSearchVariableActionTypeNames[] = kSearchVariableActionTypeNamesArray;

void
ASearchVariable::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Action"), kSearchVariableActionTypeNames[int(mAction)] );
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	dictionary->Set( CFSTR("Search"), CFBridgingToCFString(mSearch) );
	dictionary->Set( CFSTR("Captures"), (CFArrayRef)CFBridgingToCF(mCaptures) );
}

void
ASearchVariable::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mAction = dictionary->GetEnumClass( CFSTR("Action"), kSearchVariableActionTypeNames, SearchVariableActionType::kDEFAULT );

	CFStringRef newVariable;
	dictionary->Get( CFSTR("Variable"), newVariable, CFSTR("Variable") );
	check( newVariable );
	mVariable = newVariable->Copy();
	
	CFStringRef newSearch;
	dictionary->Get( CFSTR("Search"), newSearch, CFSTR("") );
	check( newSearch );
	mSearch = newSearch->GetNS();

	mCaptures = nil;
	CFArrayRef newCaptures;
	dictionary->Get( CFSTR("Captures"), newCaptures, CFArrayRef(NULL) );
	if ( newCaptures ) {
		mCaptures = newCaptures->GetNS();
	}
	SetRequiredCaptures();
}

