/*
 *  ASekectMenuItem.cpp
 *  Keyboard Maestro
 *
 *  Created Unknown
 *  Copyright (c) 2008 Stairways Software. All rights reserved.
 */

#import "KMDefines.h"
#import "ASelectMenuItem.h"
#import "XAXMenuBar.h"
#import "XCoreFoundation.h"
#import "XAXApplicationElement.h"
#import "XAXMenuItemElement.h"
#import "XRunningApplicationsList.h"
#import "XErrorHandler.h"
#import "ProcessSerialNumberUtils.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "KMVariables.h"
#if KMF_IS_ENGINE
#import "KMVariables.h"
#import "KMNotifications.h"
#import "NSStringExtensions.h"
#import "Growl/GrowlApplicationBridge.h"
#import "HExecuteActionListManager.h"
#endif
#if KMF_EDITOR_WINDOW
#import "ASelectMenuItemEditor.h"
#import "TemplateEngine.h"
#endif

static const CFStringRef kDefaultMenuPath[] = { CFSTR(""), CFSTR("") };
static const int kDefaultMenuPathCount = 2;

CFStringRef ASelectMenuItem::CopyActionDescriptionString() const
{
	static NSString* appfrontname = [@"hkSelectMenuItemFrontAppName" localized];
	NSString* appname = appfrontname;
	if ( GetApplicationTargetType() == ApplicationTargetApplicationType::kSpecific ) {
		XApplicationIdentifier app = GetTargetApplication();
		if ( app.HasValue() ) {
			appname = app.GetName();
		}
	}
	NSString* result;
	if ( IsEmpty( [mMenuItemPath lastObject] ) ) {
		if ( mMenuItemPath.count > 2 ) {
			result = [NSString stringWithFormat:[@"hkShowMenuItemTitleFormat" localized], [mMenuItemPath objectAtIndex:mMenuItemPath.count-2], [mMenuItemPath objectAtIndex:0], appname];
		} else {
			result = [NSString stringWithFormat:[@"hkShowMenuBarItemTitleFormat" localized], [mMenuItemPath objectAtIndex:0], appname];
		}
	} else {
		result = [NSString stringWithFormat:[@"hkSelectMenuItemTitleFormat" localized], [mMenuItemPath lastObject], [mMenuItemPath objectAtIndex:0], appname];
	}
	return (CFStringRef)CFBridgingRetain( result );
}

Boolean
ASelectMenuItem::GetMenuKeyCommand( UInt32 &modifiers, UniChar &ch ) const
{
	Boolean result = false;
	
	// Must be specific running target
	if ( GetApplicationTargetType() != ApplicationTargetApplicationType::kSpecific ) return false;
	check( GetTargetApplication().HasValue() );
	ProcessSerialNumber psn;
	if ( !mTargetApplication.GetCurrentlyTargetedPSN( psn ) ) return false;
	
	XAXMenuBar menubar = XAXApplicationElement( psn ).GetMenuBar();
	if ( !!menubar ) {
		AXUIElementRef menuitem = menubar.CopyMenuItemWithPath( (CFArrayRef)CFBridgingToCF(mMenuItemPath) );
		if ( menuitem ) {
			result = XAXMenuItemElement( menuitem ).GetMenuKeyCommand( modifiers, ch );
			CFRelease( menuitem );
		}
	}
	return result;
}

#if KMF_IS_ENGINE

static NSString* PerformMenu( Boolean stopOnFailure, AXUIElementRef menubar, NSArray* menupath, CFStringRef command )
{
	XErrorHandlerMaybeLogFormat3( "SelectMenu", @"PerformMenu on %@ path %@ command %@", menubar, menupath, command );
	CFContainer<AXUIElementRef> menuitem = XAXMenuBar(menubar).CopyMenuItemWithPath( (CFArrayRef)CFBridgingToCF(menupath) );
	NSString* cancelMessage = @"SelectMenuItemNotFound";
	if ( menuitem ) {
		//				XAXUIElement(menuitem).DumpAttribueNames();
		CFContainer<CFBooleanRef> isEnabled = XAXUIElement(menuitem).CopyAttributeValueAsBoolean( kAXEnabledAttribute );
		if ( stopOnFailure && isEnabled.Defined() && !isEnabled->Get() ) {
			cancelMessage = @"SelectMenuItemDisabled";
		} else {
			{
				CFArrayRef names;
				OSStatus e = AXUIElementCopyActionNames( menuitem, &names );
				if ( e != noErr ) names = NULL;
				XErrorHandlerMaybeLogFormat3( "SelectMenu", @"AXUIElementCopyActionNames %@ gives %ld %@", AXUIElementRef(menuitem), long(e), names );
				CFXRelease( names );
			}
			OSStatus err = menuitem->PerformAction( command );
			if ( err == noErr ) {
				cancelMessage = nil;
			} else {
				cancelMessage = @"SelectMenuItemFailed";
			}
		}
	}
	XErrorHandlerMaybeLogFormat1( "SelectMenu", @"PerformMenu Result is %@", cancelMessage );
	return cancelMessage;
}

static NSString* PickOrShowMenu( Boolean stopOnFailure, AXUIElementRef menubar, NSArray* menupath, Boolean* showing )
{
	if ( menupath.count < 2 || !IsEmpty( menupath.lastObject ) ) {
		return PerformMenu( stopOnFailure, menubar, menupath, kAXPressAction );
	}
	
	NSString* cancelMessage = nil;
	for ( NSUInteger c = 1; c <= menupath.count - 1; c++ ) { // Must go through at least once!
		NSArray* path = [menupath subarrayWithRange:NSMakeRange( 0, c )];
		cancelMessage = PerformMenu( stopOnFailure, menubar, path, kAXPressAction );
		if ( cancelMessage ) return cancelMessage;
	}
	*showing = true;
	return nil;
}

float
ASelectMenuItem::DoAction()
{
	Boolean didSetFrontProcess = false;
	NSString* cancelMessage = @"SelectMenuApplicationNotFound";
	OSStatus err = -1;
	if ( AXAPIEnabled() ) {
		ProcessSerialNumber psn, frontPSN;
		
		err = GetFrontProcess( &frontPSN );
		check_noerr( err );
		if ( err == noErr ) {
			if ( !mTargetApplication.GetCurrentlyTargetedPSN( psn ) ) {
				err = procNotFound;
				NSString* name = mTargetApplication.GetDisplayName();
				if ( !IsEmpty( name ) ) {
					cancelMessage = [NSString stringWithFormat:[@"SelectMenuApplicationNotFoundApp" localized], name];
				}
			}
		}
		if ( err == noErr ) {
			if ( !PSN::SameProcess( psn, frontPSN ) ) {
				bool background;
				if ( !XRunningApplicationsList::FindApplicationWithPSN( kFindApplicationDontValidate, psn, background ) || !background ) {
					err = SetFrontProcessWithOptions( &psn, kSetFrontProcessFrontWindowOnly );
					check_noerr( err );
					didSetFrontProcess = err == noErr;
				}
			}
		}

		Boolean showing = false;
		if ( err == noErr ) {
			err = menuNotFoundErr;
			
	//		XAXApplicationElement( psn ).DumpTree();
			
			XAXApplicationElement app( psn );
			if ( app ) {
				CFContainer<AXUIElementRef> menubar = app.CopyAttributeValueAsElement( kAXMenuBarAttribute );
//				check( menubar );
				cancelMessage = @"SelectMenuBarNotFound";
				if ( menubar ) {
					cancelMessage = PickOrShowMenu( mStopOnFailure, menubar, mMenuItemPath, &showing );
				}
			}
		}
		if ( didSetFrontProcess && !showing ) {
			verify_noerr( SetFrontProcessWithOptions( &frontPSN, 0 ) );
		}
	} else {
		cancelMessage = @"SelectMenuItemRequiresAccess";
	}
	
	if ( cancelMessage ) {
		cancelMessage = [NSString stringWithFormat:[cancelMessage localized], [mMenuItemPath componentsJoinedByString:@"->"], (long)err ];
		XErrorHandlerMaybeLogFormat1( "SelectMenu", CFSTR("%@"), cancelMessage );
	}
	if ( !mStopOnFailure ) {
		cancelMessage = nil;
	}
	if ( cancelMessage != nil ) {
		HExecuteActionListManager::CancelThisExecutionGroup();
	}
	KMVariables::Set( kSelectMenuResultVariable, cancelMessage ? cancelMessage : @"OK", KMNotifications::kLoadVariableFromEngineAction );
	if ( cancelMessage ) {
		[GrowlApplicationBridge 
		 notifyWithTitle:CFBridgingToNS(kmEngineName)
		 description:[NSString stringWithFormat:[@"MacroCancelledFormat" localized], cancelMessage]
		 notificationName:@"Macro Cancelled"
		 iconData:nil priority:0 isSticky:NO clickContext:nil];		
	}
	return didSetFrontProcess ? 10 : 2;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASelectMenuItem::ActionController() const
{
	return [ASelectMenuItemEditor class];
}

NSString*
ASelectMenuItem::GetActionHTMLDescriptionString() const
{
	NSString* appname = nil;
	
	switch ( GetApplicationTargetType() ) {
		case ApplicationTargetApplicationType::kSpecific: {
			XApplicationIdentifier app = GetTargetApplication();
			if ( app.HasValue() ) {
				appname = app.GetName();
			}
			break;
		}
		case ApplicationTargetApplicationType::kFront:
			break;
	}
	if ( !appname ) {
		appname = [@"hkSelectMenuItemFrontAppName" localized];
	}

	NSString* format;
	if ( IsEmpty( [mMenuItemPath lastObject] ) ) {
		format = [@"hkShowMenuItemHTMLTitleFormat" localized];
	} else {
		format = [@"hkSelectMenuItemHTMLTitleFormat" localized];
	}

	return [NSString stringWithFormat:format, appname];
}

static NSString* kActionDisplayHTML = @R"RAW(
{{ select }}: {{ menu | htmlencode }}<br />
 {%if stop %}Stop macro if menu cannot be selected.<br />{% /if %}
 Store action result in variable ‘{{ result | htmlencode }}’.
)RAW";

NSString*
ASelectMenuItem::GetActionExtraHTML() const
{
	NSArray* items = mMenuItemPath;
	if ( IsEmpty( [mMenuItemPath lastObject] ) ) {
		items = [mMenuItemPath subarrayWithRange:NSMakeRange(0, mMenuItemPath.count-1)];
	}

	NSDictionary *variables = @{
									@"select" : IsEmpty( [mMenuItemPath lastObject] ) ? @"Show" : @"Select",
									@"menu" : [items componentsJoinedByString:@" ⇢ "],
									@"stop" : @(mStopOnFailure),
									@"result" : kSelectMenuResultVariable,
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

void
ASelectMenuItem::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	mTargetApplication.SaveToDictionary( dictionary, CFSTR("TargetingType"), CFSTR("TargetApplication") );
	dictionary->Set( CFSTR("Menu"), CFBridgingToCF(mMenuItemPath) );
	dictionary->SetBoolean( CFSTR("StopOnFailure"), mStopOnFailure );
}

void
ASelectMenuItem::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mTargetApplication.InitializeFromDictionary( dictionary, CFSTR("TargetingType"), CFSTR("TargetApplication") );
	CFArrayRef menupath;
	XAssert( dictionary->Get( CFSTR("Menu"), menupath ) );
	mMenuItemPath = CFBridgingToNS(menupath);
	dictionary->GetBoolean( CFSTR("StopOnFailure"), mStopOnFailure, kDefaultSelectMenuItemStopOnFailure );
}

