//
//  ASendMIDIMessage.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Mon 15 Mar 2010.
//  Copyright (c) 2010 Stairways Software. All rights reserved.
//

#import "ASendMIDIMessage.h"
#import "MIDISender.h"
#if KMF_EDITOR_WINDOW
#import "ASendMIDIMessageEditor.h"
#endif

const int kMiddleCNote = 60;
const int kDefaultChannel = 1;

static inline UInt32 DefaultNoteForMIDIMessageType( const MIDIMessageType& inType )
{
	UInt32 note;
	switch ( inType ) {
		case MIDIMessageType::kControlChange:
			note = 0;
			break;
		case MIDIMessageType::kNoteOff:
			note = kMiddleCNote;
			break;
		case MIDIMessageType::kNoteOn:
			note = kMiddleCNote;
			break;
	}
	return note;
}

static inline UInt32 DefaultValueForMIDIMessageType( const MIDIMessageType& inType )
{
	UInt32 value;
	switch ( inType ) {
		case MIDIMessageType::kControlChange:
			value = 63;
			break;
		case MIDIMessageType::kNoteOff:
			value = 0;
			break;
		case MIDIMessageType::kNoteOn:
			value = 127;
			break;
	}
	return value;
}

ASendMIDIMessage::ASendMIDIMessage()
	:	HBaseAction(),
		mMIDIMessageType( MIDIMessageType::kDEFAULT ),
		mChannel( kDefaultChannel ),
		mNote( DefaultNoteForMIDIMessageType( MIDIMessageType::kDEFAULT ) ),
		mValue( DefaultValueForMIDIMessageType( MIDIMessageType::kDEFAULT ) )
{
#if KMF_IS_ENGINE
	MIDISender::Start();
#endif
}

ASendMIDIMessage::ASendMIDIMessage( const ASendMIDIMessage& inAction )
	:	HBaseAction( inAction ),
		mMIDIMessageType( inAction.mMIDIMessageType ),
		mChannel( inAction.mChannel ),
		mNote( inAction.mNote ),
		mValue( inAction.mValue )
{
#if KMF_IS_ENGINE
	MIDISender::Start();
#endif
}

ASendMIDIMessage::ASendMIDIMessage( const MIDIMessageType& inType )
	:	HBaseAction(),
		mMIDIMessageType( inType ),
		mChannel( kDefaultChannel ),
		mNote( DefaultNoteForMIDIMessageType( inType ) ),
		mValue( DefaultValueForMIDIMessageType( inType ) )
{
#if KMF_IS_ENGINE
	MIDISender::Start();
#endif
}

CFStringRef
ASendMIDIMessage::CopyActionDescriptionString() const
{
	CFStringRef		theTitleKey;
	
	switch ( mMIDIMessageType ) {
		case MIDIMessageType::kNoteOn:
			theTitleKey = CFSTR("hkSendMIDIMessageNoteOnTitleFormat");
			break;
		case MIDIMessageType::kNoteOff:
			theTitleKey = CFSTR("hkSendMIDIMessageNoteOffTitleFormat");
			break;
		case MIDIMessageType::kControlChange:
			theTitleKey = CFSTR("hkSendMIDIMessageControlChangeTitleFormat");
			break;
	}

	CFContainer<CFStringRef> format = theTitleKey->CopyLocalized();
	return CFString::CreateWithFormat( format, mChannel, mNote, mValue );
}

NSString*
ASendMIDIMessage::GetNewActionTitle() const
{
	NSString* theTitleKey;
	switch ( mMIDIMessageType ) {
		case MIDIMessageType::kNoteOn:
			theTitleKey = @"hkSendMIDIMessageNoteOnTitle";
			break;
		case MIDIMessageType::kNoteOff:
			theTitleKey = @"hkSendMIDIMessageNoteOffTitle";
			break;
		case MIDIMessageType::kControlChange:
			theTitleKey = @"hkSendMIDIMessageControlChangeTitle";
			break;
	}
	return NSLOCALIZE( theTitleKey );
}

#if KMF_IS_ENGINE

float
ASendMIDIMessage::DoAction()
{
	float time = 5;
	switch ( mMIDIMessageType ) {
		case MIDIMessageType::kNoteOn:
			MIDISender::SendNoteOn( mChannel, mNote, mValue );
			break;
		case MIDIMessageType::kNoteOff:
			MIDISender::SendNoteOff( mChannel, mNote );
			break;
		case MIDIMessageType::kControlChange:
			MIDISender::SendControlChange( mChannel, mNote, mValue );
			break;
	}
	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASendMIDIMessage::ActionController() const
{
	return [ASendMIDIMessageEditor class];
}

#endif

#pragma mark streaming

static const CFStringRef kMIDIMessageTypeNames[] = kMIDIMessageTypeNamesArray;

void
ASendMIDIMessage::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("MIDIMessageType"), kMIDIMessageTypeNames[int(mMIDIMessageType)] );
	dictionary->SetNumber( CFSTR("Channel"), mChannel );
	dictionary->SetNumber( CFSTR("Note"), mNote );
	dictionary->SetNumber( CFSTR("Value"), mValue );
}

void
ASendMIDIMessage::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mMIDIMessageType = dictionary->GetEnumClass( CFSTR("MIDIMessageType"), kMIDIMessageTypeNames, MIDIMessageType::kDEFAULT );
	if ( !dictionary->Get( CFSTR("Channel"), mChannel ) ) mChannel = kDefaultChannel;
	if ( mChannel < 1 || mChannel > 16 ) mChannel = kDefaultChannel;
	if ( !dictionary->Get( CFSTR("Note"), mNote ) ) mNote = DefaultNoteForMIDIMessageType( mMIDIMessageType );
	if ( mNote > 127 ) mNote = DefaultNoteForMIDIMessageType( mMIDIMessageType );
	if ( !dictionary->Get( CFSTR("Value"), mValue ) ) mValue = DefaultValueForMIDIMessageType( mMIDIMessageType );
	if ( mValue > 127 ) mValue = DefaultValueForMIDIMessageType( mMIDIMessageType );
}

