//
//  ASetFileAttribute.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Wed 11 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "ASetFileAttribute.h"
#import "FileReference.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "XErrorHandler.h"
#import "NSImageExtensions.h"
#import "Calculate.h"
#import "TokenText.h"
#import "NSStringRegularExpressions.h"
#import "AGetFileAttribute.h"

#if KMF_EDITOR_WINDOW
#import "ASetFileAttributeEditor.h"
#import "TemplateEngine.h"
#endif

ASetFileAttribute::ASetFileAttribute()
	:	HBaseAction(),
		mAttribute( FileAttributeType::kDEFAULT ),
		mPath( CFSTR("") ),
		mText( CFSTR("NOW()")->Copy() )
{
	static_assert( FileAttributeType::kDEFAULT == FileAttributeType::kModificationDate, "FileAttributeType::kDEFAULT must be kModificationDate" );
}

CFStringRef
ASetFileAttribute::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> format = CFSTR("kmSetFileAttributeTitleFormat")->CopyLocalized();
	CFContainer<CFStringRef> attribute = AGetFileAttribute::CopyAttributeName( mAttribute );
	return CFString::CreateWithFormat( format, CFStringRef(attribute), CFStringRef(mText) );
}

NSString*
ASetFileAttribute::GetVariableValue() const
{
	AGetFileAttribute* action = new AGetFileAttribute();
	action->SetAttribute( mAttribute );
	action->SetPath( mPath );
	CFContainer<CFStringRef> result = action->CopyTransformationResult( NULL, NULL );
//	NSLog( @"ASetFileAttribute::GetVariableValue result is %@", result ? result->GetNS() : nil );
	return result ? result->GetNS() : nil;
}

CFStringRef
ASetFileAttribute::CopyTransformationResult( NSUInteger* errorPos, float* outTime ) const
{
	bool calculate = false;
	bool processtokens = false;
	switch ( mAttribute ) {
		case FileAttributeType::kCreationDate:
		case FileAttributeType::kModificationDate:
		case FileAttributeType::kOwnerID:
		case FileAttributeType::kGroupID:
		case FileAttributeType::kPosixPermissions:
		case FileAttributeType::kExtensionHidden:
			calculate = true;
			break;
		case FileAttributeType::kHFSCreatorCode:
		case FileAttributeType::kHFSTypeCode:
			if ( [mText->GetNS() isMatchedByRegex:@"^[ -~][ -~][ -~][ -~]$"] ) {
				processtokens = true;
			} else {
				calculate = true;
			}
			break;
		case FileAttributeType::kFileType:
		case FileAttributeType::kFileSize:
		case FileAttributeType::kOwnerName:
		case FileAttributeType::kGroupName:
		case FileAttributeType::kParent:
		case FileAttributeType::kName:
		case FileAttributeType::kBaseName:
		case FileAttributeType::kExtension:
			processtokens = true;
			break;
	}
	if ( calculate ) {
		NSString* result = Calculate::Calculate( mText->GetNS(), errorPos );
		if ( outTime ) *outTime = 10;
		return (CFStringRef)CFBridgingRetain(result);
	} else if ( processtokens ) {
		float time;
		CFContainer<CFStringRef> result = TokenText::CopyResult( mText, time );
		if ( errorPos ) *errorPos = 0;
		if ( outTime ) *outTime = time;
		return result->Copy();
	} else {
		if ( outTime ) *outTime = 2;
		return mText->Copy();
	}
}

#if KMF_IS_ENGINE

static NSString* kFileAttributeNSAttributeKeyArray[] = kFileAttributeNSAttributeKeyNamesArray;

float
ASetFileAttribute::DoAction()
{
	NSUInteger errorPos;
	float time;	
	NSString* result = GetTransformationResult( &errorPos, &time );
	time += 10;
	
	NSString* path = FileReference::PathToActualPath( mPath->GetNS() );

	if ( result && [path isMatchedByRegex:@"^/.+"] ) {
		id value = nil;
		
		switch ( mAttribute ) {
			case FileAttributeType::kFileType:
			case FileAttributeType::kFileSize:
			case FileAttributeType::kParent:
			case FileAttributeType::kName:
			case FileAttributeType::kBaseName:
			case FileAttributeType::kExtension:
				check( false );
				break;
			case FileAttributeType::kCreationDate:
			case FileAttributeType::kModificationDate: {
				// date
				double v;
				if ( Calculate::CalculateValue( result, v ) ) {
					value = [NSDate dateWithTimeIntervalSince1970:v];
				} else {
					check( false ); // calculation should have failed previously
				}
				break;
			}
			case FileAttributeType::kOwnerID:
			case FileAttributeType::kGroupID:
			case FileAttributeType::kPosixPermissions:
			case FileAttributeType::kExtensionHidden: {
				// number
				double v;
				if ( Calculate::CalculateValue( result, v ) ) {
					value = [NSNumber numberWithInteger:v];
				} else {
					check( false ); // calculation should have failed previously
				}
				break;
			}
			case FileAttributeType::kOwnerName:
			case FileAttributeType::kGroupName:
				// string
				value = result;
				break;
			case FileAttributeType::kHFSCreatorCode:
			case FileAttributeType::kHFSTypeCode: {
				// type code
				OSType t;
				double v;
				if ( [result isMatchedByRegex:@"^[ -~][ -~][ -~][ -~]$"] ) {
					t = [result osType];
				} else if ( Calculate::CalculateValue( result, v ) ) {
					t = v;
				} else {
					t = 0;
				}
				value = [NSNumber numberWithUnsignedLong:t];
				break;
			}
		}
		
		if ( value ) {
			NSError* error;
			NSDictionary* attributes = [NSDictionary dictionaryWithObject:value forKey:kFileAttributeNSAttributeKeyArray[int(mAttribute)]];
			if ( ![[NSFileManager defaultManager] setAttributes:attributes ofItemAtPath:path error:&error] ) {
				XErrorHandlerMaybeLogFormat1( "display", CFSTR("Set File Attributes failed with error %@"), error );
			}
		}
	} else {
		XErrorHandlerMaybeLogFormat2("always", CFSTR("Set File Attribute failed calculation %@ with error at %d"), mText->GetNS(), errorPos );
	}
	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASetFileAttribute::ActionController() const
{
	return [ASetFileAttributeEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
File: {{ file | htmlencode }}<br />
)RAW";

NSString*
ASetFileAttribute::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"file" : mPath->GetNS(),
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kFileAttributeTypeNames[] = kFileAttributeTypeNamesArray;

void
ASetFileAttribute::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Attribute"), kFileAttributeTypeNames[int(mAttribute)] );
	dictionary->Set( CFSTR("Path"), CFStringRef(mPath) );
	dictionary->Set( CFSTR("Text"), CFStringRef(mText) );
}

void
ASetFileAttribute::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	bool good = dictionary->CanGetEnumClass( CFSTR("Attribute"), kFileAttributeTypeNames, mAttribute );
	XAssert( good );

	CFStringRef value;
	dictionary->Get( CFSTR("Text"), value, CFSTR("") );
	check( value );
	mText = value->Copy();
	FileReference::LoadPathFromDictionary( dictionary, mPath, NULL, CFSTR("Path") );
}
