/*
 *  ASimulateKeystroke.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Sun Dec 16 2001.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "ASimulateKeystroke.h"
#import "KMDefines.h"
#import "SystemDefinedKeys.h"
#import "ASelectMenuItem.h"
#import "AInsertText.h"

#if KMF_IS_ENGINE
#import "UCGRemoteControl.h"
#endif
#if KMF_EDITOR_WINDOW
#import "ASimulateKeystrokeEditor.h"
#endif

const Boolean kDefaultReleaseAll = false;

ASimulateKeystroke::ASimulateKeystroke()
	:	HBaseAction(),
		mKeyCode( kVK_Return ),
		mModifiers( 0 ),
		mReleaseAll( kDefaultReleaseAll )
{
}
ASimulateKeystroke::ASimulateKeystroke( const ASimulateKeystroke& inAction )
	:	HBaseAction( inAction ),
		mKeyCode( inAction.mKeyCode ),
		mModifiers( inAction.mModifiers ),
		mReleaseAll( inAction.mReleaseAll )
{
}

ASimulateKeystroke::~ASimulateKeystroke()
{
}

HBaseAction*
ASimulateKeystroke::Merge( HBaseAction* action )
{
	check( action );
	HBaseAction *newAction = NULL;
	ASelectMenuItem* selectMenuAction = dynamic_cast<ASelectMenuItem*>( action );
	if ( selectMenuAction ) {
		UniChar ch;
		UInt32 modifiers;
		if ( selectMenuAction->GetMenuKeyCommand( modifiers, ch ) && IsKeystroke( modifiers, SystemDefinedKeys::GetKeyCodeFromDisplayUniChar( ch ) ) ) {
			newAction = Clone(); // selectMenuAction->Clone();
		}
	} else {
		ASimulateKeystroke* simulateKeyAction = dynamic_cast<ASimulateKeystroke*>( action );
		if ( simulateKeyAction ) {
			if ( ((mModifiers & ~UInt32(shiftKey)) == 0) && ((simulateKeyAction->mModifiers & ~UInt32(shiftKey)) == 0) ) {
				UniChar ch1 = SystemDefinedKeys::GetNormalUniCharFromKeyCode( mKeyCode, 0, (mModifiers & shiftKey) ? ShiftStyle::kShift : ShiftStyle::kNormal );
				UniChar ch2 = SystemDefinedKeys::GetNormalUniCharFromKeyCode( simulateKeyAction->mKeyCode, 0, (simulateKeyAction->mModifiers & shiftKey) ? ShiftStyle::kShift : ShiftStyle::kNormal );
				if (' ' <= ch1 && ch1 <= 126 && ' ' <= ch2 && ch2 <= 126 ) {
					AInsertText* insertTextAction = new AInsertText( InsertTextActionType::kByTyping );
					insertTextAction->SetPlainText( [NSString stringWithFormat:@"%C%C", ch1, ch2] );
					newAction = insertTextAction;
				}
			}
		}
	}
	
	return newAction;
}

void
ASimulateKeystroke::SetSimulateKeystroke(
	const CGKeyCode& inKeyCode,
	const UInt32& inModifiers )
{
	mKeyCode = inKeyCode;
	mModifiers = inModifiers;
}
	
	
CFStringRef
ASimulateKeystroke::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> keystrokeStr = SystemDefinedKeys::CopyKeystrokeText( GetKeyCode(), GetModifiers(), KeyTextFormat::kLong );
	CFContainer<CFStringRef> formatStr = CFSTR("kmSimulateKeystrokeTitleFormat")->CopyLocalized();
	CFContainer<CFStringRef> theString = CFString::CreateWithFormat( formatStr, CFStringRef(keystrokeStr) );
	return theString->Copy();
}

Boolean
ASimulateKeystroke::IsKeystroke( UInt32 inModifiers, UInt32 inKeyCode ) const
{
	return ( ( inModifiers == GetModifiers() )&&( inKeyCode == GetKeyCode() ) );
}

#if KMF_IS_ENGINE

static EventTime gSimulateKeystrokeDelay;

float
ASimulateKeystroke::DoAction()
{
	gSimulateKeystrokeDelay = UCGRemoteControl::GetSimulateKeystrokeDelay( mModifiers, mKeyCode );
	UCGRemoteControl::SimulateKeystroke( mModifiers, mKeyCode, mReleaseAll );
	mActionTime = GetCurrentEventTime();
	return mModifiers == 0 ? 0.25 : 1;
}

Boolean
ASimulateKeystroke::IsActionCompleted()
{
	return GetCurrentEventTime() > mActionTime + gSimulateKeystrokeDelay;
}


#endif

#if KMF_EDITOR_WINDOW

Class
ASimulateKeystroke::ActionController() const
{
	return [ASimulateKeystrokeEditor class];
}

#endif

#pragma mark streaming

void
ASimulateKeystroke::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->SetNumber( CFSTR("Modifiers"), mModifiers );
	dictionary->SetNumber( CFSTR("KeyCode"), mKeyCode );
	dictionary->SetBoolean( CFSTR("ReleaseAll"), mReleaseAll );
}

void
ASimulateKeystroke::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	XAssert( dictionary->Get( CFSTR("Modifiers"), mModifiers ) );
	XAssert( dictionary->Get( CFSTR("KeyCode"), mKeyCode ) );
	dictionary->GetBoolean( CFSTR("ReleaseAll"), mReleaseAll, kDefaultReleaseAll );
}
