//
//  ASubstring.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 5 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "ASubstring.h"
#import "FileReference.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#import "CSClipboardDomain.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "XErrorHandler.h"
#import "NSImageExtensions.h"
#import "NSStringRegularExpressions.h"
#import "NSAttributedStringExtensions.h"

#if KMF_IS_ENGINE
#import "KMVariables.h"
#import "KMNotifications.h"
#import "UnitTestFrame.h"
#import "MMath.h"
#endif

#if KMF_EDITOR_WINDOW
#import "ASubstringEditor.h"
#import "TemplateEngine.h"
#endif

static const CFStringRef kStringRangeTypeNames[] = kStringRangeTypeNamesArray;

CFStringRef
ASubstring::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> format;
	CFContainer<CFStringRef> result;
	switch ( mSource ) {
		case SubstringSourceType::kClipboard: {
			result = CFSTR("kmSubstringClipboard")->CopyLocalized();
			break;
		}
		case SubstringSourceType::kNamedClipboard: {
			format = CFSTR("kmSubstringNamedClipboardFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, mClipboardReference.GetClipboardName() );
			break;
		}
		case SubstringSourceType::kVariable: {
			format = CFSTR("kmSubstringVariableFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, CFStringRef(mVariable) );
			break;
		}
	}
	return result->Copy();
}

#if KMF_IS_ENGINE

float
ASubstring::DoAction()
{
	float time = 5;

	NSAttributedString* theAttributedText = nil;
	CSClipboardDataSet*	pClip = NULL;
	switch ( mSource ) {
		case SubstringSourceType::kClipboard: {
			XClipboardDataSet* oldScrap = XClipboardHistory::CopyCurrent();
			if ( oldScrap ) {
				theAttributedText = oldScrap->AttributedString();
				delete oldScrap;
			}
			if ( !theAttributedText ) {
				theAttributedText = [[NSAttributedString alloc] initWithString:[@"NoClipboardText" localized]];
				theAttributedText.wasPlainText = true;
			}
			break;
		}
		case SubstringSourceType::kNamedClipboard: {
			mClipboardReference.UpdateClipboardReference();
			pClip = mClipboardReference.GetClipboardDataSet();
			if ( pClip ) {
				theAttributedText = pClip->AttributedString();
			}
			if ( !theAttributedText ) {
				theAttributedText = [[NSAttributedString alloc] initWithString:[@"NoClipboardText" localized]];
				theAttributedText.wasPlainText = true;
			}
			break;
		}
		case SubstringSourceType::kVariable: {
			NSString* value = KMVariables::GetWithPassword( mVariable.GetNS() );
			theAttributedText = [[NSAttributedString alloc] initWithString:value];
			theAttributedText.wasPlainText = true;
			break;
		}
	}

	NSMutableAttributedString* result = [theAttributedText mutableCopy];
	result.wasPlainText = theAttributedText.wasPlainText;

	NSRange match;
	if ( StringRange::FindMatch( match, mStringRange, result.string, mFirst, mSecond ) ) {
		if ( mDelete ) {
			[result.mutableString replaceCharactersInRange:match withString:@""];
		} else {
			if ( match.location + match.length < result.string.length ) {
				[result.mutableString replaceCharactersInRange:NSMakeRange( match.location + match.length, result.string.length - (match.location + match.length) ) withString:@""];
			}
			if ( match.location > 0 ) {
				[result.mutableString replaceCharactersInRange:NSMakeRange( 0, match.location ) withString:@""];
			}
		}
		switch ( mSource ) {
			case SubstringSourceType::kClipboard: {
				XClipboardHistory::PlaceInClipboard( result.wasPlainText ? XClipboardDataSet( result.string ) : XClipboardDataSet( result ) );
				break;
			}
			case SubstringSourceType::kNamedClipboard: {
				if ( pClip ) {
					*pClip = result.wasPlainText ? XClipboardDataSet( result.string ) : XClipboardDataSet( result );
					CSClipboardDomain::GetClipboardDomain()->SaveClipData( KMNotifications::kLoadClipDataFromEngineAction );
				}
				break;
			}
			case SubstringSourceType::kVariable: {
				KMVariables::Set( mVariable.GetNS(), result.string, KMNotifications::kLoadVariableFromEngineAction );
				break;
			}
		}
	} else {
		XErrorHandlerMaybeLogFormat2( "display", CFSTR("Failed to get range for expressions %@ and %@"), mFirst, mSecond );
	}

	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
ASubstring::ActionController() const
{
	return [ASubstringEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
 {% if range %}{{ range }}.{% /if %}
)RAW";

NSString*
ASubstring::GetActionExtraHTML() const
{
	check( mStringRange != StringRangeType::kAll );
	NSString* range = [[NSString stringWithFormat:@"kSubstringRange%@%@", (mDelete ? @"Delete" : @""), kStringRangeTypeNames[int(mStringRange)]] localized];
	if ( NeedSecond() ) {
		range = [NSString stringWithFormat:range, mFirst, mSecond];
	} else {
		range = [NSString stringWithFormat:range, mFirst];
	}
	NSDictionary *variables = @{
									@"range" : range,
	};

	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kSubstringSourceTypeNames[] = kSubstringSourceTypeNamesArray;

void
ASubstring::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Source"), kSubstringSourceTypeNames[int(mSource)] );
	dictionary->Set( CFSTR("StringRangeType"), kStringRangeTypeNames[int(mStringRange)] );
	mClipboardReference.SaveToDictionary( dictionary );
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	dictionary->SetBoolean( CFSTR( "Delete" ), mDelete );
	dictionary->Set( CFSTR("First"), CFBridgingToCFString(mFirst) );
	dictionary->Set( CFSTR("Second"), CFBridgingToCFString(mSecond) );
}

void
ASubstring::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	bool good;
	good = dictionary->CanGetEnumClass( CFSTR("Source"), kSubstringSourceTypeNames, mSource );
	XAssert( good );
	
	good = dictionary->CanGetEnumClass( CFSTR("StringRangeType"), kStringRangeTypeNames, mStringRange );
	XAssert( good );

	mClipboardReference.InitializeFromDictionary( dictionary );
	CFStringRef value;
	dictionary->Get( CFSTR("Variable"), value, CFSTR("Variable") );
	check( value );
	mVariable = value->Copy();
	dictionary->GetBoolean( CFSTR( "Delete" ), mDelete, false );

	dictionary->Get( CFSTR("First"), value, CFSTR("2") );
	check( value );
	mFirst = value->GetNS();

	dictionary->Get( CFSTR("Second"), value, CFSTR("4") );
	check( value );
	mSecond = value->GetNS();
}

#if KMF_IS_ENGINE

static void Test( NSString* source, StringRangeType t, Boolean del, NSString* answer, NSString* first, NSString* second = @"0" )
{
	NSString* kVariable = @"Test Frame";
	
	KMVariables::Set( kVariable, source, nil );
	ASubstring* action = new ASubstring();
	action->SetSource( SubstringSourceType::kVariable );
	action->SetVariable( kVariable );
	action->SetStringRange( t );
	action->SetDelete( del );
	action->SetFirst( first );
	action->SetSecond( second );
	action->DoAction();
	NSString* value = KMVariables::GetWithPassword( kVariable );
	if ( ![value isEqualToString:answer] ) {
		NSLog( @"source %@, StringRange %d, del %d, answer %@ != %@, first %@, second %@", source, int(t), del, answer, value, first, second );
	}
	UNITTEST( [value isEqualToString:answer] );
}

static UnitTestFrame gTest( ^{
	UInt64 start = MMath::HighPrecisionMicroseconds();
	Test( @"abcdefgh", StringRangeType::kTo, false, @"abc", @"3" );
	Test( @"abcdefgh", StringRangeType::kFrom, false, @"efgh", @"5" );
	Test( @"abcdefgh", StringRangeType::kLast, false, @"fgh", @"3" );
	Test( @"abcdefgh", StringRangeType::kRange, false, @"cdef", @"3", @"4" );
	Test( @"abcdefgh", StringRangeType::kFromTo, false, @"cd", @"3", @"4" );
	Test( @"abcdefgh", StringRangeType::kTo, true, @"defgh", @"3" );
	Test( @"abcdefgh", StringRangeType::kFrom, true, @"abcd", @"5" );
	Test( @"abcdefgh", StringRangeType::kLast, true, @"abcde", @"3" );
	Test( @"abcdefgh", StringRangeType::kRange, true, @"abgh", @"3", @"4" );
	Test( @"abcdefgh", StringRangeType::kFromTo, true, @"abefgh", @"3", @"4" );
	UInt64 total = MMath::HighPrecisionMicroseconds() - start;
	XErrorHandlerMaybeLogFormat1( "always", @"String Range UnitTestFrame Time used %f", total / 1000000.0 );
});

#endif
