/*
 *  ASystemAction.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Tue May 07 2002.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "ASystemAction.h"
#import "KMDefines.h"
#import "SystemDefinedKeys.h"
#import "KMIconRegistry.h"
#import "XIcon.h"
#import <unistd.h>
#import "XSystemProperties.h"
#import "MacroPaletteController.h"
#import "NSAppleScriptExtensions.h"
#import <DiscRecording/DRDevice.h>
#import "XErrorHandler.h"
#import "NSStringExtensions.h"
#import "ProcessSerialNumberUtils.h"
#import "StatusMenu.h"
#import "AsyncNSTask.h"

static const CFStringRef kSystemActionNames[] = kSystemActionNamesArray;

CFStringRef
ASystemAction::CopyActionTitle() const
{
	NSString* theTitleKey = [NSString stringWithFormat:@"hkSystemAction%@Title", kSystemActionNames[int(mSystemAction)]];
	return (CFStringRef)CFBridgingRetain([theTitleKey localized]);
}

XIcon
ASystemAction::GetActionDescriptionXIcon() const
{
	XIcon		result;
	
	switch ( GetSystemAction() ) {
		case SystemActionType::kOpenCloseCDROM:
			result = XIcon( KMIconRegistry::kEjectMediaIcon );
			break;
		case SystemActionType::kMuteVolume: {
				result = XIcon( XIcon(KMIconRegistry::kSoundIcon), XIcon(KMIconRegistry::kQuitBadgeIcon) );
			break;
		}
		case SystemActionType::kSystemBeep:
		case SystemActionType::kIncreaseVolume:
		case SystemActionType::kDecreaseVolume:
			result = XIcon( KMIconRegistry::kSoundIcon );
			break;
		case SystemActionType::kSleep:
		case SystemActionType::kRestart:
		case SystemActionType::kShutdown:
		case SystemActionType::kLogOut:
		case SystemActionType::kLoginWindow:
			result = XIcon( KMIconRegistry::kFinderIcon );
			break;
		case SystemActionType::kIncreaseBrightness:
		case SystemActionType::kDecreaseBrightness:
		case SystemActionType::kActivateScreenSaver:
		case SystemActionType::kStopScreenSaver:
		case SystemActionType::kSleepDisplay:
		case SystemActionType::kWakeDisplay:
			result = XIcon( KMIconRegistry::kDisplay );
			break;
		case SystemActionType::kShowMacroPalette:
		case SystemActionType::kHideMacroPalette:
		case SystemActionType::kToggleMacroPalette:
			result = XIcon( KMIconRegistry::kWindowIcon );
			break;
		case SystemActionType::kShowStatusMenu:
			result = XIcon( KMIconRegistry::kSelectMenuItemIcon );
			break;
	}
	
	return result;
}

#if KMF_IS_ENGINE

static void OpenCloseCDROM()
{
	NSArray* devices = [DRDevice devices];

//	for ( DRDevice* device in devices ) {
//		NSLog( @"device is %@", device );
//		NSLog( @"device info is %@", device.info );
//		NSLog( @"device status is %@", device.status );
//	}

	BOOL done = false;
	if ( !done ) {
		for ( DRDevice* device in devices ) {
			if ( [device.status[DRDeviceIsTrayOpenKey] integerValue] != 0 ) {
				if ( [device closeTray] ) {
					done = true;
					break;
				}
			}
		}
	}

	if ( !done ) {
		for ( DRDevice* device in devices ) {
			if ( [device.status[DRDeviceMediaStateKey] isEqual:DRDeviceMediaStateMediaPresent] ) {
				if ( [device ejectMedia] ) {
					done = true;
					break;
				}
			}
		}
	}

	if ( !done ) {
		for ( DRDevice* device in devices ) {
			if ( [device.info[DRDeviceLoadingMechanismCanOpenKey] integerValue] != 0 ) {
				if ( [device openTray] ) {
					done = true;
					break;
				}
			}
		}
	}

	if ( !done ) {
		XErrorHandlerMaybeLogMessage( "display", "OpenCloseCDROM could not find a device to open/close" );
	}
}

float
ASystemAction::DoAction()
{
	float time = 0;
	switch ( GetSystemAction() ) {
		case SystemActionType::kSleep:
			PSN::SendSystemCoreEventAE( kAESleep );
			time = 2;
			break;
		case SystemActionType::kRestart:
			PSN::SendSystemCoreEventAE( kAERestart );
			time = 2;
			break;
		case SystemActionType::kShutdown:
			PSN::SendSystemCoreEventAE( kAEShutDown );
			time = 2;
			break;
		case SystemActionType::kLogOut:
			PSN::SendSystemCoreEventAE( kAEReallyLogOut );
			time = 2;
			break;
		case SystemActionType::kLoginWindow: {
			AsyncNSTask* task = [[AsyncNSTask alloc] init];
			[task launchWithCommandAndArguments:@"/System/Library/CoreServices/Menu Extras/User.menu/Contents/Resources/CGSession", @"-suspend", nil];
			time = 2;
			break;
		}
		
		case SystemActionType::kOpenCloseCDROM:
			OpenCloseCDROM();
			time = 2;
			break;
		case SystemActionType::kMuteVolume:
			SystemDefinedKeys::PostHardwareKeystroke( HardwareKeyType::kSoundMute );
			time = 2;
			break;
		case SystemActionType::kIncreaseVolume:
			SystemDefinedKeys::PostHardwareKeystroke( HardwareKeyType::kSoundIncreaseVolume );
			time = 2;
			break;
		case SystemActionType::kDecreaseVolume:
			SystemDefinedKeys::PostHardwareKeystroke( HardwareKeyType::kSoundDecreaseVolume );
			time = 2;
			break;
		case SystemActionType::kIncreaseBrightness:
			SystemDefinedKeys::PostHardwareKeystroke( HardwareKeyType::kBrightnessIncrease );
			time = 2;
			break;
		case SystemActionType::kDecreaseBrightness:
			SystemDefinedKeys::PostHardwareKeystroke( HardwareKeyType::kBrightnessDecrease );
			time = 2;
			break;

		case SystemActionType::kActivateScreenSaver:
			(void)[NSAppleScript executeWithSource:@"delay 1\r tell application \"System Events\" to start current screen saver"];
			time = 2;
			break;
		case SystemActionType::kStopScreenSaver:
			(void)[NSAppleScript executeWithSource:@"tell application \"System Events\" to stop (current screen saver)"];
			time = 2;
			break;
		case SystemActionType::kSleepDisplay: {
			io_registry_entry_t entry = IORegistryEntryFromPath( kIOMasterPortDefault, "IOService:/IOResources/IODisplayWrangler" );
			if ( entry ) {
				IORegistryEntrySetCFProperty( entry, CFSTR("IORequestIdle"), kCFBooleanTrue );
				IOObjectRelease( entry );
			}
			time = 2;
			break;
		}
		case SystemActionType::kWakeDisplay: {
			io_registry_entry_t entry = IORegistryEntryFromPath( kIOMasterPortDefault, "IOService:/IOResources/IODisplayWrangler" );
			if ( entry ) {
				IORegistryEntrySetCFProperty( entry, CFSTR("IORequestIdle"), kCFBooleanFalse );
				IOObjectRelease( entry );
			}
			time = 2;
			break;
		}
			
		case SystemActionType::kSystemBeep:
			XErrorHandler::Beep();
			break;

		case SystemActionType::kShowMacroPalette:
			[MacroPaletteController showPalette];
			break;
		case SystemActionType::kHideMacroPalette:
			[MacroPaletteController hidePalette];
			break;
		case SystemActionType::kToggleMacroPalette:
			[MacroPaletteController togglePalette];
			break;
			
		case SystemActionType::kShowStatusMenu:
			StatusMenu::ShowMenu();
			break;
	}
	return time;
}

#endif

#pragma mark streaming

void
ASystemAction::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("SystemAction"), kSystemActionNames[int(mSystemAction)] );
}

void
ASystemAction::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mSystemAction = dictionary->GetEnumClass( CFSTR("SystemAction"), kSystemActionNames, SystemActionType::kDEFAULT );
}

