//
//  ATransformImage.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Tue 1 May 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "ATransformImage.h"
#import "XException.h"
#import "KMDefines.h"
#import "XErrorHandler.h"
#import "NSStringRegularExpressions.h"
#import "Calculate.h"
#import "TokenText.h"
#import "CSClipboardDomain.h"
#import "NSStringExtensions.h"

#import "XAXApplicationElement.h"
#import "XAXWindowElement.h"
#import "XAXWindowElementList.h"
#import "ScreenInfo.h"
#import "NSImageExtensions.h"
#import "NSStringExtensions.h"

#if KMF_EDITOR_WINDOW
#import "ATransformImageEditor.h"
#endif
#if KMF_IS_ENGINE
#import "XRunningApplicationsList.h"
#import "KMNotifications.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#endif

ATransformImage::ATransformImage( TransformImageActionType inAction )
	:	mTransformAction( inAction )

{
}

NSString*
ATransformImage::GetNewActionTitle() const
{
	NSString* title;
	switch ( mTransformAction ) {
		case TransformImageActionType::kFlipHorizontal:
		case TransformImageActionType::kFlipVertical:
			title = [@"kmTransformImageActionFlip" localized];
			break;
		case TransformImageActionType::kRotateCW:
		case TransformImageActionType::kRotate180:
		case TransformImageActionType::kRotateCCW:
		case TransformImageActionType::kFreeRotate:
			title = [@"kmTransformImageActionRotate" localized];
			break;
	}
	
	return title;
}

CFStringRef
ATransformImage::CopyActionDescriptionString() const
{
	CFStringRef description;
	switch ( mTransformAction ) {
		case TransformImageActionType::kFlipHorizontal:
			description = CFSTR("kmTransformImageActionFlipHorizontal")->CopyLocalized();
			break;
		case TransformImageActionType::kFlipVertical:
			description = CFSTR("kmTransformImageActionFlipVertical")->CopyLocalized();
			break;
		case TransformImageActionType::kRotateCW:
			description = CFSTR("kmTransformImageActionRotateCW")->CopyLocalized();
			break;
		case TransformImageActionType::kRotate180:
			description = CFSTR("kmTransformImageActionRotate180")->CopyLocalized();
			break;
		case TransformImageActionType::kRotateCCW:
			description = CFSTR("kmTransformImageActionRotateCCW")->CopyLocalized();
			break;
		case TransformImageActionType::kFreeRotate: {
			CFContainer<CFStringRef> format = CFSTR("kmTransformImageActionFreeRotate")->CopyLocalized();
			description = CFString::CreateWithFormat( format, mRotation );
			break;
		}
	}
	
	return description;
}

#if KMF_IS_ENGINE

float
ATransformImage::DoAction()
{	
	const XClipboardDataSet* clip = mTarget.GetClipboardDataSet();
	if ( clip ) {
		NSImage* image = clip->Image();
		if ( image ) {
			switch ( mTransformAction ) {
				case TransformImageActionType::kFlipHorizontal:
					image = [image imageFlippedHorizontally];
					break;
				case TransformImageActionType::kFlipVertical:
					image = [image imageFlippedVertically];
					break;
				case TransformImageActionType::kRotateCW:
					image = [image imageRotatedByDegrees:270];
					break;
				case TransformImageActionType::kRotate180:
					image = [image imageRotatedByDegrees:180];
					break;
				case TransformImageActionType::kRotateCCW:
					image = [image imageRotatedByDegrees:90];
					break;
				case TransformImageActionType::kFreeRotate: {
					NSRect imageRect;
					imageRect.origin = NSZeroPoint;
					imageRect.size = image.size;
					NSDictionary* calculationContext = @{ @"IMAGE" : NSStringFromRect(imageRect) };
					double rotate;
					if ( !Calculate::CalculateValue( mRotation, rotate, calculationContext ) ) {
						XErrorHandlerMaybeLogFormat1("always", CFSTR("Transform Image Rotate %@ failed"), mRotation );
					} else {
						image = [image imageRotatedByDegrees:-rotate];
					}
					break;
				}
			}
			if ( image ) {
				mTarget.WriteClipboardDataSet( XClipboardDataSet( image ) );
			}
		} else {
			XErrorHandlerMaybeLogMessage("always", CFSTR("Transform Image cannot get source image") );
		}
	}
	
	return 10;
}


#endif

#if KMF_EDITOR_WINDOW

Class
ATransformImage::ActionController() const
{
	return [ATransformImageEditor class];
}

NSString*
ATransformImage::GetActionHTMLDescriptionString() const
{
	NSString* key;
	switch ( mTransformAction ) {
		case TransformImageActionType::kFlipHorizontal:
			key = @"kmTransformImageActionHTMLFlipHorizontal";
			break;
		case TransformImageActionType::kFlipVertical:
			key = @"kmTransformImageActionHTMLFlipVertical";
			break;
		case TransformImageActionType::kRotateCW:
			key = @"kmTransformImageActionHTMLRotateCW";
			break;
		case TransformImageActionType::kRotate180:
			key = @"kmTransformImageActionHTMLRotate180";
			break;
		case TransformImageActionType::kRotateCCW:
			key = @"kmTransformImageActionHTMLRotateCCW";
			break;
		case TransformImageActionType::kFreeRotate: {
			key = @"kmTransformImageActionHTMLFreeRotate";
			break;
		}
	}
	if ( mTransformAction == TransformImageActionType::kFreeRotate ) {
		return [NSString stringWithFormat:[key localized], mTarget.GetDisplayName(), mRotation];
	} else {
		return [NSString stringWithFormat:[key localized], mTarget.GetDisplayName()];
	}
}

#endif

#pragma mark streaming

static const CFStringRef kTransformImageActionTypeNames[] = kTransformImageActionTypeNamesArray;

void
ATransformImage::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("TransformAction"), kTransformImageActionTypeNames[int(mTransformAction)] );
	mTarget.SaveToDictionary( dictionary, CFSTR("Target") );
	dictionary->Set( CFSTR("Rotation"), CFBridgingToCFString(mRotation) );
}

void
ATransformImage::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mTransformAction = dictionary->GetEnumClass( CFSTR("TransformAction"), kTransformImageActionTypeNames, TransformImageActionType::kDEFAULT );
	mTarget.InitializeFromDictionary( dictionary, CFSTR("Target") );
	CFStringRef value;
	dictionary->Get( CFSTR("Rotation"), value, CFBridgingToCFString(kTransformImageDefaultRotation) );
	mRotation = value->GetNS();
}
