//
//  AUntil.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 25 Mar 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "AUntil.h"
#import "MMacro.h"
#if KMF_EDITOR_WINDOW
#import "AUntilEditor.h"
#import "TemplateEngine.h"
#else
#import "KMPreferences.h"
#endif
#if KMF_IS_ENGINE
#import "NSStringExtensions.h"
#import "HExecuteActionListManager.h"
#endif

AUntil::AUntil()
	:	HBaseAction()
{
}

AUntil::AUntil( const AUntil& inAction )
	:	HBaseAction( inAction ),
		mConditions( inAction.mConditions ),
		mActionList( inAction.mActionList )
{
}

AUntil::~AUntil()
{
#if KMF_IS_ENGINE
	HExecuteActionListManager::CancelExecution( mTrackingID );
#endif
}

CFStringRef
AUntil::CopyActionDescriptionString() const
{
	return CFSTR("kmUntilTitle")->CopyLocalized();
}

#if KMF_IS_ENGINE
float
AUntil::DoAction()
{
	mTrackingID = [NSUUID UUID];
	mExecuting = true;
	HExecuteActionListManager::Execute( ExecuteActionListDisplay::kHide, @"Hide", mActionList, mTrackingID );
	return 10;
}

Boolean
AUntil::IsActionCompleted()
{
	ConditionTestResult result = ConditionTestResult::kInProgress;
	
	if ( mExecuting && !HExecuteActionListManager::IsExecuting( mTrackingID ) ) {
		mExecuting = false;
	}
	if ( !mExecuting ) {
		if ( mConditions.GetCount() == 0 ) {
			result = ConditionTestResult::kTrue;
		} else {
			result = mConditions.ConditionResult();
		}
		if ( result == ConditionTestResult::kFalse ) {
			HExecuteActionListManager::Execute( ExecuteActionListDisplay::kHide, @"Hide", mActionList, mTrackingID );
			mExecuting = true;
		}
	}
	
	return result == ConditionTestResult::kTrue;
}

void
AUntil::ActionAborted()
{
	HExecuteActionListManager::CancelExecution( mTrackingID );
	mConditions.ConditionTestAbort();
}

#endif

#if KMF_EDITOR_WINDOW

Class
AUntil::ActionController() const
{
	return [AUntilEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
{% if actions.@count != 0 %}
	<ul class="IconList">
		{% for action in actions %}
			<li>{{ action }}</li>
		{% /for %}
	</ul>
{% /if %}
{% if conditions.@count != 0 %}
	<ul class="IconList"><li>
		<h3>{{ until }}</h3>
		<ul class="IconList">
			{% for condition in conditions %}
				<li><img src={{ condition.icon | inlineimage: 16 }} alt="Condition icon" width="16" height="16" /><span class="IconTitle">{{ condition.description | htmlencode }}</span>{% if condition.extraicon %}<img src={{ condition.extraicon | inlineimage: 16 }} alt="Swatch" width="16" height="16" />{% /if %}</li>
			{% /for %}
		</ul>
	</li></ul>
{% /if %}
)RAW";

NSString*
AUntil::GetActionHTMLDescriptionString() const
{
	if ( mActionList.GetCount() == 0 && mConditions.GetCount() == 0 ) return @"Until With No Actions and No Conditions Will Do Nothing";
	if ( mConditions.GetCount() == 0 ) return @"Until With No Conditions Will Execute the Following Actions Once";
	return @"Execute the Following Actions";
}

NSString*
AUntil::GetActionExtraHTML() const
{
	NSDictionary *variables = @{
									@"conditions" : mConditions.GetConditionDisplayContent(),
									@"actions" : mActionList.GetDisplayContent(),
									@"until" : [NSString stringWithFormat:@"%@ %@",GetActionTitle(),mConditions.GetConditionDisplayText()],
								};
	NSString *result = [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
	return result;
}


#endif

BOOL
AUntil::MatchesTextSearch( NSString* searchString ) const
{
	if ( HBaseAction::MatchesTextSearch( searchString ) ) return YES;
	if ( mConditions.MatchesTextSearch( searchString ) ) return YES;
	return mActionList.MatchesTextSearch( searchString );
}

#pragma mark streaming

void
AUntil::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	SafeCFPropertyListRef conditions = (SafeCFPropertyListRef)mConditions.CopyToPropertyList();
	dictionary->Set( CFSTR("Conditions"), conditions );
	conditions->Release();
	SafeCFPropertyListRef actions = mActionList.CopyActionsToPropertyList();
	dictionary->Set( CFSTR("Actions"), actions );
	CFRelease( actions );
}

void
AUntil::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	SafeCFPropertyListRef conditions;
	XAssert( dictionary->Get( CFSTR("Conditions"), conditions ) );
	mConditions.LoadFromPropertyList( conditions );
	SafeCFPropertyListRef actions = (SafeCFPropertyListRef)dictionary->Get( CFSTR("Actions") );
	XAssert( actions );
	mActionList.LoadActionsFromPropertyList( actions );
}

