//
//  AUseVariable.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 21 Apr 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "AUseVariable.h"
#import "KMDefines.h"

#if KMF_IS_ENGINE
#import "Calculate.h"
#import "UCGRemoteControl.h"
#import "XErrorHandler.h"
#import "XContainer.h"
#import "XAXWindowElement.h"
#import "XAXWindowElementList.h"
#import "XAXApplicationElement.h"
#import "NSAppleScriptExtensions.h"
#import "XAXWindowElementList.h"
#import "XRunningApplicationsList.h"
#import "NSStringRegularExpressions.h"
#endif

#if KMF_EDITOR_WINDOW
#import "AUseVariableEditor.h"
#endif


AUseVariable::AUseVariable()
	:	HBaseAction(),
		mAction( UseVariableActionType::kDEFAULT ),
		mVariable( CFSTR("Variable")->Copy() )
{
}

AUseVariable::AUseVariable( const AUseVariable& inAction )
	:	HBaseAction( inAction ),
		mAction( inAction.mAction ),
		mVariable( inAction.mVariable )
{
}

AUseVariable::~AUseVariable()
{
}

CFStringRef
AUseVariable::CopyActionDescriptionString() const
{
	CFStringRef key;
	switch ( mAction ) {
		case UseVariableActionType::kSetMouse:
			key = CFSTR("UseVariableActionType::kSetMouse");
			break;
		case UseVariableActionType::kSetWindowPosition:
			key = CFSTR("UseVariableActionType::kSetWindowPosition");
			break;
		case UseVariableActionType::kSetWindowSize:
			key = CFSTR("UseVariableActionType::kSetWindowSize");
			break;
		case UseVariableActionType::kSetWindowFrame:
			key = CFSTR("UseVariableActionType::kSetWindowFrame");
			break;
		case UseVariableActionType::kSetWindowByName:
			key = CFSTR("UseVariableActionType::kSetWindowByName");
			break;
		case UseVariableActionType::kSetWindowByNameContains:
			key = CFSTR("UseVariableActionType::kSetWindowByNameContains");
			break;
		case UseVariableActionType::kSetWindowByNameMatches:
			key = CFSTR("UseVariableActionType::kSetWindowByNameMatches");
			break;
		case UseVariableActionType::kSetApplicationByName:
			key = CFSTR("UseVariableActionType::kSetApplicationByName");
			break;
		case UseVariableActionType::kSetApplicationByNameContains:
			key = CFSTR("UseVariableActionType::kSetApplicationByNameContains");
			break;
		case UseVariableActionType::kSetApplicationByNameMatches:
			key = CFSTR("UseVariableActionType::kSetApplicationByNameMatches");
			break;
		case UseVariableActionType::kSetSystemVolume:
			key = CFSTR("UseVariableActionType::kSetSystemVolume");
			break;
	}
	
	CFContainer<CFStringRef> format = key->CopyLocalized();
	return CFString::CreateWithFormat( format, CFStringRef(mVariable) );
}


#if KMF_IS_ENGINE

float
AUseVariable::DoAction()
{
	float time = 10;
	double v1, v2, v3, v4;

	switch ( mAction ) {
		case UseVariableActionType::kSetMouse:
			if ( Calculate::CalculateVariablePair( mVariable->GetNS(), v1, v2 ) ) {
				UCGRemoteControl::PostMouse( 0, CGPointMake( v1, v2 ), CGPointMake( v1, v2 ), true, 0, kCGMouseButtonLeft );
			} else {
				XErrorHandlerMaybeLogFormat1( "always", CFSTR("Use Variable to Set Mouse Location failed for variable %@"), CFStringRef(mVariable) );
			}
			break;
		case UseVariableActionType::kSetWindowPosition:
		case UseVariableActionType::kSetWindowSize:
		case UseVariableActionType::kSetWindowFrame: {
			XAXWindowElement windowElement( XAXApplicationElement().GetFocussedOrMainWindow() );
			bool good = false;
			if ( !!windowElement ) {
				NSDictionary* calculationContext = nil;
				NSRect r;
				if ( windowElement.GetPosition( r.origin ) && windowElement.GetSize( r.size ) ) {
					calculationContext = @{ @"WINDOW" : NSStringFromRect(r) };
				}
				if ( mAction == UseVariableActionType::kSetWindowPosition ) {
					if ( Calculate::CalculateVariablePair( mVariable->GetNS(), v1, v2, calculationContext ) ) {
						windowElement.SetPosition( CGPointMake( v1, v2 ) );
						good = true;
					}
				} else if ( mAction == UseVariableActionType::kSetWindowSize ) {
					if ( Calculate::CalculateVariablePair( mVariable->GetNS(), v1, v2, calculationContext ) ) {
						windowElement.SetSize( CGSizeMake( v1, v2 ) );
						good = true;
					}
				} else {
					check( mAction == UseVariableActionType::kSetWindowFrame );
					if ( Calculate::CalculateVariableQuad( mVariable->GetNS(), v1, v2, v3, v4 ) ) {
						windowElement.SetPosition( CGPointMake( v1, v2 ) );
						windowElement.SetSize( CGSizeMake( v3, v4 ) );
						windowElement.SetPosition( CGPointMake( v1, v2 ) );
						windowElement.SetSize( CGSizeMake( v3, v4 ) );
					}
				}
			}

			if ( !good ) {
				NSString* actionName = nil;
				if ( mAction == UseVariableActionType::kSetWindowPosition ) {
					actionName = @"Set Window Position";
				} else if ( mAction == UseVariableActionType::kSetWindowSize ) {
					actionName = @"Set Window Size";
				} else {
					check( mAction == UseVariableActionType::kSetWindowFrame );
					actionName = @"Set Window Frame";
				}
				XErrorHandlerMaybeLogFormat2( "always", CFSTR("Use Variable to %@ failed for variable %@"), actionName, CFStringRef(mVariable) );
			}

			break;
		}
		case UseVariableActionType::kSetWindowByName:
		case UseVariableActionType::kSetWindowByNameContains:
		case UseVariableActionType::kSetWindowByNameMatches: {
			NSString* value = KMVariables::GetWithPassword( mVariable->GetNS() );
            XAXApplicationElement app;
			bool match = false;
			if ( app && value ) {
                XAXWindowElementList theWindowList( app );
				for ( CFIndex i = 0; i < theWindowList.GetCount(); i++ ) {
					CFContainer<CFStringRef> winName = theWindowList[i].CopyTitle( NULL );
					if ( winName.Defined() ) {
						switch ( mAction ) {
							case UseVariableActionType::kSetMouse:
							case UseVariableActionType::kSetWindowPosition:
							case UseVariableActionType::kSetWindowSize:
							case UseVariableActionType::kSetWindowFrame:
							case UseVariableActionType::kSetApplicationByName:
							case UseVariableActionType::kSetApplicationByNameContains:
							case UseVariableActionType::kSetApplicationByNameMatches:
							case UseVariableActionType::kSetSystemVolume:
								check( false );
								[[clang::fallthrough]];
							case UseVariableActionType::kSetWindowByName:
								match = winName->IsEqual( CFBridgingToCFString(value), kCFCompareLocalized );
								break;
							case UseVariableActionType::kSetWindowByNameContains:
								match = winName->Contains( CFBridgingToCFString(value), kCFCompareCaseInsensitive );
								break;
							case UseVariableActionType::kSetWindowByNameMatches:
								match = [winName->GetNS() isMatchedByRegex:value];
								break;
						}
					}
					if ( match ) {
						theWindowList[i].SetMainWindow();
						break;
					}
				}
			}
			if ( !match ) {
				XErrorHandlerMaybeLogFormat1( "always", CFSTR("Use Variable to Set Window By Name failed for variable %@"), CFStringRef(mVariable) );
			}
			break;
		}
		case UseVariableActionType::kSetApplicationByName:
		case UseVariableActionType::kSetApplicationByNameContains:
		case UseVariableActionType::kSetApplicationByNameMatches: {
			NSString* value = KMVariables::GetWithPassword( mVariable->GetNS() );
			ProcessSerialNumber psn;
			bool match = false;
			if ( value ) {
				switch ( mAction ) {
					case UseVariableActionType::kSetMouse:
					case UseVariableActionType::kSetWindowPosition:
					case UseVariableActionType::kSetWindowSize:
					case UseVariableActionType::kSetWindowFrame:
					case UseVariableActionType::kSetWindowByName:
					case UseVariableActionType::kSetWindowByNameContains:
					case UseVariableActionType::kSetWindowByNameMatches:
					case UseVariableActionType::kSetSystemVolume:
						check( false );
						[[clang::fallthrough]];
					case UseVariableActionType::kSetApplicationByName:
						match = XRunningApplicationsList::FindApplicationWithName( kFindApplicationDoValidate, CFBridgingToCFString(value), FindApplicationWithNameMethodType::kIs, psn );
						break;
					case UseVariableActionType::kSetApplicationByNameContains:
						match = XRunningApplicationsList::FindApplicationWithName( kFindApplicationDoValidate, CFBridgingToCFString(value), FindApplicationWithNameMethodType::kContains, psn );
						break;
					case UseVariableActionType::kSetApplicationByNameMatches:
						match = XRunningApplicationsList::FindApplicationWithName( kFindApplicationDoValidate, CFBridgingToCFString(value), FindApplicationWithNameMethodType::kMatches, psn );
						break;
				}
			}
			if ( match ) {
				verify_noerr( SetFrontProcessWithOptions( &psn, kSetFrontProcessFrontWindowOnly ) );
			} else {
				XErrorHandlerMaybeLogFormat1( "always", CFSTR("Use Variable to Set Application By Name failed for variable %@"), CFStringRef(mVariable) );
			}
			break;
		}
		case UseVariableActionType::kSetSystemVolume:
			if ( Calculate::CalculateVariable( mVariable->GetNS(), v1 ) ) {
				(void)[NSAppleScript executeWithSource:[NSString stringWithFormat:@"set volume output volume %lf", v1]];
			} else {
				XErrorHandlerMaybeLogFormat1( "always", CFSTR("Use Variable to Set Window Frame failed for variable %@"), CFStringRef(mVariable) );
			}
			break;
	}

	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AUseVariable::ActionController() const
{
	return [AUseVariableEditor class];
}

#endif

#pragma mark streaming

static const CFStringRef kUseVariableActionTypeNames[] = kUseVariableActionTypeNamesArray;

void
AUseVariable::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Action"), kUseVariableActionTypeNames[int(mAction)] );
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
}

void
AUseVariable::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	mAction = dictionary->GetEnumClass( CFSTR("Action"), kUseVariableActionTypeNames, UseVariableActionType::kDEFAULT );
	CFStringRef value;
	dictionary->Get( CFSTR("Variable"), value, CFSTR("Variable") );
	check( value );
	mVariable = value->Copy();
}
