//
//  AWhile.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Fri 18 Mar 2011.
//  Copyright (c) 2011 Stairways Software. All rights reserved.
//

#import "AWhile.h"
#import "MMacro.h"
#if KMF_EDITOR_WINDOW
#import "AWhileEditor.h"
#import "TemplateEngine.h"
#else
#import "KMPreferences.h"
#endif
#if KMF_IS_ENGINE
#import "NSStringExtensions.h"
#import "HExecuteActionListManager.h"
#endif

AWhile::AWhile()
	:	HBaseAction()
{
}

AWhile::AWhile( const AWhile& inAction )
	:	HBaseAction( inAction ),
		mConditions( inAction.mConditions ),
		mActionList( inAction.mActionList )
{
}

AWhile::~AWhile()
{
#if KMF_IS_ENGINE
	HExecuteActionListManager::CancelExecution( mTrackingID );
#endif
}

CFStringRef
AWhile::CopyActionDescriptionString() const
{
	return CFSTR("kmWhileTitle")->CopyLocalized();
}

#if KMF_IS_ENGINE
float
AWhile::DoAction()
{
	mTrackingID = [NSUUID UUID];
	mExecuting = false;
	return 10;
}

Boolean
AWhile::IsActionCompleted()
{
	if ( mConditions.GetCount() == 0 ) return true;
	
	ConditionTestResult result = ConditionTestResult::kInProgress;
	
	if ( mExecuting && !HExecuteActionListManager::IsExecuting( mTrackingID ) ) {
		mExecuting = false;
	}
	if ( !mExecuting ) {
		result = mConditions.ConditionResult();
		if ( result == ConditionTestResult::kTrue ) {
			HExecuteActionListManager::Execute( ExecuteActionListDisplay::kHide, @"Hide", mActionList, mTrackingID );
			mExecuting = true;
		}
	}
	
	return result == ConditionTestResult::kFalse;
}

void
AWhile::ActionAborted()
{
	HExecuteActionListManager::CancelExecution( mTrackingID );
	mConditions.ConditionTestAbort();
}

#endif

#if KMF_EDITOR_WINDOW

Class
AWhile::ActionController() const
{
	return [AWhileEditor class];
}

static NSString* kActionDisplayHTML = @R"RAW(
<ul class="IconList">
	{% for condition in conditions %}
		<li><img src={{ condition.icon | inlineimage: 16 }} alt="Condition icon" width="16" height="16" /><span class="IconTitle">{{ condition.description | htmlencode }}</span>{% if condition.extraicon %}<img src={{ condition.extraicon | inlineimage: 16 }} alt="Swatch" width="16" height="16" />{% /if %}</li>
	{% /for %}
 <li>
{% if actions.@count == 0 %}
	<h3>Do nothing.</h3>
{% else %}
	<h3>Execute the Following Actions:</h3>
	<ul class="IconList">
		{% for action in actions %}
			<li>{{ action }}</li>
		{% /for %}
	</ul>
{% /if %}
 </li>
 </ul>
)RAW";

NSString*
AWhile::GetActionHTMLDescriptionString() const
{
	if ( mConditions.GetCount() == 0 ) return @"While With No Conditions Will Do Nothing";
	return [NSString stringWithFormat:@"While %@",mConditions.GetConditionDisplayText()];
}

NSString*
AWhile::GetActionExtraHTML() const
{
	if ( mConditions.GetCount() == 0 ) return @"";

	NSDictionary *variables = @{
									@"conditions" : mConditions.GetConditionDisplayContent(),
									@"actions" : mActionList.GetDisplayContent(),
								};
	NSString *result = [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
	return result;
}

#endif

BOOL
AWhile::MatchesTextSearch( NSString* searchString ) const
{
	if ( HBaseAction::MatchesTextSearch( searchString ) ) return YES;
	if ( mConditions.MatchesTextSearch( searchString ) ) return YES;
	return mActionList.MatchesTextSearch( searchString );
}

#pragma mark streaming

void
AWhile::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	SafeCFPropertyListRef conditions = (SafeCFPropertyListRef)mConditions.CopyToPropertyList();
	dictionary->Set( CFSTR("Conditions"), conditions );
	conditions->Release();
	SafeCFPropertyListRef actions = mActionList.CopyActionsToPropertyList();
	dictionary->Set( CFSTR("Actions"), actions );
	CFRelease( actions );
}

void
AWhile::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	SafeCFPropertyListRef conditions;
	XAssert( dictionary->Get( CFSTR("Conditions"), conditions ) );
	mConditions.LoadFromPropertyList( conditions );
	SafeCFPropertyListRef actions = (SafeCFPropertyListRef)dictionary->Get( CFSTR("Actions") );
	XAssert( actions );
	mActionList.LoadActionsFromPropertyList( actions );
}

