//
//  AWriteFile.h
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 5 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "HBaseAction.h"
#import "KMVariables.h"
#import "ClipboardReference.h"
#import "StyledTextSource.h"

enum class WriteFileSourceType { // WARNING: tag valued in AWriteFile.xib match this enum
	kClipboard,
	kNamedClipboard,
	kVariable,
	kPlainText,
	kStyledText,
	kDEFAULT = kClipboard,
};
#define kWriteFileSourceTypeNamesArray {CFSTR("Clipboard"), CFSTR("NamedClipboard"), CFSTR("Variable"), CFSTR("Text"), CFSTR("StyledText"), NULL}

enum class WriteFileFormatType { // WARNING: tag valued in AWriteFile.xib match this enum
	kPlainText,
	kRTF,
	kRTFD,
	kSimpleText,
	kHTML,
	kWebArchive,
	kWordDoc,
	kWordML,
	kOfficeOpenXML,
	kOpenDocument,
	kPNG,
	kTIFF, // NSImageCompressionMethod
	kJPEG, // NSImageCompressionFactor
	kBMP,
	kGIF,
	kDEFAULT = WriteFileFormatType::kPlainText, 
	kLastText = WriteFileFormatType::kOpenDocument, 
	kFirstImage = WriteFileFormatType::kLastText+1, 
}; // WARNING: These keys are also used in Localized.strings
#define kWriteFileFormatTypeNamesArray {CFSTR("PlainText"), CFSTR("RTF"), CFSTR("RTFD"), CFSTR("SimpleText"), CFSTR("HTML"), CFSTR("WordDoc"), CFSTR("WordML"), CFSTR("OfficeOpenXML"), CFSTR("WebArchive"), CFSTR("OpenDocument"), CFSTR("PNG"), CFSTR("TIFF"), CFSTR("JPEG"), CFSTR("BMP"), CFSTR("GIF"), NULL}

enum class WriteFileEncodingType { // WARNING: tag valued in AWriteFile.xib match this enum
	kMacRoman,
	kLatin1,
	kUTF8,
	kUTF16,
	kDEFAULT = kUTF8, 
}; // WARNING: These keys are also used in Localized.strings
#define kWriteFileEncodingTypeNamesArray {CFSTR("MacRoman"), CFSTR("Latin1"), CFSTR("UTF8"), CFSTR("UTF16"), NULL}
#define kWriteFileEncodingTypeMacEncodingArray {NSMacOSRomanStringEncoding, NSISOLatin1StringEncoding, NSUTF8StringEncoding, NSUTF16StringEncoding, NULL}

class AWriteFile
	:	public HBaseAction
{

private:

	WriteFileSourceType				mSource = WriteFileSourceType::kDEFAULT;
	ClipboardReference				mClipboardReference;
	StyledTextSource				mStyledText;
	VariableContainer				mVariable = CFSTR("Variable");
	CFContainer<CFStringRef>		mDestination = CFSTR("");
	WriteFileFormatType				mFormat = WriteFileFormatType::kPlainText;
	WriteFileEncodingType			mEncoding = WriteFileEncodingType::kUTF8;
	Boolean							mAppend = false;

	CFStringRef						CopyActionDescriptionShortString() const;

public:

    AWriteFile() = default;
	AWriteFile( const AWriteFile& inAction ) = default;
	AWriteFile( bool append );
    virtual ~AWriteFile() = default;

	virtual AWriteFile*				Clone() const { return new AWriteFile( *this ); }
	virtual AWriteFile*				Create() const { return new AWriteFile(); }
	
	virtual HActionType				GetActionType() const { return HActionType::kWriteFile; }
	virtual	CFStringRef				CopyActionDescriptionString() const;
	
	virtual NSString*				GetNewActionTitle() const;

#if KMF_EDITOR_WINDOW
	virtual void					UpdateToDefaultValues() { mClipboardReference.UpdateClipboardReference(); }
#endif

#if KMF_IS_ENGINE
    virtual float					DoAction();
#endif
#if KMF_EDITOR_WINDOW
	virtual Class					ActionController() const;
	virtual	NSString*				GetActionHTMLDescriptionString() const;
	virtual	NSString*				GetActionExtraHTML() const;
#endif

	bool							CanWriteImage() const { return mSource != WriteFileSourceType::kVariable && mSource != WriteFileSourceType::kPlainText && mSource != WriteFileSourceType::kStyledText && !mAppend; }
	bool							CanAppend() const { return mFormat == WriteFileFormatType::kPlainText; }
	
	WriteFileSourceType				GetSource() const { return mSource; }
	void							SetSource( WriteFileSourceType inSource ) { mSource = inSource; }
	
	virtual void					SetVariable( NSString* inVariable ) { check( inVariable ); mVariable = [inVariable copy]; }
	virtual NSString*				GetVariable() const { return mVariable->GetNS(); }
	
	NSString*						GetClipboardUID() const { return mClipboardReference.GetClipboardUID(); }
	void							SetClipboardUID( NSString* inUID ) { mClipboardReference.SetClipboardUID( inUID ); }

	CFStringRef						CopyDestination() const { return mDestination->Copy(); }
	NSString*						GetDestination() const { return mDestination->GetNS(); }
	void							SetDestination( CFStringRef destination ) { mDestination = destination->Copy(); }
	
	void							SetPlainText( NSString* inString ) { mStyledText.SetPlainText( inString ); }
	void							SetStyledText( NSAttributedString* inString ) { mStyledText.SetStyledText( inString ); }

	NSString*						GetSourceString() const { return mStyledText.GetString(); }
	NSAttributedString*				GetSourceAttributedString() const { return mStyledText.GetAttributedString(); }

	Boolean							GetUseStyledText() const { return mStyledText.GetUseStyledText(); }
	
	WriteFileFormatType				GetFormat() const { return mFormat; }
	void							SetFormat( WriteFileFormatType inFormat ) { mFormat = inFormat; }
	
	WriteFileEncodingType			GetEncoding() const { return mEncoding; }
	void							SetEncoding( WriteFileEncodingType inEncoding ) { mEncoding = inEncoding; }
	
	Boolean							GetAppend() const { return mAppend; }
	void							SetAppend( Boolean inAppend ) { mAppend = inAppend; }
	
	virtual void					SaveToDictionary( CFMutableDictionaryRef dictionary ) const;
	virtual void					InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException );
	
};

