//
//  AWriteFile.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 5 Apr 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "AWriteFile.h"
#import "FileReference.h"
#import "XClipboardDataSet.h"
#import "XClipboardHistory.h"
#import "CSClipboardDomain.h"
#import "NSStringExtensions.h"
#import "NSObjectExtensions.h"
#import "XErrorHandler.h"
#import "NSImageExtensions.h"
#import "NSStringRegularExpressions.h"

#if KMF_IS_ENGINE
#import "KMVariables.h"
#import "KMNotifications.h"
#import "TokenText.h"
#endif

#if KMF_EDITOR_WINDOW
#import "AWriteFileEditor.h"
#import "TemplateEngine.h"
#endif

static const CFStringRef kWriteFileFormatTypeNames[] = kWriteFileFormatTypeNamesArray;
static const CFStringRef kWriteFileEncodingTypeNames[] = kWriteFileEncodingTypeNamesArray;

AWriteFile::AWriteFile( bool append )
	:	HBaseAction(),
		mSource( append ? WriteFileSourceType::kPlainText : WriteFileSourceType::kDEFAULT ),
		mAppend( append )
{
}

NSString*
AWriteFile::GetNewActionTitle() const
{
	return mAppend ? [@"hWriteFileFormatType::kAppend" localized] : [@"hWriteFileFormatType::kWrite" localized];
};

CFStringRef
AWriteFile::CopyActionDescriptionShortString() const
{
	CFContainer<CFStringRef> format;
	CFContainer<CFStringRef> result;
	CFContainer<CFStringRef> operation;
	if ( mAppend ) {
		operation = CFSTR("kmWriteOperationAppend")->CopyLocalized();
	} else {
		operation = CFSTR("kmWriteOperationNormal")->CopyLocalized();
	}
	switch ( mSource ) {
		case WriteFileSourceType::kClipboard: {
			format = CFSTR("kmWriteFileFromClipboardTitleFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, CFStringRef(operation) );
			break;
		}
		case WriteFileSourceType::kNamedClipboard: {
			format = CFSTR("kmWriteFileFromNamedClipboardTitleFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, CFStringRef(operation), mClipboardReference.GetClipboardName() );
			break;
		}
		case WriteFileSourceType::kVariable: {
			format = CFSTR("kmWriteFileFromVariableTitleFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, CFStringRef(operation), CFStringRef(mVariable) );
			break;
		}
		case WriteFileSourceType::kPlainText: {
			format = CFSTR("kmWriteFileFromPlainTextTitleFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, CFStringRef(operation) );
			break;
		}
		case WriteFileSourceType::kStyledText: {
			format = CFSTR("kmWriteFileFromStyledTextTitleFormat")->CopyLocalized();
			result = CFString::CreateWithFormat( format, CFStringRef(operation) );
			break;
		}
	}
	return result->Copy();
}

CFStringRef
AWriteFile::CopyActionDescriptionString() const
{
	CFContainer<CFStringRef> shortstr = CopyActionDescriptionShortString();
	CFContainer<CFStringRef> format = CFSTR("kmWriteFileMakeLongTitleFormat")->CopyLocalized();
	return CFString::CreateWithFormat( format, CFStringRef(shortstr) );
}

#if KMF_IS_ENGINE

static const NSStringEncoding gEncodingForWriteFileEncoding[] = kWriteFileEncodingTypeMacEncodingArray;

float
AWriteFile::DoAction()
{
	float time = 10;

	NSString* destination = FileReference::PathToActualPath( mDestination->GetNS() );
	if ( ![destination isMatchedByRegex:@"^/.+"] ) {
		XErrorHandlerMaybeLogMessage( "display", CFSTR("No file destination specified for Write File action") );
		return 0;
	}
	
	WriteFileFormatType format = mFormat;
	if ( !CanWriteImage() && format >= WriteFileFormatType::kFirstImage ) {
		format = WriteFileFormatType::kPlainText;
	}
	if ( mAppend ) {
		format = WriteFileFormatType::kPlainText;
	}
	if ( format >= WriteFileFormatType::kFirstImage ) {
		NSImage* image = nil;
		switch ( mSource ) {
			case WriteFileSourceType::kClipboard: {
				XClipboardDataSet* oldScrap = XClipboardHistory::CopyCurrent();
				if ( oldScrap ) {
					image = oldScrap->Image();
					delete oldScrap;
				}
				break;
			}
			case WriteFileSourceType::kNamedClipboard: {
				mClipboardReference.UpdateClipboardReference();
				CSClipboardDataSet*	pClip = mClipboardReference.GetClipboardDataSet();
				if ( pClip ) {
					image = pClip->Image();
				}
				break;
			}
			case WriteFileSourceType::kPlainText:
			case WriteFileSourceType::kStyledText:
			case WriteFileSourceType::kVariable:
				check( false );
				break;
		}
		if ( image ) {
			NSBitmapImageFileType bitmapFormat;
			switch ( mFormat ) {
				case WriteFileFormatType::kPlainText:
				case WriteFileFormatType::kRTF:
				case WriteFileFormatType::kRTFD:
				case WriteFileFormatType::kSimpleText:
				case WriteFileFormatType::kHTML:
				case WriteFileFormatType::kWebArchive:
				case WriteFileFormatType::kWordDoc:
				case WriteFileFormatType::kWordML:
				case WriteFileFormatType::kOfficeOpenXML:
				case WriteFileFormatType::kOpenDocument:
					check( false );
					[[clang::fallthrough]];
				case WriteFileFormatType::kPNG:
					bitmapFormat = NSPNGFileType;
					break;
				case WriteFileFormatType::kTIFF:
					bitmapFormat = NSTIFFFileType;
					break;
				case WriteFileFormatType::kJPEG:
					bitmapFormat = NSJPEGFileType;
					break;
				case WriteFileFormatType::kBMP:
					bitmapFormat = NSBMPFileType;
					break;
				case WriteFileFormatType::kGIF:
					bitmapFormat = NSGIFFileType;
					break;
			}
			NSData* imageData = [image dataRepresentationOfType: bitmapFormat];
			[imageData writeToFile:destination atomically:NO];        
		}
	} else {
		NSAttributedString* theAttributedText = nil;
		switch ( mSource ) {
			case WriteFileSourceType::kClipboard: {
				XClipboardDataSet* oldScrap = XClipboardHistory::CopyCurrent();
				if ( oldScrap ) {
					theAttributedText = oldScrap->AttributedString();
					delete oldScrap;
				}
				if ( !theAttributedText ) {
					theAttributedText = [[NSAttributedString alloc] initWithString:[@"NoClipboardText" localized]];
				}
				break;
			}
			case WriteFileSourceType::kNamedClipboard: {
				mClipboardReference.UpdateClipboardReference();
				CSClipboardDataSet*	pClip = mClipboardReference.GetClipboardDataSet();
				if ( pClip ) {
					theAttributedText = pClip->AttributedString();
				}
				if ( !theAttributedText ) {
					theAttributedText = [[NSAttributedString alloc] initWithString:[@"NoClipboardText" localized]];
				}
				break;
			}
			case WriteFileSourceType::kVariable: {
				CFContainer<CFStringRef> theTokenText = CFString::CreateWithFormat( CFSTR("%%Variable%%%@%%"), (CFStringRef)mVariable );
				float t;
				CFContainer<CFStringRef> theText = TokenText::CopyResult( theTokenText, t );
				time += t;
				theAttributedText = [[NSAttributedString alloc] initWithString:theText->GetNS()];
				break;
			}
			case WriteFileSourceType::kPlainText: {
				NSMutableAttributedString* t = [[NSMutableAttributedString alloc] initWithString:mStyledText.GetString()];
				time += TokenText::Process( t );
				theAttributedText = [t copy];
				break;
			}
			case WriteFileSourceType::kStyledText: {
				NSMutableAttributedString* t = mStyledText.CopyMutableAttributedString();
				time += TokenText::Process( t );
				theAttributedText = [t copy];
				break;
			}
		}
		if ( mAppend ) {
			[theAttributedText.string appendToFile:destination encoding:gEncodingForWriteFileEncoding[int(mEncoding)]];
		} else {
			NSMutableDictionary* attributes = [NSMutableDictionary dictionary];
			switch ( format ) {
				case WriteFileFormatType::kPNG:
				case WriteFileFormatType::kTIFF:
				case WriteFileFormatType::kJPEG:
				case WriteFileFormatType::kBMP:
				case WriteFileFormatType::kGIF:
					check( false );
					[[clang::fallthrough]];
				case WriteFileFormatType::kPlainText:
					[attributes setObject:NSPlainTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					[attributes setObject:[NSNumber numberWithInt:gEncodingForWriteFileEncoding[int(mEncoding)]] forKey:NSCharacterEncodingDocumentAttribute];
					break;
				case WriteFileFormatType::kRTF:
					[attributes setObject:NSRTFTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
				case WriteFileFormatType::kRTFD:
					[attributes setObject:NSRTFDTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
				case WriteFileFormatType::kSimpleText:
					[attributes setObject:NSMacSimpleTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
				case WriteFileFormatType::kHTML:
					[attributes setObject:NSHTMLTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
				case WriteFileFormatType::kWebArchive:
					[attributes setObject:NSWebArchiveTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
				case WriteFileFormatType::kWordDoc:
					[attributes setObject:NSDocFormatTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
				case WriteFileFormatType::kWordML:
					[attributes setObject:NSWordMLTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
				case WriteFileFormatType::kOfficeOpenXML:
					[attributes setObject:NSOfficeOpenXMLTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
				case WriteFileFormatType::kOpenDocument:
					[attributes setObject:NSOpenDocumentTextDocumentType forKey:NSDocumentTypeDocumentAttribute];
					break;
			}
			NSError* error1 = nil;
			NSFileWrapper* filewrapper = [theAttributedText fileWrapperFromRange:NSMakeRange( 0, theAttributedText.length ) documentAttributes:attributes error:&error1];
			if ( error1 ) {
				XErrorHandlerMaybeLogFormat1( "display", CFSTR("Failed to get file wrapper from NSAttributedString with error %@"), error1 );
			}
			if ( filewrapper ) {
				NSError* error2 = nil;
				[filewrapper writeToURL:[destination urlForPath] options:NSFileWrapperWritingAtomic originalContentsURL:nil error:&error2];
				if ( error2 ) {
					XErrorHandlerMaybeLogFormat1( "display", CFSTR("Failed to write the file wrapper from NSAttributedString with error %@"), error2 );
				}
			}

		}
		
		

	}
	return time;
}

#endif

#if KMF_EDITOR_WINDOW

Class
AWriteFile::ActionController() const
{
	return [AWriteFileEditor class];
}

NSString*
AWriteFile::GetActionHTMLDescriptionString() const
{
	CFContainer<CFStringRef> shortstr = CopyActionDescriptionShortString();
	return shortstr->GetNS();
}

static NSString* kActionDisplayHTML = @R"RAW(
{{ text }}
To file: {{ file | htmlencode }}<br />
With format {{ format | htmlencode }}{% if encoding %} ({{encoding}}{% /if %}).<br />
)RAW";

NSString*
AWriteFile::GetActionExtraHTML() const
{
	NSString* text;
	switch ( mSource ) {
		case WriteFileSourceType::kClipboard:
		case WriteFileSourceType::kNamedClipboard:
		case WriteFileSourceType::kVariable:
			text = @"";
			break;
		case WriteFileSourceType::kPlainText:
		case WriteFileSourceType::kStyledText:
			text = mStyledText.GetDisplayHTMLContent();
	}
	NSString* format = [[NSString stringWithFormat:@"WriteFileFormat%@", kWriteFileFormatTypeNames[int(mFormat)]] localized];
	NSString* encoding = @"";
	if ( mFormat == WriteFileFormatType::kPlainText ) {
		encoding = [[NSString stringWithFormat:@"WriteFileEncoding%@", kWriteFileEncodingTypeNames[int(mEncoding)]] localized];
	}
	NSDictionary *variables = @{
									@"text" : text,
									@"file" : mDestination->GetNS(),
									@"format" : format,
									@"encoding" : encoding,
	};
	return [[TemplateEngine sharedDetailViewTemplateEngine] processTemplate:kActionDisplayHTML withVariables:variables];
}

#endif

#pragma mark streaming

static const CFStringRef kWriteFileSourceTypeNames[] = kWriteFileSourceTypeNamesArray;

void
AWriteFile::SaveToDictionary( CFMutableDictionaryRef dictionary ) const
{
	dictionary->Set( CFSTR("Source"), kWriteFileSourceTypeNames[int(mSource)] );
	dictionary->Set( CFSTR("Format"), kWriteFileFormatTypeNames[int(mFormat)] );
	dictionary->Set( CFSTR("Encoding"), kWriteFileEncodingTypeNames[int(mEncoding)] );
	dictionary->Set( CFSTR("Variable"), CFStringRef(mVariable) );
	mStyledText.SaveToDictionary( dictionary );
	mClipboardReference.SaveToDictionary( dictionary );
	FileReference::SavePathToDictionary( dictionary, mDestination->GetNS(), CFSTR("Destination") );
	dictionary->SetBoolean( CFSTR( "Append" ), mAppend );
}

void
AWriteFile::InitializeFromDictionary( CFDictionaryRef dictionary ) throw ( XException )
{
	bool good;
	good = dictionary->CanGetEnumClass( CFSTR("Source"), kWriteFileSourceTypeNames, mSource );
	XAssert( good );
	
	good = dictionary->CanGetEnumClass( CFSTR("Format"), kWriteFileFormatTypeNames, mFormat );
	XAssert( good );

	good = dictionary->CanGetEnumClass( CFSTR("Encoding"), kWriteFileEncodingTypeNames, mEncoding );
	XAssert( good );

	CFStringRef value;
	dictionary->Get( CFSTR("Variable"), value, CFSTR("Variable") );
	check( value );
	mVariable = value->Copy();
	mStyledText.InitializeFromDictionary( dictionary );
	mClipboardReference.InitializeFromDictionary( dictionary );
	FileReference::LoadPathFromDictionary( dictionary, mDestination, NULL, CFSTR("Destination") );
	dictionary->GetBoolean( CFSTR( "Append" ), mAppend, false );
}
