//
//  ClipboardReference.cpp
//  Keyboard Maestro
//
//  Created by Peter N Lewis on Thu 20 Sep 2012.
//  Copyright (c) 2012 Stairways Software. All rights reserved.
//

#import "ClipboardReference.h"
#import "NSStringExtensions.h"
#import "CSClipboardDomain.h"
#import "KMNotifications.h"
#import "MMacroPreferenceFile.h"

static NSString* gDefaultName = [@"DefaultClipboardName" localized];
static NSCountedSet* gClipboardReferenceUIDs = [NSCountedSet new];

static void Use( NSString* uid )
{
	if ( !IsEmpty( uid ) ) {
		[gClipboardReferenceUIDs addObject:uid];
	}
}

static void Unuse( NSString* uid )
{
	if ( !IsEmpty( uid ) ) {
		[gClipboardReferenceUIDs removeObject:uid];
	}
}

bool
ClipboardReference::ClipboardUIDInUse( NSString* uid )
{
	check( !IsEmpty( uid ) );
	return [gClipboardReferenceUIDs member:uid];
}

ClipboardReference::ClipboardReference()
{
	mClipboardUID = @"";
	mClipboardRedundantDisplayName = gDefaultName;
	Use( mClipboardUID );
};

ClipboardReference::ClipboardReference( const ClipboardReference& inObj )
{
	Unuse( mClipboardUID );
	mClipboardUID = inObj.mClipboardUID;
	mClipboardRedundantDisplayName = inObj.mClipboardRedundantDisplayName;
	Use( mClipboardUID );
	UpdateClipboardReference();
}

ClipboardReference::~ClipboardReference()
{
	Unuse( mClipboardUID );
}

void
ClipboardReference::UpdateClipboardReference()
{
	Unuse( mClipboardUID );
	Boolean changed = false;
	if ( IsEmpty( mClipboardUID ) ) {
		mClipboardUID = [NSString uuidString];
		changed = true;
	}
	CFContainer<CFStringRef> redundandName = CSClipboardDomain::StrictCopyClipboardNameByUID( CFBridgingToCFString( mClipboardUID ) );
	if ( redundandName ) {
		if ( ![mClipboardRedundantDisplayName isEqualToString:redundandName->GetNS()] ) {
			changed = true;
			mClipboardRedundantDisplayName = redundandName->GetNS();
		}
	} else {
		if ( IsEmpty( mClipboardRedundantDisplayName ) ) {
			mClipboardRedundantDisplayName = gDefaultName;
		}
		CSClipboardDataSet* pExistingClipData = CSClipboardDomain::GetClipboardDomain()->GetClipDataSetByDisplayName( CFBridgingToCFString( mClipboardRedundantDisplayName ) );
		if ( pExistingClipData ) {
			mClipboardUID = pExistingClipData->GetUID();
		} else {
			std::unique_ptr<CSClipboardDataSet> pNewClipData( new CSClipboardDataSet( CFBridgingToCFString( mClipboardRedundantDisplayName ), kCSClipboardDataSetAddDefaultContent ) );
			CSClipboardDomain::GetClipboardDomain()->InsertClipDataSet( pNewClipData.get(), KMNotifications::kLoadClipDataFromDefaultCreate );
			mClipboardUID = pNewClipData->GetUID();
			//			NSLog( @"CSClipboardDomain::UpdateClipboardReference CREATE uid %@ name %@", inOutUID.Get(), inOutRedundantDisplayName.Get() );
		}
		changed = true;
	}
	if ( changed ) {
		MMacroPreferenceFile::SetLoadAllMacrosNeedsSave();
	}
	Use( mClipboardUID );
}

void
ClipboardReference::SetClipboardUID( NSString* uid )
{
	Unuse( mClipboardUID );
	mClipboardUID = uid;
	UpdateClipboardReference();
}

ClipboardReference&
ClipboardReference::operator=( const ClipboardReference& inObj )
{
	Unuse( mClipboardUID );
	mClipboardUID = inObj.mClipboardUID;
	mClipboardRedundantDisplayName = inObj.mClipboardRedundantDisplayName;
	Use( mClipboardUID );
	return *this;
}

NSString*
ClipboardReference::GetClipboardName() const
{
	CFContainer<CFStringRef> name = CSClipboardDomain::CopyClipboardNameByUID( CFBridgingToCFString( mClipboardUID ) );
	return name.Defined() ? name->GetNS() : mClipboardRedundantDisplayName;
}

CSClipboardDataSet*
ClipboardReference::GetClipboardDataSet() const
{
	return CSClipboardDomain::GetClipboardDomain()->GetClipDataSetByUID( CFBridgingToCFString( mClipboardUID ) );
}

static NSString* kDefaultClipboardDictionaryUIDKey = @"Name"; // WARNING: This is CFSTR("Name") - yikes!
static NSString* kDefaultClipboardDictionaryRedundantDisplayName = @"RedundandDisplayName"; // WARNING: This is CFSTR("RedundandDisplayName") misspelled - yikes!

void
ClipboardReference::SaveToDictionary( CFMutableDictionaryRef dictionary, NSString* uidKey, NSString* redundantNameKey ) const
{
	if ( !uidKey ) uidKey = kDefaultClipboardDictionaryUIDKey;
	if ( !redundantNameKey ) redundantNameKey = kDefaultClipboardDictionaryRedundantDisplayName;
	ClipboardReference save = *this;
	save.UpdateClipboardReference();
	dictionary->Set( CFBridgingToCFString(uidKey), CFBridgingToCFString(save.mClipboardUID) );
	dictionary->Set( CFBridgingToCFString(redundantNameKey), CFBridgingToCFString(save.mClipboardRedundantDisplayName) );
}

void
ClipboardReference::InitializeFromDictionary( CFDictionaryRef dictionary, NSString* uidKey, NSString* redundantNameKey ) throw ( XException )
{
	if ( !uidKey ) uidKey = kDefaultClipboardDictionaryUIDKey;
	if ( !redundantNameKey ) redundantNameKey = kDefaultClipboardDictionaryRedundantDisplayName;

	CFStringRef temp;
	dictionary->Get( CFBridgingToCFString(uidKey), temp, CFStringRef(NULL) );
	mClipboardUID = temp ? temp->GetNS() : nil;
	dictionary->Get( CFBridgingToCFString(redundantNameKey), temp, CFStringRef(NULL) );
	mClipboardRedundantDisplayName = temp ? temp->GetNS() : nil;

	UpdateClipboardReference();
}


