/*
 *  HActionList.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Tue Dec 11 2001.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "HActionList.h"
#import "HBaseAction.h"
#import "AActivateApplication.h"
#import "KMNotifications.h"
#import "XNotificationCenter.h"
#import "ADontRotateStatusMenu.h"
#import "NSObjectExtensions.h"

HActionList::~HActionList()
{
}

HActionList::HActionList( SafeCFPropertyListRef properties ) throw( XException )
{
	LoadActionsFromPropertyList( properties );
}

void
HActionList::Empty()
{
	mActionList.clear();
}

bool
HActionList::DontRotateStatusMenu() const
{
	if ( GetCount() < 1 ) return false;
	ADontRotateStatusMenu* action = dynamic_cast<ADontRotateStatusMenu*>( (HBaseAction*)*(mActionList.begin()) );
	if ( action ) return true;
	return false;
}

void
HActionList::AppendAction( HBaseAction* inAction )
{
	mActionList.push_back( HBaseActionClonePtr(inAction) );
}

static inline HBaseActionClonePtrListIterator IteratorForIndex( HBaseActionClonePtrList& list, CFIndex inIndex )
{
	HBaseActionClonePtrListIterator it = list.begin();
	for ( CFIndex i = 0; i < inIndex; i++ ) {
		check( it != list.end() );
		it++;
	}
	return it;
}

static inline HBaseActionClonePtrListConstIterator IteratorForIndex( const HBaseActionClonePtrList& list, CFIndex inIndex )
{
	HBaseActionClonePtrListConstIterator it = list.begin();
	for ( CFIndex i = 0; i < inIndex; i++ ) {
		check( it != list.end() );
		it++;
	}
	return it;
}

void
HActionList::InsertActionAt( const CFIndex& inIndex, HBaseAction* inAction )
{
	auto it = IteratorForIndex( mActionList, inIndex );
	mActionList.insert( it, HBaseActionClonePtr(inAction) );
}

HBaseAction*
HActionList::GetAction( const CFIndex& inIndex ) const
{
	auto it = IteratorForIndex( mActionList, inIndex );
	if (it != mActionList.end()) {
		return *it;
	} else {
		return NULL;
	}
}

void
HActionList::RelocateActionToIndex( const CFIndex& inCurIndex, const CFIndex& inNewIndex )
{
	if ( inCurIndex != inNewIndex ) {
		auto src = IteratorForIndex( mActionList, inCurIndex );
		auto dst = IteratorForIndex( mActionList, inNewIndex );
		if ( inCurIndex < inNewIndex ) {
			dst++;
		}
		mActionList.splice( dst, mActionList, src );
	}
}

void
HActionList::ExchangeActionsAtIndices( const CFIndex& inIndex1, const CFIndex& inIndex2 )
{
	auto it1 = IteratorForIndex( mActionList, inIndex1 );
	auto it2 = IteratorForIndex( mActionList, inIndex2 );
	std::swap( *it1, *it2 );
}

void
HActionList::SetActionAtIndex( const CFIndex& inIndex, HBaseAction* inAction )
{
	auto it = IteratorForIndex( mActionList, inIndex );
	*it = HBaseActionClonePtr(inAction);
}

void
HActionList::RemoveActionAtIndex( const CFIndex& inIndex )
{
	auto it = IteratorForIndex( mActionList, inIndex );
	mActionList.erase( it );
}

void
HActionList::DoForEveryAction( HBaseAction::ActionBlock inApplier )
{
	for ( auto it : mActionList ) {
		inApplier( it );
	}
}

void
HActionList::DoForEveryActionAndSubAction( HBaseAction::ActionBlock inApplier )
{
	DoForEveryAction( ^( HBaseAction* inAction ){
		inApplier( inAction );
		inAction->DoForEverySubAction( inApplier );
	});
}

#if KMF_IS_EDITOR

void
HActionList::Execute( NSString* displayName, NSString* trackingID )
{
	CFContainer<SafeCFPropertyListRef> actionsPList = CopyActionsToPropertyList();

	NSMutableDictionary* userInfo = [NSMutableDictionary new];
	userInfo[KMNotifications::kExecuteActionsKey] = CFBridgingToNS(actionsPList);
	if ( !IsEmpty( displayName ) ) {
		userInfo[KMNotifications::kExecuteActionsDisplayKey] = displayName;
	}
	if ( !IsEmpty( trackingID ) ) {
		userInfo[KMNotifications::kExecuteActionsTrackingIDKey] = trackingID;
	}
	XNotificationCenter::PostNotification( KMNotifications::kDoExecuteActions, userInfo );
}

void
HActionList::ExecuteAction( NSString* displayName, const HBaseAction* trialAction )
{
	HActionList actions;
	actions.AppendAction( trialAction->Clone() );
	actions.Execute( displayName );
}

NSArray*
HActionList::GetDisplayContent() const
{
	NSMutableArray* actions = [NSMutableArray array];
	
	for ( int i = 0; i < GetCount(); i++ ) {
		[actions addObject:GetAction( i )->ActionDisplayHTML()];
	}
	
	return actions;
}

#endif

BOOL
HActionList::MatchesTextSearch( NSString* searchString ) const
{
	for ( int i = 0; i < GetCount(); i++ ) {
		if ( GetAction( i )->MatchesTextSearch( searchString ) ) return YES;
	}
	return NO;
}

SafeCFPropertyListRef
HActionList::CopyActionsToPropertyList() const
{
	CFMutableArrayRef actions = CFArray::CreateMutable();
	for ( CFIndex i = 0; i < GetCount(); i++ ) {
		HBaseAction* theAction = GetAction( i );
		SafeCFPropertyListRef actionPList = theAction->CopyToPropertyList( false );
		actions->Append( actionPList );
		actionPList->Release();
	}
	
	return (SafeCFPropertyListRef)actions;
}


void
HActionList::LoadActionsFromPropertyList( SafeCFPropertyListRef properties ) throw( XException )
{
	CFArrayRef actions = (CFArrayRef)properties;
	XAssert( actions && actions->IsCorrectTypeID() );
	for ( CFIndex i = 0; i < actions->GetCount(); i++ ) {
		SafeCFPropertyListRef actionProperties = (SafeCFPropertyListRef)actions->Get( i );
		try {
			HBaseAction* pAction = HBaseAction::CreateFromPropertyList( actionProperties );
			AppendAction( pAction );
		} catch ( XException ) {
		}
	}
}
