/*
 *  HExecuteActionListManager.cpp
 *  Keyboard Maestro
 *
 *  Created by Michael Kamprath on Sun Apr 21 2002.
 *  Copyright (c) 2004 Stairways Software. All rights reserved.
 */

#import "HExecuteActionListManager.h"
#import "MMacro.h"
#import "MMacroGroup.h"
#import "HBaseAction.h"
#import "TBaseTrigger.h"
#import "THotKeyTrigger.h"
#import "XErrorHandler.h"
#import "HTimeSaved.h"
#import "THotKeyTriggerManager.h"
#import "TTypedStringTrigger.h"
#import "SystemDefinedKeys.h"
#import "StatusMenu.h"
#import "KMPreferences.h"
#import "ASimulateKeystroke.h"
#import "XClipboardHistory.h"
#import "KMPreferences.h"
#import "XEventLoopTimer.h"
#import "HActionList.h"
#import "Growl/GrowlApplicationBridge.h"
#import "NSStringExtensions.h"
#import "MacroDebugger.h"
#import "ADebuggerAction.h"
#import "KMMacroLastExecuted.h"
#import "ReverseWrapper.h"
#import "AExecuteMacro.h"

#import <list>

class HExecuteActionListEntry;

static const EventTimerInterval kInitialInterval = 10*kEventDurationMillisecond;
static const EventTimerInterval kTimerInterval = 1*kEventDurationMillisecond;

static const float kDefaultInterActionDelayTime = 0.0; // Time to wait between actions

static NSUInteger gExecutionEntryOrder = 1;

static HExecuteActionListEntry* gCurrentExecuteActionListEntry = NULL;

@interface HExecutionDebugState : NSObject

@property (nonatomic, assign) ExecutingEntryDebugState debugState;
@property (nonatomic, assign) NSUInteger debugStepEntryOrder;

@end

@implementation HExecutionDebugState

@synthesize debugState;
@synthesize debugStepEntryOrder;

@end

NSMutableDictionary* gExecutionDebugStates = nil;

class HExecuteActionListEntry {
private:

	NSUUID*					mTrackingID;
	NSUUID*					mExecutionGroup;
	NSString*				mExecutionMacroName;
	NSString*				mExecutionMacroGroupName;
	mutable HBaseAction*	mExecutionAction = NULL;
	mutable NSString*		mExecutionActionDescription = nil;

	UInt32					mTrippedKeyCode = hkUndefinedKeyCode;
	UInt32					mTrippedModifiers = 0;
	TrippedCaseBehaviour	mTrippedCase = TrippedCaseBehaviour::kNone;
	
	HActionList				mActionList;
	
	CFIndex					mCurActionIndex = -1;
	EventTime				mCurActionExecuteTime = 0;
	
	Boolean					mFirstAction = true;
	Boolean					mDelayingBetweenActions = false;
	EventTime				mNextActionExecuteTime;

	NSUInteger				mExecutionEntryOrder = gExecutionEntryOrder++;

	void					HandleDoAction( HBaseAction* inAction );
	
	HExecuteActionListEntry( const HExecuteActionListEntry& inObj ) = delete;
	HExecuteActionListEntry& operator= ( const HExecuteActionListEntry& inObj ) = delete;
	
public:
	
	HExecuteActionListEntry( ExecuteActionListDisplay display, const MMacro* inMacro, TBaseTrigger* inTrippedTrigger, NSUUID* trackingID = NULL );
	HExecuteActionListEntry( ExecuteActionListDisplay display, NSString* displayName, const HActionList& inActionList, NSUUID* trackingID = NULL );
	HExecuteActionListEntry( ExecuteActionListDisplay display, NSString* displayName, const HBaseAction* inTrialAction, NSUUID* trackingID = NULL );
	virtual ~HExecuteActionListEntry();

	Boolean					mDisplayInList = false;
	
	bool					mCancel = false;
	bool					mSkipLoop = false;

	bool					ContinueExecution();
	
	bool					IsTrackingID( NSUUID* trackingID ) const;
	bool					IsExecutionGroup( NSUUID* executionGroup ) const;
	
	bool					IsExecutingExecuteMacro() const;
	bool					IsExecutingLoop() const;

	NSUUID*					GetTrackingID() const { return mTrackingID; }
	NSUUID*					GetExecutionGroup() const { return mExecutionGroup; }
	NSString*				GetExecutionMacroName() const { return mExecutionMacroName; }
	NSString*				GetExecutionMacroGroupName() const { return mExecutionMacroGroupName; }
	HBaseAction*			GetExecutionAction() const;
	NSString*				GetExecutionActionDescription() const;
	UInt32					GetTrippedKeyCode() const { return mTrippedKeyCode; }
	UInt32					GetTrippedModifiers()const  { return mTrippedModifiers; }
	TrippedCaseBehaviour	GetTrippedCase() const { return mTrippedCase; }
	NSUInteger				GetExecutionEntryOrder() const { return mExecutionEntryOrder; }

	bool					DontRotateStatusMenu() const;

};

@implementation HExecutionEntry

@synthesize name;
@synthesize executionGroup;
@synthesize debugState;
@synthesize actions;
@synthesize order;

- (NSComparisonResult) compare: (HExecutionEntry*) other;
{
	NSComparisonResult result = [self.name compare:other.name options:NSCaseInsensitiveSearch|NSDiacriticInsensitiveSearch|NSForcedOrderingSearch];
	if ( result == NSOrderedSame ) {
		result = (NSComparisonResult)IntegerCFCompare( self.order, other.order );
	}
	return result;
}

- (BOOL) isEqual: (id) other;
{
	if ( ![other isKindOfClass:[self class]] ) return NO;
	return [self compare:other] == NSOrderedSame;
}

- (NSUInteger) hash;
{
	return self.order;
}

- (void) cancelExecutionGroup;
{
	HExecuteActionListManager::CancelExecutionGroup( self.executionGroup );
}

- (void) setExecutionDebugState: (ExecutingEntryDebugState) state;
{
	HExecuteActionListManager::SetExecutionDebugState( self.executionGroup, state );
}

- (NSString*) description;
{
	return [NSString stringWithFormat:@"HExecutionEntry %@ [%@] action %@ (debug state: %d)", self.name, self.executionGroup, self.actions, self.debugState];
}

@end

static EventTime
GetInterActionDelay()
{
	float delay;
	Get( KMPreferences::kInterActionDelay, &delay, kDefaultInterActionDelayTime );
	return delay * kEventDurationSecond;
}

static void AddExecutionDebugStates( NSUUID* executionGroup )
{
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		gExecutionDebugStates = [NSMutableDictionary new];
	});
	HExecutionDebugState* state = [HExecutionDebugState new];
	state.debugState = MacroDebugger::PauseNewMacros() ? ExecutingEntryDebugState::kPaused : ExecutingEntryDebugState::kRunning;
	[gExecutionDebugStates setObject:state forKey:executionGroup];
	XErrorHandlerMaybeLogFormat2( "HExecuteActionListEntry", @"AddExecutionDebugStates for macro %@ with state %d", executionGroup, state.debugState );
}

static void RemoveExecutionDebugStates( NSUUID* executionGroup )
{
	XErrorHandlerMaybeLogFormat1( "HExecuteActionListEntry", @"RemoveExecutionDebugStates for macro %@", executionGroup );
	[gExecutionDebugStates removeObjectForKey:executionGroup];
}

HExecuteActionListEntry::HExecuteActionListEntry( ExecuteActionListDisplay display, const MMacro* inMacro, TBaseTrigger* inTrippedTrigger, NSUUID* trackingID )
	:	mTrackingID( trackingID ),
		mExecutionGroup( gCurrentExecuteActionListEntry ? gCurrentExecuteActionListEntry->GetExecutionGroup() : [NSUUID UUID] ),
		mExecutionMacroName( inMacro->GetName() ),
		mExecutionMacroGroupName( inMacro->GetOwningGroup()->GetName() ),
		mActionList( inMacro->GetActionList() ),
		mDisplayInList( display == ExecuteActionListDisplay::kDisplay )
{
	if ( mDisplayInList ) { // Never add a named macro to an execution group
		mExecutionGroup = [NSUUID UUID];
		AddExecutionDebugStates( mExecutionGroup );
	}
	check( !mDisplayInList || mExecutionMacroName );
	const THotKeyTrigger* pHotKeyTrigger = dynamic_cast<const THotKeyTrigger*>(inTrippedTrigger);
	if ( pHotKeyTrigger ) {
		mTrippedKeyCode = pHotKeyTrigger->GetKeyCode();
		mTrippedModifiers = pHotKeyTrigger->GetModifiers();
	}
	const TTypedStringTrigger* pTypedStringTrigger = dynamic_cast<const TTypedStringTrigger*>(inTrippedTrigger);
	if ( pTypedStringTrigger ) {
		mTrippedCase = pTypedStringTrigger->GetTrippedCaseBehaviour();
	}
	int simulateDeletes = inTrippedTrigger ? inTrippedTrigger->SimulateDeletes() : 0;
	if ( simulateDeletes ) {
		XErrorHandlerMaybeLogFormat1( "TypedString", CFSTR("Simulate %d backspaces before macro"), simulateDeletes );
	}
	while ( simulateDeletes > 0 ) {
		ASimulateKeystroke* action = new ASimulateKeystroke();
		action->SetSimulateKeystroke( kVK_Delete, 0 );
		mActionList.InsertActionAt( 0, action );
		--simulateDeletes;
	}
	
	XErrorHandlerMaybeLogFormat1( "HExecuteActionListEntry", @"HExecuteActionListEntry::HExecuteActionListEntry create from macro %@", mExecutionMacroName );
}

HExecuteActionListEntry::HExecuteActionListEntry( ExecuteActionListDisplay display, NSString* displayName, const HActionList& inActionList, NSUUID* trackingID )
	:	mTrackingID( trackingID ),
		mExecutionGroup( gCurrentExecuteActionListEntry ? gCurrentExecuteActionListEntry->GetExecutionGroup() : [NSUUID UUID] ),
		mExecutionMacroName( gCurrentExecuteActionListEntry ? gCurrentExecuteActionListEntry->GetExecutionMacroName() : displayName ),
		mExecutionMacroGroupName( gCurrentExecuteActionListEntry ? gCurrentExecuteActionListEntry->GetExecutionMacroGroupName() : displayName ),
		mActionList( inActionList ),
		mDisplayInList( display == ExecuteActionListDisplay::kDisplay )
{
	check( !mDisplayInList || !gCurrentExecuteActionListEntry );
	if ( mDisplayInList ) {
		AddExecutionDebugStates( mExecutionGroup );
	}
	XErrorHandlerMaybeLogMessage( "HExecuteActionListEntry", CFSTR("HExecuteActionListEntry::HExecuteActionListEntry create from action list") );
}

HExecuteActionListEntry::HExecuteActionListEntry( ExecuteActionListDisplay display, NSString* displayName, const HBaseAction* inTrialAction, NSUUID* trackingID )
	:	mTrackingID( trackingID ),
		mExecutionGroup( gCurrentExecuteActionListEntry ? gCurrentExecuteActionListEntry->GetExecutionGroup() : [NSUUID UUID] ),
		mExecutionMacroName( gCurrentExecuteActionListEntry ? gCurrentExecuteActionListEntry->GetExecutionMacroName() : displayName ),
		mExecutionMacroGroupName( gCurrentExecuteActionListEntry ? gCurrentExecuteActionListEntry->GetExecutionMacroGroupName() : displayName ),
		mDisplayInList( display == ExecuteActionListDisplay::kDisplay )
{
	if ( mDisplayInList ) {
		AddExecutionDebugStates( mExecutionGroup );
	}
	XErrorHandlerMaybeLogMessage( "HExecuteActionListEntry", CFSTR("HExecuteActionListEntry::HExecuteActionListEntry create from action") );
	mActionList.AppendAction( inTrialAction->Clone() );
}

HExecuteActionListEntry::~HExecuteActionListEntry()
{
	if ( mDisplayInList ) {
		RemoveExecutionDebugStates( mExecutionGroup );
	}
	XErrorHandlerMaybeLogMessage( "HExecuteActionListEntry", CFSTR("HExecuteActionListEntry::HExecuteActionListEntry destroy") );
}

bool
HExecuteActionListEntry::DontRotateStatusMenu() const
{
	return mActionList.DontRotateStatusMenu();
}

bool
HExecuteActionListEntry::IsTrackingID( NSUUID* trackingID ) const
{
	check( trackingID );
	return [mTrackingID isEqual:trackingID];
}

bool
HExecuteActionListEntry::IsExecutionGroup( NSUUID* executionGroup ) const
{
	check( executionGroup && mExecutionGroup );
	return [mExecutionGroup isEqual:executionGroup];
}

bool
HExecuteActionListEntry::IsExecutingExecuteMacro() const
{
	return dynamic_cast<AExecuteMacro*>(GetExecutionAction()) != NULL;
}

bool
HExecuteActionListEntry::IsExecutingLoop() const
{
	HBaseAction* a = GetExecutionAction();
	if ( !a ) return false;
	return a->IsLoop();
}

NSString*
HExecuteActionListEntry::GetExecutionActionDescription() const
{
	if ( mExecutionActionDescription ) return mExecutionActionDescription;

	int actionIndex = mCurActionIndex;
	if ( actionIndex < 0 ) {
		actionIndex = 0;
	}
	while ( actionIndex < mActionList.GetCount() && !mActionList[actionIndex]->IsActive() ) {
		actionIndex++;
	}
	if ( actionIndex < mActionList.GetCount() ) {
		mExecutionAction = mActionList[actionIndex];
		mExecutionActionDescription = mExecutionAction->GetActionDescriptionString();
	} else {
		mExecutionAction = NULL;
		mExecutionActionDescription = @"";
	}
	return mExecutionActionDescription;
}

HBaseAction*
HExecuteActionListEntry::GetExecutionAction() const
{
	(void) GetExecutionActionDescription();
	return mExecutionAction;
}

void
HExecuteActionListEntry::HandleDoAction( HBaseAction* inAction )
{
	try {
		float time = inAction->DoAction();
		HTimeSaved::Add( time );
		XClipboardHistory::Sync();
	} catch ( XException actionErr ) {
		XErrorHandler::LogException( actionErr, CFSTR("HExecuteActionListEntry::HandleDoAction()") );
	} catch(...) {
		XErrorHandlerMaybeLogMessage( "always", CFSTR("Caught Unknown Error while doing action in HExecuteActionListEntry::HandleDoAction()") );
	}
}

bool
HExecuteActionListEntry::ContinueExecution()
{
	HExecutionDebugState* state = [gExecutionDebugStates objectForKey:mExecutionGroup];
	ExecutingEntryDebugState debugState = state ? state.debugState : ExecutingEntryDebugState::kRunning;
	NSUInteger debugStepEntryOrder = state.debugStepEntryOrder;

	if ( debugState == ExecutingEntryDebugState::kPaused ) {
		HBaseAction* nextAction = GetExecutionAction();
		if ( nextAction && dynamic_cast<ADebuggerAction*>(nextAction) ) {
			debugState = ExecutingEntryDebugState::kStepInto;
		}
	}

	if ( debugState == ExecutingEntryDebugState::kPaused ) return true;

	bool	destroyMe = false;
	bool	executeNextAction = false;
	bool	checkDebugPause = false;

	gCurrentExecuteActionListEntry = this;
	if ( mDelayingBetweenActions ) {
		mCurActionExecuteTime = GetCurrentEventTime();
		mDelayingBetweenActions = (mNextActionExecuteTime > mCurActionExecuteTime);
		if ( !mDelayingBetweenActions ) {
			HandleDoAction( mActionList[mCurActionIndex] );
			mFirstAction = false;
		}
	} else if ( mCurActionIndex >= mActionList.GetCount() ) {
		check( mCurActionIndex == -1 ); // Could happen if the macro has no actions
		destroyMe = true;
	} else if ( mCurActionIndex < 0 ) {
		executeNextAction = true;
		if ( debugState == ExecutingEntryDebugState::kStepInto && debugStepEntryOrder != GetExecutionEntryOrder() ) {
			checkDebugPause = true;
		}
	} else if ( mSkipLoop ) {
		mSkipLoop = true;
		mActionList[mCurActionIndex]->ActionAborted();
		executeNextAction = true;
		checkDebugPause = true;
	} else if ( mActionList[mCurActionIndex]->IsActionCompleted() ) {
		executeNextAction = true;
		checkDebugPause = true;
// mCancel may be set now!
	} else if ( !mCancel && mActionList[mCurActionIndex]->HasTimeOut() && GetCurrentEventTime() > mCurActionExecuteTime + mActionList[mCurActionIndex]->TimeOutPeriod() ) {
		// timed out
		if ( mActionList[mCurActionIndex]->TimeOutAbortsMacro() ) {
			[GrowlApplicationBridge 
			 notifyWithTitle:CFBridgingToNS(kmEngineName)
			 description:[NSString stringWithFormat:[@"MacroCancelledFormat" localized], [@"TimeoutCancel" localized]]
			 notificationName:@"Macro Cancelled"
			 iconData:nil priority:0 isSticky:NO clickContext:nil];
			XErrorHandlerMaybeLogFormat1( "always", [@"MacroCancelledFormat" localized], [@"TimeoutCancel" localized] );
			destroyMe = true;
			HExecuteActionListManager::CancelThisExecutionGroup();
		} else {
			mActionList[mCurActionIndex]->ActionAborted();
			executeNextAction = true;
			checkDebugPause = true;
		}
	}
	
	if ( !mCancel && !destroyMe && executeNextAction ) {
		mCurActionIndex++;
		mExecutionActionDescription = nil; // remove cache
		mExecutionAction = NULL; // remove cache
		while ( mCurActionIndex < mActionList.GetCount() && !mActionList[mCurActionIndex]->IsActive() ) {
			mCurActionIndex++;
		}
		if ( mCurActionIndex < mActionList.GetCount() ) {
			mCurActionExecuteTime = GetCurrentEventTime();
			mNextActionExecuteTime = mCurActionExecuteTime + GetInterActionDelay();
			mDelayingBetweenActions = !mFirstAction && (mNextActionExecuteTime > mCurActionExecuteTime);
			if ( checkDebugPause ) {
				switch( debugState ) {
					case ExecutingEntryDebugState::kPaused:
						check( false );
						[[clang::fallthrough]];
					case ExecutingEntryDebugState::kRunning:
					case ExecutingEntryDebugState::kStepOut:
						break;
					case ExecutingEntryDebugState::kStepInto:
						mDelayingBetweenActions = true;
						HExecuteActionListManager::SetExecutionDebugState( GetExecutionGroup(), ExecutingEntryDebugState::kPaused );
						break;
					case ExecutingEntryDebugState::kStepOver:
						if ( debugStepEntryOrder == GetExecutionEntryOrder() ) {
							mDelayingBetweenActions = true;
							HExecuteActionListManager::SetExecutionDebugState( GetExecutionGroup(), ExecutingEntryDebugState::kPaused );
						}
						break;
				}
			}
			if ( !mDelayingBetweenActions ) {
				HandleDoAction( mActionList[mCurActionIndex] );
				mFirstAction = false;
			}
		} else {
			destroyMe = true;
		}
	}
	if ( destroyMe ) {
		switch( debugState ) {
			case ExecutingEntryDebugState::kPaused:
				check( false );
				[[clang::fallthrough]];
			case ExecutingEntryDebugState::kRunning:
				break;
			case ExecutingEntryDebugState::kStepOut:
				if ( debugStepEntryOrder == GetExecutionEntryOrder() ) {
					HExecuteActionListManager::SetExecutionDebugState( GetExecutionGroup(), ExecutingEntryDebugState::kPaused );
				}
				break;
			case ExecutingEntryDebugState::kStepInto:
				HExecuteActionListManager::SetExecutionDebugState( GetExecutionGroup(), ExecutingEntryDebugState::kPaused );
				break;
			case ExecutingEntryDebugState::kStepOver:
				if ( debugStepEntryOrder == GetExecutionEntryOrder() ) {
					HExecuteActionListManager::SetExecutionDebugState( GetExecutionGroup(), ExecutingEntryDebugState::kPaused );
				}
				break;
		}
	}
	gCurrentExecuteActionListEntry = NULL;
	
	return !mCancel && !destroyMe;
}

using namespace HExecuteActionListManager;

class HExecuteActionListManagerExecutor final
	:	public XEventLoopTimer
{
private:
	
	bool mStartedRotating;

	std::list<std::unique_ptr<HExecuteActionListEntry>> mEntries;
	
	HExecuteActionListManagerExecutor( const HExecuteActionListManagerExecutor& inObj ) = delete;
	HExecuteActionListManagerExecutor& operator= ( const HExecuteActionListManagerExecutor& inObj ) = delete;
	HExecuteActionListManagerExecutor() = delete;

protected:
	
	void		HandleTimerEvent();
	
public:
	
	HExecuteActionListManagerExecutor( std::unique_ptr<HExecuteActionListEntry> entry );
	virtual ~HExecuteActionListManagerExecutor();
	
	void		CheckStart();
	void		AddEntry( std::unique_ptr<HExecuteActionListEntry> entry );
	bool		IsExecuting( NSUUID* trackingID ) const;
	NSArray*	ExecutingList( ExecutingListSortOrder sort );
	void		CancelExecution( NSUUID* trackingID );
	void		CancelAllExecution();
	void		CancelAllOtherExecutionGroups( NSUUID* executionGroup );
	void		CancelThisExecutionGroup( NSUUID* executionGroup );
	void		CancelJustThisMacro( NSUUID* executionGroup );
	void		CancelJustThisLoop( NSUUID* executionGroup );
	void		SetExecutionDebugState( NSUUID* executionGroup, ExecutingEntryDebugState state );
	void		SetExecutionDebugStateAll( ExecutingEntryDebugState state, NSUUID* execeptExecutionGroup = nil );
	bool		DontRotateStatusMenu() const;
};

HExecuteActionListManagerExecutor* gExecutor = NULL;

HExecuteActionListManagerExecutor::HExecuteActionListManagerExecutor( std::unique_ptr<HExecuteActionListEntry> entry )
	:	XEventLoopTimer( GetMainEventLoop(), kInitialInterval, kTimerInterval ),
		mStartedRotating( false )
{
	AddEntry( std::move(entry) );
}

HExecuteActionListManagerExecutor::~HExecuteActionListManagerExecutor()
{
	gExecutor = NULL;
	if ( mStartedRotating ) {
		StatusMenu::StopExecuting();
	}
}

bool
HExecuteActionListManagerExecutor::DontRotateStatusMenu() const
{
	for ( auto& i : mEntries ) {
		if ( i->DontRotateStatusMenu() ) {
			return true;
		}
	}
	return false;
}

void
HExecuteActionListManagerExecutor::CheckStart()
{
	UInt32 maxcount;
	KMPreferences::Get( KMPreferences::kMaximumSimultaneousMacrosID, &maxcount, 50 );
	if ( mEntries.size() > maxcount ) {
		XErrorHandlerMaybeLogFormat1( "always", CFSTR("More than %d simultaneous macros - aborting everything"), maxcount );
		CancelAllExecution();
	} else {
		if ( !mStartedRotating && !DontRotateStatusMenu() ) {
			mStartedRotating = true;
			StatusMenu::StartExecuting();
		}
	}
}

void
HExecuteActionListManagerExecutor::AddEntry( std::unique_ptr<HExecuteActionListEntry> entry )
{
	mEntries.push_back( std::move(entry) );
	CheckStart();
	StatusMenu::UpdateCancelMenu();
}

void
HExecuteActionListManagerExecutor::CancelAllExecution()
{
	for ( auto& i : mEntries ) {
		i->mCancel = true;
	}
}

void
HExecuteActionListManagerExecutor::CancelAllOtherExecutionGroups( NSUUID* executionGroup )
{
	for ( auto& i : mEntries ) {
		if ( !i->IsExecutionGroup( executionGroup ) ) {
			i->mCancel = true;
		}
	}
}

void
HExecuteActionListManagerExecutor::CancelThisExecutionGroup( NSUUID* executionGroup )
{
	for ( auto& i : mEntries ) {
		if ( i->IsExecutionGroup( executionGroup ) ) {
			i->mCancel = true;
		}
	}
}

void
HExecuteActionListManagerExecutor::CancelJustThisMacro( NSUUID* executionGroup )
{
	for ( auto& i : ReverseWrapper(mEntries) ) {
		if ( i->IsExecutionGroup( executionGroup ) ) {
			if ( i->IsExecutingExecuteMacro() ) {
				break;
			}
			i->mCancel = true;
		}
	}
}

void
HExecuteActionListManagerExecutor::CancelJustThisLoop( NSUUID* executionGroup )
{
	for ( auto& i : ReverseWrapper(mEntries) ) {
		if ( i->IsExecutionGroup( executionGroup ) ) {
			if ( i->IsExecutingLoop() ) {
				i->mSkipLoop = true;
				break;
			}
			i->mCancel = true;
		}
	}
}

void
HExecuteActionListManagerExecutor::SetExecutionDebugState( NSUUID* executionGroup, ExecutingEntryDebugState inState )
{
	HExecutionDebugState* state = [gExecutionDebugStates objectForKey:executionGroup];
	check( state );
	if ( state ) {
		state.debugState = inState;
		for ( auto& i : mEntries ) {
			if ( i->IsExecutionGroup( executionGroup ) ) {
				state.debugStepEntryOrder = i->GetExecutionEntryOrder();
			}
		}
	}
}

void
HExecuteActionListManagerExecutor::SetExecutionDebugStateAll( ExecutingEntryDebugState debugState, NSUUID* execeptExecutionGroup )
{
	for ( NSUUID* key in gExecutionDebugStates.allKeys ) {
		HExecutionDebugState* state = gExecutionDebugStates[key];
		if ( !execeptExecutionGroup || ![execeptExecutionGroup isEqual:key] ) {
			state.debugState = debugState;
			for ( auto& i : mEntries ) {
				if ( i->IsExecutionGroup( key ) ) {
					state.debugStepEntryOrder = i->GetExecutionEntryOrder();
				}
			}
		}
	}
}

bool
HExecuteActionListManagerExecutor::IsExecuting( NSUUID* trackingID ) const
{
	check( trackingID );
	
	bool found = false;
	for ( auto& i : mEntries ) {
		if ( !i->mCancel && i->IsTrackingID( trackingID ) ) {
			found = true;
			break;
		}
	}
	return found;
}

NSArray*
HExecuteActionListManagerExecutor::ExecutingList( ExecutingListSortOrder sort )
{
	NSMutableArray* entries = [NSMutableArray new];
	for ( auto& i : mEntries ) {
		if ( !i->mCancel && i->mDisplayInList ) {
			HExecutionEntry* entry = [HExecutionEntry new];
			entry.name = i->GetExecutionMacroName();
			entry.executionGroup = i->GetExecutionGroup();
			check( entry.executionGroup );
			HExecutionDebugState* state = [gExecutionDebugStates objectForKey:entry.executionGroup];
			check( state );
			entry.debugState = state ? state.debugState : ExecutingEntryDebugState::kRunning;
			entry.order = i->GetExecutionEntryOrder();
			NSMutableArray* actions = [NSMutableArray new];
			for ( auto& j : mEntries ) {
				if ( j->IsExecutionGroup( entry.executionGroup ) ) {
					[actions addObject:j->GetExecutionActionDescription()];
				}
			}
			entry.actions = actions;
			[entries addObject:entry];
		}
	}
	switch ( sort ) {
		case ExecutingListSortOrder::kAlphabetical:
			[entries sortUsingComparator:^( HExecutionEntry* e1, HExecutionEntry* e2 ) {
				return_from_block [e1 compare:e2];
			}];
			break;
		case ExecutingListSortOrder::kExecution:
			// This should be a NOP since the mEntries is ordered by execution order already
			[entries sortUsingComparator:^( HExecutionEntry* e1, HExecutionEntry* e2 ) {
				return_from_block (NSComparisonResult) IntegerCFCompare( e1.order, e2.order );
			}];
			break;
	}

	return entries;
}

void
HExecuteActionListManagerExecutor::CancelExecution( NSUUID* trackingID )
{
	check( trackingID );

	for ( auto& i : mEntries ) {
		if ( !i->mCancel && i->IsTrackingID( trackingID ) ) {
			i->mCancel = true;
			break;
		}
	}
}

void
HExecuteActionListManagerExecutor::HandleTimerEvent()
{
	bool haveSynced = false;
	for ( auto& i : mEntries ) {
		if ( !i->mCancel ) {
			if ( !haveSynced ) {
				XClipboardHistory::Sync();
				haveSynced = true;
			}
			if ( !i->ContinueExecution() ) {
				i->mCancel = true;
			}
		}
	}
	int orgSize = mEntries.size();
	mEntries.remove_if( []( std::unique_ptr<HExecuteActionListEntry>& i ){
		return_from_block i->mCancel;
	});
	int finalSize = mEntries.size();
	if ( finalSize == 0 ){
		delete this;
	}
	if ( orgSize != finalSize ) {
		StatusMenu::UpdateCancelMenu();
	}
	MacroDebugger::Update();
}

static void
AddEntry( std::unique_ptr<HExecuteActionListEntry> entry )
{
	if ( !gExecutor ) {
		gExecutor = new HExecuteActionListManagerExecutor( std::move(entry) );
	} else {
		gExecutor->AddEntry( std::move(entry) );
	}
}

void
HExecuteActionListManager::Execute( ExecuteActionListDisplay display, const MMacro* inMacro, TBaseTrigger* inTrippedTrigger, NSUUID* trackingID )
{
	KMMacroLastExecuted::Executed( inMacro->GetUID() );
	AddEntry( std::unique_ptr<HExecuteActionListEntry>( new HExecuteActionListEntry( display, inMacro, inTrippedTrigger, trackingID ) ) );
}

void
HExecuteActionListManager::Execute( ExecuteActionListDisplay display, NSString* displayName, const HActionList& inActionList, NSUUID* trackingID )
{
	if ( inActionList.GetCount() == 0 ) return;
	
	AddEntry( std::unique_ptr<HExecuteActionListEntry>( new HExecuteActionListEntry( display, displayName, inActionList, trackingID ) ) );
}

void
HExecuteActionListManager::Execute( ExecuteActionListDisplay display, NSString* displayName, const HBaseAction* inTrialAction, NSUUID* trackingID )
{
	AddEntry( std::unique_ptr<HExecuteActionListEntry>( new HExecuteActionListEntry( display, displayName, inTrialAction, trackingID ) ) );
}

bool
HExecuteActionListManager::IsExecuting()
{
	return gExecutor != NULL;
}

bool
HExecuteActionListManager::IsExecuting( NSUUID* trackingID )
{
	if ( !gExecutor ) return false;
	
	return trackingID && gExecutor->IsExecuting( trackingID );
}

NSArray*
HExecuteActionListManager::ExecutingList( ExecutingListSortOrder sort )
{
	if ( !gExecutor ) return @[];

	return gExecutor->ExecutingList( sort );
}

void
HExecuteActionListManager::CancelExecution( NSUUID* trackingID )
{
	if ( !gExecutor ) return;
	
	if ( trackingID ) {
		gExecutor->CancelExecution( trackingID );	
	}
}

void
HExecuteActionListManager::CancelAllOtherExecutionGroups()
{
	verify_action( gExecutor, return );
	verify_action( gCurrentExecuteActionListEntry, return );
	
	gExecutor->CancelAllOtherExecutionGroups( gCurrentExecuteActionListEntry->GetExecutionGroup() );
}

void
HExecuteActionListManager::CancelThisExecutionGroup()
{
	verify_action( gExecutor, return );
	verify_action( gCurrentExecuteActionListEntry, return );

	gExecutor->CancelThisExecutionGroup( gCurrentExecuteActionListEntry->GetExecutionGroup() );
}

void
HExecuteActionListManager::CancelJustThisMacro()
{
	verify_action( gExecutor, return );
	verify_action( gCurrentExecuteActionListEntry, return );

	gExecutor->CancelJustThisMacro( gCurrentExecuteActionListEntry->GetExecutionGroup() );
}

void
HExecuteActionListManager::CancelJustThisLoop()
{
	verify_action( gExecutor, return );
	verify_action( gCurrentExecuteActionListEntry, return );

	gExecutor->CancelJustThisLoop( gCurrentExecuteActionListEntry->GetExecutionGroup() );
}

NSUUID*
HExecuteActionListManager::GetThisExecutionGroup()
{
	verify_action( gExecutor, return nil );
	verify_action( gCurrentExecuteActionListEntry, return nil );

	return gCurrentExecuteActionListEntry->GetExecutionGroup();
}

void
HExecuteActionListManager::CancelExecutionGroup( NSUUID* executionGroup )
{
	if ( gExecutor ) {
		gExecutor->CancelThisExecutionGroup( executionGroup );
	}
}

void
HExecuteActionListManager::SetExecutionDebugState( NSUUID* executionGroup, ExecutingEntryDebugState state )
{
	if ( gExecutor ) {
		gExecutor->SetExecutionDebugState( executionGroup, state );
	}
}

void
HExecuteActionListManager::SetExecutionPausedAll( BOOL paused, NSUUID* execeptExecutionGroup )
{
	if ( gExecutor ) {
		gExecutor->SetExecutionDebugStateAll( paused ? ExecutingEntryDebugState::kPaused : ExecutingEntryDebugState::kRunning, execeptExecutionGroup );
	}
}
void
HExecuteActionListManager::SetExecutionDebugStateAll( ExecutingEntryDebugState state, NSUUID* execeptExecutionGroup )
{
	if ( gExecutor ) {
		gExecutor->SetExecutionDebugStateAll( state, execeptExecutionGroup );
	}
}

void
HExecuteActionListManager::CancelAllExecution()
{
	if ( gExecutor ) {
		gExecutor->CancelAllExecution();
	}
}

UInt32
HExecuteActionListManager::GetTrippedKeyCode()
{
	verify_action( gCurrentExecuteActionListEntry, return hkUndefinedKeyCode );
	
	return gCurrentExecuteActionListEntry->GetTrippedKeyCode();
}

UInt32
HExecuteActionListManager::GetTrippedModifiers()
{
	verify_action( gCurrentExecuteActionListEntry, return 0 );
	
	return gCurrentExecuteActionListEntry->GetTrippedModifiers();
}

TrippedCaseBehaviour
HExecuteActionListManager::GetTrippedCase()
{
	verify_action( gCurrentExecuteActionListEntry, return TrippedCaseBehaviour::kNone );
	
	return gCurrentExecuteActionListEntry->GetTrippedCase();
}

NSString*
HExecuteActionListManager::ExecutingMacroName()
{
	verify_action( gCurrentExecuteActionListEntry, return @"" );
	return gCurrentExecuteActionListEntry->GetExecutionMacroName();
}

NSString*
HExecuteActionListManager::ExecutingMacroGroupName()
{
	if ( !gCurrentExecuteActionListEntry ) {
		return @"";
	}
	return gCurrentExecuteActionListEntry->GetExecutionMacroGroupName();
}
